#include "Cgasf.h"
#include "CompositeGridOperators.h"
#include "Ogshow.h"
#include "Ogen.h"
#include "Ogmg.h"
#include "AsfParameters.h"
#include "App.h"
// #include "turbulenceModels.h"


//\begin{>>CompositeGridSolverInclude.tex}{\subsection{initialize}} 
int Cgasf::
setupGridFunctions()
//=========================================================================================
// /Description:
//    initialize a CompositeGridSolver.
//
//\end{CompositeGridSolverInclude.tex}  
//=========================================================================================
{
  assert( current==0 );
  GridFunction & solution = gf[current];
  CompositeGrid & cg = *solution.u.getCompositeGrid();

  cg.update(MappedGrid::THEcenter | MappedGrid::THEvertex | 
	    MappedGrid::THEinverseVertexDerivative ); // **** fix this for rectangular ****
  
  if( pp==NULL )
    pp=new realCompositeGridFunction;
  if( ppx==NULL )
    ppx=new realCompositeGridFunction;

  Range all;
  CompositeGridOperators & operators = *gf[current].u.getOperators();

  p().updateToMatchGrid(cg,all,all,all); p()=0.;
  p().setOperators(operators);
  pressureRightHandSide.updateToMatchGrid(cg,all,all,all); pressureRightHandSide=0.;
  pressureRightHandSide.setOperators(operators);
  px().updateToMatchGrid(cg,all,all,all);
  px().setOperators(operators);
  int stencilSize=int( pow(3,cg.numberOfDimensions())+1 ); 
  coeff.updateToMatchGrid(cg,stencilSize,all,all,all);  // add one for interpolation
  
  coeff.setOperators(operators); 
  
  coeff.setIsACoefficientMatrix(TRUE,stencilSize);  
  operators.setStencilSize(stencilSize);
  
  if( implicitSolver==NULL )
    {
      numberOfImplicitSolvers=1;
      implicitSolver= new Oges [numberOfImplicitSolvers];
    }
  implicitSolver[0].setGrid(cg);
  implicitSolver[0].setOrderOfAccuracy(2);
  implicitSolver[0].setCoefficientArray(coeff);
  
  if( parameters.dbase.get<AsfParameters::TestProblems >("testProblem")!=AsfParameters::standard )
    assignTestProblem(gf[current]);   // ************************ is this ok ? *****

  return DomainSolver::setupGridFunctions();
}



