*
*  cgcns command file: compute errors for flow in the twilight zone
* 
* Usage:
*    cgcns [-noplot] tz.cmd -g=<grid> -tf=<final time> -tp=<tPlot> -tz=[poly|trig] -degreex=<> -degreet=<> ...
*                           -cnsVariation=<> -show=<show file> -go=[go|halt|og]
*       $cnsVariation : jameson, godunov, nonconservative
* 
* Examples:
*     cgcns tz.cmd -g=square20 -tf=1. -tp=.1 
*     cgcns tz.cmd -g=cice2.hdf -tf=1. -tp=.1 
*     cgcns tz.cmd -g=box20 -tf=1. -tp=.1 
* -- nononservative:
*     cgcns tz.cmd -g=square10 -cnsVariation=nonconservative -tf=1. -tp=.1 -degreex=2 -degreet=2 [exact]
* 
* 
* --- set default values for parameters ---
$grid="square20.hdf"; $show = " "; $backGround="square"; $noplot=""; $cnsVariation="godunov"; 
$tFinal=1.; $tPlot=.1; $cfl=1.; $debug=1; $tol=.2; $x0=.5; $dtMax=1.e10; $nbz=2; 
$tz = "poly"; $degreex=2; $degreet=2; $fx=1.; $fy=1.; $fz=1.; $ft=1.;
$ratio=2;  $nrl=1;  # refinement ratio and number of refinement levels
$go="halt"; 
* 
* ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"l=i"=> \$nrl,"r=i"=> \$ratio, "tf=f"=>\$tFinal,"debug=i"=> \$debug, \
            "tp=f"=>\$tPlot, "bg=s"=>\$backGround,"show=s"=>\$show,"noplot=s"=>\$noplot,"tz=s"=>\$tz,\
            "degreex=i"=>\$degreex, "degreet=i"=>\$degreet,"fx=f"=>\$fx,"fy=f"=>\$fy,"fz=f"=>\$fz,"ft=f"=>\$ft,\
            "cnsVariation=s"=>\$cnsVariation,"go=s"=>\$go );
* -------------------------------------------------------------------------------------------------
if( $tz eq "poly" ){ $tz="turn on polynomial"; }else{ $tz="turn on trigonometric"; }
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
if( $cnsVariation eq "godunov" ){ $cnsVariation="compressible Navier Stokes (Godunov)"; }
if( $cnsVariation eq "jameson" ){ $cnsVariation="compressible Navier Stokes (Jameson)"; }  
if( $cnsVariation eq "nonconservative" ){ $cnsVariation="compressible Navier Stokes (non-conservative)"; }  
*
* Here is the overlapping grid to use:
$grid 
*
  $cnsVariation
*    compressible Navier Stokes (Godunov)  
**  compressible Navier Stokes (Jameson)
*
*   one step
*   add extra variables
*     1
  exit
* -- twilightzone options:
  $tz
  degree in space $degreex
  degree in time $degreet
  frequencies (x,y,z,t)   $fx $fy $fz $ft
* 
*
*  do not use iterative implicit interpolation
*
  final time $tFinal
  times to plot $tPlot
* no plotting
  plot and always wait
*
  show file options
    compressed
    * specify the max number of parallel hdf sub-files: 
    OBPSF:maximum number of parallel sub-files 2
    open
     $show
    frequency to flush
      20
    exit
* -- specify which variables will appear in the show file:
*     showfile options...
*     OBPSF:show variable: rho 1
*     OBPSF:show variable: u 0
*     OBPSF:show variable: v 0
*     OBPSF:show variable: w 0
*     OBPSF:show variable: T 0
*     OBPSF:show variable: Mach Number 0
*     OBPSF:show variable: p 1
*     close show file options
* 
  * no plotting
  reduce interpolation width
    2
  boundary conditions
    all=dirichletBoundaryCondition
    * all=noSlipWall uniform(T=.3572)
#    all=slipWall 
#     $backGround(0,0)=superSonicInflow uniform(r=2.6667,u=1.25,e=10.119,s=0)
#     $backGround(1,0)=superSonicOutflow
#     $backGround(0,1)=slipWall
#     $backGround(1,1)=slipWall
    done
*
  pde parameters
    mu
     0.0
    kThermal
     0.0
    heat release
      0.
    rate constant
      0.
   reciprocal activation energy
     1.
  done
  debug $debug
*
  turn off adaptive grids
*   save error function to the show file
**  show amr error function
  order of AMR interpolation
      2
  error threshold
     $tol 
  regrid frequency
    $regrid=$nbz*$ratio;
    $regrid
  change error estimator parameters
    default number of smooths
      1
    set scale factors     
      2 1 1 1 1 
    done
    exit
  change adaptive grid parameters
    refinement ratio
      $ratio
    default number of refinement levels
      $nrl
    number of buffer zones
      $nbz
*    width of proper nesting 
*      1
    grid efficiency
      .7 .5 
  exit
*
#   initial conditions
# *    smooth step function
#      step function
#       $xStep
# *        5.
#       r=2.6667 u=1.25 e=10.119 
#       r=1. e=1.786 s=0.
#   continue
continue
*
$go
