// This file automatically generated from ims.bC with bpp.
#include "DomainSolver.h"
#include "CompositeGridOperators.h"
#include "GridCollectionOperators.h"
#include "interpPoints.h"
#include "SparseRep.h"
#include "ExposedPoints.h"
#include "Ogen.h"
#include "App.h"
#include "ParallelUtility.h"
#include "Oges.h"
#include "OgesParameters.h"
#include "AdamsPCData.h"
#include "gridFunctionNorms.h"

//==================================================================
//            Implicit Multi-step Method
//            --------------------------
//
//==================================================================

static bool useNewExposedPoints=true;

// here are some bpp macros that are used for the explicit and implicit predictor-corrector methods
// This file contains some macros that are shared amongst the different predictor-corrector methods

// ===============================================================================
//    Perform the initialization step for the PC method
//
//  /METHOD (input) : name of the method: adamsPC or implicitPC
// ===============================================================================



// =======================================================================================================
//    Macro to move the grids at the start of a PC time step.
// =======================================================================================================

// This next declaration also appears in advance.bC
#define updateOpt EXTERN_C_NAME(updateopt)
extern "C"
{
      void updateOpt(const int &nd1a,const int &nd1b,const int &nd2a,const int &nd2b,
                                    const int &nd3a,const int &nd3b,const int &nd4a,const int &nd4b, 		  const int &mask,real &u1, const real&u2,  
                                    const real&ut1, const real&ut2, const real&ut3, const real&ut4, 
                                    const int &ipar, const real& rpar, int & ierr );
}


#define ForBoundary(side,axis)   for( axis=0; axis<mg.numberOfDimensions(); axis++ ) for( side=0; side<=1; side++ )

//\begin{>>CompositeGridSolverInclude.tex}{\subsection{buildImplicitSolvers}} 
void DomainSolver::
buildImplicitSolvers(CompositeGrid & cg)
// ==========================================================================================
// /Description:
//     Determine the number and type of implicit solvers needed. Depending on the boundary
//
//  1) If the equations are decoupled and the boundary conditions for all components are 
//     the same then we can form one scalar implicit system.
//  2) If the equations are decoupled and the boundary conditions are not the same but 
//     decoupled then we can solve separate scalar implicit systems.
//  3) If the boundary conditions or equations are coupled then we solve a implicit system.  
//
//\end{CompositeGridSolverInclude.tex}  
// ==========================================================================================
{
    printF("DomainSolver::buildImplicitSolvers:ERROR: base class function called! This function should\n"
                  "              be re-written by a derived class!\n");
    Overture::abort("error");

}


//\begin{>>CompositeGridSolverInclude.tex}{\subsection{implicitSolve}} 
void DomainSolver::
implicitSolve(const real & dt0,
            	      GridFunction & cgf1,
                            GridFunction & cgf0)
// ==========================================================================================
// /Description:
//    The implicit method can be optionally used on only some grids. To implement this
//   approach we simply create a sparse matrix that is just the identity matrix on grids that
// are advanced explicitly but equal to the standard implicit matrix on grids that are advance
// implicitly:
// \begin{verbatim}
//         I - \nu \alpha \dt \Delta      on implicit grids
//         I                              on explicit grids
// \end{verbatim}
// If the form of the boundary conditions for the different components of $\uv$ are the same
// then we can build a single scalar matrix that can be used to advance each component, one after
// the other. If the boundary conditions are not of the same form then we build a matrix for
// a system of equations for the velocity components $(u,v,w)$.
//
// /dt0 (input) : time step used to build the implicit matrix.
// /cgf1 (input/output) : On input holds the right-hand-side for the implicit equations; on output
//    holds the solution.
// /cgf0 (input) : current best approximation to the solution. Used as initial guess for iterative
//   solvers and used for linearization.
//\end{CompositeGridSolverInclude.tex}  
// ==========================================================================================
{
    
    printF("DomainSolver::implicitSolve:ERROR: base class function called! This function should\n"
                  "              be re-written by a derived class!\n");
    Overture::abort("error");

}

// ===========================================================================================
//      *** Time-step update macro ***
// 
//         u1 <- u0 + AB1*UA0 + AB2*UB0 [ + DTI*UTIMPLICIT ]  (add last terms for implicit grids)
// ===========================================================================================


//\begin{>>CompositeGridSolverInclude.tex}{\subsection{advanceImplicitMultiStep}} 
void DomainSolver::
advanceImplicitMultiStep( real & t0, real & dt0, int & numberOfSubSteps, int & init, int initialStep  )
// ======================================================================================
//  /Description:
//     Time step using an implicit method on the viscous terms and and a multi-step
//    method on the rest.
//
//  This approach should be efficient for moving grids since a larger time step is taken
// and the grid need to be regenerated fewer times. The grid needs only be moved on the
// predictor step. Then one or more corrections can be applied.
//
//
//  Method:
//   Suppose we are solving a PDE:
//           u_t = f(u,x,t)  + F(x,t)
//   That we have split into an explicit part, fe(u),  and implicit part, A*u:
//           u_t = fe(u) + A u  + F(x,t)
//
//   Predictor (2nd-order): (implicit part involves u(p) )
//
//      (u(p)-u(n))/dt = 1.5*( fe(u(n)) + g(n) ) - .5*( fe(u(n-1))+g(n-1) )
//                       + A(  u(p)+u(n) )/2   + G(n)
//   where for TZ flow with exact soluton, ue, the forcing functions g and G are 
//             g(n) = ue_t - fe(ue)
//             G(n) = -[ A( ue(n+1)+ue(n) )/2 ]
// 
//   Corrector:
//      (u(c)-u(n))/dt = 5*( fe(u(p)) + g(p) ) + .5*( fe(u(n))+g(n) )  
//                            + A( u(c)+u(n) )/2   + G(n)
//   
//  Note: 
//           utImplicit = A*u(n)/2  + G(n) 
//  which is computed (once per step) by getUt when implicitOption=computeImplicitTermsSeparately
// 
//\end{CompositeGridSolverInclude.tex}  
// ==========================================================================================
{
    FILE *& debugFile =parameters.dbase.get<FILE* >("debugFile");
    FILE *& pDebugFile =parameters.dbase.get<FILE* >("pDebugFile");

    if( debug() & 4 )
        printF(" $$$$$$ DomainSolver::advanceImplicitMultiStep $$$$$$$$\n");

    assert( parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")==Parameters::implicit );

    assert( parameters.dbase.get<int >("orderOfPredictorCorrector")==2 );  // for now we just have 2nd-order in time

//  bool init=false;
    if( !parameters.dbase.get<DataBase >("modelData").has_key("AdamsImplicitData") )
    {
    // this must be the initial call to this routine
//    init=true;
        parameters.dbase.get<DataBase >("modelData").put<AdamsPCData>("AdamsImplicitData");
    }
//  if( init!=init_ )
//  {
//    printF("**** advanceImplicitMultiStep;WARNING: init!=init_ !! *************\n");
//    init=init_;  // do this for now
//  }
    

    AdamsPCData & adamsData = parameters.dbase.get<DataBase >("modelData").get<AdamsPCData>("AdamsImplicitData");
    
    real & dtb=adamsData.dtb;
    int &mab0 =adamsData.mab0, &mab1=adamsData.mab1, &mab2=adamsData.mab2;
    int &nab0 =adamsData.nab0, &nab1=adamsData.nab1, &nab2=adamsData.nab2, &nab3=adamsData.nab3;
    int &ndt0=adamsData.ndt0;
    real *dtp = adamsData.dtp;

    const int & predictorOrder = parameters.dbase.get<int>("predictorOrder");
    if( predictorOrder<0 || predictorOrder>2 )
    {
        if( init )
            printF("advanceImplicitMultiStep: ERROR: predictorOrder=%i! Will use default.\n",predictorOrder);
    }
    if( init && debug() & 1 )
    {
        printF("advanceImplicitMultiStep:INFO: predictorOrder=%i ( =0 -> use default, order=2)\n",predictorOrder);
    }
    
    int numberOfCorrections;
    if( parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod")==Parameters::adamsBashforth2 )
        numberOfCorrections=0; //kkc 060301 given the assertion above, this should never happen, right?
    else
        numberOfCorrections=parameters.dbase.get<int>("numberOfPCcorrections"); 
  
    
    if( debug() & 2 )
        fprintf(debugFile," *** Entering advanceImplicitMultiStep: t0=%e, dt0=%e *** \n",t0,dt0);
  
    
  // For moving grids we keep gf[mab0], gf[mab1] and gf[mab2]
  // For non-moving grids we keep gf[mab0], gf[mab1] and we set mab2==mab1

    const int numberOfGridFunctions =  movingGridProblem() ? 3 : 2; 

    mab2 = (mab0 -1 + numberOfGridFunctions) % numberOfGridFunctions;

  // 
    int mNew = mab2;    // new     : gf[mNew] : will hold u(t+dt)
    int mCur = mab0;    // current : gf[mCur] : holds u(t) 
    int mOld = mab1;    // old     : gf[mOld] : holds u(t-dt) if numberOfGridFunctions==3 otherwise mOld=mNew
    

    Parameters::ImplicitOption & implicitOption = parameters.dbase.get<Parameters::ImplicitOption >("implicitOption");
    
    implicitOption=Parameters::doNotComputeImplicitTerms; // no need to compute during initialization
    parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")=Parameters::crankNicolson;
    int iparam[10];
    real rparam[10];

    RealCompositeGridFunction & uti = fn[2];   // ****
    
    if( debug() & 2 )
    {
        fprintf(debugFile," *** Entering advanceImplicitMultiStep: t0=%e, dt0=%e *** \n",t0,dt0);
    }
  
    
    int grid;
    Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2]; 
    Range N = parameters.dbase.get<Range >("Rt");   // time dependent variables
    RealArray error(numberOfComponents()+3); 
    
  // real time0=getCPU();

    int numberOfExtraPressureTimeLevels=0;
    
    if( init )
    {
    // Form the matrix for implicit time stepping (optionally save the solution used for linearization)
    // NOTE: the matrix will only be generated the first time through or if parameters.dbase.get<int >("globalStepNumber") % refactorFrequency == 0 
    // We want to factor the matrix here since it may be needed for computing the RHS
        formMatrixForImplicitSolve(dt,gf[mNew], gf[mCur] );

    // **** To initialize the method we need to compute du/dt at times t and t-dt *****

    // this is a macro (pcMacros.h):
        if( movingGridProblem() )
        { 
            getGridVelocity( gf[mCur],t0 );
        }
        if( parameters.dbase.get<int >("orderOfTimeExtrapolationForPressure")!=-1 )
        {
            if( parameters.dbase.get<int >("orderOfPredictorCorrector")==2 && parameters.dbase.get<int >("orderOfTimeExtrapolationForPressure")>1 &&
                    poisson!=NULL && poisson->isSolverIterative()  )
            {
        // orderOfTimeExtrapolationForPressure==1 :  p(t+dt) = 2*p(t) - p(t-dt)
        //                                      2 :  p(t+dt) = 3*p(t) - 3*p(t-dt) + p(t-2*dt)
                assert( previousPressure==NULL );
                assert( !parameters.isMovingGridProblem() );  // fix for this case
                numberOfExtraPressureTimeLevels = parameters.dbase.get<int >("orderOfTimeExtrapolationForPressure") - 1;
                printf(" ***initPC: allocate %i extra grid functions to store the pressure at previous times ****\n",
                 	   numberOfExtraPressureTimeLevels);
                previousPressure = new realCompositeGridFunction [numberOfExtraPressureTimeLevels];
                for( int i=0; i<numberOfExtraPressureTimeLevels; i++ )
                {
                    previousPressure[i].updateToMatchGrid(gf[mCur].cg);
                }
            }
        }
        if( parameters.dbase.get<bool >("twilightZoneFlow") )
        {
            OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
            if( parameters.dbase.get<int >("orderOfPredictorCorrector")==4 )
            {
        // for the fourth-order PC method, first compute u.t(t-2*dt) and u.t(t-3*dt)
                int m,grid;
                for( m=0; m<=1; m++ )
                {
                    real tp=t0-(m+2)*dt0;
                    if( movingGridProblem() )
                    {
    	// move gf[mOld] to t-(m+2)*dt
              	moveGrids( t0,t0,tp,dt0,gf[mCur],gf[mCur],gf[mOld] );   // Is this correct? dt0?   
              	gf[mOld].u.updateToMatchGrid(gf[mOld].cg); // *wdh* 040826
                    }
                    gf[mOld].t=tp;
                    e.assignGridFunction( gf[mOld].u,tp );
                    updateStateVariables(gf[mOld]); // *wdh* 080204 
                    if( parameters.useConservativeVariables() )
              	gf[mOld].primitiveToConservative();
                    const int nab=(nab2+m) % 4;
                    for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
                    {
              	rparam[0]=tp;
              	rparam[1]=tp; // tforce
              	rparam[2]=tp+dt0; // tImplicit
              	iparam[0]=grid;
              	iparam[1]=gf[mOld].cg.refinementLevelNumber(grid);
              	iparam[2]=numberOfStepsTaken;
              	getUt(gf[mOld].u[grid],gf[mOld].getGridVelocity(grid),fn[nab][grid],iparam,rparam,
                    	      uti[grid],&gf[mOld].cg[grid]);
                    }
                }
            }
      // get solution and derivative at t-dt
            if( movingGridProblem() )
            {
        // move gf[mOld] to t-dt
                if( debug() & 2 )
                    fPrintF(parameters.dbase.get<FILE* >("debugFile"),"implicitPC: take an initial step backwards\n");
             // display(gf[mOld].cg[0].vertex()(I1,I2,I3,0),sPrintF(" gf[mOld] vertex before move back at t=%e",gf[mOld].t),
             //                  parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
                moveGrids( t0,t0,t0-dt0,dt0,gf[mCur],gf[mCur],gf[mOld] );          // this will set gf[mOld].t=t-dt
                if( debug() & 8 )
                {
                    for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
                    {
                        if( parameters.gridIsMoving(grid) )
              	{
                	  display(gf[mOld].cg[grid].vertex()(I1,I2,I3,0),sPrintF("\n *** PC: AFTER moveGrids:  gf[mOld] grid=%i vertex after move back t=%e",grid,gf[mOld].t),
                      		  parameters.dbase.get<FILE* >("debugFile"),"%10.7f ");
              	}
                    }
                }
                gf[mOld].u.updateToMatchGrid(gf[mOld].cg); // make sure the grid is correct, vertex used in TZ  *wdh* 040826
            }
            else
                gf[mOld].t=t0-dt0; 
            int grid;
            if( false )
            {
                for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
                {
                    MappedGrid & c = gf[mOld].cg[grid];
                    getIndex(c.dimension(),I1,I2,I3);
          // *** we need to move grid gf[mOld] to t-dt and initialize gf[mOld].u (at least the pressure)
                    Range Na(0,parameters.dbase.get<int >("numberOfComponents")-1);
                    gf[mOld].u[grid](I1,I2,I3,Na)=e(c,I1,I2,I3,Na,t0-dt0); 
                }
            }
            else
            {
        // *wdh* 050412
                e.assignGridFunction( gf[mOld].u,t0-dt0 );
                updateStateVariables(gf[mOld]); // *wdh* 080204 
            }
            fn[nab0]=0.; 
            fn[nab1]=0.; 
            for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
            {
                rparam[0]=gf[mOld].t;
                rparam[1]=gf[mOld].t; // tforce
                rparam[2]=gf[mCur].t-gf[mOld].t; // tImplicit
                iparam[0]=grid;
                iparam[1]=gf[mOld].cg.refinementLevelNumber(grid);
                iparam[2]=numberOfStepsTaken;
                getUt(gf[mOld].u[grid],gf[mOld].getGridVelocity(grid),fn[nab1][grid],iparam,rparam,
                	  uti[grid],&gf[mOld].cg[grid]);
            }
           // display(fn[nab1][0],sPrintF("ut(t-dt) from getUt at t=%e\n",gf[mOld].t),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
            if( false ) // **** wdh **** 040824
            {
                for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
                {
                    MappedGrid & c = gf[mOld].cg[grid];
                    getIndex(c.dimension(),I1,I2,I3);
                    Range Na(0,parameters.dbase.get<int >("numberOfComponents")-1);
                    fn[nab1][grid](I1,I2,I3,Na)=e.t(c,I1,I2,I3,Na,t0-dt0); 
                    if( parameters.gridIsMoving(grid) )
                    { // add on gDot.grad(u)
              	const realArray & gridVelocity = gf[mOld].getGridVelocity(grid);
              	const int na=parameters.dbase.get<int >("uc"), nb=na+c.numberOfDimensions()-1;   // ***** watch out ***
              	for( int n=na; n<=nb; n++ )
              	{
                	  fn[nab1][grid](I1,I2,I3,n)+=gridVelocity(I1,I2,I3,0)*e.x(c,I1,I2,I3,n,t0-dt0)+
                  	    gridVelocity(I1,I2,I3,1)*e.y(c,I1,I2,I3,n,t0-dt0);
                	  if( c.numberOfDimensions()>2 )
                  	    fn[nab1][grid](I1,I2,I3,n)+=gridVelocity(I1,I2,I3,2)*e.z(c,I1,I2,I3,n,t0-dt0);
              	}
                    }
                }
                display(fn[nab1][0],sPrintF("ut(t-dt) from TZ at t=%e\n",gf[mOld].t),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
            }
            if( numberOfExtraPressureTimeLevels>0 )
            {
        // get extra time levels for extrapolating the pressure in time (as an initial guess for iterative solvers)
                for( int i=0; i<numberOfExtraPressureTimeLevels; i++ )
                {
          // if orderOfPC==2 : we need p(t-2*dt), p(t-3*dt) ...
          //             ==4 : we need p(t-5*dt), ... **check**
                    const real tp = t0-dt0*(i+parameters.dbase.get<int >("orderOfPredictorCorrector"));
                    realCompositeGridFunction & pp = previousPressure[i];
                    Range all;
                    for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
                    {
              	e.gd( pp[grid],0,0,0,0,all,all,all,parameters.dbase.get<int >("pc"),tp);
                    }
                }
            }
            if( debug() & 8 || debug() & 64 )
            {
                for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
                {
                    aString buff;
                    display(gf[mOld].u[grid],sPrintF(buff,"\n ****implicitPC: Init:gf[mOld].u grid=%i : du/dt(t) t=%9.3e",grid,gf[mOld].t),
                                    parameters.dbase.get<FILE* >("debugFile"),"%9.3e ");
                    display(fn[nab1][grid],sPrintF(buff,"\n ****implicitPC: Init:fn[nab1] grid=%i : du/dt(t) t=%9.3e",grid,gf[mOld].t),
                                    parameters.dbase.get<FILE* >("debugFile"),"%9.3e ");
                    display(gf[mOld].getGridVelocity(grid),sPrintF("adams:init: gridVelocity[%i] at t=%9.3e\n",grid,gf[mOld].t),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
                    display(gf[mOld].cg[grid].vertex(),sPrintF("adams:init: gf[mOld].cg[%i].vertex at t=%9.3e\n",grid,gf[mOld].t),parameters.dbase.get<FILE* >("debugFile"),"%7.4f ");
                }
                if( parameters.isMovingGridProblem() )
                {
                    determineErrors( gf[mOld].u,gf[mOld].gridVelocity, gf[mOld].t, 0, error,
                           		       sPrintF(" adams:init: errors in u at t=%9.3e (t0-dt0=%9.3e)\n",gf[mOld].t,t0-dt0) );
                    fn[nab1].updateToMatchGrid(gf[mOld].cg);  // for moving grid TZ to get errors correct
                    determineErrors( fn[nab1],gf[mOld].gridVelocity, gf[mOld].t, 1, error,
                           		       sPrintF(" adams:init: errors in ut (fn[nab1]) at t=%9.3e (t0-dt0=%9.3e)\n",gf[mOld].t,t0-dt0) );
                }
            }
        }
        else  // ****** NOT twilightZoneFlow
        {
            printF(" **************** implicitPC: still need correct initial values for du/dt(t-dt)  ****** \n");
            printF(" **************** use values from du/dt(t)                                  ****** \n");
            fn[nab0]=0.;
            fn[nab1]=0.;
            if( parameters.useConservativeVariables() )
                gf[mCur].primitiveToConservative();
      // if( parameters.isAdaptiveGridProblem() )
      //   gf[mOld].u.updateToMatchGrid(gf[mOld].cg);  // 040928 -- why is this needed here ?
            if( debug() & 8 )
            {
                printF(" PC: init: gf[mOld].u.numberOfGrids=%i \n",gf[mOld].u.numberOfGrids());
                printF(" PC: init: gf[mOld].cg.numberOfComponentGrids=%i \n",gf[mOld].cg.numberOfComponentGrids());
                printF(" PC: init: gf[mCur].u.numberOfGrids=%i \n",gf[mCur].u.numberOfGrids());
                printF(" PC: init: gf[mCur].cg.numberOfComponentGrids=%i \n",gf[mCur].cg.numberOfComponentGrids());
            }
      // *wdh* 050319 gf[mOld].u=gf[mCur].u;  // 990903 give initial values to avoid NAN's at ghost points for CNS
            assign(gf[mOld].u,gf[mCur].u);  // 990903 give initial values to avoid NAN's at ghost points for CNS
            gf[mOld].form=gf[mCur].form;
            for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
            {
                rparam[0]=gf[mOld].t;
                rparam[1]=gf[mOld].t; // tforce
                rparam[2]=gf[mCur].t-gf[mOld].t; // tImplicit
                iparam[0]=grid;
                iparam[1]=gf[mOld].cg.refinementLevelNumber(grid);
                iparam[2]=numberOfStepsTaken;
    //     mappedGridSolver[grid]->getUt(gf[mOld].u[grid],gf[mOld].getGridVelocity(grid),fn[nab1][grid],iparam,rparam,
    // 				  uti[grid],&gf[mOld].cg[grid]);
                getUt(gf[mOld].u[grid],gf[mOld].getGridVelocity(grid),fn[nab1][grid],iparam,rparam,
                                  				  uti[grid],&gf[mOld].cg[grid]);
            }
            for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
            {
                MappedGrid & c = gf[mOld].cg[grid];
                getIndex(c.dimension(),I1,I2,I3);
        // fn[nab1][grid](I1,I2,I3,N)=fn[nab0][grid](I1,I2,I3,N);
                if( parameters.dbase.get<int >("orderOfPredictorCorrector")==4 )
                {
                    for( int m=0; m<=1; m++ )
                    {
              	const int nab=(mOld+m+1) % 4;
    	// *wdh* 050319 fn[nab][grid](I1,I2,I3,N)=fn[nab0][grid](I1,I2,I3,N);
              	assign(fn[nab][grid],fn[nab0][grid],I1,I2,I3,N);
                    }
                }
            }
        }
        if( parameters.dbase.get<int >("orderOfAccuracy")==4 )
        {
            OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
            const int numberOfGhostLines=2;
            Range V(parameters.dbase.get<int >("uc"),parameters.dbase.get<int >("uc")+parameters.dbase.get<int >("numberOfDimensions")-1);
            const int numberOfPreviousValuesOfPressureToSave=
                parameters.dbase.get<int >("orderOfPredictorCorrector")==2 ? 1 : 3;
            for( int m=0; m<numberOfPreviousValuesOfPressureToSave; m++ )
            {
                real tp=t0-(m+2)*dt0;
                assert( nab0==0 );
                const int nab=(nab0+m);
                for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
                {
                    MappedGrid & c = gf[mOld].cg[grid];
                    const IntegerArray & gridIndexRange = c.gridIndexRange();
                    getIndex(c.dimension(),I1,I2,I3);
          // save p for use when extrapolating in time
          //    ua(.,.,.,pc)= p(t-2*dt)  (for 2nd/4th order)
          //    ub(.,.,.,pc)= p(t-3*dt)  (for 4th order)
          //    uc(.,.,.,pc)= p(t-4*dt)  (for 4th order)
                    if( parameters.dbase.get<bool >("twilightZoneFlow") )
                    {
    	// *wdh* 050416 fn[nab][grid](I1,I2,I3,parameters.dbase.get<int >("pc"))=e(c,I1,I2,I3,parameters.dbase.get<int >("pc"),tp);  
            //  fn[nab][grid](I1,I2,I3,parameters.dbase.get<int >("pc"))=e(c,I1,I2,I3,parameters.dbase.get<int >("pc"),tp);
                        e.gd(fn[nab][grid],0,0,0,0,I1,I2,I3,parameters.dbase.get<int >("pc"),tp);
    	//  display(fn[nab][grid],"fn[nab][grid] after assigning for fourth order",parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
                    }
                    else
              	fn[nab][grid](I1,I2,I3,parameters.dbase.get<int >("pc"))=gf[mOld].u[grid](I1,I2,I3,parameters.dbase.get<int >("pc")); // *** fix this ****
          // We also extrapolate, in time, the ghost values of u -- used in the BC's
                    getIndex(gridIndexRange,I1,I2,I3,numberOfGhostLines);
                    int side,axis;
                    for( axis=0; axis<parameters.dbase.get<int >("numberOfDimensions"); axis++ )
                    {
              	for( side=0; side<=1; side++ )
              	{
                	  const int is=1-2*side;
                	  if( c.boundaryCondition(side,axis)>0 )
                	  {
    	    // set values on the two ghost lines
                  	    if( side==0 )
                    	      Iv[axis]=Range(gridIndexRange(side,axis)-2,gridIndexRange(side,axis)-1);
                  	    else
                    	      Iv[axis]=Range(gridIndexRange(side,axis)+1,gridIndexRange(side,axis)+2);
                  	    if( parameters.dbase.get<bool >("twilightZoneFlow") )
                  	    {
    	      // *wdh* 050416 fn[nab][grid](I1,I2,I3,V)=e(c,I1,I2,I3,V,tp);
                  // fn[nab][grid](I1,I2,I3,V)=e(c,I1,I2,I3,V,tp);
                  // display(fn[nab][grid],"fn[nab][grid] before assign V on ghost",parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
                                    e.gd(fn[nab][grid],0,0,0,0,I1,I2,I3,V,tp);
                  // display(fn[nab][grid],"fn[nab][grid] after assign V on ghost",parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
                  	    }
                  	    else
                    	      fn[nab][grid](I1,I2,I3,V)=gf[mOld].u[grid](I1,I2,I3,V); // ***** fix this ****
                	  }
              	}
    	// set back to gridIndexRange to avoid re-doing corners: *** is this ok for 3D ???
              	Iv[axis]=Range(gridIndexRange(0,axis),gridIndexRange(1,axis));
                    }
                }
            }
        } // end if( parameters.dbase.get< >("orderOfAccuracyInSpace")==4 )
        dtb=dt0;    // delta t to go from ub to ua
        dtp[0]=dt0;
        dtp[1]=dt0;
        dtp[2]=dt0;
        dtp[3]=dt0;
        dtp[4]=dt0;
    //       if( debug() & 8 )
    //       {
    //         fprintf(parameters.dbase.get<FILE* >("debugFile")," advance Adams PC: ut at t0=%e\n",t0);
    //         outputSolution( fn[nab0],t0 );
    //         fprintf(parameters.dbase.get<FILE* >("debugFile")," advance Adams PC: Errors in ut at t0=%e\n",t0);
    //         determineErrors( fn[1],t0-dt0 );
    //       }
        init=false;

    }
    else
    {
        if( dtb!=dt0 )
        {
            if( debug() & 2 ) printF("advanceImplicitMultiStep dtb!=dt0 : re-initialize\n");
            parameters.dbase.get<int >("initializeImplicitTimeStepping")=true;
        }
        else
        {
            if( debug() & 2 )
                printF("advanceImplicitMultiStep dtb==dt0 : do not re-initialize\n");
        }
    }

  // * ----
//   GridFunction gfl(&parameters);  // **** to hold current guess for nonlinear problems
//   gfl.updateToMatchGrid(gf[mCur].cg);
//   gfl.u= gf[mCur].u;
  // * ----

    for( int mst=1; mst<=numberOfSubSteps; mst++ )
    {
        parameters.dbase.get<int >("globalStepNumber")++;
        
        realCompositeGridFunction & ua = fn[nab0];   // pointer to du/dt
        realCompositeGridFunction & ub = fn[nab1];   // pointer to du(t-dt)/dt or du/dt(t+dt)
        
        real ab1,ab2;
        if( predictorOrder==1 )
        { // first order predictor
            ab1=dt0;
            ab2=0.;
        }
        else 
        { // 2nd -order predictor
            ab1= dt0*(1.+dt0/(2.*dtb));  // for AB predictor
            ab2= -dt0*dt0/(2.*dtb);
        }

        const real am1=.5*dt0;                  // for AM corrector
        const real am2=.5*dt0;


        for( int correction=0; correction<=numberOfCorrections; correction++ )
        {
            if( correction>1  && debug() & 1 )
                printF("ims: correction=%i\n",correction);
            
      // Predictor-Corrector. First time predict, subsequent times correct.
      //
      //  correction==0 :
      //       ---Adams-Bashforth Predictor
      //           u(*) <- u(t) + ab1*du/dt +ab2*du(t-dtb)/dt
      //  i.e.     gf[1]<- gf[mCur]+ ab1*ua    +ab2*ub
      //      
      //            The constants ab1 and ab2 are
      //                 ab1 = dt*( 1.+dt/(2*dtb) )   = (3/2)*dt if dtb=dt
      //                 ab2 = -dt*(  dt/(2*dtb) )    =-(1/2)*dt if dtb=dt
      //            Determined by extrapolation to time t+dt/2 from the
      //            times of ua and ub
      //
      // correction>0 :
      //       ---Adams Moulton Corrector
      //          u(t+dt) <- u(t) + dt* ( (1/2) du(*)/dt + (1/2) du(t)/dt )
      //          gf[mNew]  gf[mCur]              ub               ua

      // We only need to compute the "explicit" part of the implicit terms once for correction==0: 
      // These values are stored in utImplicit 
            implicitOption =correction==0 ? Parameters::computeImplicitTermsSeparately : 
                                                                            Parameters::doNotComputeImplicitTerms;

            if( correction==0 )
            {
	// ------------------------------------------------------
	// ----------------- Moving Grids -----------------------
	// ------------------------------------------------------

      	bool useNew=false;
      	bool gridWasAdapted=false;

            if( movingGridProblem() )
            {
                checkArrays(" adamsPC : before move grids"); 
                if( debug() & 8 )
                    printf(" adamsPC: before moveTheGridsMacro: t0=%9.3e, gf[mNew].t=%9.3e, gf[mNew].gridVelocityTime=%9.3e\n",
                     	   t0,gf[mNew].t,gf[mNew].gridVelocityTime);
        // generate gf[mNew] from gf[mCur] (compute grid velocity on gf[mCur] and gf[mNew]
                moveGrids( t0,t0,t0+dt0,dt0,gf[mCur],gf[mCur],gf[mNew] ); 
                checkArrays(" adamsPC : after move grids");
                if( parameters.isAdaptiveGridProblem() )
                {
          // both moving and AMR 
                    parameters.dbase.get<Ogen* >("gridGenerator")->updateRefinement(gf[mNew].cg);
                }
                if( debug() & 16 )
                {
                    if( twilightZoneFlow() )
                    {
                        fprintf(parameters.dbase.get<FILE* >("debugFile"),"\n ---> adamsPC : Errors in u after moveGrids t=%e  \n",gf[mCur].t);
                        determineErrors( gf[mCur] );
                    }
                }
                real cpu0=getCPU();
                gf[mNew].cg.rcData->interpolant->updateToMatchGrid( gf[mNew].cg );  
                parameters.dbase.get<RealArray>("timing")(Parameters::timeForUpdateInterpolant)+=getCPU()-cpu0;
                cpu0=getCPU();
                gf[mNew].u.getOperators()->updateToMatchGrid(gf[mNew].cg); 
                parameters.dbase.get<RealArray>("timing")(Parameters::timeForUpdateOperators)+=getCPU()-cpu0;
                if( debug() & 4 ) printf("adamsPC : step: update gf[mNew] for moving grids...\n");
                if( debug() & 16 )
                {
                    if( twilightZoneFlow() )
                    {
                        fprintf(parameters.dbase.get<FILE* >("debugFile"),"\n ---> adamsPC: Errors in u before updateForMovingGrids t=%e  \n",gf[mCur].t);
                        fprintf(parameters.dbase.get<FILE* >("debugFile"),"*** mCur=%i mNew=%i numberOfGridFunctions=%i *** \n",
                        	      mCur,mNew,numberOfGridFunctions);
                        determineErrors( gf[mCur] );
                    }
                }
                updateForMovingGrids(gf[mNew]);
        // ****      gf[mNew].u.updateToMatchGrid( gf[mNew].cg );  
                if( debug() & 16 )
                {
                    if( twilightZoneFlow() )
                    {
                        fprintf(parameters.dbase.get<FILE* >("debugFile"),"\n ---> adamsPC: Errors in u after updateForMovingGrids t=%e  \n",gf[mCur].t);
                        determineErrors( gf[mCur] );
                    }
                }
        // get values for exposed points on gf[mCur]
                if( parameters.useConservativeVariables() )
                    gf[mCur].primitiveToConservative();  // *wdh* 010318
                cpu0=getCPU();
                if( useNewExposedPoints )
                {
                    if( Parameters::checkForFloatingPointErrors )
                        checkSolution(gf[mCur].u,"Before interp exposed points",true);
                    if( debug() & 16 )
                    {
                        if( twilightZoneFlow() )
                        {
                  	fprintf(parameters.dbase.get<FILE* >("debugFile"),"\n ---> adamsPC: Errors in u BEFORE interp exposed t=%e  \n",gf[mCur].t);
                  	determineErrors( gf[mCur] );
                        }
                    }
          // parameters.dbase.get<int >("stencilWidthForExposedPoints")=5; // ****************** TEMP *****
                    ExposedPoints exposedPoints;
                    exposedPoints.setAssumeInterpolationNeighboursAreAssigned(parameters.dbase.get<int >("extrapolateInterpolationNeighbours"));
                    exposedPoints.initialize(gf[mCur].cg,gf[mNew].cg,parameters.dbase.get<int >("stencilWidthForExposedPoints"));
                    exposedPoints.interpolate(gf[mCur].u,(twilightZoneFlow() ? parameters.dbase.get<OGFunction* >("exactSolution") : NULL),t0);
                    if( debug() & 16 )
                    {
                        if( twilightZoneFlow() )
                        {
                  	fprintf(parameters.dbase.get<FILE* >("debugFile"),"\n ---> adamsPC: Errors in u AFTER interp exposed t=%e  \n",gf[mCur].t);
                  	determineErrors( gf[mCur] );
                        }
                    }
      //          // We may need values of ub at interpolation points, or exposed? ***
      //          if( interpolateUt ) // &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
      //  	{
      //  	  // *** no exposedPoints.interpolate(ub,NULL,t0);  // *wdh* 040827
      //  	}
                    if( useNew )
                    {
            // fixup du/dt(t-dt)
                        if( debug() & 4 )
                            printf(" ---- fixup exposed points of du/dt(t-dt) ----- \n");
                        ExposedPoints exposedPoints;
            //            
            // exposedPoints.setExposedPointType(ExposedPoints::exposedDiscretization);
                        exposedPoints.setAssumeInterpolationNeighboursAreAssigned(parameters.dbase.get<int >("extrapolateInterpolationNeighbours"));
                        exposedPoints.initialize(gf[mOld].cg,gf[mNew].cg,parameters.dbase.get<int >("stencilWidthForExposedPoints"));
                        exposedPoints.interpolate(gf[mOld].u,(twilightZoneFlow() ? parameters.dbase.get<OGFunction* >("exactSolution") : NULL),gf[mOld].t);
            // For now recompute du/dt(t-dt) using the mask values from cg(t+dt)
                        for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
                        {
                  	if( gridWasAdapted || exposedPoints.getNumberOfExposedPoints(grid)>0 )
                  	{
                                if( debug() & 2 )
                      	    printf(" ---- adamsPC: recompute du/dt(t-dt) for grid=%i t-dt = %9.3e  (%i exposed)-----\n",grid,gf[mOld].t,
                           		   exposedPoints.getNumberOfExposedPoints(grid));
      	  // This is only necesssary if there are exposed points on this grid
                    	  rparam[0]=gf[mOld].t;
                    	  rparam[1]=gf[mOld].t;
                    	  rparam[2]=gf[mCur].t; // tImplicit
                    	  iparam[0]=grid;
                    	  iparam[1]=gf[mOld].cg.refinementLevelNumber(grid);
                    	  iparam[2]=numberOfStepsTaken;
      // 	  mappedGridSolver[grid]->getUt(gf[mOld].u[grid],gf[mOld].getGridVelocity(grid),ub[grid],iparam,rparam,
      // 					uti[grid],&gf[mNew].cg[grid]);
                    	  getUt(gf[mOld].u[grid],gf[mOld].getGridVelocity(grid),ub[grid],iparam,rparam,
                        		uti[grid],&gf[mNew].cg[grid]);
                  	}
                        }
                        if( debug() & 4 )
                        {	
                  	if( twilightZoneFlow() )
                  	{
                    	  fprintf(parameters.dbase.get<FILE* >("debugFile")," ***adamsPC: gf[mOld] after interp exposed, gf[mOld].t=%e",
                                                gf[mOld].t);
                    	  for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
                    	  {
                      	    display(gf[mOld].u[grid],sPrintF("\n ****gf[mOld].u[grid=%i]",grid),parameters.dbase.get<FILE* >("debugFile"),"%7.1e ");
                    	  }
                    	  determineErrors( gf[mOld] );
                    	  fprintf(parameters.dbase.get<FILE* >("debugFile")," ***adamsPC: du/dt(t-dt)  after interp exposed, gf[mOld].t=%e",
                                                gf[mOld].t);
                    	  for( grid=0; grid<gf[mOld].cg.numberOfComponentGrids(); grid++ )
                    	  {
                      	    display(ub[grid],sPrintF("\n ****ub[grid=%i]: du/dt(t-dt)",grid),parameters.dbase.get<FILE* >("debugFile"),"%7.1e ");
                    	  }
                    	  determineErrors( ub,gf[mOld].gridVelocity, gf[mOld].t, 1, error );
                  	}
                        }
                    }
                    if( Parameters::checkForFloatingPointErrors )
                        checkSolution(gf[mCur].u,"adamsPC: After interp exposed points",true);
                }
                else
                {
                    interpolateExposedPoints(gf[mCur].cg,gf[mNew].cg,gf[mCur].u, 
                                   			     (twilightZoneFlow() ? parameters.dbase.get<OGFunction* >("exactSolution") : NULL),t0,
                                   			     false,Overture::nullIntArray(),Overture::nullIntegerDistributedArray(),
                                   			     parameters.dbase.get<int >("stencilWidthForExposedPoints") ); 
                }
                if( twilightZoneFlow() && false ) // **** wdh **** 
                {
          // for testing ***
                    OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
                    int grid=0; 
                    MappedGrid & c = gf[mNew].cg[grid];
                    getIndex(c.dimension(),I1,I2,I3);
                    ub[grid](I1,I2,I3,N)=e.t(c,I1,I2,I3,N,t0-dt0); 
                }
                parameters.dbase.get<RealArray>("timing")(Parameters::timeForInterpolateExposedPoints)+=getCPU()-cpu0;
        // compute dudt now -- after exposed points have been computed!
                checkArrays(" adamsPC :  after moving grids update"); 
                if( debug() & 16 )
                {
                    if( twilightZoneFlow() )
                    {
                        fprintf(parameters.dbase.get<FILE* >("debugFile"),"\n ---> adamsPC: Errors in u after move grids t=%e  \n",gf[mCur].t);
                        determineErrors( gf[mCur] );
                    }
                }
                if( debug() & 16 )
                    printf(" adamsPC: AFTER moveTheGridsMacro: t0=%9.3e, gf[mNew].t=%9.3e, gf[mNew].gridVelocityTime=%9.3e\n",
                     	   t0,gf[mNew].t,gf[mNew].gridVelocityTime);
            }
	// we need to rebuild the implicit time stepping matrix.
        	if( movingGridProblem() )
                    parameters.dbase.get<int >("initializeImplicitTimeStepping")=true;


            }

      // Optionally refactor the matrix : if parameters.dbase.get<int >("globalStepNumber") % refactorFrequency == 0 
      // (We need to do this after the grids have moved but before dudt is evaluated (for nonlinear problems)
            if( correction==0 && (parameters.dbase.get<int >("initializeImplicitTimeStepping") || parameters.dbase.get<int >("globalStepNumber")>0) )
      	formMatrixForImplicitSolve(dt0,gf[mNew], gf[mCur] );

            const int maba = correction==0 ? mCur : mNew;
            const int naba = correction==0 ? nab0 : nab1;
      // fn[nab0] <- du/dt(t0)  or fn[nab1] <- du/dt(t+dt0)
            for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
            {
      	rparam[0]=gf[maba].t;
      	rparam[1]=gf[maba].t; // tforce
      	rparam[2]=gf[maba].t+dt0; // tImplicit
      	iparam[0]=grid;
      	iparam[1]=gf[maba].cg.refinementLevelNumber(grid);
      	iparam[2]=numberOfStepsTaken;

      	getUt(gf[maba].u[grid],gf[maba].getGridVelocity(grid),
            	      fn[naba][grid],iparam,rparam,uti[grid],&gf[mNew].cg[grid]);
            }

            addArtificialDissipation(gf[maba].u,dt0);	// add "implicit" dissipation to u 

            if( debug() & 16 || debug() & 64 )
            {
      	for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
      	{
        	  display(ua[grid],"\n ****ua: du/dt(t)",debugFile);
        	  display(ub[grid],"\n ****ub: du/dt(t-dt)",debugFile);
      	}
            }
            if( debug() & 16 )
            {
      	aString label = sPrintF(" ImplicitMS: errors in ut (ua) at t=%e, correction=%i \n",t0,correction);
      	determineErrors( ua,gf[mCur].gridVelocity, t0, 1, error,label );
                real tub = correction==0 ? t0-dtb : t0+dt0;
      	label = sPrintF(" ImplicitMS: errors in ut (ub) at t=%e, correction=%i \n",tub,correction);
      	determineErrors( ub,gf[mNew].gridVelocity, tub, 1, error,label );
            }

      //  --------------------------------------------------------
      //  --- Assign the explicit or implicit time-step update ---
      //  --------------------------------------------------------
            for( grid=0; grid<gf[mNew].cg.numberOfComponentGrids(); grid++ )
            {
        // const real dti = dt0; 
                const real dti = (1.-parameters.dbase.get<real >("implicitFactor"))*dt0;

      	if( correction==0 )
      	{ // u1 <- u0 + ab1*ua0 + ab2*ub0 [ + dti*utImplicit ]  (add last terms for implicit grids)
                #ifdef USE_PPP
                    RealArray u1;  getLocalArrayWithGhostBoundaries(gf[mNew].u[grid],u1);
                    RealArray u0;  getLocalArrayWithGhostBoundaries(gf[mCur].u[grid],u0);
                    RealArray ua0; getLocalArrayWithGhostBoundaries(ua[grid],ua0);
                    RealArray ub0; getLocalArrayWithGhostBoundaries(ub[grid],ub0);
                    RealArray utImplicit; getLocalArrayWithGhostBoundaries(uti[grid],utImplicit);
                    const intSerialArray & mask1 = gf[mNew].cg[grid].mask().getLocalArray();
                #else
                    RealDistributedArray & u1 = gf[mNew].u[grid];
                    RealDistributedArray & u0 = gf[mCur].u[grid];
                    RealDistributedArray & ua0 = ua[grid];
                    RealDistributedArray & ub0 = ub[grid];
                    RealDistributedArray & utImplicit = uti[grid];
                    const intSerialArray & mask1 = gf[mNew].cg[grid].mask(); 
                #endif
                    getIndex(gf[mNew].cg[grid].extendedIndexRange(),I1,I2,I3);
                    int n1a,n1b,n2a,n2b,n3a,n3b;
                    bool ok = ParallelUtility::getLocalArrayBounds(gf[mCur].u[grid],u0,I1,I2,I3,n1a,n1b,n2a,n2b,n3a,n3b);
                    if( !ok ) continue;
          // const intArray & mask1 = gf[mNew].cg[grid].mask();
                    int ierr=0;
                    const int maskOption=0; // assign pts where mask>0
                    int ipar[]={0,maskOption,n1a,n1b,n2a,n2b,n3a,n3b,N.getBase(),N.getBound()}; //
                    real rpar[5]={0.,0.,0.,0.,0.};
                    real *ut1p, *ut2p, *ut3p, *ut4p;
                    if( parameters.getGridIsImplicit(grid) )
                        ipar[0]=3;  // add three extra "ut" terms if grid is advanced implicitly
                    else    
                        ipar[0]=2;  // add two extra "ut" terms
                    rpar[0]=ab1; rpar[1]=ab2; rpar[2]=dti;
                    ut1p=ua0.getDataPointer();
                    ut2p=ub0.getDataPointer();
                    ut3p=utImplicit.getDataPointer();
                    ut4p=ut3p;
                    updateOpt(u0.getBase(0),u0.getBound(0),u0.getBase(1),u0.getBound(1),
                          	    u0.getBase(2),u0.getBound(2),u0.getBase(3),u0.getBound(3),
                          	    *mask1.getDataPointer(),  
                          	    *u0.getDataPointer(),*u1.getDataPointer(), 
                          	    *ut1p, *ut2p, *ut3p, *ut4p, 
                          	    ipar[0], rpar[0], ierr );
      	}
      	else
      	{ // u1 <- u0 + am1*ub0 + am2*ua0 [ + dti*utImplicit ] (add last terms for implicit grids)
                #ifdef USE_PPP
                    RealArray u1;  getLocalArrayWithGhostBoundaries(gf[mNew].u[grid],u1);
                    RealArray u0;  getLocalArrayWithGhostBoundaries(gf[mCur].u[grid],u0);
                    RealArray ua0; getLocalArrayWithGhostBoundaries(ua[grid],ua0);
                    RealArray ub0; getLocalArrayWithGhostBoundaries(ub[grid],ub0);
                    RealArray utImplicit; getLocalArrayWithGhostBoundaries(uti[grid],utImplicit);
                    const intSerialArray & mask1 = gf[mNew].cg[grid].mask().getLocalArray();
                #else
                    RealDistributedArray & u1 = gf[mNew].u[grid];
                    RealDistributedArray & u0 = gf[mCur].u[grid];
                    RealDistributedArray & ua0 = ua[grid];
                    RealDistributedArray & ub0 = ub[grid];
                    RealDistributedArray & utImplicit = uti[grid];
                    const intSerialArray & mask1 = gf[mNew].cg[grid].mask(); 
                #endif
                    getIndex(gf[mNew].cg[grid].extendedIndexRange(),I1,I2,I3);
                    int n1a,n1b,n2a,n2b,n3a,n3b;
                    bool ok = ParallelUtility::getLocalArrayBounds(gf[mCur].u[grid],u0,I1,I2,I3,n1a,n1b,n2a,n2b,n3a,n3b);
                    if( !ok ) continue;
          // const intArray & mask1 = gf[mNew].cg[grid].mask();
                    int ierr=0;
                    const int maskOption=0; // assign pts where mask>0
                    int ipar[]={0,maskOption,n1a,n1b,n2a,n2b,n3a,n3b,N.getBase(),N.getBound()}; //
                    real rpar[5]={0.,0.,0.,0.,0.};
                    real *ut1p, *ut2p, *ut3p, *ut4p;
                    if( parameters.getGridIsImplicit(grid) )
                        ipar[0]=3;  // add three extra "ut" terms if grid is advanced implicitly
                    else    
                        ipar[0]=2;  // add two extra "ut" terms
                    rpar[0]=am1; rpar[1]=am2; rpar[2]=dti;
                    ut1p=ub0.getDataPointer();
                    ut2p=ua0.getDataPointer();
                    ut3p=utImplicit.getDataPointer();
                    ut4p=ut3p;
                    updateOpt(u0.getBase(0),u0.getBound(0),u0.getBase(1),u0.getBound(1),
                          	    u0.getBase(2),u0.getBound(2),u0.getBase(3),u0.getBound(3),
                          	    *mask1.getDataPointer(),  
                          	    *u0.getDataPointer(),*u1.getDataPointer(), 
                          	    *ut1p, *ut2p, *ut3p, *ut4p, 
                          	    ipar[0], rpar[0], ierr );
      	}
            }

            if( correction==0 )
            {
        // printF(" +++ ims: gf[mNew].t=%9.3e --> change to t0+dt0=%9.3e +++\n",gf[mNew].t,t0+dt0);
                gf[mNew].t=t0+dt0;  // gf[mNew] now lives at this time
            }
            

      // *** assign boundary conditions for the implicit method 
            applyBoundaryConditionsForImplicitTimeStepping( gf[mNew] ); // ***** gf[mNew].gridVelocity must be correct here
        
            if( debug() & 4 )
            {
      	aString label = sPrintF(" ***ImplicitMS: RHS Before implicitSolve t=%e\n",gf[mNew].t);
      	if( twilightZoneFlow() )
      	{
        	  gf[mNew].u.display(label,debugFile,"%8.5f ");
      	}
                label = sPrintF(" ***ImplicitMS: Errors in rhs gf before implicitSolve t=%e\n",gf[mNew].t);
      	determineErrors( gf[mNew],label );
            }

      // **** fix this *** we could refactor for each correction here !
//       if( mst>1 || correction>0 )
//       {
//         // Optionally refactor the matrix : if parameters.dbase.get<int >("globalStepNumber") % refactorFrequency == 0 
// 	formMatrixForImplicitSolve(dt0,gf[mNew], gf[mCur] );
//       }
            

      // ------------------------------------
      // --- Solve the implicit equations ---
      // ------------------------------------
  
            implicitSolve( dt0,gf[mNew], gf[mCur] );  // gf[mNew]=RHS  gf[mCur]=used for initial guess and linearization

      // * ---
//       implicitSolve( dt0,gf[mNew], gfl );  // gf[mNew]=RHS  gf[mCur]=used for initial guess and linearization
//       gfl.u=gf[mNew].u;


            if( false )
            {
                realCompositeGridFunction diff;
      	diff=gf[mNew].u-gf[mCur].u;
                Range & Rt = parameters.dbase.get<Range >("Rt");
      	printF(" After implicit solve: max-diff(u-uGuess) =");
      	for( int n=Rt.getBase(); n<=Rt.getBound(); n++ )
      	{
                    int maskOption=0, extra=1;
        	  real maxDiff = maxNorm(diff, n, maskOption, extra );
                    real mxd = 0;
        	  for( int grid=0; grid<gf[mNew].cg.numberOfComponentGrids(); grid++ )
          	    mxd = max( mxd, max(fabs(gf[mNew].u[grid]-gf[mCur].u[grid])));
        	  printF(" n=%i : %8.2e (all=%8.2e), ",n,maxDiff,mxd);
      	}
      	printF("\n");
            }

            if( debug() & 4 )
            {
      	if( twilightZoneFlow() )
      	{
        	  gf[mNew].u.display(sPrintF("ImplicitMS: gf[mNew].u after implicitSolve but BEFORE BC's (t=%8.2e)",
                             				     gf[mNew].t),debugFile,"%8.5f ");
      	}
      	aString label = sPrintF(" ***ImplicitMS: after implicitSolve but BEFORE BC's, t=%e, correction=%i\n",gf[mNew].t,correction);
      	determineErrors( gf[mNew],label );
            }

      // apply explicit BC's  --- > really only have to apply to implicit grids I think?
            applyBoundaryConditions(gf[mNew]);   // ***** gf[mNew].gridVelocity must be correct here!


            updateStateVariables( gf[mNew],1 );  

            if( debug() & 4 )
            {
      	if( twilightZoneFlow() )
      	{
        	  gf[mNew].u.display(sPrintF("ImplicitMS: gf[mNew].u after implicitSolve and BC's (t=%8.2e)",
                             				     gf[mNew].t),debugFile,"%8.5f ");
      	}
      	aString label = sPrintF(" ***ImplicitMS: after implicitSolve and BC's, t=%e, correction=%i\n",gf[mNew].t,correction);
      	determineErrors( gf[mNew],label );
            }

      // extrapolate p in time as an initial guess for iterative solvers
            if( correction==0 && 
        	  ( parameters.dbase.has_key("extrapolatePoissonSolveInTime") && 
          	    poisson->isSolverIterative() ) )
            {
      	const int & pc = parameters.dbase.get<int >("pc");
      	assert( pc>= 0 );
      	for( grid=0; grid<gf[mCur].cg.numberOfComponentGrids(); grid++ )
      	{
	  // *wdh* 050424 getIndex(gf[mCur].cg[grid].gridIndexRange(),I1,I2,I3);  // **** why not include ghost pts??
        	  getIndex(gf[mCur].cg[grid].dimension(),I1,I2,I3);  // **** why not include ghost pts??
	  // note that initially gf[mNew](.,.,.,pc) = p(t-dt)
                    #ifdef USE_PPP
          	    realSerialArray uNew; getLocalArrayWithGhostBoundaries(gf[mNew].u[grid],uNew);
          	    realSerialArray uCur; getLocalArrayWithGhostBoundaries(gf[mCur].u[grid],uCur);
          	    realSerialArray uOld; getLocalArrayWithGhostBoundaries(gf[mOld].u[grid],uOld);
                        bool ok = ParallelUtility::getLocalArrayBounds(gf[mNew].u[grid],uNew,I1,I2,I3); 
                        if( !ok ) continue;
        	  #else
                        const realSerialArray & uNew=gf[mNew].u[grid];
                        const realSerialArray & uCur=gf[mCur].u[grid];
                        const realSerialArray & uOld=gf[mOld].u[grid];
                    #endif	    
        	  
        	  uNew(I1,I2,I3,pc)=(ab1/dt0)*uCur(I1,I2,I3,pc)+(ab2/dt0)*uOld(I1,I2,I3,pc);
	  // gf[mNew].u[grid](I1,I2,I3,pc)=(ab1/dt0)*gf[mCur].u[grid](I1,I2,I3,pc)+(ab2/dt0)*gf[mOld].u[grid](I1,I2,I3,pc);
      	}
            }
      // e.g. for variable density, update p eqn here     
            bool updateSolutionDependentEquations = correction==0;  
            solveForTimeIndependentVariables( gf[mNew],updateSolutionDependentEquations ); 

            if( debug() & 8 )
            {
      	aString label =sPrintF(" ImplicitMS: Errors after pressure solve, t0+dt0: t0=%e, dt0=%e  \n",t0,dt0);
      	determineErrors( gf[mNew],label );
            }

      // correct for forces on moving bodies  *wdh* 040913
            if( movingGridProblem() )  // may only need to do this for correction<numberOfCorrections
      	correctMovingGrids( t0,t0+dt0,gf[mCur],gf[mNew] ); 

      // * ----

//       if( correction>=2 )
//       {
// 	parameters.dbase.get<int >("globalStepNumber")++;
//       }

//       if( parameters.dbase.get<int >("globalStepNumber") % 10 == 0 )
//       {
// 	if( parameters.dbase.get<int>("useNewImplicitMethod")==1 )
// 	{
// 	  realCompositeGridFunction & residual = uti;  // save residual here -- check this 
// 	  getResidual( t0+dt0 ,dt0,gf[mNew],residual );
// 	  saveSequenceInfo(t0+dt0,residual);
// 	}
            
//       }


        } // end corrections
        
    // permute (mab0,mab1,mab2) 
        mab0 = (mab0-1 + numberOfGridFunctions) % numberOfGridFunctions;
        mab1 = (mab1-1 + numberOfGridFunctions) % numberOfGridFunctions;
    // mab2 is always 1 "ahead" of mab0 
        mab2 = (mab0-1 + numberOfGridFunctions) % numberOfGridFunctions;
    // XXX     ^ should that be a 2 ?? kkc 060301  -- this is ok *wdh* 071121
        mNew=mab2;
        mCur=mab0;
        mOld=mab1;

        if( parameters.dbase.get<int >("orderOfPredictorCorrector")==2 )
        {
            nab0 = (nab0+1) % 2;
            nab1 = (nab1+1) % 2;
        }
        else
        {
            nab0 = (nab0-1 +4) % 4;
            nab1 = (nab1-1 +4) % 4;
            nab2 = (nab2-1 +4) % 4;
            nab3 = (nab3-1 +4) % 4;
        }

        ndt0=(ndt0-1 +5)%5;  // for dtp[]
//      // switch mab0 <-> mab1
//      mab0 = (mab0+1) % 2;
//      mab1 = (mab1+1) % 2;
            
        dtb=dt0;
        t0+=dt0;

        if( parameters.dbase.get<int >("globalStepNumber") % 10 == 0 )
        {
      // residual = u.t = fn[nab0]+uti
      // saveSequenceInfo(t0,evaluate(fn[nab0]+uti));   // 070704 *wdh* turn this off for now -- this is not correct in general 

      // new way:
            if( parameters.dbase.get<int>("useNewImplicitMethod")==1 )
            {
      	realCompositeGridFunction & residual = uti;  // save residual here -- check this 
                getResidual( t0,dt0,gf[mab0],residual );
      	saveSequenceInfo(t0,residual);
            }
            
        }
        
        output( gf[mab0],parameters.dbase.get<int >("globalStepNumber")+1 ); // output to files, user defined output

        const int zeroUnusedPointsAfterThisManySteps=20;
        if( (mst==numberOfSubSteps || (mst % zeroUnusedPointsAfterThisManySteps)==0) &&  // mst starts at 1
      	parameters.dbase.get<int >("extrapolateInterpolationNeighbours")==0 )
        {
      // *note* we cannot fixup unused if we extrapolate interp. neighbours since these values will be zeroed out!
      // (esp. important for viscoPlastic model -- linearized solution becomes corrupted)

            if( debug() & 2 ) printF(" ************** ims.bC fixupUnusedPoints ************\n");
            
      // zero out unused points to keep them from getting too big ** is this needed?? ****
            for( int m=0; m<=1; m++ )
            {
	// ** gf[m].u.zeroUnusedPoints(coeff);
      	fixupUnusedPoints(gf[m].u);
            }
        }
        
    } // end  substeps

  // update the current solution:  
    current = mab0;
    
  // tm(2)+=getCPU()-time0;
}


//\begin{>>CompositeGridSolverInclude.tex}{\subsection{applyBoundaryConditionsForImplicitTimeStepping}} 
int DomainSolver::
applyBoundaryConditionsForImplicitTimeStepping(GridFunction & cgf )
// ======================================================================================
//  /Description:
//     On implicit grids, apply boundary conditions to the rhs side grid function used in the implicit solve;
//  on explicit grids apply the normal explicit boundary conditions. 
// /cgf (input) : use this grid function as the right-hand-side. cgf.t should be the time corresponding to the
//      next time step.
//\end{CompositeGridSolverInclude.tex}  
// ==========================================================================================
{
    int grid;

    for( grid=0; grid<cgf.cg.numberOfComponentGrids(); grid++ )
    {
    // clean this up 
        if( parameters.getGridIsImplicit(grid) )
        {
            applyBoundaryConditionsForImplicitTimeStepping(cgf.u[grid],
                                         						     gf[current].u[grid],  // -- fix this -- should be uL
                                         						     cgf.getGridVelocity(grid),
                                         						     cgf.t,
                                         						     parameters.dbase.get<int >("scalarSystemForImplicitTimeStepping"),grid );
        }
        else
        { // apply explicit BC's **** could be trouble if these require interpolation points ??   **********
            applyBoundaryConditions(cgf.t,cgf.u[grid],cgf.getGridVelocity(grid),grid);
        }


    }
    
    return 0;
}


#include "Integrate.h"


bool DomainSolver::
isImplicitMatrixSingular( realCompositeGridFunction &uL )
{
    return false;
}

int  DomainSolver::
addConstraintEquation( Parameters &parameters, Oges& solver, 
                   		       realCompositeGridFunction &coeff, 
                   		       realCompositeGridFunction &ucur, 
                   		       realCompositeGridFunction &rhs, const int &numberOfComponents) 
{
    printF("DomainSolver::addConstraintEquation should never be called!");
    Overture::abort("error");
    return 0;
}

//\begin{>>CompositeGridSolverInclude.tex}{\subsection{implicitSolve}} 
void DomainSolver::
formMatrixForImplicitSolve(const real & dt0,
                     			   GridFunction & cgf1,
                     			   GridFunction & cgf0 )
// ==========================================================================================
// /Description: This function was once part of implicitSolve.  It was
// broken out to allow the construction of the matrix independently of
// the actual solve.  Basically all the work is done to initialize the
// implicit time stepping.  The implicit method can be optionally used
// on only some grids. To implement this approach we simply create a
// sparse matrix that is just the identity matrix on grids that are
// advanced explicitly but equal to the standard implicit matrix on
// grids that are advance implicitly: 
//  \begin{verbatim} 
//  I - \nu \alpha \dt \Delta on implicit grids 
//  I on explicit grids 
//  \end{verbatim} 
// If the form of the boundary conditions for the different components of
// $\uv$ are the same then we can build a single scalar matrix that
// can be used to advance each component, one after the other. If the
// boundary conditions are not of the same form then we build a matrix
// for a system of equations for the velocity components $(u,v,w)$.
//
// Note that originally cgf1 from implicitSolve was used to get the time,
// grid, and operators.  We are now using whatever is passed in as "u" to
// this function.  The operators should be the same (?) and the time is
// used in the debug output.  What about the grid though? It can change 
// due to AMR (used with implicit?) as well as from the grid velocity.
// /dt0 (input) : time step used to build the implicit matrix.
// /cgf1 (input) : holds the RHS 
// /cgf0 (input) : holds the current state of the solution (used for linearization)
//\end{CompositeGridSolverInclude.tex}  
// ==========================================================================================
{
    printf("DomainSolver::formMatrixForImplicitSolve:ERROR: base class function called. This function should be over-ridden\n");
    Overture::abort("error");
}

int
DomainSolver::
setOgesBoundaryConditions( GridFunction &cgf, IntegerArray & boundaryConditions, RealArray &boundaryConditionData,
                                                      const int imp )
// ===================================================================================================================
// /Description:
//   Assign the boundaryCondition data for passing to Oges (predfined equations) when it builds the implicit system.
//
// This function is called by DomainSolver::formMatrixForImplicitSolve
// 
//  /cgf (input) : A grid function holding the current grid.
//  /boundaryConditions (output) : boundary conditions for Oges
//  /boundaryConditionData (output) : boundary condition data for Oges 
//  /imp (input) : the number of the implicit system being solved
// 
// ====================================================================================================================
{
    CompositeGrid & cg = cgf.cg;

    printf("DomainSolver::setOgesBoundaryConditions:ERROR: base class function called. This function should be over-ridden\n");
    Overture::abort("error");

    return 0;
}
