#include "Parameters.h"
#include "FlowSolutions.h"
#include "GenericGraphicsInterface.h"

// Define new user-defined options here:
namespace
{
enum UserDefinedKnownSolutionEnum
{
  unknownSolution=0,
  specifiedPiston=1,
  forcedPiston=2
};
}


int Parameters::
getUserDefinedKnownSolution(real t, MappedGrid & mg, realArray & ua, 
			    const Index & I1, const Index &I2, const Index &I3 )
// ==========================================================================================
//  /Description:
//     Evaluate a user defined known solution.
// ==========================================================================================
{
  if( ! dbase.get<DataBase >("modelData").has_key("userDefinedKnownSolutionData") )
  {
    printf("getUserDefinedKnownSolution:ERROR: sub-directory `userDefinedKnownSolutionData' not found!\n");
    Overture::abort("error");
  }
  DataBase & db =  dbase.get<DataBase >("modelData").get<DataBase>("userDefinedKnownSolutionData");

  UserDefinedKnownSolutionEnum & userKnownSolution = db.get<UserDefinedKnownSolutionEnum>("userKnownSolution");
  real *rpar = db.get<real[10]>("rpar");

  if( userKnownSolution==specifiedPiston )
  {

//   const int rc=ipar[0];
//   const int uc=ipar[1];
//   const int vc=ipar[2];
//   const int tc=ipar[3];
//   const int pc=ipar[4];  // fill in if >=0 
//   const int sc=ipar[5];  // fill in if >=0 
  
//   const real gamma   =rpar[0];
//   const real Rg      =rpar[1];
//   const real rho0    =rpar[2];
//   const real u0      =rpar[3];
//   const real v0      =rpar[4];
//   const real p0      =rpar[5];
//   const real t       =rpar[6];
//   const real ag      =rpar[7];  // g(t)= ap*t^pg  (piston with specfied motion)
//   const real pg      =rpar[8];
//   const real mass    =rpar[9];  // for hydrodynamically forced piston
//   const real height  =rpar[10]; // for hydrodynamically forced piston

    FlowSolutions flowSolutions;

    int ipar[]={  dbase.get<int >("rc"), dbase.get<int >("uc"), dbase.get<int >("vc"), dbase.get<int >("tc"),-1,-1 }; // 

    // do this for now -- fix this ----------------------------------------- ****************************
    const real rho0=1.4, v0=0.,  p0=1,  te0=1./rho0;
    const real a0 = sqrt( dbase.get<real >("gamma")*p0/rho0);
    const real u0 = 1.5*a0;

    real mass=1., height=1.;

    // real pg=4., ag=-.5/pg;  // ** hard code for now -- fix this ---  (define  dbase.get<real >("a") "parameter" and look it up here)
    real ag=rpar[0], pg=rpar[1];
    
    real rpar[]={ dbase.get<real >("gamma"), dbase.get<real >("Rg"), rho0,u0,v0, p0,t,ag,pg,mass,height}; //

    // evaluate the known solution for flow in an expanding channel
    flowSolutions.getFlowSolution(FlowSolutions::specifiedPistonMotion, mg,ua,ipar,rpar,I1,I2,I3);
  }
  else if( userKnownSolution==forcedPiston )
  {
    FlowSolutions flowSolutions;

    int ipar[]={  dbase.get<int >("rc"), dbase.get<int >("uc"), dbase.get<int >("vc"), dbase.get<int >("tc"),-1,-1 }; // 

    // do this for now -- fix this ----------------------------------------- ****************************
    const real rho0=1.4, v0=0., p0=1, te0=1./rho0;
    const real a0 = sqrt( dbase.get<real >("gamma")*p0/rho0);
    const real u0 = 1.5*a0;


    real pg=4., ag=-.5/pg;  // not used
    
    real mass=1., height=1.;
    mass = rpar[0], height=rpar[1];
     
    real rpar[]={ dbase.get<real >("gamma"), dbase.get<real >("Rg"), rho0, u0,v0, p0,t,ag,pg,mass,height}; //

    // evaluate the known solution for flow in an expanding channel
    flowSolutions.getFlowSolution(FlowSolutions::forcedPistonMotion, mg,ua,ipar,rpar,I1,I2,I3);
  }
  else
  {
    printf("getUserDefinedKnownSolution:ERROR: unknown value for userDefinedKnownSolution=%i\n",
      userKnownSolution);
    //kkc 070130 I think the above is what was intended      userDefinedKnownSolution);
    Overture::abort("EROR");
  }
  
  return 0;
}



//\begin{>>MovingGridsSolverInclude.tex}{\subsection{updateUserDefinedMotion}} 
int Parameters::
updateUserDefinedKnownSolution(GenericGraphicsInterface & gi)
// ==========================================================================================
// /Description: 
//   This function is called to set the user defined know solution.
// 
// /Return value: >0 : known solution was chosen, 0 : no known solution was chosen
//\end{MovingGridsSolverInclude.tex}  
// ==========================================================================================
{
  // Make  dbase.get<real >("a") sub-directory in the data-base to store variables used here
  if( ! dbase.get<DataBase >("modelData").has_key("userDefinedKnownSolutionData") )
     dbase.get<DataBase >("modelData").put<DataBase>("userDefinedKnownSolutionData");
  DataBase & db =  dbase.get<DataBase >("modelData").get<DataBase>("userDefinedKnownSolutionData");

  if( !db.has_key("userKnownSolution") )
  {
    db.put<UserDefinedKnownSolutionEnum>("userKnownSolution",unknownSolution);
    db.put<real[10]>("rpar");
  }
  UserDefinedKnownSolutionEnum & userKnownSolution = db.get<UserDefinedKnownSolutionEnum>("userKnownSolution");
  real *rpar = db.get<real[10]>("rpar");

  const aString menu[]=
    {
      "no known solution",
      "specified piston motion",
      "forced piston motion",
      "done",
      ""
    }; 
  gi.appendToTheDefaultPrompt("userDefinedKnownSolution>");
  aString answer;
  for( ;; ) 
  {

    int response=gi.getMenuItem(menu,answer,"Choose the movement");
    
    if( answer=="done" || answer=="exit" )
    {
      break;
    }
    else if( answer=="no known solution" )
    {
      userKnownSolution=unknownSolution;
    }
    else if( answer=="specified piston motion" )
    {
      userKnownSolution=specifiedPiston; 
      rpar[1]=4.; rpar[0]=-.5/rpar[1];   // default : p=4, ap=-.5/p

      printf("specified piston motion: x(t) = a*t^p\n");
      gi.inputString(answer,"Enter a,p for x(t) = a*t^p");
      sScanF(answer,"%e %e",&rpar[0],&rpar[1]);
      printf("The specified piston motion parameters are a=%e, p=%e\n",rpar[0],rpar[1]);
    }
    else if( answer=="forced piston motion" )
    {
      userKnownSolution=forcedPiston; 
      rpar[0]=1.; rpar[1]=1.; 

      gi.inputString(answer,"Enter mass,height");
      sScanF(answer,"%e %e",&rpar[0],&rpar[1]);
      printf("forced piston: mass=%e, height=%e\n",rpar[0],rpar[1]);
    }
    else
    {
      cout << "unknown response=[" << answer << "]\n";
      gi.stopReadingCommandFile();
    }
    
  }

  gi.unAppendTheDefaultPrompt();
  return userKnownSolution;
}
