*
* cgins: Heated cylinder in a gravitational field (Boussinesq flow)
*       
*  cgins [-noplot] heatedCyl -g=<name> -tf=<tFinal> -tp=<tPlot> -solver=<yale/best> -ts=<implicit> -debug=<num> ..,
*                     -nu=<num> -Prandtl=<num> -bg=<backGround> -show=<name> -implicitFactor=<num>
* 
* Examples:
*    cgins heatedCyl -g="cice2.order2.hdf" -nu=.05 -tf=10. -tp=1. -ts=implicit
*    cgins heatedCyl -g="cice4.order2.hdf" -nu=.01 -tf=10. -tp=1. -ts=implicit
*    srun -N1 -n1 -ppdebug $cginsp heatedCyl -g="cice2.order2.hdf" -nu=.05 -tf=10. -tp=1. -ts=implicit -solver=best
*   
*   --- flow inside a disk: 
*    cgins noplot heatedCyl -g="sice2.order2.hdf" -nu=.05 -tf=10. -tp=.1 -go=og -dtMax=.001
*    cgins noplot heatedCyl -g="sice2.order2.hdf" -nu=.05 -tf=10. -tp=.1 -ts=implicit -dtMax=.05 -go=og
*    cgins noplot heatedCyl -g="square5.hdf" -nu=.05 -tf=10. -tp=.1 -ts=implicit -dtMax=.01 -go=og
* 
* --- set default values for parameters ---
* 
$grid="cic.hdf"; $ts="adams PC"; $noplot=""; $backGround="square"; $vIn=-.1; $T0=0.; 
$tFinal=20.; $tPlot=.1; $dtMax=.1; $degreeSpace=2; $degreeTime=2; $show=" "; $debug=1; 
$nu=.1; $Prandtl=.72; $thermalExpansivity=.1; $Twall=1.;  $implicitFactor=.5; $implicitVariation="viscous"; 
$tz=0; # turn on tz here
$gravity = "0 -10. 0."; $cdv=1.; $cDt=.25; 
* $solver="choose best iterative solver";
$solver="yale";
* 
*
* ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"tf=f"=>\$tFinal,"implicitFactor=f"=>\$implicitFactor, "model=s"=>\$model,\
 "tp=f"=>\$tPlot, "solver=s"=>\$solver, "tz=s"=>\$tz, "show=s"=>\$show,"order=i"=>\$order,"debug=i"=>\$debug, \
 "ts=s"=>\$ts,"nu=f"=>\$nu,"Prandtl=f"=>\$Prandtl,"cfl=f"=>\$cfl, "bg=s"=>\$backGround, "noplot=s"=>\$noplot,\
 "go=s"=>\$go,"dtMax=f"=>\$dtMax,"cDt=f"=>\$cDt,"iv=s"=>\$implicitVariation  );
* -------------------------------------------------------------------------------------------------
if( $solver eq "best" ){ $solver="choose best iterative solver"; }
if( $order eq "2" ){ $order = "second order accurate"; }else{ $order = "fourth order accurate"; }
if( $model eq "ins" ){ $model = "incompressible Navier Stokes"; }else\
                     { $model = "incompressible Navier Stokes\n Boussinesq model"; }
if( $implicitVariation eq "viscous" ){ $implicitVariation = "useNewImplicitMethod\n implicitViscous"; }\
elsif( $implicitVariation eq "adv" ){ $implicitVariation = "useNewImplicitMethod\n implicitAdvectionAndViscous"; }\
elsif( $implicitVariation eq "full" ){ $implicitVariation = "useNewImplicitMethod\n implicitFullLinearized"; }\
else{ $implicitVariation = "implicitFullLinearized"; }
*
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
$kThermal=$nu/$Prandtl;
*
* specify the grid: 
$grid
* 
  incompressible Navier Stokes
  Boussinesq model
*   define real parameter kappa $kThermal
  define real parameter thermalExpansivity $thermalExpansivity
  define real parameter adcBoussinesq 0. 
  continue
* 
  OBTZ:polynomial
  OBTZ:twilight zone flow $tz
  degree in space $degreeSpace
  degree in time $degreeTime
* 
* choose time stepping method:
  $ts
*   
  * number of PC corrections 5
  $implicitVariation
  implicit factor $implicitFactor 
* 
  choose grids for implicit
    all=implicit
   done
* 
*
  final time $tFinal
  times to plot $tPlot
  dtMax $dtMax
  pde parameters
    nu  $nu
    kThermal $kThermal
    gravity
      $gravity
    OBPDE:divergence damping  $cdv 
    OBPDE:cDt div damping $cDt
   done
* 
  debug $debug
***
  show file options
    compressed
      open
      $show
    frequency to flush
      4
    exit
***
  pressure solver options
     $solver
     * these tolerances are chosen for PETSc
     relative tolerance
       1.e-4
     absolute tolerance
       1.e-6
    exit
  implicit time step solver options
     $solver
     * these tolerances are chosen for PETSc
     relative tolerance
       1.e-5
     absolute tolerance
       1.e-7
    exit
* 
  boundary conditions
    all=noSlipWall
    Annulus(0,1)=noSlipWall, uniform(T=$Twall)
    Annulus(1,1)=noSlipWall, uniform(T=$Twall)
    * $backGround(1,1)=inflowWithVelocityGiven, uniform(p=1.,v=$vIn)
    * $backGround(0,1)=outflow
  done
* 
  initial conditions
  if( $tz eq "0" ){ $commands="uniform flow\n" . "p=1., u=0., v=$vIn, T=$T0\n" . "continue"; }else{ $commands="continue";}
    $commands
* 
  continue
  plot:T
  $go
  
