#include "Cgins.h"
#include "PlotStuff.h"
#include "Ogshow.h"
#include "ParallelUtility.h"
#include "display.h"
#include "CgSolverUtil.h"
int 
getLineFromFile( FILE *file, char s[], int lim);

int
main(int argc, char *argv[])
{
  Overture::start(argc,argv);  // initialize Overture and A++/P++
  const int myid=Communication_Manager::My_Process_Number;

  int plotOption=true;
  bool smartRelease=false;
  bool reportMemory=false;
  bool loadBalance=false;
  
  aString commandFileName="";
  if( argc > 1 )
  { // look at arguments for "noplot" or some other name
    aString line;
    for( int i=1; i<argc; i++ )
    {
      line=argv[i];
      if( line=="-noplot" || line=="noplot" )
        plotOption=false;
      else if( line=="-nopause" || line=="-abortOnEnd" || line=="-nodirect" ||
               line=="-readCollective" || line=="-writeCollective" ||
               line=="nopause" || line=="abortOnEnd" || line=="nodirect" )
        continue; // these commands are processed by getGraphicsInterface below 
      else if( line=="memory" )
      {
	reportMemory=true;
        Diagnostic_Manager::setTrackArrayData(TRUE);
      }
      else if( line=="loadBalance" || line=="-loadBalance" )
      {
	loadBalance=true;
      }
      else if( line=="release" )
      {
        smartRelease=true;
        printF("*** turn on smart release of memory ***\n");
        Diagnostic_Manager::setSmartReleaseOfInternalMemory( ON );
      }
      else if( commandFileName=="" )
      {
        commandFileName=line;    
        printF("cgins: reading commands from file [%s]\n",(const char*)commandFileName);
      }
      
    }
  }
  else
    printF("Usage: `cgins [options][file.cmd]' \n"
            "     options:                            \n" 
            "          noplot:   run without graphics \n" 
            "          nopause: do not pause \n" 
            "          abortOnEnd: abort if command file ends \n" 
            "          memory:   run with A++ memory tracking\n" 
            "          release:  run with A++ smart release of memeory\n"
            "     file.cmd: read this command file \n");


  
  GenericGraphicsInterface & ps = *Overture::getGraphicsInterface("cgins",false,argc,argv);

  // By default start saving the command file called "cgins.cmd"
  aString logFile="cgins.cmd";
  ps.saveCommandFile(logFile);
  printF("User commands are being saved in the file `%s'\n",(const char *)logFile);

  ps.appendToTheDefaultPrompt("cgins>");

  // read from a command file if given
  if( commandFileName!="" )
  {
    printF("read command file =%s\n",(const char*)commandFileName);
    ps.readCommandFile(commandFileName);
  }

  aString nameOfShowFile;

#if 0
  aString nameOfOGFile, nameOfShowFile;
  ps.inputString(nameOfOGFile,"Enter the name of the (old) overlapping grid file:");

  printF("Try to read the overlapping grid file : %s\n",(const char*)nameOfOGFile);
  
  // create and read in a CompositeGrid
  #ifdef USE_PPP
    // On Parallel machines always add at least this many ghost lines on local arrays
    const int numGhost=2;
    MappedGrid::setMinimumNumberOfDistributedGhostLines(numGhost);


  #endif
  // Reduce the width of extrapolation for interp neighbours
  // and increase parallel ghost boundary width to 2 -- this will mean that we should
  // get the same result independent of np
  GenericMappedGridOperators::setDefaultMaximumWidthForExtrapolateInterpolationNeighbours(3);

  CompositeGrid m;
  getFromADataBase(m,nameOfOGFile);
  CompositeGrid & cg = m.numberOfMultigridLevels()==1 ? m : m.multigridLevel[0];
  if( m.numberOfMultigridLevels() > 1 )
    m.update(CompositeGrid::THEmultigridLevel);
#else

  CompositeGrid cg;
  readOrBuildTheGrid(ps, cg, loadBalance);

#endif


  Interpolant interpolant(cg);
  interpolant.setImplicitInterpolationMethod(Interpolant::iterateToInterpolate);

  #ifdef USE_PPP
  if( !interpolant.interpolationIsExplicit() )
  {
    printf("cgins:ERROR: The parallel composite grid interpolator needs explicit interpolation ****\n");
    Overture::abort();
  }
  #endif

  // cout << "cgins: numberOfGrids = " << cg.numberOfGrids() << endl;
  // cout << "cgins: numberOfComponentGrids = " << cg.numberOfComponentGrids() << endl;

  // cout << "cgins: nameOfShowFile = " << nameOfShowFile << endl;
  
   Ogshow *show=NULL;
//  ** old way:
//   ps.inputString(nameOfShowFile,"Enter the name of the (new) show file (enter `none' for none):");
//   if( nameOfShowFile=="none" )
//     nameOfShowFile="";
//   if( nameOfShowFile!="" )
//     show = new Ogshow( nameOfShowFile,".",false );
  

//  CgSolver & solver = *new CgSolver(cg,&ps,show,plotOption);

  Cgins & solver = *new Cgins(cg,&ps,show,plotOption);

  solver.setParametersInteractively();

//   assert( solver.compositeGridSolver!=NULL && solver.compositeGridSolver[0]!=NULL );
//   pDomainSolver = solver.compositeGridSolver[0]; // do this for now

  solver.solve();
  solver.printStatistics();

  ps.unAppendTheDefaultPrompt();
//  delete show;

  if( reportMemory )
    Diagnostic_Manager::report();

  if(  myid==0 && false ) 
  {
    printf("\n +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      cg[grid].displayComputedGeometry();
    printf(" +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  }


//  delete pDomainSolver;

  delete &solver;  // do this so we shutdown PETSc (if used).

  Overture::finish();          
  if( smartRelease )
  {
    int totalNumberOfArrays=GET_NUMBER_OF_ARRAYS;
    if( totalNumberOfArrays>0 )
    {
      printf("\n**** WARNING: Number of A++ is %i >0 \n",totalNumberOfArrays);
    }
  }
  
  return 0;
}


