// This file automatically generated from insBC.bC with bpp.
#include "Cgins.h"
#include "Parameters.h"
#include "turbulenceModels.h"
#include "Insbc4WorkSpace.h"
#include "App.h"
#include "ParallelUtility.h"

#define ForBoundary(side,axis)   for( axis=0; axis<mg.numberOfDimensions(); axis++ ) for( side=0; side<=1; side++ )

void
applyFourthOrderBoundaryConditions( realMappedGridFunction & u0, real t, int grid, Insbc4WorkSpace & ws,
                            				    IntegerArray & ipar, RealArray & rpar, OGFunction & exact, Parameters & parameters );

// This next include file defines the setTemperatureBC macro
// ----------------------------------------------------------------------------
// Macro: Apply BC's on the Temperature 
// 
//   There are 3 cases: 
//      (1) apply a dirichlet BC                       (OPTION=dirichlet)
//      (2) extrapolate ghost pts on dirichlet BC's     (OPTION=extrapolateGhost)
//      (3) apply a mixed BC                           (OPTION=mixed)
// 
// Macro args:
// 
// tc : component to assign
// NAME : name of of the calling function (for comments)
// BCNAME : noSlipWall, inflowWithVelocityGiven etc. 
// OPTION: dirichlet, mixed, extrapolateGhost
// ----------------------------------------------------------------------------


//    Mixed-derivative BC for component i: 
//          mixedCoeff(i)*u(i) + mixedNormalCoeff(i)*u_n(i) = mixedRHS(i)
#define mixedRHS(component,side,axis,grid)         bcData(component+numberOfComponents*(0),side,axis,grid)
#define mixedCoeff(component,side,axis,grid)       bcData(component+numberOfComponents*(1),side,axis,grid)
#define mixedNormalCoeff(component,side,axis,grid) bcData(component+numberOfComponents*(2),side,axis,grid)

//\begin{>>MappedGridSolverInclude.tex}{\subsection{applyBoundaryConditionsINS}} 
int Cgins::
applyBoundaryConditions(const real & t, realMappedGridFunction & u, 
                  			realMappedGridFunction & gridVelocity,
                  			const int & grid,
                  			const int & option /* =-1 */,
                  			realMappedGridFunction *puOld /* =NULL */,  
                  			realMappedGridFunction *pGridVelocityOld /* =NULL */,
                  			const real & dt /* =-1. */ )
//=========================================================================================
// /Description:
//   Apply boundary conditions for the incompressibleNavierStokes (explicit time stepping).
// 
// /t (input):
// /u (input/output) : apply to this grid function.
// /gridIsMoving (input) : true if this grid is moving.
// /gridVelocity (input) : the grid velocity if gridIsMoving==true.
// /variableBoundaryData (input) : true if there is boundary data that depends on the position along the boundary.
// /boundaryData (input) : boundary data used if variableBoundaryData==true.
// grid (input) : the grid number if this MappedGridFunction is part of a CompositeGridFunction.
// option (input): not used here.
//
// /Note:
// ***Remember to also change the BC routine for implicit time stepping if changes are made here
// applyBoundaryConditionsForImplicitTimeStepping
//
// /NOTE on the bcData array: 
//  Boundary condition parameter values are stored in the array bcData. Let nc=numberOfComponents,
//  then the values 
//
//          bcData(i,side,axis,grid)  : i=0,1,...,nc-1 
//
//  would normally represent the RHS values for dirichlet BC's on component i, such as
//            u(i1,i2,i3,i) = bcData(i,side,axis,grid) 
// 
//  For a Mixed-derivative boundary condition, the parameters (a0,a1,a2) in the mixed BC:
//               a1*u(i1,i2,i3,i) + a2*u(i1,i2,i3,i)_n = a0
//  are stored in
//          a_j = bcData(i+nc*(j),side,axis,grid),  j=0,1,2
// 
//  Thus bcData(i,side,axis,grid) still holds the RHS value for the mixed-derivative condition
// 
//\end{MappedGridSolverInclude.tex}  
//=========================================================================================
{
    if( parameters.dbase.get<bool>("simulateGridMotionOnly") ) return 0;

    real time0=getCPU();

    if( debug() & 4 )
    {
        printF(">>>>> Cgins::applyBoundaryConditions <<<<<<<\n");
    }
    

    checkArrayIDs(" insBC: start"); 

    MappedGrid & mg = *u.getMappedGrid();
    const int numberOfDimensions = mg.numberOfDimensions();
    
    const bool isRectangular = mg.isRectangular();
    
  // *** turn off for stretched c-grid at outflow 
    bool applyDivergenceBoundaryCondition=true; // false; // true;
    bool applyDivergenceBoundaryConditionAtOutflow=true;
    
//   MappedGrid & mg = *u.getMappedGrid();
//   printf("applyBoundaryConditionsINS: grid=%i variableBoundaryData=%i\n",grid,variableBoundaryData);
//   display(mg.boundaryCondition(),sPrintF(buff,"grid=%i applyBoundaryConditionsINS: mg.boundaryCondition()",grid));
      
    const bool gridIsMoving = parameters.gridIsMoving(grid);

    const int uc = parameters.dbase.get<int >("uc");
    const int vc = parameters.dbase.get<int >("vc");
    const int wc = parameters.dbase.get<int >("wc");
    const int tc = parameters.dbase.get<int >("tc");
    const int & nc = parameters.dbase.get<int >("nc");
    const int orderOfAccuracy=parameters.dbase.get<int >("orderOfAccuracy");
    assert( orderOfAccuracy==2 || orderOfAccuracy==4 );
    
    const RealArray & bcData = parameters.dbase.get<RealArray>("bcData");

    typedef int BoundaryCondition;
    
    const BoundaryCondition & noSlipWall = Parameters::noSlipWall;
    const BoundaryCondition & slipWall   = Parameters::slipWall;
    const BoundaryCondition & inflowWithVelocityGiven = InsParameters::inflowWithVelocityGiven;
    const BoundaryCondition & inflowWithPressureAndTangentialVelocityGiven 
                              = InsParameters::inflowWithPressureAndTangentialVelocityGiven;
    const BoundaryCondition & outflow = InsParameters::outflow;
    const BoundaryCondition & tractionFree = InsParameters::tractionFree;
    const BoundaryCondition & symmetry = Parameters::symmetry;
    const BoundaryCondition & dirichletBoundaryCondition = Parameters::dirichletBoundaryCondition;
    const BoundaryCondition & axisymmetric = Parameters::axisymmetric;
    
  // make some shorter names for readability
    BCTypes::BCNames dirichlet             = BCTypes::dirichlet,
                                      neumann               = BCTypes::neumann,
                                      mixed                 = BCTypes::mixed,
                                      extrapolate           = BCTypes::extrapolate,
                                      normalComponent       = BCTypes::normalComponent,
                 //   aDotU                 = BCTypes::aDotU,
                                      generalizedDivergence = BCTypes::generalizedDivergence,
                                      tangentialComponent   = BCTypes::tangentialComponent,
                                      vectorSymmetry        = BCTypes::vectorSymmetry,
                                      allBoundaries         = BCTypes::allBoundaries; 


  // Here is the array that defines the domain interfaces, interfaceType(side,axis,grid) 
    const IntegerArray & interfaceType = parameters.dbase.get<IntegerArray >("interfaceType");
    const BoundaryCondition & interfaceBoundaryCondition = Parameters::interfaceBoundaryCondition;


    bool assignSlipWall=false;
    bool assignNoSlipWall=false;
    bool assignInflowWithVelocityGiven=false;
    bool assignOutflow=false;
    bool assignTractionFree=false;
    bool assignAxisymmetric=false;
    bool assignSymmetry=false;
    bool assignDirichletBoundaryCondition=false;
    bool assignInflowWithPressureAndTangentialVelocityGiven=false;
    
    int side,axis;
    for( axis=0; axis<mg.numberOfDimensions(); axis++ )
    {
        for( side=0; side<=1; side++ )
        {
            int bc=mg.boundaryCondition(side,axis);
            switch (bc)
            {
            case 0 : break;
            case -1: break;
            case Parameters::slipWall:                   assignSlipWall=true; break;
            case Parameters::noSlipWall :                assignNoSlipWall=true; break;
            case InsParameters::inflowWithVelocityGiven:    assignInflowWithVelocityGiven=true; break;
            case InsParameters::outflow:                    assignOutflow=true; break;
            case InsParameters::tractionFree:               assignTractionFree=true; break;
            case Parameters::axisymmetric:               assignAxisymmetric=true; break;
            case Parameters::symmetry :                  assignSymmetry=true; break;
            case Parameters::dirichletBoundaryCondition: assignDirichletBoundaryCondition=true; break;
            case InsParameters::inflowWithPressureAndTangentialVelocityGiven :
      	assignInflowWithPressureAndTangentialVelocityGiven=true; break;
            default: 
                printf("insBC:ERROR: unknown boundary condition =%i on grid %i, side=%i, axis=%i\n",bc,grid,side,axis);
                throw "error";
            Overture::abort("error");
            }
        }
    }

    const int numberOfGhostPointsNeeded = parameters.numberOfGhostPointsNeeded();

  // **************************************************************************
  //  apply boundary conditions in order of increasing priority (so corners
  //    take the values from the bc that is applied last)
  // **************************************************************************

    const int numberOfComponents =parameters.dbase.get<int >("numberOfComponents");

    Range C(0,numberOfComponents-1);  // ***** is this correct ******
    Range V = Range(uc,uc+numberOfDimensions-1);
    const Range & Rt = parameters.dbase.get<Range >("Rt"); // time dependent parameters (u,v,w,[T]). 
    
    BoundaryConditionParameters extrapParams;
    BoundaryConditionParameters bcParams;

    BoundaryData::BoundaryDataArray & pBoundaryData = parameters.getBoundaryData(grid);

    const InsParameters::PDEModel & pdeModel = parameters.dbase.get<InsParameters::PDEModel >("pdeModel");
    
    const bool assignTemperature = pdeModel==InsParameters::BoussinesqModel ||
                                                                  pdeModel==InsParameters::viscoPlasticModel;
    

    const Parameters::TimeSteppingMethod & timeSteppingMethod = 
                                            parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod");
    

    Range twoPhaseFlowComponents;
    if( pdeModel==InsParameters::twoPhaseFlowModel )
    {
        assert( nc==(tc+1) ) ;
      twoPhaseFlowComponents=Range(tc,nc);
    }

    if( timeSteppingMethod==Parameters::steadyStateRungeKutta && 
            t>0. )  // apply all boundary conditions at t=0
    {
    // *****************************************************
    // ***********STEADY STATE CASE*************************
    // *****************************************************

    // we only need to apply a limited number of BC's for the steady state solver since most
    // have already been done.


        if( assignSlipWall )
        {
      // on a slip wall we need to extrapolate points that lie outside interpolation pts on the bndry.
            u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent0, slipWall,0.,t);
            if( numberOfDimensions==3 )
      	u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent1, slipWall,0.,t);

            if( !isRectangular ) // rectangular case is already done
                u.applyBoundaryCondition(V,BCTypes::generalizedDivergence,Parameters::slipWall,0.,t); 	
        }
        
        if( assignNoSlipWall )
        {
            
            if( false )
            { // for testing
      	BoundaryConditionParameters bcParams;
      	bcParams.lineToAssign=1;
      	u.applyBoundaryCondition(vc,dirichlet,noSlipWall,0.,t,bcParams);
                u.applyBoundaryCondition(V,BCTypes::generalizedDivergence,noSlipWall,0.,t); 	
            }
            else
            {
                u.applyBoundaryCondition(V,extrapolate,noSlipWall,0.,t);
                u.applyBoundaryCondition(V,BCTypes::generalizedDivergence,noSlipWall,0.,t); 	
            }
            
      // NOTE: dirichlet or neumann BC's for T are already done in the lineSolver
        }
        
        if( assignDirichletBoundaryCondition ) // *wdh* added 070605
        {
            u.applyBoundaryCondition(C,dirichlet,dirichletBoundaryCondition,0.,t,bcParams);
            u.applyBoundaryCondition(C,extrapolate,dirichletBoundaryCondition,0.,t,bcParams);

        }
        
        
        if( assignOutflow && 
                parameters.dbase.get<int>("outflowOption")==0 && 
                parameters.dbase.get<int >("checkForInflowAtOutFlow")==1 )
        {
      // *wdh* 030603 ** add these for Kyle's bug ??
            const int orderOfExtrapolation=extrapParams.orderOfExtrapolation;
            extrapParams.orderOfExtrapolation=2;
            u.applyBoundaryCondition(V,extrapolate,    outflow,0.,t,extrapParams);

      // **check for local inflow at an outflow boundary**
      // where( inflow ) give u.n=0
            Index I1,I2,I3;
        
            if( !parameters.dbase.get<bool >("twilightZoneFlow") &&  assignOutflow && orderOfAccuracy==2 )
            {
        
      	for( axis=0; axis<mg.numberOfDimensions(); axis++ )
      	{
        	  for( side=Start; side<=End; side++ )
        	  {
          	    if( mg.boundaryCondition(side,axis)==outflow )
          	    {
            	      RealDistributedArray & normal  = mg.vertexBoundaryNormal(side,axis);  

            	      getGhostIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);
            	      intArray & mask = bcParams.mask();
            	      mask.redim(I1,I2,I3);   // mask lives on ghost line.
            	      getBoundaryIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);
            	      if( isRectangular )
            	      {
            		mask = u(I1,I2,I3,uc+axis)*(2*side-1) <0.;
            	      }
            	      else
            	      {
            		if( mg.numberOfDimensions()==2 )
            		{
              		  mask = (u(I1,I2,I3,uc)*normal(I1,I2,I3,0)+
                    			  u(I1,I2,I3,vc)*normal(I1,I2,I3,1)) <0; 
            		}
            		else
            		{
              		  mask = (u(I1,I2,I3,uc)*normal(I1,I2,I3,0)+
                    			  u(I1,I2,I3,vc)*normal(I1,I2,I3,1)+
                    			  u(I1,I2,I3,wc)*normal(I1,I2,I3,2)) <0; 
            		}
            	      }
      	
            	      int count=sum(mask);
            	      if( count>0 )
            	      {
            		if( debug() & 4 )
              		  printf("insBC: number of outflow points that are inflow = %i\n",count);
            		bcParams.setUseMask(TRUE);
		// u.applyBoundaryCondition(V,neumann,outflow,0.,t,bcParams);
            		u.applyBoundaryCondition(V,neumann,BCTypes::boundary(side,axis),0.,t,bcParams);
            		bcParams.setUseMask(FALSE);
            	      }
          	    }
        	  }
      	}
            }



            extrapParams.orderOfExtrapolation=orderOfExtrapolation;  // reset
            u.applyBoundaryCondition(V,generalizedDivergence,outflow,0.,t);
        }
        
        
        
        if( assignInflowWithVelocityGiven )
        {

      // the inflow value can be time dependent
            u.applyBoundaryCondition(V,dirichlet,inflowWithVelocityGiven,bcData,pBoundaryData,t,
                         			       Overture::defaultBoundaryConditionParameters(),grid);


            bool assignGhostWithDirichlet=false;
            
            for( axis=0; axis<mg.numberOfDimensions(); axis++ )
            {
      	for( side=0; side<=1; side++ )
      	{ // this could fail if there are two inflow sides!
                    if( mg.boundaryCondition(side,axis)==inflowWithVelocityGiven &&
                            parameters.bcType(side,axis,grid)==Parameters::blasiusProfile  )
        	  {
          	    assignGhostWithDirichlet=true;
        	  }
      	}
            }
            if( assignGhostWithDirichlet )
            {
        // this only works for Blasius (or parabolic inflow)

        // ** added by Kyle for Blasius inflow ***
      	BoundaryConditionParameters gDirParams;
      	gDirParams.lineToAssign=1;
      	u.applyBoundaryCondition(V,dirichlet,inflowWithVelocityGiven,bcData,pBoundaryData,t, gDirParams,grid);
            }
            else
            {
                u.applyBoundaryCondition(V,extrapolate,inflowWithVelocityGiven,0.,t);
    
                u.applyBoundaryCondition(V,generalizedDivergence,inflowWithVelocityGiven,0.,t);
        // add for 4th order dissipation
                if( parameters.dbase.get<bool >("useFourthOrderArtificialDiffusion") )
      	{
        	  u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent0,inflowWithVelocityGiven,0.,t);
        	  if( numberOfDimensions==3 )
          	    u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent1,inflowWithVelocityGiven,0.,t);
      	}
            }

      // NOTE: dirichlet or neumann BC's for T are already done in the lineSolver

        }
        
        if( assignNoSlipWall )
        {  // Extrapolate ghost points on dirichlet BC's for T 
      // NOTE: dirichlet or neumann BC's for T are already done in the lineSolver
              if( assignTemperature )
              {
                  FILE *& debugFile  =  parameters.dbase.get<FILE* >("debugFile");
                  FILE *& pDebugFile =  parameters.dbase.get<FILE* >("pDebugFile");
                  ForBoundary(side,axis)
                  {
                      if( mg.boundaryCondition(side,axis)==noSlipWall )
                      {
                          if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                          { // This is an interface between domains
               // for now we only know about interfaces at no-slip walls: 
                              assert( mg.boundaryCondition(side,axis)==noSlipWall );
      	 // what about BC's applied at t=0 before the boundary data is set ??
      	 // if( parameters.dbase.get<int >("globalStepNumber") < 2 ) continue; // ********************* TEMP *****
      	 // if this is an iterface we should turn off the TZ forcing for the boundary condition since we want
      	 // to use the boundary data instead.
                              #ifdef USE_PPP
                                  realSerialArray uLocal;  getLocalArrayWithGhostBoundaries(u,uLocal);
                              #else
                                  const realSerialArray & uLocal = u;
                              #endif
                   	 Index Ib1,Ib2,Ib3;
                   	 getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
                   	 bool ok = ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3);
                   	 if( debug() & 4 )
                   	 {
                     	   printP("insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                          		  side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                     	   fprintf(pDebugFile,"insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                                     			       side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                     	   if( pBoundaryData[side][axis]==NULL )
                     	   {
                       	     if( !ok )
                                          fprintf(pDebugFile," insBC:applyBC on T: pBoundaryData[side][axis] == NULL! \n");
                                      else
                                          fprintf(pDebugFile," insBC:applyBC on T: ERROR: pBoundaryData[side][axis] == NULL! \n");
                     	   }
                     	   else
                     	   {
      	     // RealArray & bd = *pBoundaryData[side][axis];
      	     // ::display(bd,"boundaryData",pDebugFile,"%8.2e ");
                     	   }
                   	 }
                              assert( mixedCoeff(tc,side,axis,grid)!=0. || mixedNormalCoeff(tc,side,axis,grid)!=0. );
                   	 if( ok && pBoundaryData[side][axis]==NULL )
                   	 {
                 // At t=0 when the initial conditions are being set-up (and initial conditions projected) there may
                 // not be a boundaryData array created yet for the interface. We thus create one and fill in some default values
                 // based on the current solution
                                  RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                                  bd=0.;
                                  #ifdef USE_PPP
                       	     const RealArray & normal = mg.vertexBoundaryNormalArray(side,axis);
                                  #else
                       	     const RealArray & normal = mg.vertexBoundaryNormal(side,axis);
                                  #endif
                     	   real a0=mixedCoeff(tc,side,axis,grid);
                     	   real a1=mixedNormalCoeff(tc,side,axis,grid);
                 // bd(Ib1,Ib2,Ib3,tc)=a0*uLocal(Ib1,Ib2,Ib3,tc);    // Assume that T.n=0 at the interface at t=0 -- could do better --
                                  MappedGridOperators & op = *(u.getOperators());
                                  Range N(tc,tc);
                                  RealArray ux(Ib1,Ib2,Ib3,N), uy(Ib1,Ib2,Ib3,N);
                     	   op.derivative(MappedGridOperators::xDerivative,uLocal,ux,Ib1,Ib2,Ib3,N);
                     	   op.derivative(MappedGridOperators::yDerivative,uLocal,uy,Ib1,Ib2,Ib3,N);
                                  if( mg.numberOfDimensions()==2 )
                     	   {
                                      bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                        				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                     	   }
                     	   else
                     	   {
                                      RealArray uz(Ib1,Ib2,Ib3,N);
                       	     op.derivative(MappedGridOperators::zDerivative,uLocal,uz,Ib1,Ib2,Ib3,N);
                       	     bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                        				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)+
                                        				    normal(Ib1,Ib2,Ib3,2)*uz(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                     	   }
                                  const bool twilightZoneFlow = parameters.dbase.get<bool >("twilightZoneFlow");
                     	   if( false && twilightZoneFlow ) //  *******************************************************
                     	   {
                                      fprintf(pDebugFile," insBC:applyBC on T: ********** Set TRUE value for bd at t=%8.2e ************\n",t);
                       	     const bool rectangular= mg.isRectangular() && !twilightZoneFlow;
                       	     realArray & x= mg.center();
            #ifdef USE_PPP
                       	     realSerialArray xLocal; 
                       	     if( !rectangular || twilightZoneFlow ) 
                         	       getLocalArrayWithGhostBoundaries(x,xLocal);
            #else
                       	     const realSerialArray & xLocal = x;
            #endif
                       	     OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
                       	     realSerialArray ue(Ib1,Ib2,Ib3), uex(Ib1,Ib2,Ib3), uey(Ib1,Ib2,Ib3); 
                       	     e.gd( ue ,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,0,Ib1,Ib2,Ib3,tc,t);
                       	     e.gd( uex,xLocal,mg.numberOfDimensions(),rectangular,0,1,0,0,Ib1,Ib2,Ib3,tc,t);
                       	     e.gd( uey,xLocal,mg.numberOfDimensions(),rectangular,0,0,1,0,Ib1,Ib2,Ib3,tc,t);
                       	     real a0=mixedCoeff(tc,side,axis,grid), a1=mixedNormalCoeff(tc,side,axis,grid);
                       	     if( mg.numberOfDimensions()==2 )
                       	     {
                         	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                          				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                       	     }
                       	     else
                       	     {
                         	       realSerialArray uez(Ib1,Ib2,Ib3);
                         	       e.gd( uez,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,1,Ib1,Ib2,Ib3,tc,t);
                         	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                          				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)+
                                          				      normal(Ib1,Ib2,Ib3,2)*uez(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                       	     }
                     	   } // *******************************************
                   	 }
                   	 if( pBoundaryData[side][axis]!=NULL )
                     	   u.getOperators()->setTwilightZoneFlow( false );
                   	 else
                   	 {
                                  if( t>0. || debug() & 4 )
                     	   {
                                      if( ok )
                                 	       printP("$$$$ insBC:applyBC:INFO:interface but no boundaryData, t=%9.3e\n",t);
                     	   }
                   	 }
                          }
                          if( debug() & 4 )
                   	 printF("++++insBC: noSlipWall: (grid,side,axis)=(%i,%i,%i) : "
                        		"  BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
                        		"  BC: u: %3.2f*u+%3.2f*u.n=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
                        		grid,side,axis, 
                        		mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
                        		mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
                        		mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
                     	   );
      //        if( mixedNormalCoeff(tc,side,axis,grid)!=0. ) // coeff of T.n is non-zero
      //        {
      // 	 mixedBoundaryConditionOnTemperature=true;
      // 	 if( debug() & 4 )
      // 	   printF("++++insBC: noSlipWall:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
      // 		  "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
      //                   "  BC: u: %3.2f*u+%3.2f*u.n=%3.2f=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
      // 		  grid,side,axis, 
      //                   mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
      //                   mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
      //                   mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
      // 	     );
      //        }
                          if( mixedNormalCoeff(tc,side,axis,grid)==0. ) // coeff of T.n 
                          {
      	 // Dirichlet
                   	 u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),0.,t);
      // 	      u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
      // 				       bcParams,grid);
                          }
                          else
                          {
      	 // Mixed or Neumann -- 
                          }
                          if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                          { // reset TZ
                   	 u.getOperators()->setTwilightZoneFlow( parameters.dbase.get<bool >("twilightZoneFlow") );
                          }
                      } // end if bc = noSlipWall
                  } // end for boundary
         // ************ try this ********* 080909
         // u.updateGhostBoundaries(); // this is done in finish boundary conditions now
              } // end if assignTemperature

        }
        if( assignInflowWithVelocityGiven )
        {
      // Extrapolate ghost points on dirichlet BC's for T 
      // NOTE: dirichlet or neumann BC's for T are already done in the lineSolver
              if( assignTemperature )
              {
                  FILE *& debugFile  =  parameters.dbase.get<FILE* >("debugFile");
                  FILE *& pDebugFile =  parameters.dbase.get<FILE* >("pDebugFile");
                  ForBoundary(side,axis)
                  {
                      if( mg.boundaryCondition(side,axis)==inflowWithVelocityGiven )
                      {
                          if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                          { // This is an interface between domains
               // for now we only know about interfaces at no-slip walls: 
                              assert( mg.boundaryCondition(side,axis)==noSlipWall );
      	 // what about BC's applied at t=0 before the boundary data is set ??
      	 // if( parameters.dbase.get<int >("globalStepNumber") < 2 ) continue; // ********************* TEMP *****
      	 // if this is an iterface we should turn off the TZ forcing for the boundary condition since we want
      	 // to use the boundary data instead.
                              #ifdef USE_PPP
                                  realSerialArray uLocal;  getLocalArrayWithGhostBoundaries(u,uLocal);
                              #else
                                  const realSerialArray & uLocal = u;
                              #endif
                   	 Index Ib1,Ib2,Ib3;
                   	 getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
                   	 bool ok = ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3);
                   	 if( debug() & 4 )
                   	 {
                     	   printP("insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                          		  side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                     	   fprintf(pDebugFile,"insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                                     			       side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                     	   if( pBoundaryData[side][axis]==NULL )
                     	   {
                       	     if( !ok )
                                          fprintf(pDebugFile," insBC:applyBC on T: pBoundaryData[side][axis] == NULL! \n");
                                      else
                                          fprintf(pDebugFile," insBC:applyBC on T: ERROR: pBoundaryData[side][axis] == NULL! \n");
                     	   }
                     	   else
                     	   {
      	     // RealArray & bd = *pBoundaryData[side][axis];
      	     // ::display(bd,"boundaryData",pDebugFile,"%8.2e ");
                     	   }
                   	 }
                              assert( mixedCoeff(tc,side,axis,grid)!=0. || mixedNormalCoeff(tc,side,axis,grid)!=0. );
                   	 if( ok && pBoundaryData[side][axis]==NULL )
                   	 {
                 // At t=0 when the initial conditions are being set-up (and initial conditions projected) there may
                 // not be a boundaryData array created yet for the interface. We thus create one and fill in some default values
                 // based on the current solution
                                  RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                                  bd=0.;
                                  #ifdef USE_PPP
                       	     const RealArray & normal = mg.vertexBoundaryNormalArray(side,axis);
                                  #else
                       	     const RealArray & normal = mg.vertexBoundaryNormal(side,axis);
                                  #endif
                     	   real a0=mixedCoeff(tc,side,axis,grid);
                     	   real a1=mixedNormalCoeff(tc,side,axis,grid);
                 // bd(Ib1,Ib2,Ib3,tc)=a0*uLocal(Ib1,Ib2,Ib3,tc);    // Assume that T.n=0 at the interface at t=0 -- could do better --
                                  MappedGridOperators & op = *(u.getOperators());
                                  Range N(tc,tc);
                                  RealArray ux(Ib1,Ib2,Ib3,N), uy(Ib1,Ib2,Ib3,N);
                     	   op.derivative(MappedGridOperators::xDerivative,uLocal,ux,Ib1,Ib2,Ib3,N);
                     	   op.derivative(MappedGridOperators::yDerivative,uLocal,uy,Ib1,Ib2,Ib3,N);
                                  if( mg.numberOfDimensions()==2 )
                     	   {
                                      bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                        				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                     	   }
                     	   else
                     	   {
                                      RealArray uz(Ib1,Ib2,Ib3,N);
                       	     op.derivative(MappedGridOperators::zDerivative,uLocal,uz,Ib1,Ib2,Ib3,N);
                       	     bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                        				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)+
                                        				    normal(Ib1,Ib2,Ib3,2)*uz(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                     	   }
                                  const bool twilightZoneFlow = parameters.dbase.get<bool >("twilightZoneFlow");
                     	   if( false && twilightZoneFlow ) //  *******************************************************
                     	   {
                                      fprintf(pDebugFile," insBC:applyBC on T: ********** Set TRUE value for bd at t=%8.2e ************\n",t);
                       	     const bool rectangular= mg.isRectangular() && !twilightZoneFlow;
                       	     realArray & x= mg.center();
            #ifdef USE_PPP
                       	     realSerialArray xLocal; 
                       	     if( !rectangular || twilightZoneFlow ) 
                         	       getLocalArrayWithGhostBoundaries(x,xLocal);
            #else
                       	     const realSerialArray & xLocal = x;
            #endif
                       	     OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
                       	     realSerialArray ue(Ib1,Ib2,Ib3), uex(Ib1,Ib2,Ib3), uey(Ib1,Ib2,Ib3); 
                       	     e.gd( ue ,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,0,Ib1,Ib2,Ib3,tc,t);
                       	     e.gd( uex,xLocal,mg.numberOfDimensions(),rectangular,0,1,0,0,Ib1,Ib2,Ib3,tc,t);
                       	     e.gd( uey,xLocal,mg.numberOfDimensions(),rectangular,0,0,1,0,Ib1,Ib2,Ib3,tc,t);
                       	     real a0=mixedCoeff(tc,side,axis,grid), a1=mixedNormalCoeff(tc,side,axis,grid);
                       	     if( mg.numberOfDimensions()==2 )
                       	     {
                         	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                          				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                       	     }
                       	     else
                       	     {
                         	       realSerialArray uez(Ib1,Ib2,Ib3);
                         	       e.gd( uez,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,1,Ib1,Ib2,Ib3,tc,t);
                         	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                          				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)+
                                          				      normal(Ib1,Ib2,Ib3,2)*uez(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                       	     }
                     	   } // *******************************************
                   	 }
                   	 if( pBoundaryData[side][axis]!=NULL )
                     	   u.getOperators()->setTwilightZoneFlow( false );
                   	 else
                   	 {
                                  if( t>0. || debug() & 4 )
                     	   {
                                      if( ok )
                                 	       printP("$$$$ insBC:applyBC:INFO:interface but no boundaryData, t=%9.3e\n",t);
                     	   }
                   	 }
                          }
                          if( debug() & 4 )
                   	 printF("++++insBC: inflowWithVelocityGiven: (grid,side,axis)=(%i,%i,%i) : "
                        		"  BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
                        		"  BC: u: %3.2f*u+%3.2f*u.n=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
                        		grid,side,axis, 
                        		mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
                        		mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
                        		mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
                     	   );
      //        if( mixedNormalCoeff(tc,side,axis,grid)!=0. ) // coeff of T.n is non-zero
      //        {
      // 	 mixedBoundaryConditionOnTemperature=true;
      // 	 if( debug() & 4 )
      // 	   printF("++++insBC: inflowWithVelocityGiven:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
      // 		  "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
      //                   "  BC: u: %3.2f*u+%3.2f*u.n=%3.2f=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
      // 		  grid,side,axis, 
      //                   mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
      //                   mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
      //                   mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
      // 	     );
      //        }
                          if( mixedNormalCoeff(tc,side,axis,grid)==0. ) // coeff of T.n 
                          {
      	 // Dirichlet
                   	 u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),0.,t);
      // 	      u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
      // 				       bcParams,grid);
                          }
                          else
                          {
      	 // Mixed or Neumann -- 
                          }
                          if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                          { // reset TZ
                   	 u.getOperators()->setTwilightZoneFlow( parameters.dbase.get<bool >("twilightZoneFlow") );
                          }
                      } // end if bc = inflowWithVelocityGiven
                  } // end for boundary
         // ************ try this ********* 080909
         // u.updateGhostBoundaries(); // this is done in finish boundary conditions now
              } // end if assignTemperature
        }
    
        if( parameters.dbase.get<Parameters::TurbulenceModel >("turbulenceModel")!=Parameters::noTurbulenceModel )
        {
            turbulenceModelBoundaryConditions(t,u,parameters,grid,pBoundaryData);
        }
        
        if( parameters.dbase.get<int >("extrapolateInterpolationNeighbours")!=0 ||  // new way 
                parameters.dbase.get<bool >("useFourthOrderArtificialDiffusion") ||     // get rid of these checks
                pdeModel==InsParameters::viscoPlasticModel  )
        {
            assert( parameters.dbase.get<int >("extrapolateInterpolationNeighbours")!=0 );

      //  NOTE: The visco-plastic model uses 2 ghost lines since the coefficient of viscosity depends on the first derivatives of u.

            assert( parameters.dbase.get<int >("extrapolateInterpolationNeighbours") );  // consistency check
            const int orderOfExtrapolationSaved=extrapParams.orderOfExtrapolation;

            extrapParams.orderOfExtrapolation=3; // *********

      // u.applyBoundaryCondition(V,BCTypes::extrapolateInterpolationNeighbours);
            u.applyBoundaryCondition(C,BCTypes::extrapolateInterpolationNeighbours,allBoundaries,0.,t,extrapParams);

      // *** assign the 2nd ghost line ***
  
            extrapParams.orderOfExtrapolation=3; // *********
            extrapParams.ghostLineToAssign=2;

            if( true )
            {
	// printF("insBC: extrapolate 2nd ghost line on all boundaries...\n");
      	u.applyBoundaryCondition(C,extrapolate,allBoundaries,0.,t,extrapParams);
            }
            
      // reset: 
            extrapParams.ghostLineToAssign=1;
            extrapParams.orderOfExtrapolation=orderOfExtrapolationSaved;

        }
        else
        {
            assert( !parameters.dbase.get<int >("extrapolateInterpolationNeighbours") );  // consistency check
        }
        
        if( parameters.dbase.get<int >("orderOfAccuracy")==2 ) // ** moved from generic applyBC - 060907
            u.finishBoundaryConditions();

        parameters.dbase.get<RealArray>("timing")(Parameters::timeForBoundaryConditions)+=getCPU()-time0;
        return 0;
    }
    




  // ********************************************************************
  // ************Non-steady state case***********************************
  // ********************************************************************

    char buff[200];

  // the dirichletBoundaryCondition is for testing TZ flow.
    if( assignDirichletBoundaryCondition )
    {
        if( debug() & 32  )
        {
            display(u,sPrintF(buff,"insBC: u before assignDirichletBoundaryCondition, grid=%i, t=%e",grid,t),
                    parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
        }

        u.applyBoundaryCondition(Rt,dirichlet,dirichletBoundaryCondition,0.,t);
        checkArrayIDs(" insBC: after dirichlet"); 

        if( debug() & 32  )
        {
            display(u,sPrintF(buff,"insBC: u after dirichlet, grid=%i, t=%e",grid,t),
                    parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
        }

        u.applyBoundaryCondition(Rt,extrapolate,dirichletBoundaryCondition,0.,t);
        checkArrayIDs(" insBC: after extrapolate (1)"); 
        if( debug() & 32  )
        {
            display(u,sPrintF(buff,"insBC: u after extrapolate, assignDirichletBoundaryCondition, grid=%i, t=%e",grid,t),
                    parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
        }


    }
    

  // ** new format *** first do all dirichlet type boundary conditions



  // either use boundaryData or use bcData.
    if( assignInflowWithVelocityGiven )
    {
    // *NOTE* cannot assign on extended range unless we increase the size of the boundaryData
        u.applyBoundaryCondition(V,dirichlet,inflowWithVelocityGiven,bcData,pBoundaryData,t,bcParams,grid);
    // *wdh* 080517 : 
//     if( assignTemperature )
//     {
//       u.applyBoundaryCondition(tc,dirichlet,inflowWithVelocityGiven,bcData,pBoundaryData,t,bcParams,grid);
//     }

    // assign dirichlet BC's on T here -- other BC's are done later
          if( assignTemperature )
          {
              FILE *& debugFile  =  parameters.dbase.get<FILE* >("debugFile");
              FILE *& pDebugFile =  parameters.dbase.get<FILE* >("pDebugFile");
              ForBoundary(side,axis)
              {
                  if( mg.boundaryCondition(side,axis)==inflowWithVelocityGiven )
                  {
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // This is an interface between domains
             // for now we only know about interfaces at no-slip walls: 
                          assert( mg.boundaryCondition(side,axis)==noSlipWall );
    	 // what about BC's applied at t=0 before the boundary data is set ??
    	 // if( parameters.dbase.get<int >("globalStepNumber") < 2 ) continue; // ********************* TEMP *****
    	 // if this is an iterface we should turn off the TZ forcing for the boundary condition since we want
    	 // to use the boundary data instead.
                          #ifdef USE_PPP
                              realSerialArray uLocal;  getLocalArrayWithGhostBoundaries(u,uLocal);
                          #else
                              const realSerialArray & uLocal = u;
                          #endif
               	 Index Ib1,Ib2,Ib3;
               	 getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
               	 bool ok = ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3);
               	 if( debug() & 4 )
               	 {
                 	   printP("insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                      		  side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   fprintf(pDebugFile,"insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                                 			       side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   if( pBoundaryData[side][axis]==NULL )
                 	   {
                   	     if( !ok )
                                      fprintf(pDebugFile," insBC:applyBC on T: pBoundaryData[side][axis] == NULL! \n");
                                  else
                                      fprintf(pDebugFile," insBC:applyBC on T: ERROR: pBoundaryData[side][axis] == NULL! \n");
                 	   }
                 	   else
                 	   {
    	     // RealArray & bd = *pBoundaryData[side][axis];
    	     // ::display(bd,"boundaryData",pDebugFile,"%8.2e ");
                 	   }
               	 }
                          assert( mixedCoeff(tc,side,axis,grid)!=0. || mixedNormalCoeff(tc,side,axis,grid)!=0. );
               	 if( ok && pBoundaryData[side][axis]==NULL )
               	 {
               // At t=0 when the initial conditions are being set-up (and initial conditions projected) there may
               // not be a boundaryData array created yet for the interface. We thus create one and fill in some default values
               // based on the current solution
                              RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                              bd=0.;
                              #ifdef USE_PPP
                   	     const RealArray & normal = mg.vertexBoundaryNormalArray(side,axis);
                              #else
                   	     const RealArray & normal = mg.vertexBoundaryNormal(side,axis);
                              #endif
                 	   real a0=mixedCoeff(tc,side,axis,grid);
                 	   real a1=mixedNormalCoeff(tc,side,axis,grid);
               // bd(Ib1,Ib2,Ib3,tc)=a0*uLocal(Ib1,Ib2,Ib3,tc);    // Assume that T.n=0 at the interface at t=0 -- could do better --
                              MappedGridOperators & op = *(u.getOperators());
                              Range N(tc,tc);
                              RealArray ux(Ib1,Ib2,Ib3,N), uy(Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::xDerivative,uLocal,ux,Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::yDerivative,uLocal,uy,Ib1,Ib2,Ib3,N);
                              if( mg.numberOfDimensions()==2 )
                 	   {
                                  bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                 	   else
                 	   {
                                  RealArray uz(Ib1,Ib2,Ib3,N);
                   	     op.derivative(MappedGridOperators::zDerivative,uLocal,uz,Ib1,Ib2,Ib3,N);
                   	     bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,2)*uz(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                              const bool twilightZoneFlow = parameters.dbase.get<bool >("twilightZoneFlow");
                 	   if( false && twilightZoneFlow ) //  *******************************************************
                 	   {
                                  fprintf(pDebugFile," insBC:applyBC on T: ********** Set TRUE value for bd at t=%8.2e ************\n",t);
                   	     const bool rectangular= mg.isRectangular() && !twilightZoneFlow;
                   	     realArray & x= mg.center();
        #ifdef USE_PPP
                   	     realSerialArray xLocal; 
                   	     if( !rectangular || twilightZoneFlow ) 
                     	       getLocalArrayWithGhostBoundaries(x,xLocal);
        #else
                   	     const realSerialArray & xLocal = x;
        #endif
                   	     OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
                   	     realSerialArray ue(Ib1,Ib2,Ib3), uex(Ib1,Ib2,Ib3), uey(Ib1,Ib2,Ib3); 
                   	     e.gd( ue ,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uex,xLocal,mg.numberOfDimensions(),rectangular,0,1,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uey,xLocal,mg.numberOfDimensions(),rectangular,0,0,1,0,Ib1,Ib2,Ib3,tc,t);
                   	     real a0=mixedCoeff(tc,side,axis,grid), a1=mixedNormalCoeff(tc,side,axis,grid);
                   	     if( mg.numberOfDimensions()==2 )
                   	     {
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                   	     else
                   	     {
                     	       realSerialArray uez(Ib1,Ib2,Ib3);
                     	       e.gd( uez,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,1,Ib1,Ib2,Ib3,tc,t);
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,2)*uez(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                 	   } // *******************************************
               	 }
               	 if( pBoundaryData[side][axis]!=NULL )
                 	   u.getOperators()->setTwilightZoneFlow( false );
               	 else
               	 {
                              if( t>0. || debug() & 4 )
                 	   {
                                  if( ok )
                             	       printP("$$$$ insBC:applyBC:INFO:interface but no boundaryData, t=%9.3e\n",t);
                 	   }
               	 }
                      }
                      if( debug() & 4 )
               	 printF("++++insBC: inflowWithVelocityGiven: (grid,side,axis)=(%i,%i,%i) : "
                    		"  BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
                    		"  BC: u: %3.2f*u+%3.2f*u.n=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
                    		grid,side,axis, 
                    		mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
                    		mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
                    		mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
                 	   );
    //        if( mixedNormalCoeff(tc,side,axis,grid)!=0. ) // coeff of T.n is non-zero
    //        {
    // 	 mixedBoundaryConditionOnTemperature=true;
    // 	 if( debug() & 4 )
    // 	   printF("++++insBC: inflowWithVelocityGiven:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
    // 		  "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
    //                   "  BC: u: %3.2f*u+%3.2f*u.n=%3.2f=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
    // 		  grid,side,axis, 
    //                   mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
    //                   mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
    //                   mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
    // 	     );
    //        }
                      if( mixedNormalCoeff(tc,side,axis,grid)==0. ) // coeff of T.n 
                      {
    	 // Dirichlet
               	 u.applyBoundaryCondition(tc,dirichlet,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
                                  				  bcParams,grid);
                      }
                      else
                      {
    	 // Mixed or Neumann -- 
                      }
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // reset TZ
               	 u.getOperators()->setTwilightZoneFlow( parameters.dbase.get<bool >("twilightZoneFlow") );
                      }
                  } // end if bc = inflowWithVelocityGiven
              } // end for boundary
       // ************ try this ********* 080909
       // u.updateGhostBoundaries(); // this is done in finish boundary conditions now
          } // end if assignTemperature

        if( pdeModel==InsParameters::twoPhaseFlowModel )
        {
            u.applyBoundaryCondition(twoPhaseFlowComponents,dirichlet,inflowWithVelocityGiven,bcData,
                                                              pBoundaryData,t,bcParams,grid);
        }
        
    }
    
  // assigned extended boundaries for 4th order:
    bcParams.extraInTangentialDirections= orderOfAccuracy==2 ? 0 : 2;

//   if( variableBoundaryData )
//   {
//     // boundaryData.display("insBC: variableBoundaryData=TRUE, boundaryData");
//     u.applyBoundaryCondition(V,dirichlet,  inflowWithVelocityGiven,bcData,pBoundaryData,t);
//     // u.applyBoundaryCondition(V,dirichlet,  inflowWithVelocityGiven,boundaryData,t);
//   }
//   else
//   {
//     getTimeDependentBoundaryConditions( t, grid); // ************ fix this *****
//     u.applyBoundaryCondition(V,dirichlet,  inflowWithVelocityGiven,bcData,t,Overture::defaultBoundaryConditionParameters(),grid);
//   }
    if( false )
    {
        display(u,sPrintF(buff,"u after inflowWithVelocityGiven, grid=%i, t=%e",grid,t),parameters.dbase.get<FILE* >("debugFile"),"%5.2f ");
    }
        
    if( assignSlipWall )
    {
        if( gridIsMoving )
            u.applyBoundaryCondition(V,normalComponent,slipWall,gridVelocity,t,bcParams);
        else
            u.applyBoundaryCondition(V,normalComponent,slipWall,0.,t,bcParams);
    }


  // old:  const  int nc=parameters.dbase.get<int >("numberOfComponents");
    
    if( assignNoSlipWall )
    {
        if( gridIsMoving )
            u.applyBoundaryCondition(V,dirichlet,noSlipWall,gridVelocity,t,bcParams);
        else
            u.applyBoundaryCondition(V,dirichlet,noSlipWall,bcData,t,bcParams,grid);

    // assign dirichlet BC's on T here -- other BC's are done later
          if( assignTemperature )
          {
              FILE *& debugFile  =  parameters.dbase.get<FILE* >("debugFile");
              FILE *& pDebugFile =  parameters.dbase.get<FILE* >("pDebugFile");
              ForBoundary(side,axis)
              {
                  if( mg.boundaryCondition(side,axis)==noSlipWall )
                  {
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // This is an interface between domains
             // for now we only know about interfaces at no-slip walls: 
                          assert( mg.boundaryCondition(side,axis)==noSlipWall );
    	 // what about BC's applied at t=0 before the boundary data is set ??
    	 // if( parameters.dbase.get<int >("globalStepNumber") < 2 ) continue; // ********************* TEMP *****
    	 // if this is an iterface we should turn off the TZ forcing for the boundary condition since we want
    	 // to use the boundary data instead.
                          #ifdef USE_PPP
                              realSerialArray uLocal;  getLocalArrayWithGhostBoundaries(u,uLocal);
                          #else
                              const realSerialArray & uLocal = u;
                          #endif
               	 Index Ib1,Ib2,Ib3;
               	 getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
               	 bool ok = ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3);
               	 if( debug() & 4 )
               	 {
                 	   printP("insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                      		  side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   fprintf(pDebugFile,"insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                                 			       side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   if( pBoundaryData[side][axis]==NULL )
                 	   {
                   	     if( !ok )
                                      fprintf(pDebugFile," insBC:applyBC on T: pBoundaryData[side][axis] == NULL! \n");
                                  else
                                      fprintf(pDebugFile," insBC:applyBC on T: ERROR: pBoundaryData[side][axis] == NULL! \n");
                 	   }
                 	   else
                 	   {
    	     // RealArray & bd = *pBoundaryData[side][axis];
    	     // ::display(bd,"boundaryData",pDebugFile,"%8.2e ");
                 	   }
               	 }
                          assert( mixedCoeff(tc,side,axis,grid)!=0. || mixedNormalCoeff(tc,side,axis,grid)!=0. );
               	 if( ok && pBoundaryData[side][axis]==NULL )
               	 {
               // At t=0 when the initial conditions are being set-up (and initial conditions projected) there may
               // not be a boundaryData array created yet for the interface. We thus create one and fill in some default values
               // based on the current solution
                              RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                              bd=0.;
                              #ifdef USE_PPP
                   	     const RealArray & normal = mg.vertexBoundaryNormalArray(side,axis);
                              #else
                   	     const RealArray & normal = mg.vertexBoundaryNormal(side,axis);
                              #endif
                 	   real a0=mixedCoeff(tc,side,axis,grid);
                 	   real a1=mixedNormalCoeff(tc,side,axis,grid);
               // bd(Ib1,Ib2,Ib3,tc)=a0*uLocal(Ib1,Ib2,Ib3,tc);    // Assume that T.n=0 at the interface at t=0 -- could do better --
                              MappedGridOperators & op = *(u.getOperators());
                              Range N(tc,tc);
                              RealArray ux(Ib1,Ib2,Ib3,N), uy(Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::xDerivative,uLocal,ux,Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::yDerivative,uLocal,uy,Ib1,Ib2,Ib3,N);
                              if( mg.numberOfDimensions()==2 )
                 	   {
                                  bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                 	   else
                 	   {
                                  RealArray uz(Ib1,Ib2,Ib3,N);
                   	     op.derivative(MappedGridOperators::zDerivative,uLocal,uz,Ib1,Ib2,Ib3,N);
                   	     bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,2)*uz(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                              const bool twilightZoneFlow = parameters.dbase.get<bool >("twilightZoneFlow");
                 	   if( false && twilightZoneFlow ) //  *******************************************************
                 	   {
                                  fprintf(pDebugFile," insBC:applyBC on T: ********** Set TRUE value for bd at t=%8.2e ************\n",t);
                   	     const bool rectangular= mg.isRectangular() && !twilightZoneFlow;
                   	     realArray & x= mg.center();
        #ifdef USE_PPP
                   	     realSerialArray xLocal; 
                   	     if( !rectangular || twilightZoneFlow ) 
                     	       getLocalArrayWithGhostBoundaries(x,xLocal);
        #else
                   	     const realSerialArray & xLocal = x;
        #endif
                   	     OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
                   	     realSerialArray ue(Ib1,Ib2,Ib3), uex(Ib1,Ib2,Ib3), uey(Ib1,Ib2,Ib3); 
                   	     e.gd( ue ,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uex,xLocal,mg.numberOfDimensions(),rectangular,0,1,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uey,xLocal,mg.numberOfDimensions(),rectangular,0,0,1,0,Ib1,Ib2,Ib3,tc,t);
                   	     real a0=mixedCoeff(tc,side,axis,grid), a1=mixedNormalCoeff(tc,side,axis,grid);
                   	     if( mg.numberOfDimensions()==2 )
                   	     {
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                   	     else
                   	     {
                     	       realSerialArray uez(Ib1,Ib2,Ib3);
                     	       e.gd( uez,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,1,Ib1,Ib2,Ib3,tc,t);
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,2)*uez(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                 	   } // *******************************************
               	 }
               	 if( pBoundaryData[side][axis]!=NULL )
                 	   u.getOperators()->setTwilightZoneFlow( false );
               	 else
               	 {
                              if( t>0. || debug() & 4 )
                 	   {
                                  if( ok )
                             	       printP("$$$$ insBC:applyBC:INFO:interface but no boundaryData, t=%9.3e\n",t);
                 	   }
               	 }
                      }
                      if( debug() & 4 )
               	 printF("++++insBC: noSlipWall: (grid,side,axis)=(%i,%i,%i) : "
                    		"  BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
                    		"  BC: u: %3.2f*u+%3.2f*u.n=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
                    		grid,side,axis, 
                    		mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
                    		mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
                    		mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
                 	   );
    //        if( mixedNormalCoeff(tc,side,axis,grid)!=0. ) // coeff of T.n is non-zero
    //        {
    // 	 mixedBoundaryConditionOnTemperature=true;
    // 	 if( debug() & 4 )
    // 	   printF("++++insBC: noSlipWall:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
    // 		  "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
    //                   "  BC: u: %3.2f*u+%3.2f*u.n=%3.2f=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
    // 		  grid,side,axis, 
    //                   mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
    //                   mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
    //                   mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
    // 	     );
    //        }
                      if( mixedNormalCoeff(tc,side,axis,grid)==0. ) // coeff of T.n 
                      {
    	 // Dirichlet
               	 u.applyBoundaryCondition(tc,dirichlet,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
                                  				  bcParams,grid);
                      }
                      else
                      {
    	 // Mixed or Neumann -- 
                      }
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // reset TZ
               	 u.getOperators()->setTwilightZoneFlow( parameters.dbase.get<bool >("twilightZoneFlow") );
                      }
                  } // end if bc = noSlipWall
              } // end for boundary
       // ************ try this ********* 080909
       // u.updateGhostBoundaries(); // this is done in finish boundary conditions now
          } // end if assignTemperature

    // display(u,"u after dirichlet noSlipWall",parameters.dbase.get<FILE* >("debugFile"));
    }
    
    bcParams.extraInTangentialDirections=0; // reset
    


    if( assignOutflow  && orderOfAccuracy==2 )
    {
    // outflow:
    // (1) extrapolate (u,v,w,p)    (default case)
    // (2) set alpha p + beta p.n =   (done in assignPressureRHS)

        if( parameters.dbase.get<int>("outflowOption")==1 ||
                parameters.dbase.get<int >("checkForInflowAtOutFlow")==2 )
        {
      // expect inflow at an outflow boundary -- use Neumann BC instead of extrapolating
      // printF("insBC: apply neumann BC at outflow\n");
            if( true )
            {
      	for( int n=0; n<numberOfDimensions; n++ )
      	{
        	  u.applyBoundaryCondition(uc+n,neumann,outflow,0.,t);
      	}
            }
            else
            {
                u.applyBoundaryCondition(V,neumann,outflow,0.,t);
            }
            
        }
        else
        {
      // default outflow BC's
            const int orderOfExtrapolation=extrapParams.orderOfExtrapolation;
            if( parameters.dbase.get<int >("orderOfExtrapolationForOutflow")>= 0 )
      	extrapParams.orderOfExtrapolation=parameters.dbase.get<int >("orderOfExtrapolationForOutflow");
            u.applyBoundaryCondition(Rt,extrapolate,outflow,0.,t,extrapParams);
            extrapParams.orderOfExtrapolation=orderOfExtrapolation;  // reset
        }
    }
    
    if( assignTractionFree )
    {
    // tractionFree:
    //   ** for now just apply a neumann BC **
        u.applyBoundaryCondition(Rt,neumann,tractionFree,0.,t);  
    // u.applyBoundaryCondition(V,extrapolate,tractionFree,0.,t);  
    }
    
  // **check for local inflow at an outflow boundary**
  // where( inflow ) give u.n=0
    Index I1,I2,I3;
        
    if( !parameters.dbase.get<bool >("twilightZoneFlow") &&  
            assignOutflow && orderOfAccuracy==2 && 
            parameters.dbase.get<int>("outflowOption")==0 && 
            parameters.dbase.get<int >("checkForInflowAtOutFlow")==1 )
    {
    // check for inflow at the outflow boundary    
        for( axis=0; axis<mg.numberOfDimensions(); axis++ )
        {
            for( side=Start; side<=End; side++ )
            {
      	if( mg.boundaryCondition(side,axis)==outflow )
      	{
        	  RealDistributedArray & normal  = mg.vertexBoundaryNormal(side,axis);  

        	  getGhostIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);
        	  intArray & mask = bcParams.mask();
        	  mask.redim(I1,I2,I3);   // mask lives on ghost line.
        	  getBoundaryIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);
        	  if( isRectangular )
        	  {
          	    mask = u(I1,I2,I3,uc+axis)*(2*side-1) <0.;
        	  }
        	  else
        	  {
          	    if( mg.numberOfDimensions()==2 )
          	    {
            	      mask = (u(I1,I2,I3,uc)*normal(I1,I2,I3,0)+
                  		      u(I1,I2,I3,vc)*normal(I1,I2,I3,1)) <0; 
          	    }
          	    else
          	    {
            	      mask = (u(I1,I2,I3,uc)*normal(I1,I2,I3,0)+
                  		      u(I1,I2,I3,vc)*normal(I1,I2,I3,1)+
                  		      u(I1,I2,I3,wc)*normal(I1,I2,I3,2)) <0; 
          	    }
        	  }
      	
        	  int count=sum(mask);
        	  if( count>0 )
        	  {
          	    if( debug() & 4 )
            	      printf("insBC: number of outflow points that are inflow = %i\n",count);
          	    bcParams.setUseMask(TRUE);
	    // u.applyBoundaryCondition(V,neumann,outflow,0.,t,bcParams);
          	    u.applyBoundaryCondition(V,neumann,BCTypes::boundary(side,axis),0.,t,bcParams);
          	    bcParams.setUseMask(FALSE);
        	  }
      	}
            }
        }
    }
    
    if( assignSlipWall && orderOfAccuracy==2 )
    {
    // finish slipWall
    // (2) vector symmetry (is this really true on a curved wall??)
    // (3) div(u)=0 (done further below)

        if( true ) // 061015: use this again to be consistent with implicit time stepping BC's
            u.applyBoundaryCondition(V,vectorSymmetry, slipWall,0.,t);   
        else
        { // use this 981130
            u.applyBoundaryCondition(V,extrapolate,    slipWall,0.,t);
            if( true )
            {
      	u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent0, slipWall,0.,t);
      	if( numberOfDimensions==3 )
        	  u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent1, slipWall,0.,t);
            }
        }

        if( assignTemperature )
        {
            u.applyBoundaryCondition(tc,neumann,slipWall,0.,t);
        }
        if( pdeModel==InsParameters::twoPhaseFlowModel )
        {
            u.applyBoundaryCondition(twoPhaseFlowComponents,neumann,slipWall,0.,t);
        }    
    }
    
  // Before we can apply a generalizedDivergence at a corner we need to first get some values
  // at all ghostpoints  -- therefore we first extrapolate all remaining BC's 

    if( assignInflowWithVelocityGiven && orderOfAccuracy==2 )
    {
    // inflowWithVelocityGiven:
    // (1) set (u,v,w)=
    // (2) extrapolate (u,v,w,p)
    // (3) set div(u)=0.   (done further below)
//  u.applyBoundaryCondition(V,dirichlet,  inflowWithVelocityGiven,inflowWithVelocityGivenData,t);
        u.applyBoundaryCondition(Rt,extrapolate,inflowWithVelocityGiven,0.,t);

          if( assignTemperature )
          {
              FILE *& debugFile  =  parameters.dbase.get<FILE* >("debugFile");
              FILE *& pDebugFile =  parameters.dbase.get<FILE* >("pDebugFile");
              ForBoundary(side,axis)
              {
                  if( mg.boundaryCondition(side,axis)==inflowWithVelocityGiven )
                  {
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // This is an interface between domains
             // for now we only know about interfaces at no-slip walls: 
                          assert( mg.boundaryCondition(side,axis)==noSlipWall );
    	 // what about BC's applied at t=0 before the boundary data is set ??
    	 // if( parameters.dbase.get<int >("globalStepNumber") < 2 ) continue; // ********************* TEMP *****
    	 // if this is an iterface we should turn off the TZ forcing for the boundary condition since we want
    	 // to use the boundary data instead.
                          #ifdef USE_PPP
                              realSerialArray uLocal;  getLocalArrayWithGhostBoundaries(u,uLocal);
                          #else
                              const realSerialArray & uLocal = u;
                          #endif
               	 Index Ib1,Ib2,Ib3;
               	 getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
               	 bool ok = ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3);
               	 if( debug() & 4 )
               	 {
                 	   printP("insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                      		  side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   fprintf(pDebugFile,"insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                                 			       side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   if( pBoundaryData[side][axis]==NULL )
                 	   {
                   	     if( !ok )
                                      fprintf(pDebugFile," insBC:applyBC on T: pBoundaryData[side][axis] == NULL! \n");
                                  else
                                      fprintf(pDebugFile," insBC:applyBC on T: ERROR: pBoundaryData[side][axis] == NULL! \n");
                 	   }
                 	   else
                 	   {
    	     // RealArray & bd = *pBoundaryData[side][axis];
    	     // ::display(bd,"boundaryData",pDebugFile,"%8.2e ");
                 	   }
               	 }
                          assert( mixedCoeff(tc,side,axis,grid)!=0. || mixedNormalCoeff(tc,side,axis,grid)!=0. );
               	 if( ok && pBoundaryData[side][axis]==NULL )
               	 {
               // At t=0 when the initial conditions are being set-up (and initial conditions projected) there may
               // not be a boundaryData array created yet for the interface. We thus create one and fill in some default values
               // based on the current solution
                              RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                              bd=0.;
                              #ifdef USE_PPP
                   	     const RealArray & normal = mg.vertexBoundaryNormalArray(side,axis);
                              #else
                   	     const RealArray & normal = mg.vertexBoundaryNormal(side,axis);
                              #endif
                 	   real a0=mixedCoeff(tc,side,axis,grid);
                 	   real a1=mixedNormalCoeff(tc,side,axis,grid);
               // bd(Ib1,Ib2,Ib3,tc)=a0*uLocal(Ib1,Ib2,Ib3,tc);    // Assume that T.n=0 at the interface at t=0 -- could do better --
                              MappedGridOperators & op = *(u.getOperators());
                              Range N(tc,tc);
                              RealArray ux(Ib1,Ib2,Ib3,N), uy(Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::xDerivative,uLocal,ux,Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::yDerivative,uLocal,uy,Ib1,Ib2,Ib3,N);
                              if( mg.numberOfDimensions()==2 )
                 	   {
                                  bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                 	   else
                 	   {
                                  RealArray uz(Ib1,Ib2,Ib3,N);
                   	     op.derivative(MappedGridOperators::zDerivative,uLocal,uz,Ib1,Ib2,Ib3,N);
                   	     bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,2)*uz(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                              const bool twilightZoneFlow = parameters.dbase.get<bool >("twilightZoneFlow");
                 	   if( false && twilightZoneFlow ) //  *******************************************************
                 	   {
                                  fprintf(pDebugFile," insBC:applyBC on T: ********** Set TRUE value for bd at t=%8.2e ************\n",t);
                   	     const bool rectangular= mg.isRectangular() && !twilightZoneFlow;
                   	     realArray & x= mg.center();
        #ifdef USE_PPP
                   	     realSerialArray xLocal; 
                   	     if( !rectangular || twilightZoneFlow ) 
                     	       getLocalArrayWithGhostBoundaries(x,xLocal);
        #else
                   	     const realSerialArray & xLocal = x;
        #endif
                   	     OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
                   	     realSerialArray ue(Ib1,Ib2,Ib3), uex(Ib1,Ib2,Ib3), uey(Ib1,Ib2,Ib3); 
                   	     e.gd( ue ,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uex,xLocal,mg.numberOfDimensions(),rectangular,0,1,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uey,xLocal,mg.numberOfDimensions(),rectangular,0,0,1,0,Ib1,Ib2,Ib3,tc,t);
                   	     real a0=mixedCoeff(tc,side,axis,grid), a1=mixedNormalCoeff(tc,side,axis,grid);
                   	     if( mg.numberOfDimensions()==2 )
                   	     {
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                   	     else
                   	     {
                     	       realSerialArray uez(Ib1,Ib2,Ib3);
                     	       e.gd( uez,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,1,Ib1,Ib2,Ib3,tc,t);
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,2)*uez(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                 	   } // *******************************************
               	 }
               	 if( pBoundaryData[side][axis]!=NULL )
                 	   u.getOperators()->setTwilightZoneFlow( false );
               	 else
               	 {
                              if( t>0. || debug() & 4 )
                 	   {
                                  if( ok )
                             	       printP("$$$$ insBC:applyBC:INFO:interface but no boundaryData, t=%9.3e\n",t);
                 	   }
               	 }
                      }
                      if( debug() & 4 )
               	 printF("++++insBC: inflowWithVelocityGiven: (grid,side,axis)=(%i,%i,%i) : "
                    		"  BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
                    		"  BC: u: %3.2f*u+%3.2f*u.n=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
                    		grid,side,axis, 
                    		mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
                    		mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
                    		mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
                 	   );
    //        if( mixedNormalCoeff(tc,side,axis,grid)!=0. ) // coeff of T.n is non-zero
    //        {
    // 	 mixedBoundaryConditionOnTemperature=true;
    // 	 if( debug() & 4 )
    // 	   printF("++++insBC: inflowWithVelocityGiven:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
    // 		  "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
    //                   "  BC: u: %3.2f*u+%3.2f*u.n=%3.2f=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
    // 		  grid,side,axis, 
    //                   mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
    //                   mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
    //                   mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
    // 	     );
    //        }
                      if( mixedNormalCoeff(tc,side,axis,grid)==0. ) // coeff of T.n 
                      {
    	 // Dirichlet
                      }
                      else
                      {
    	 // Mixed or Neumann -- 
    	 // Mixed BC or Neumann
               	 real a0=mixedCoeff(tc,side,axis,grid);
               	 real a1=mixedNormalCoeff(tc,side,axis,grid);
               	 bcParams.a.redim(3);
               	 if( a0==0. && a1==1. )
               	 {
                 	   if( debug() & 4 )
                   	     printF("++++insBC: inflowWithVelocityGiven:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : apply neumannBC\n",
                        		    grid,side,axis);
    //                 real b0=bcData(tc+2,side,axis,grid);
    // 		u.applyBoundaryCondition(tc,neumann,BCTypes::boundary(side,axis),b0,t); // b0 ignored??
                 	   bcParams.a(0)=a0;
                 	   bcParams.a(1)=a1;
                 	   bcParams.a(2)=mixedRHS(tc,side,axis,grid);  // this is not used -- this does not work
                 	   if( false )
                 	   {  // **** TEMP FIX ****
                   	     u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),0.,t);
                 	   }
                 	   else
                 	   {
                   	     u.applyBoundaryCondition(tc,neumann,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
                                      				      bcParams,grid);
                 	   }
               	 }
               	 else
               	 {
                 	   if( debug() & 4 )
                 	   {
                   	     fPrintF(pDebugFile,"++++insBC:inflowWithVelocityGiven:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
                        		    "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f (t=%8.2e)\n",
                         		     grid,side,axis,a0,a1,bcData(tc,side,axis,grid),t);
                 	   }
                 	   if( debug() & 4 )
                 	   {
                                  #ifndef USE_PPP
                        	      Index Ib1,Ib2,Ib3;
                    	      getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
                                    RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                    	      ::display(bd(Ib1,Ib2,Ib3,tc),"insBC:inflowWithVelocityGiven:T: RHS for mixed BC: bd(Ib1,Ib2,Ib3,tc)",pDebugFile,"%5.2f ");
                        	      Index Ig1,Ig2,Ig3;
    	      // getGhostIndex(mg.gridIndexRange(),side,axis,Ig1,Ig2,Ig3);
                    	      getIndex(mg.gridIndexRange(),Ig1,Ig2,Ig3,1);
                                    ::display(u(Ig1,Ig2,Ig3,tc),"insBC:inflowWithVelocityGiven:T: BEFORE mixed BC: u(I1,I2,I3,tc)",pDebugFile,"%5.2f ");
                                  #endif
                 	   }
                 	   bcParams.a(0)=a0;
                 	   bcParams.a(1)=a1;
                 	   bcParams.a(2)=mixedRHS(tc,side,axis,grid); 
                 	   if( false )
                 	   {  // **** TEMP FIX ****
                   	     u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),0.,t);
                 	   }
                 	   else
                 	   {
                   	     u.applyBoundaryCondition(tc,mixed,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
                                      				      bcParams,grid);
                 	   }
                 	   if( debug() & 4 )
                 	   {
                                  #ifndef USE_PPP
                        	      Index Ig1,Ig2,Ig3;
    	      // getGhostIndex(mg.gridIndexRange(),side,axis,Ig1,Ig2,Ig3);
                    	      getIndex(mg.gridIndexRange(),Ig1,Ig2,Ig3,1);
                                    ::display(u(Ig1,Ig2,Ig3,tc),"insBC:inflowWithVelocityGiven:T: AFTER mixed BC: u(I1,I2,I3,tc)",pDebugFile,"%5.2f ");
                                  #endif
                 	   }
               	 }
                      }
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // reset TZ
               	 u.getOperators()->setTwilightZoneFlow( parameters.dbase.get<bool >("twilightZoneFlow") );
                      }
                  } // end if bc = inflowWithVelocityGiven
              } // end for boundary
       // ************ try this ********* 080909
       // u.updateGhostBoundaries(); // this is done in finish boundary conditions now
          } // end if assignTemperature
    }
    
    if( assignNoSlipWall && orderOfAccuracy==2 )
    {
    // noSlipWall:
    // (1) set (u,v,w)=
    // (2) extrapolate (u,v,w,p)
    // (3) set div(u)=0. (done further below)

        u.applyBoundaryCondition(Rt,extrapolate,noSlipWall,0.,t);

    // extrapParams.dbase.get< >("orderOfExtrapolation")=4; // *****  why??
    // u.applyBoundaryCondition(V,extrapolate,noSlipWall,0.,t,extrapParams);

    // display(u,"u before mixed noSlipWall",parameters.dbase.get<FILE* >("debugFile"));

          if( assignTemperature )
          {
              FILE *& debugFile  =  parameters.dbase.get<FILE* >("debugFile");
              FILE *& pDebugFile =  parameters.dbase.get<FILE* >("pDebugFile");
              ForBoundary(side,axis)
              {
                  if( mg.boundaryCondition(side,axis)==noSlipWall )
                  {
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // This is an interface between domains
             // for now we only know about interfaces at no-slip walls: 
                          assert( mg.boundaryCondition(side,axis)==noSlipWall );
    	 // what about BC's applied at t=0 before the boundary data is set ??
    	 // if( parameters.dbase.get<int >("globalStepNumber") < 2 ) continue; // ********************* TEMP *****
    	 // if this is an iterface we should turn off the TZ forcing for the boundary condition since we want
    	 // to use the boundary data instead.
                          #ifdef USE_PPP
                              realSerialArray uLocal;  getLocalArrayWithGhostBoundaries(u,uLocal);
                          #else
                              const realSerialArray & uLocal = u;
                          #endif
               	 Index Ib1,Ib2,Ib3;
               	 getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
               	 bool ok = ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3);
               	 if( debug() & 4 )
               	 {
                 	   printP("insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                      		  side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   fprintf(pDebugFile,"insBC:applyBC: apply a mixed BC on the interface (side,axis,grid)=(%i,%i,%i) %f*T + %f*T.n \n",
                                 			       side,axis,grid,mixedCoeff(tc,side,axis,grid),mixedNormalCoeff(tc,side,axis,grid));
                 	   if( pBoundaryData[side][axis]==NULL )
                 	   {
                   	     if( !ok )
                                      fprintf(pDebugFile," insBC:applyBC on T: pBoundaryData[side][axis] == NULL! \n");
                                  else
                                      fprintf(pDebugFile," insBC:applyBC on T: ERROR: pBoundaryData[side][axis] == NULL! \n");
                 	   }
                 	   else
                 	   {
    	     // RealArray & bd = *pBoundaryData[side][axis];
    	     // ::display(bd,"boundaryData",pDebugFile,"%8.2e ");
                 	   }
               	 }
                          assert( mixedCoeff(tc,side,axis,grid)!=0. || mixedNormalCoeff(tc,side,axis,grid)!=0. );
               	 if( ok && pBoundaryData[side][axis]==NULL )
               	 {
               // At t=0 when the initial conditions are being set-up (and initial conditions projected) there may
               // not be a boundaryData array created yet for the interface. We thus create one and fill in some default values
               // based on the current solution
                              RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                              bd=0.;
                              #ifdef USE_PPP
                   	     const RealArray & normal = mg.vertexBoundaryNormalArray(side,axis);
                              #else
                   	     const RealArray & normal = mg.vertexBoundaryNormal(side,axis);
                              #endif
                 	   real a0=mixedCoeff(tc,side,axis,grid);
                 	   real a1=mixedNormalCoeff(tc,side,axis,grid);
               // bd(Ib1,Ib2,Ib3,tc)=a0*uLocal(Ib1,Ib2,Ib3,tc);    // Assume that T.n=0 at the interface at t=0 -- could do better --
                              MappedGridOperators & op = *(u.getOperators());
                              Range N(tc,tc);
                              RealArray ux(Ib1,Ib2,Ib3,N), uy(Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::xDerivative,uLocal,ux,Ib1,Ib2,Ib3,N);
                 	   op.derivative(MappedGridOperators::yDerivative,uLocal,uy,Ib1,Ib2,Ib3,N);
                              if( mg.numberOfDimensions()==2 )
                 	   {
                                  bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                 	   else
                 	   {
                                  RealArray uz(Ib1,Ib2,Ib3,N);
                   	     op.derivative(MappedGridOperators::zDerivative,uLocal,uz,Ib1,Ib2,Ib3,N);
                   	     bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*ux(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,1)*uy(Ib1,Ib2,Ib3,tc)+
                                    				    normal(Ib1,Ib2,Ib3,2)*uz(Ib1,Ib2,Ib3,tc)) + a0*uLocal(Ib1,Ib2,Ib3,tc);
                 	   }
                              const bool twilightZoneFlow = parameters.dbase.get<bool >("twilightZoneFlow");
                 	   if( false && twilightZoneFlow ) //  *******************************************************
                 	   {
                                  fprintf(pDebugFile," insBC:applyBC on T: ********** Set TRUE value for bd at t=%8.2e ************\n",t);
                   	     const bool rectangular= mg.isRectangular() && !twilightZoneFlow;
                   	     realArray & x= mg.center();
        #ifdef USE_PPP
                   	     realSerialArray xLocal; 
                   	     if( !rectangular || twilightZoneFlow ) 
                     	       getLocalArrayWithGhostBoundaries(x,xLocal);
        #else
                   	     const realSerialArray & xLocal = x;
        #endif
                   	     OGFunction & e = *(parameters.dbase.get<OGFunction* >("exactSolution"));
                   	     realSerialArray ue(Ib1,Ib2,Ib3), uex(Ib1,Ib2,Ib3), uey(Ib1,Ib2,Ib3); 
                   	     e.gd( ue ,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uex,xLocal,mg.numberOfDimensions(),rectangular,0,1,0,0,Ib1,Ib2,Ib3,tc,t);
                   	     e.gd( uey,xLocal,mg.numberOfDimensions(),rectangular,0,0,1,0,Ib1,Ib2,Ib3,tc,t);
                   	     real a0=mixedCoeff(tc,side,axis,grid), a1=mixedNormalCoeff(tc,side,axis,grid);
                   	     if( mg.numberOfDimensions()==2 )
                   	     {
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                   	     else
                   	     {
                     	       realSerialArray uez(Ib1,Ib2,Ib3);
                     	       e.gd( uez,xLocal,mg.numberOfDimensions(),rectangular,0,0,0,1,Ib1,Ib2,Ib3,tc,t);
                     	       bd(Ib1,Ib2,Ib3,tc)=a1*(normal(Ib1,Ib2,Ib3,0)*uex(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,1)*uey(Ib1,Ib2,Ib3)+
                                      				      normal(Ib1,Ib2,Ib3,2)*uez(Ib1,Ib2,Ib3)) + a0*ue(Ib1,Ib2,Ib3);
                   	     }
                 	   } // *******************************************
               	 }
               	 if( pBoundaryData[side][axis]!=NULL )
                 	   u.getOperators()->setTwilightZoneFlow( false );
               	 else
               	 {
                              if( t>0. || debug() & 4 )
                 	   {
                                  if( ok )
                             	       printP("$$$$ insBC:applyBC:INFO:interface but no boundaryData, t=%9.3e\n",t);
                 	   }
               	 }
                      }
                      if( debug() & 4 )
               	 printF("++++insBC: noSlipWall: (grid,side,axis)=(%i,%i,%i) : "
                    		"  BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
                    		"  BC: u: %3.2f*u+%3.2f*u.n=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
                    		grid,side,axis, 
                    		mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
                    		mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
                    		mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
                 	   );
    //        if( mixedNormalCoeff(tc,side,axis,grid)!=0. ) // coeff of T.n is non-zero
    //        {
    // 	 mixedBoundaryConditionOnTemperature=true;
    // 	 if( debug() & 4 )
    // 	   printF("++++insBC: noSlipWall:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
    // 		  "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f,  \n"
    //                   "  BC: u: %3.2f*u+%3.2f*u.n=%3.2f=%3.2f,  v: %3.2f*v+%3.2f*v.n=%3.2f \n",
    // 		  grid,side,axis, 
    //                   mixedCoeff(tc,side,axis,grid), mixedNormalCoeff(tc,side,axis,grid), mixedRHS(tc,side,axis,grid), 
    //                   mixedCoeff(uc,side,axis,grid), mixedNormalCoeff(uc,side,axis,grid), mixedRHS(uc,side,axis,grid), 
    //                   mixedCoeff(vc,side,axis,grid), mixedNormalCoeff(vc,side,axis,grid), mixedRHS(vc,side,axis,grid)
    // 	     );
    //        }
                      if( mixedNormalCoeff(tc,side,axis,grid)==0. ) // coeff of T.n 
                      {
    	 // Dirichlet
                      }
                      else
                      {
    	 // Mixed or Neumann -- 
    	 // Mixed BC or Neumann
               	 real a0=mixedCoeff(tc,side,axis,grid);
               	 real a1=mixedNormalCoeff(tc,side,axis,grid);
               	 bcParams.a.redim(3);
               	 if( a0==0. && a1==1. )
               	 {
                 	   if( debug() & 4 )
                   	     printF("++++insBC: noSlipWall:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : apply neumannBC\n",
                        		    grid,side,axis);
    //                 real b0=bcData(tc+2,side,axis,grid);
    // 		u.applyBoundaryCondition(tc,neumann,BCTypes::boundary(side,axis),b0,t); // b0 ignored??
                 	   bcParams.a(0)=a0;
                 	   bcParams.a(1)=a1;
                 	   bcParams.a(2)=mixedRHS(tc,side,axis,grid);  // this is not used -- this does not work
                 	   if( false )
                 	   {  // **** TEMP FIX ****
                   	     u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),0.,t);
                 	   }
                 	   else
                 	   {
                   	     u.applyBoundaryCondition(tc,neumann,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
                                      				      bcParams,grid);
                 	   }
               	 }
               	 else
               	 {
                 	   if( debug() & 4 )
                 	   {
                   	     fPrintF(pDebugFile,"++++insBC:noSlipWall:adiabaticWall: (grid,side,axis)=(%i,%i,%i) : "
                        		    "Mixed BC for T: %3.2f*T+%3.2f*T.n=%3.2f (t=%8.2e)\n",
                         		     grid,side,axis,a0,a1,bcData(tc,side,axis,grid),t);
                 	   }
                 	   if( debug() & 4 )
                 	   {
                                  #ifndef USE_PPP
                        	      Index Ib1,Ib2,Ib3;
                    	      getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
                                    RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
                    	      ::display(bd(Ib1,Ib2,Ib3,tc),"insBC:noSlipWall:T: RHS for mixed BC: bd(Ib1,Ib2,Ib3,tc)",pDebugFile,"%5.2f ");
                        	      Index Ig1,Ig2,Ig3;
    	      // getGhostIndex(mg.gridIndexRange(),side,axis,Ig1,Ig2,Ig3);
                    	      getIndex(mg.gridIndexRange(),Ig1,Ig2,Ig3,1);
                                    ::display(u(Ig1,Ig2,Ig3,tc),"insBC:noSlipWall:T: BEFORE mixed BC: u(I1,I2,I3,tc)",pDebugFile,"%5.2f ");
                                  #endif
                 	   }
                 	   bcParams.a(0)=a0;
                 	   bcParams.a(1)=a1;
                 	   bcParams.a(2)=mixedRHS(tc,side,axis,grid); 
                 	   if( false )
                 	   {  // **** TEMP FIX ****
                   	     u.applyBoundaryCondition(tc,extrapolate,BCTypes::boundary(side,axis),0.,t);
                 	   }
                 	   else
                 	   {
                   	     u.applyBoundaryCondition(tc,mixed,BCTypes::boundary(side,axis),bcData,pBoundaryData,t,
                                      				      bcParams,grid);
                 	   }
                 	   if( debug() & 4 )
                 	   {
                                  #ifndef USE_PPP
                        	      Index Ig1,Ig2,Ig3;
    	      // getGhostIndex(mg.gridIndexRange(),side,axis,Ig1,Ig2,Ig3);
                    	      getIndex(mg.gridIndexRange(),Ig1,Ig2,Ig3,1);
                                    ::display(u(Ig1,Ig2,Ig3,tc),"insBC:noSlipWall:T: AFTER mixed BC: u(I1,I2,I3,tc)",pDebugFile,"%5.2f ");
                                  #endif
                 	   }
               	 }
                      }
                      if( interfaceType(side,axis,grid)!=Parameters::noInterface )
                      { // reset TZ
               	 u.getOperators()->setTwilightZoneFlow( parameters.dbase.get<bool >("twilightZoneFlow") );
                      }
                  } // end if bc = noSlipWall
              } // end for boundary
       // ************ try this ********* 080909
       // u.updateGhostBoundaries(); // this is done in finish boundary conditions now
          } // end if assignTemperature

    // display(u,"u after mixed noSlipWall",parameters.dbase.get<FILE* >("debugFile"));
    }


    if( parameters.isAxisymmetric() && assignAxisymmetric )
    {
    // cylindrically symmetric BC: v=v.yy=u.y=0
        u.applyBoundaryCondition(vc,dirichlet,axisymmetric,0.,t);
        u.applyBoundaryCondition(uc,neumann,axisymmetric,0.,t);
        extrapParams.lineToAssign=1;
        extrapParams.orderOfExtrapolation=2;

    // Extrap to higher order here needed for some reason -- this is fixed below with the BC u.x+2*u.y=0
        extrapParams.orderOfExtrapolation=4; // **** test this ****

        u.applyBoundaryCondition(vc,extrapolate,axisymmetric,0.,t,extrapParams);
    
        if( assignTemperature )
        {
            u.applyBoundaryCondition(tc,neumann,axisymmetric,0.,t);
        }
        if( pdeModel==InsParameters::twoPhaseFlowModel )
        {
            u.applyBoundaryCondition(twoPhaseFlowComponents,neumann,axisymmetric,0.,t);
        }    

    }
    if( assignSymmetry )
    {
    // symmetry BC:
        u.applyBoundaryCondition(V,vectorSymmetry,symmetry,0.,t);   
        if( assignTemperature )
        {
            u.applyBoundaryCondition(tc,BCTypes::evenSymmetry,symmetry,0.,t);
        }
        if( pdeModel==InsParameters::twoPhaseFlowModel )
        {
            u.applyBoundaryCondition(twoPhaseFlowComponents,BCTypes::evenSymmetry,symmetry,0.,t);
        }    
    }

    
  // *wdh* 000929 : also need to update periodic boundaries here
    u.periodicUpdate();


    checkArrayIDs(" insBC: before generalizedDivergence"); 

    if( orderOfAccuracy==2 )
    {
        if( assignInflowWithVelocityGiven )
        {
      // *note* that when this condition is applied on adjacent boundaries the ghost points next to corners
      //  are assigned in a certain order and some symmetry will be lost.
            u.applyBoundaryCondition(V,generalizedDivergence,inflowWithVelocityGiven,0.,t);
        }
    
        if( !parameters.isAxisymmetric() )
        {
            if( assignNoSlipWall )
      	u.applyBoundaryCondition(V,generalizedDivergence,noSlipWall,0.,t);
        }
        else
        {
      // div(u) = u.x + v.y + v/y = 0
      //   For y>0 and noSlipWall (u=v=0) --> u.x + v.y = 0

            if( assignNoSlipWall )
      	u.applyBoundaryCondition(V,generalizedDivergence,noSlipWall,0.,t);

            if( assignAxisymmetric )
            {
	// BC u.x + v.y + v/y =0  at y=0: u.x + 2 v.y =0
      	bcParams.a.redim(3);
      	bcParams.a(0)=1.;
      	bcParams.a(1)=2.;
      	bcParams.a(2)=0.;
      	u.applyBoundaryCondition(V,generalizedDivergence,axisymmetric,0.,t,bcParams);
            }
        }
    }
    
  // display(u,sPrintF(buff,"u after generalized divergence insBC grid=%i",grid),parameters.dbase.get<FILE* >("debugFile"));

    if( (true || numberOfDimensions!=3)  &&   // **** for now turn this off in 3D -- problesm outside interp pts*******
            applyDivergenceBoundaryCondition && assignSlipWall && orderOfAccuracy==2 ) 
    {
    // on a slip wall we need to extrapolate points that lie outside 
    // interpolation pts on the bndry. (*wdh* 061015)
        u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent0, slipWall,0.,t);
        if( numberOfDimensions==3 )
            u.applyBoundaryCondition(V,BCTypes::normalDerivativeOfTangentialComponent1, slipWall,0.,t);
        
        u.applyBoundaryCondition(V,generalizedDivergence,slipWall,0.,t); 
    }
    
    if( assignInflowWithPressureAndTangentialVelocityGiven && orderOfAccuracy==2 )
    {
    //  inflowWithPressureAndTangentialVelocityGiven
    //     give tangetial velocity = 0 
    //     extrapolate (u,v,w)
    //     set div(u)=0
        u.applyBoundaryCondition(V,tangentialComponent,inflowWithPressureAndTangentialVelocityGiven,0.,t);
        if( assignTemperature )
        {
            u.applyBoundaryCondition(tc,dirichlet,inflowWithVelocityGiven,bcData,pBoundaryData,t,bcParams,grid);
        }
        if( pdeModel==InsParameters::twoPhaseFlowModel )
        {
            u.applyBoundaryCondition(twoPhaseFlowComponents,dirichlet,inflowWithVelocityGiven,bcData,
                                                              pBoundaryData,t,bcParams,grid);
        }    

        u.applyBoundaryCondition(Rt,extrapolate,inflowWithPressureAndTangentialVelocityGiven,0.,t);

        u.applyBoundaryCondition(V,generalizedDivergence,inflowWithPressureAndTangentialVelocityGiven,0.,t);
    }
    
  // **** if we do this we probably don't have to check for inflow points at outflow.
  // *** turn off for stretched c-grid  
  // 090221 -- SS solver can work better with div bc at outflow
  // if( applyDivergenceBoundaryCondition && orderOfAccuracy==2 )
    if( applyDivergenceBoundaryConditionAtOutflow && orderOfAccuracy==2 )
    {
        if( assignOutflow )
        {
            
       // 080909 -- turn this off for now : this BC can result in a boundary layer in div(u) at outflow
       //           which may actually increase div(u) overall (solid-fuel rods example). 
      // 090221 -- turn this back on
            if( true &&  
                    parameters.dbase.get<int>("outflowOption")==0 && !parameters.isAxisymmetric() )
            {
        // printF("insBC: apply generalizedDivergence BC at outflow\n");
      	u.applyBoundaryCondition(V,generalizedDivergence,outflow,0.,t); // ****wdh***** 990827
            }
            else
            {
	// div(u) = u.x + v.y + v/y = 0

        // *** For now do nothing in this case ***
//  	bcParams.a.redim(3);
//  	bcParams.a(0)=1.;
//  	bcParams.a(1)=2.;
//  	bcParams.a(2)=0.;
//  	u.applyBoundaryCondition(V,generalizedDivergence,axisymmetric,0.,t,bcParams);

            }
        }

        if( assignTractionFree )
            u.applyBoundaryCondition(V,generalizedDivergence,tractionFree,0.,t); 
    }
    
    checkArrayIDs(" insBC: after generalizedDivergence"); 


  // Boundary conditions for the passive scalar.
    if( parameters.dbase.get<bool >("advectPassiveScalar") )
    {
        const int sc = parameters.dbase.get<int >("sc");
        if( assignInflowWithVelocityGiven )
        {
            u.applyBoundaryCondition(sc,dirichlet,inflowWithVelocityGiven,bcData,pBoundaryData,t,
                         			       Overture::defaultBoundaryConditionParameters(),grid);
            u.applyBoundaryCondition(sc,extrapolate,inflowWithVelocityGiven,0.,t);
        }
        if( assignOutflow )
            u.applyBoundaryCondition(sc,extrapolate,outflow,0.,t);

        if( assignNoSlipWall )
            u.applyBoundaryCondition(sc,neumann,noSlipWall,bcData,pBoundaryData,t,
                         			       Overture::defaultBoundaryConditionParameters(),grid);
        if( assignSlipWall )
            u.applyBoundaryCondition(sc,neumann,slipWall,bcData,pBoundaryData,t,
                         			       Overture::defaultBoundaryConditionParameters(),grid);

    // ** u.applyBoundaryCondition(sc,extrapolate,allBoundaries,0.,t);
    }

  // if( true ) return 0; // * ok ---


  // extrapolate the neighbours of interpolation points -- these values are used
  // by the fourth-order artificial viscosity 

    if( parameters.dbase.get<int >("orderOfAccuracy")==2 )
    {
        const int discretizationWidth = mg.discretizationWidth(0);
    // if( discretizationWidth!=3 ) printf(" INSBC: discretizationWidth=%i\n",discretizationWidth);
        
        if(  discretizationWidth<5 &&
                  parameters.dbase.get<bool >("useFourthOrderArtificialDiffusion") && (parameters.dbase.get<real >("ad41")!=0. || parameters.dbase.get<real >("ad42")!=0.)  )
        { // double check
            assert( numberOfGhostPointsNeeded>=2 );
        }
        
        if( discretizationWidth<5 && 
                numberOfGhostPointsNeeded>=2 )
        {
            extrapParams.ghostLineToAssign=2;
            extrapParams.orderOfExtrapolation=parameters.dbase.get<int >("orderOfAccuracy")+1; // 3;
            u.applyBoundaryCondition(Rt,extrapolate,allBoundaries,0.,t,extrapParams);

            if( parameters.dbase.get<bool >("useImplicitFourthArtificialDiffusion") )
                extrapParams.orderOfExtrapolation=3; // parameters.dbase.get<int >("orderOfAccuracy"); // 3;
            else
                extrapParams.orderOfExtrapolation=3;
            
            assert( parameters.dbase.get<int >("extrapolateInterpolationNeighbours") );  // consistency check
            u.applyBoundaryCondition(Rt,BCTypes::extrapolateInterpolationNeighbours,allBoundaries,0.,t,extrapParams);
        }
        else if( timeSteppingMethod==Parameters::rKutta )
        {
      // semi-implicit method needs du/dt at interpolation points.
            assert( parameters.dbase.get<int >("extrapolateInterpolationNeighbours") );  // consistency check
            u.applyBoundaryCondition(Rt,BCTypes::extrapolateInterpolationNeighbours);
        }
        else
        {
            assert( !parameters.dbase.get<int >("extrapolateInterpolationNeighbours") );  // consistency check
        }
        
    }
    
      if( orderOfAccuracy==4 )
    {
    // apply BC's for fourth-order accuracy

        if( false )
        {
            bcParams.lineToAssign=1;
            u.applyBoundaryCondition(V,dirichlet,allBoundaries,0.,t,bcParams);
            bcParams.lineToAssign=2;
            u.applyBoundaryCondition(V,dirichlet,allBoundaries,0.,t,bcParams);
            bcParams.lineToAssign=1;
        }
        else
        {
      // we need initial guesses for all ghost points -- this is done elsewhere --

//         extrapParams.dbase.get< >("ghostLineToAssign")=1;
//         extrapParams.dbase.get< >("orderOfExtrapolation")=parameters.dbase.get<int >("orderOfAccuracy")+1;
//         u.applyBoundaryCondition(V,extrapolate,allBoundaries,0.,t,extrapParams);

//         extrapParams.dbase.get< >("ghostLineToAssign")=2;
//         extrapParams.dbase.get< >("orderOfExtrapolation")=parameters.dbase.get<int >("orderOfAccuracy")+1; // 3;
//         u.applyBoundaryCondition(V,extrapolate,allBoundaries,0.,t,extrapParams);


      // ************* fix this ***************************************************************
//      mg.update(MappedGrid::THEmask | MappedGrid::THEcenter | MappedGrid::THEvertex |
//	       MappedGrid::THEinverseVertexDerivative | MappedGrid::THEvertexBoundaryNormal );

// ----
            if( false )
            {
        // **** for debugging ***** set p at corners

      	const realArray & vertex = mg.vertex();
      	OGFunction & e = *parameters.dbase.get<OGFunction* >("exactSolution");
      	int side,axis;
      	for( axis=0; axis<numberOfDimensions; axis++ )
      	{
        	  for( side=0; side<=1; side++ )
        	  {
          	    if( mg.boundaryCondition(side,axis)>0 )
          	    {
            	      int iv[3]={0,0,0};// 
            	      int is[3]={0,0,0};// 
            	      int axisp1 = (axis+1)%numberOfDimensions;
            	      for( int side2=0; side2<=1; side2++ )
            	      {
            		if( mg.boundaryCondition(side2,axisp1)>0 )
            		{
		  // *** corner points
              		  iv[axis]=mg.gridIndexRange()(side,axis);
              		  iv[axisp1]=mg.gridIndexRange()(side2,axisp1);
            	      
              		  is[axis]  =1-2*side;
              		  is[axisp1]=1-2*side2;
            		
                  // set p at the corners
              		  for( int ks2=-2; ks2<=2; ks2++ )
              		  {
                    		    for( int ks1=-2; ks1<=2; ks1++ )
                    		    {
                      		      int i1=iv[0]-ks1*is[0], i2=iv[1]-ks2*is[1], i3=iv[2]-is[2];
                  		      real x0=vertex(i1,i2,i3,0);
                  		      real y0=vertex(i1,i2,i3,1);
                  		      real z0=0.;
                  		      u(i1,i2,i3,parameters.dbase.get<int >("pc"))=e(x0,y0,z0,parameters.dbase.get<int >("pc"),t);
//  		      for( int c=uc; c<uc+numberOfDimensions; c++ )
//  			u(i1,i2,i3,c)=e(x0,y0,z0,c,t);
              		  
                		    }
              		  }
                                    iv[0]=iv[1]=0;

            		}
            	      }
          	    
          	    }
        	  }
      	}
            }
// -------------

            if( assignSlipWall )
            {
        // On a slip wall use vector symmetry on both ghost lines
                bcParams.ghostLineToAssign=1;
      	u.applyBoundaryCondition(V,vectorSymmetry, slipWall,0.,t);   
      	if( assignTemperature )
        	  u.applyBoundaryCondition(tc,BCTypes::evenSymmetry, slipWall,0.,t); 
                bcParams.ghostLineToAssign=2;
      	u.applyBoundaryCondition(V,vectorSymmetry, slipWall,0.,t,bcParams);   
                bcParams.ghostLineToAssign=1;
      	if( assignTemperature )
        	  u.applyBoundaryCondition(tc,BCTypes::evenSymmetry,slipWall,0.,t,bcParams);
            }
            

            IntegerArray ipar(20);  // fix this 
            RealArray rpar(20);
            bool useWhereMask=false;
            
            ipar(0) = parameters.dbase.get<int >("debug");
            ipar(1) = parameters.dbase.get<int >("pc");
            ipar(2) = parameters.dbase.get<int >("uc");
            ipar(3) = parameters.dbase.get<int >("vc");
            ipar(4) = parameters.dbase.get<int >("wc");
            ipar(5) = grid;
            ipar(6) = parameters.dbase.get<int >("orderOfAccuracy");
            ipar(7) = parameters.gridIsMoving(grid);
            ipar(8) = useWhereMask;
            ipar(9) = parameters.getGridIsImplicit(grid);
            ipar(10)= parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod");
            ipar(11)= parameters.dbase.get<Parameters::ImplicitOption >("implicitOption");
            ipar(12)= parameters.isAxisymmetric();
            ipar(13)= parameters.dbase.get<bool >("twilightZoneFlow");

            real dx,dy,dz;
            rpar(0) = t;
            rpar(1) = dx;
            rpar(2) = dy;
            rpar(3) = dz;
            rpar(4) = parameters.dbase.get<real >("nu");
            rpar(5) = getSignForJacobian(mg); // mg.mapping().getSignForJacobian();
            rpar(6) = parameters.dbase.get<real >("anl");

      // printf(" applyFourthOrderBoundaryConditions at t=%e\n",t);
            assert( parameters.dbase.get<Insbc4WorkSpace* >("bc4workSpacePointer")!=NULL );
            
            if( assignTemperature )
            {
      	Overture::abort("insBC:ERROR: 4th-order BC's with Temperature not finished");
            }
            

            applyFourthOrderBoundaryConditions( u,t,grid,*parameters.dbase.get<Insbc4WorkSpace* >("bc4workSpacePointer"),ipar,rpar,
                                            *parameters.dbase.get<OGFunction* >("exactSolution"),parameters );

      // **** assign symmetry conditions on slip walls at corner points too
      //     *** for now reassign all points on the extended region
            if( assignSlipWall )
            {
        // On a slip wall use vector symmetry on both ghost lines
                bcParams.ghostLineToAssign=1;
                bcParams.extraInTangentialDirections=2; // include 2 ghost points
      	u.applyBoundaryCondition(V,vectorSymmetry, slipWall,0.,t,bcParams);   
      	if( assignTemperature )
        	  u.applyBoundaryCondition(tc,BCTypes::evenSymmetry, slipWall,0.,t,bcParams); 
                bcParams.ghostLineToAssign=2;
      	u.applyBoundaryCondition(V,vectorSymmetry, slipWall,0.,t,bcParams);   
      	if( assignTemperature )
        	  u.applyBoundaryCondition(tc,BCTypes::evenSymmetry, slipWall,0.,t,bcParams); 
                bcParams.ghostLineToAssign=1;
      	bcParams.extraInTangentialDirections=0; // reset
            }


      // **** for now assign BC along extended boundaries ****
            if( false )
            {
      	OGFunction & e = *parameters.dbase.get<OGFunction* >("exactSolution");
      	int side,axis;
      	for( axis=0; axis<numberOfDimensions; axis++ )
      	{
        	  for( side=0; side<=1; side++ )
        	  {
          	    if( mg.boundaryCondition(side,axis)>0 )
          	    {
            	      getBoundaryIndex(mg.gridIndexRange(),side,axis,I1,I2,I3,2);  
            	      u(I1,I2,I3,V)=e(mg,I1,I2,I3,V,t);
          	    
          	    }
        	  }
      	}
            }


        }
        
        
    }
    

//   if( parameters.dbase.get<Parameters::TurbulenceModel >("turbulenceModel")==Parameters::kEpsilon )
//   {
//     // here are some fake k-epsilon equations
//     Range KE(parameters.dbase.get<int >("kc"),parameters.dbase.get<int >("epsc"));
//     u.applyBoundaryCondition(KE,dirichlet,allBoundaries,0.,t);
//     u.applyBoundaryCondition(KE,extrapolate,allBoundaries,0.,t);
//   }

  // apply turbulence model boundary conditions
    turbulenceModelBoundaryConditions(t,u,parameters,grid,pBoundaryData);

  // display(u,sPrintF(buff,"u at end of insBC grid=%i",grid),parameters.dbase.get<FILE* >("debugFile"));

  // update corners and periodic edges
  // *** not here  u.finishBoundaryConditions();

    checkArrayIDs(" insBC: end"); 


    if( parameters.dbase.get<int >("orderOfAccuracy")==2 ) // ** moved from generic applyBC - 060907
        u.finishBoundaryConditions();

    parameters.dbase.get<RealArray>("timing")(Parameters::timeForBoundaryConditions)+=getCPU()-time0;
    return 0;
}




