*
* This file is included by other command files to define a Cgcns domain
*
* The following parameters should be set before including this file: 
*   $domainName : name of the domain to assign
*   $solverName : name given to the domain (e.g. "fluid")
*   $cnsVariation : jameson, godunov, nonconservative
*   $tz, $degreeSpace, $degreeTime : 
*   $ts
*   $mu, $kThermal, $ktc
*   $ic : specify initial condition commands
*   $bc : specify boundary condition commands
*   $cnsGodunovOrder : 1 or 2
*   $gravity : 
*   $adCns : linear artificial diffusion coefficient for the Godunov method
*   $moveCmds : commands for moving grids
*   $extraCmds : extra commands
* 
if( $mu eq "" ){ $mu=0.; }
if( $kThermal eq "" ){ $kThermal=0.; }
if( $ktc eq "" ){ $ktc=0.; }
if( $adCns eq "" ){ $adCns=0.; }
if( $boundaryPressureOffset eq "" ){ $boundaryPressureOffset=0.; }
if( $tz eq "" ){ $tz="turn off twilight zone"; }
if( $moveCmds eq "" ){ $moveCmds="*"; }
if( $extraCmds eq "" ){ $extraCmds="*"; }
if( $cnsGodunovOrder eq "" ){ $cnsGodunovOrder=2;}
if( $degreex eq "" ){ $degreex=2;}
if( $degreet eq "" ){ $degreet=2;}
if( $cnsVariation eq "" ){ $cnsVariation="godunov";}  #
* ------- start new domain ----------
setup $domainName
 set solver Cgcns
 solver name $solverName
 solver parameters
* 
if( $cnsVariation eq "godunov" ){ $pdeVariation="compressible Navier Stokes (Godunov)"; }
if( $cnsVariation eq "jameson" ){ $pdeVariation="compressible Navier Stokes (Jameson)"; }   #
if( $cnsVariation eq "nonconservative" ){ $pdeVariation="compressible Navier Stokes (non-conservative)";}  #
  $pdeVariation
*   compressible Navier Stokes (Godunov)  
*   compressible Navier Stokes (Jameson)
  exit
*
$cnsts = $ts;
if( $cnsVariation eq "godunov" ){ $cnsts="forward Euler"; } # godunov always used FE
$cnsts 
* 
  $tz
  degree in space $degreex
  degree in time $degreet
  frequencies (x,y,z,t)   $fx $fy $fz $ft
*
  reduce interpolation width
    2
* 
  pde parameters
    mu $mu
    kThermal $kThermal
    thermal conductivity $ktc
    gravity
      $gravity
    OBPDE:artificial diffusion $adCns $adCns $adCns $adCns $adCns $adCns $adCns
    * offset for the pressure when computing the force on the boundary 
    OBPDE:boundary pressure offset $boundaryPressureOffset
    OBPDE:Godunov order of accuracy $cnsGodunovOrder
* pause
  done
$setAxi = $axisymmetric ? "turn on axisymmetric flow" : "*";
$setAxi
$setAxi="";
* 
  $moveCmds
* 
*
  boundary conditions
    $bc 
*    Annulus(0,1)=noSlipWall uniform(u=.0,T=300.)
*    square(0,0)=subSonicInflow uniform(r=1.,u=$u0,T=300.)
*    square(1,0)=subSonicOutflow mixedDerivative(1.*t+1.*t.n=300.)
*    square(0,1)=slipWall
*    square(1,1)=slipWall
    done
  debug $debug
  initial conditions
  if( $tz eq "turn off twilight zone" && $ic eq "" ){ $ic="uniform flow\n" . "r=1. u=$u0 T=300."; }elsif( $ic eq "" ){ $ic="*";}
    $ic
* pause
  continue
*
  $extraCmds
* 
 continue
done
