#include "Cgmp.h"
#include "ParallelUtility.h"
#include "App.h"
#include "Interface.h"

#define FOR_4D(i1,i2,i3,i4,I1,I2,I3,I4) \
int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase(),  I4Base =I4.getBase();  \
int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(), I4Bound=I4.getBound(); \
for(i4=I4Base; i4<=I4Bound; i4++) \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

#define FOR_4(i1,i2,i3,i4,I1,I2,I3,I4) \
I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase(),  I4Base =I4.getBase();  \
I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(), I4Bound=I4.getBound(); \
for(i4=I4Base; i4<=I4Bound; i4++) \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

#define FOR_3D(i1,i2,i3,I1,I2,I3) \
int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

#define FOR_3(i1,i2,i3,I1,I2,I3) \
I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

#define FOR_3IJD(i1,i2,i3,I1,I2,I3,j1,j2,j3,J1,J2,J3) \
int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
int J1Base =J1.getBase(),   J2Base =J2.getBase(),  J3Base =J3.getBase();  \
for(i3=I3Base,j3=J3Base; i3<=I3Bound; i3++,j3++) \
for(i2=I2Base,j2=J2Base; i2<=I2Bound; i2++,j2++) \
for(i1=I1Base,j1=J1Base; i1<=I1Bound; i1++,j1++)


#define FOR_4IJD(i1,i2,i3,i4,I1,I2,I3,I4,j1,j2,j3,j4,J1,J2,J3,J4) \
int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase(),  I4Base =I4.getBase();  \
int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(), I4Bound=I4.getBound(); \
int J1Base =J1.getBase(),   J2Base =J2.getBase(),  J3Base =J3.getBase(),  J4Base =J4.getBase();  \
for(i4=I4Base,j4=J4Base; i4<=I4Bound; i4++,j4++) \
for(i3=I3Base,j3=J3Base; i3<=I3Bound; i3++,j3++) \
for(i2=I2Base,j2=J2Base; i2<=I2Bound; i2++,j2++) \
for(i1=I1Base,j1=J1Base; i1<=I1Bound; i1++,j1++)

#define FOR_4IJ(i1,i2,i3,i4,I1,I2,I3,I4,j1,j2,j3,j4,J1,J2,J3,J4) \
I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase(),  I4Base =I4.getBase();  \
I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(), I4Bound=I4.getBound(); \
J1Base =J1.getBase(),   J2Base =J2.getBase(),  J3Base =J3.getBase(),  J4Base =J4.getBase();  \
for(i4=I4Base,j4=J4Base; i4<=I4Bound; i4++,j4++) \
for(i3=I3Base,j3=J3Base; i3<=I3Bound; i3++,j3++) \
for(i2=I2Base,j2=J2Base; i2<=I2Bound; i2++,j2++) \
for(i1=I1Base,j1=J1Base; i1<=I1Bound; i1++,j1++)


bool Cgmp::
checkIfInterfacesMatch(Mapping &map1, int &dir1, int &side1, Mapping &map2, int &dir2, int &side2)
{
  bool interfacesMatch = false;
  int numberOfDimensions = map1.getRangeDimension();
  if (map1.getRangeDimension()!=map1.getDomainDimension()) return false; // then we have a problem, should be an assert? 
  realSerialArray r(1,3),x(1,3),xp(1,3); // arrays for geometric tests
  real dr = 0.01; // parameter space displacement from the boundary used when doing geometric tests

  r = x = xp = 0;
  /// invert points from grid1 onto grid2
  for ( int s=0; s<2 && !interfacesMatch; s++ ) 
  {
    r( 0,dir1) = (real)side1;
    r( 0, (dir1+1)%numberOfDimensions ) = (1-2*s)*dr;
    if ( numberOfDimensions==3 ) r( 0, (dir1+2)%numberOfDimensions ) =         dr;
	
    map1.mapS(r,x);
    r = -1;
    if ( numberOfDimensions==2 ) r(0,2) = 0;
    map2.inverseMapS(x,r);
    bool wasInvertedOntoMapping = fabs(r(0,dir2)-(side2))<100*REAL_EPSILON; // make sure the inversion happend onto the correct boundary
    for ( int a=0; a<numberOfDimensions && wasInvertedOntoMapping; a++ ) // make sure the inversion was still within the bounds of the mapping
      wasInvertedOntoMapping = wasInvertedOntoMapping && ( r(0,a)>-100*REAL_EPSILON && r(0,a)<(1+100*REAL_EPSILON) ); // should have a tolerance probably...
    if ( wasInvertedOntoMapping )
    {
      map2.mapS(r,xp);
      interfacesMatch = ( SQR(x(0,0)-xp(0,0)) + SQR(x(0,1)-xp(0,1)) + SQR(x(0,2)-x(0,2)) ) < 100*REAL_EPSILON; // ahh, my old friend, 100*REAL_EPSILON
    }

#if 0
    if ( interfacesMatch )
    {
      cout<<map1.getName(Mapping::mappingName)<<" matched to "<<map2.getName(Mapping::mappingName)<<endl;
      cout<<r(0,0)<<"   "<<r(0,1)<<endl;
      x.display("X");
      xp.display("XP");
    }
#endif

    if ( !interfacesMatch && numberOfDimensions==3 ) // check the other corner
    {
      r( 0,dir1) = (real)side1;
      r( 0, (dir1+1)%numberOfDimensions ) = (1-2*s)*dr;
      r( 0, (dir1+2)%numberOfDimensions ) =        -dr; // NOTE the - sign is the only change from the above block!!!
	    
      map1.mapS(r,x);
      r = -1;
      map2.inverseMapS(x,r);
      bool wasInvertedOntoMapping = fabs(r(0,dir2)-(side2))<100*REAL_EPSILON; // make sure the inversion happend onto the correct boundary
      for ( int a=0; a<numberOfDimensions && wasInvertedOntoMapping; a++ )
	wasInvertedOntoMapping = wasInvertedOntoMapping && ( r(0,a)>-100*REAL_EPSILON && r(0,a)<(1+100*REAL_EPSILON) ); // should have a tolerance probably...
      if ( wasInvertedOntoMapping )
      {
	map2.mapS(r,xp);
	interfacesMatch = ( SQR(x(0,0)-xp(0,0)) + SQR(x(0,1)-xp(0,1)) + SQR(x(0,2)-x(0,2)) ) < 100*REAL_EPSILON; // ahh, my old friend, 100*REAL_EPSILON
      } // if was inverted ok
	    
    } // if the previous interface did not match
  } // for each side on the face we are matching
  return interfacesMatch;
} // checkIfInterfacesMatch



void
getLocalBoundsAndBoundaryConditions( const realMappedGridFunction & a, 
                                     IntegerArray & gidLocal, 
                                     IntegerArray & dimensionLocal, 
                                     IntegerArray & bcLocal );

#define interfaceCgCm interfacecgcm_
extern "C"
{
void interfaceCgCm( const int&nd, 
		    const int&nd1a,const int&nd1b,const int&nd2a,const int&nd2b,const int&nd3a,const int&nd3b,
		    const int&gridIndexRange1, real&u1, const int&mask1,const real&rsxy1, const real&xy1, 
		    const int&boundaryCondition1, 
		    const int&md1a,const int&md1b,const int&md2a,const int&md2b,const int&md3a,const int&md3b,
		    const int&gridIndexRange2, real&u2, const int&mask2,const real&rsxy2, const real&xy2, 
		    const int&boundaryCondition2,
		    const int&ipar, const real&rpar, 
		    real&aa2, real&aa4, real&aa8, 
		    int&ipvt2, int&ipvt4, int&ipvt8,
		    int&ierr );
}




// =====================================================================================================
//
///  \brief Find the interfaces and initialize the work-space. 
/// \details Interfaces are located by faces on different domains that have an interface BC and
///    have the same share flag.
/// \param gfIndex (input) : Domain d should use the grid function : domainSolver[d]->gf[gfIndex[d]]
///
// =====================================================================================================
int Cgmp::
initializeInterfaces(std::vector<int> & gfIndex)
{
  real cpu0=getCPU();

  Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];
  Index Jv[3], &J1=Jv[0], &J2=Jv[1], &J3=Jv[2];


  // **** Make a list of grids on each side of each the interface. *****
  //      Normally there will be one grid on either side that meet at the interface. 
  //      There may be additional grids when we use AMR

  
  const int numberOfDomains = domainSolver.size();
  InterfaceList & interfaceList = parameters.dbase.get<InterfaceList>("interfaceList");
  
  FILE *& interfaceFile =parameters.dbase.get<FILE* >("interfaceFile");

  ForDomain(d1)
  {
    CompositeGrid & cg1 = domainSolver[d1]->gf[gfIndex[d1]].cg;
    const int numberOfDimensions = cg1.numberOfDimensions();
    const IntegerArray & originalBoundaryCondition1 = 
                domainSolver[d1]->parameters.dbase.get<IntegerArray>("originalBoundaryCondition");
    const IntegerArray & interfaceType1 = 
                domainSolver[d1]->parameters.dbase.get<IntegerArray >("interfaceType");

    // ** look for base grids with an interface BC ***
    for( int grid1=0; grid1<cg1.numberOfBaseGrids(); grid1++ ) 
    {

      MappedGrid & mg1 = cg1[grid1];
      const IntegerArray & bc1 = mg1.boundaryCondition();
      const IntegerArray & share1 = mg1.sharedBoundaryFlag();

      // check for interface boundary conditions
      for( int dir1=0; dir1<mg1.numberOfDimensions(); dir1++ )
      {
	for( int side1=0; side1<=1; side1++ )
	{
	  if( interfaceType1(side1,dir1,grid1)!=Parameters::noInterface )
	  {
	    // **** Find the adjacent base grid that matches this interface ****
  	    bool foundAnInterfaceForThisBoundary = false;
	    for( int d2=d1+1; d2<numberOfDomains; d2++ ) if( domainSolver[d2]!=NULL )
	    {
	      CompositeGrid & cg2 = domainSolver[d2]->gf[gfIndex[d2]].cg;
              const IntegerArray & originalBoundaryCondition2 = 
                      domainSolver[d2]->parameters.dbase.get<IntegerArray>("originalBoundaryCondition");

	      const IntegerArray & interfaceType2 = 
                domainSolver[d2]->parameters.dbase.get<IntegerArray >("interfaceType");

	      for( int grid2=0; grid2<cg2.numberOfBaseGrids(); grid2++ )
	      {

		MappedGrid & mg2 = cg2[grid2];
		const IntegerArray & bc2 = mg2.boundaryCondition();
		const IntegerArray & share2 = mg2.sharedBoundaryFlag();

		for( int dir2=0; dir2<mg2.numberOfDimensions(); dir2++ )
		{
		  for( int side2=0; side2<=1; side2++ )
		  {
		    // (1) A matching interface has the same share flag
                    // (2) must also have the same original boundary conditions (*wdh* 080219)
                    //     This is needed if there are multiple grids on the same boundary
		    // (3) it must also match geometrically at at least one boundary point, here are some examples:
		    /*
		          +---------------------------------------------------+ boundary 1
		          +---------------------------------------------------+ boundary 2 matches exactly, this is the easiest case

		          +---------------------------------------------------+ boundary 1
                                  +-----------------------------+               boundary 2 matches at both points (much check the reversed condition too)

		          +----------------------+                              boundary 1
                                  +-----------------------------+               boundary 2 matches at one point (much check the reversed condition too)

		          +----------------------+                              boundary 1
                                                 +--------------+               boundary 2 does not really match boundary 1, hence the increment dr

			  In 3D there are 8 possible checks to make (4 for each face) but we always stop when one match is found (e.g. in the second example
			  we would stop after the third check, the first point in boundary 2).
			  kkc 080425
		    */
		    if( share1(side1,dir1)==share2(side2,dir2)  && // this must be a matching interface 
			// bc2(side2,dir2)==Parameters::interfaceBoundaryCondition &&
                        interfaceType2(side2,dir2,grid2)!=Parameters::noInterface &&
                        originalBoundaryCondition1(side1,dir1,grid1)==originalBoundaryCondition2(side2,dir2,grid2) )
		    {
		      // geometric tests
		      bool interfacesMatch = false;
		      Mapping & map1 = mg1.mapping().getMapping();
		      Mapping & map2 = mg2.mapping().getMapping();
		      
		      interfacesMatch = checkIfInterfacesMatch(map1,dir1,side1,map2,dir2,side2);
		      if ( !interfacesMatch )
		      { // invert points from grid2 onto grid1
			interfacesMatch = checkIfInterfacesMatch(map2,dir2,side2,map1,dir1,side1);
		      }

		      if ( interfacesMatch ) 
		      {
			if ( foundAnInterfaceForThisBoundary ) 
			{
			  printF("--Cgmp:initializeInterface: ERROR found more than one match for (d1,grid1,side1,dir1,share)=(%i,%i,%i,%i,%i)\n"
				 "                            additional match is (d2,grid2,side2,dir2,share)=(%i,%i,%i,%i,%i)\n",
				 d1,grid1,side1,dir1,share1(side1,dir1),
				 d2,grid2,side2,dir2,share2(side2,dir2));
			  printF(" grid1=%s, grid2=%s\n",(const char*)mg1.getName(),(const char*)mg2.getName());

                          Overture::abort("error");
			}
			  
			// **** interface found *************
			if( interfaceList.size()==0 )
			  fPrintF(interfaceFile,"\n**********************************************************************\n");
			fPrintF(interfaceFile,
				"--Cgmp:initializeInterfaces: interface found: %s\n"
				"       (d1,grid1,side1,dir1,share)=(%i,%i,%i,%i,%i)  grid=%s domain=%s\n"
				"       (d2,grid2,side2,dir2,share)=(%i,%i,%i,%i,%i)  grid=%s domain=%s\n",
				(const char*)parameters.icNames[interfaceType1(side1,dir1,grid1)],
				d1,grid1,side1,dir1,share1(side1,dir1), (const char*)mg1.getName(),
				(const char*)domainSolver[d1]->getName(),
				d2,grid2,side2,dir2,share2(side2,dir2),(const char*)mg2.getName(), 
				(const char*)domainSolver[d2]->getName());

			interfaceList.push_back(InterfaceDescriptor());  // add a new interface to the list
			InterfaceDescriptor & interface =  interfaceList.back(); 
			interface.gridListSide1.push_back(GridFaceDescriptor(d1,grid1,side1,dir1));
			  
			interface.gridListSide2.push_back(GridFaceDescriptor(d2,grid2,side2,dir2));
			foundAnInterfaceForThisBoundary = true;

		      } // if found an interface and they match geometrically

		    } // if found a possible interface by matching share and boundary conditions
		  } // for side2
		} // for dir2
	      } // for grid2
	    } // for domain 2 (d2)

	    if( !foundAnInterfaceForThisBoundary )
	    {
              // If no matching interface was found then we still need to check that the 
              // interface was not matched previously ...

	      for( int inter=0; inter < interfaceList.size(); inter++ )
	      {
		InterfaceDescriptor & interfaceDescriptor = interfaceList[inter]; 

		// there may be multiple grid faces that lie on the interface:     
		for( int face=0; face<interfaceDescriptor.gridListSide1.size(); face++ )
		{
		  // GridFaceDescriptor & gridDescriptor1 = interfaceDescriptor.gridListSide1[face];
		  GridFaceDescriptor & gridDescriptor2 = interfaceDescriptor.gridListSide2[face];
                  if( gridDescriptor2.grid==grid1 && gridDescriptor2.side==side1 && gridDescriptor2.axis==dir1 )
		  {
		    foundAnInterfaceForThisBoundary=true;
		  }
		}
	      }
	      if( !foundAnInterfaceForThisBoundary )
	      {
		printF("--Cgmp:initializeInterfaces: ERROR: no matching interface found for :\n"
		       "       (d1,grid1,side1,dir1,share,bc)=(%i,%i,%i,%i,%i,%i)  (grid1=%s)\n",
		       d1,grid1,side1,dir1,share1(side1,dir1),bc1(side1,dir1),(const char*)mg1.getName());

                printF("\n");
		ForDomain(dm)
		{
		  const IntegerArray & originalBoundaryCondition = 
		    domainSolver[dm]->parameters.dbase.get<IntegerArray>("originalBoundaryCondition");
		  const IntegerArray & interfaceType = 
		    domainSolver[dm]->parameters.dbase.get<IntegerArray >("interfaceType");

		  CompositeGrid & cg = domainSolver[dm]->gf[gfIndex[dm]].cg;

		  printF("**** domain %i %s : *******\n"
			 "   side axis grid interfaceType   bc   share\n",dm,(const char*)domainSolver[dm]->getName());
		  for( int g=0; g<cg.numberOfBaseGrids(); g++ )
		  {
		    MappedGrid & mg = cg[g];
		    const IntegerArray & share = mg.sharedBoundaryFlag();
		    
		    for( int s=0; s<=1; s++) for( int d=0; d<numberOfDimensions; d++ )
		    {
		      printF("   %2i  %2i  %4i   %5i         %3i    %3i\n",s,d,g,interfaceType(s,d,g),
			     originalBoundaryCondition(s,d,g),share(s,d)  );
		    }
		  }
		}
		
		Overture::abort("error");
	      }
	    }

	  } // if an interface bc
	} // for side1
      } // for dir1
    } // for grid1
  } // for domain 1 (d1)
  if( interfaceList.size()>0 )
  {
    fPrintF(interfaceFile,"**********************************************************************\n\n");
    gridHasMaterialInterfaces=true;
  }
  else
  {
    gridHasMaterialInterfaces=false;
  }
  
  parameters.dbase.get<RealArray>("timing")(Parameters::timeForInterfaces)+=getCPU()-cpu0;
  
  return 0;
}



//===============================================================================================
// 
///  \brief: Assign the boundary conditions at an internal interface. 
/// \param gfIndex (input) : Domain d should use the grid function : domainSolver[d]->gf[gfIndex[d]]
/// \details: this routine solves the coupled heatFlux boundary conditions directly.
//
//===============================================================================================
int Cgmp::
assignInterfaceBoundaryConditions(std::vector<int> & gfIndex, 
				  const real dt )
{
  real cpu0=getCPU();

  if( !gridHasMaterialInterfaces ) return 0;

  InterfaceList & interfaceList = parameters.dbase.get<InterfaceList>("interfaceList");
  
  if( interfaceList.size()==0 )
  {
    // -- Initialize the list of interfaces --
    initializeInterfaces(gfIndex);

    if( interfaceList.size()>0 ) gridHasMaterialInterfaces=true;

    if( interfaceList.size()>0 )
    {
      printF("**** assignInterfaceBoundaryConditions:initializeInterfaces: number of interfaces =%i\n",
	     interfaceList.size());
    }
    else
    {
      printF("**** assignInterfaceBoundaryConditions:initializeInterfaces: There are NO interfaces\n");
    }
    
  }
  const int np= max(1,Communication_Manager::numberOfProcessors());
  const int & myid = parameters.dbase.get<int>("myid");
  
  const int numberOfDomains=domainSolver.size();

  Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];
  Index Jv[3], &J1=Jv[0], &J2=Jv[1], &J3=Jv[2];
  

  // ************************************************
  // ******* Apply the interface conditions. ********
  // ************************************************

  // We need to interpolate the grid function if an interface is covered by more than 1 face
  std::vector<bool> interpolateThisDomain(numberOfDomains,false);

  // loop over interfaces
  for( int inter=0; inter < interfaceList.size(); inter++ )
  {
    InterfaceDescriptor & interfaceDescriptor = interfaceList[inter]; 

    // there may be multiple grid faces that lie on the interface:     
    for( int face=0; face<interfaceDescriptor.gridListSide1.size(); face++ )
    {
      GridFaceDescriptor & gridDescriptor1 = interfaceDescriptor.gridListSide1[face];
      GridFaceDescriptor & gridDescriptor2 = interfaceDescriptor.gridListSide2[face];
      
      const int d1=gridDescriptor1.domain, 
                grid1=gridDescriptor1.grid, side1=gridDescriptor1.side, dir1=gridDescriptor1.axis;
      const int d2=gridDescriptor2.domain, 
                grid2=gridDescriptor2.grid, side2=gridDescriptor2.side, dir2=gridDescriptor2.axis;

      assert( d1>=0 && d1<numberOfDomains && d2>=0 && d2<numberOfDomains );

      GridFunction & gf1 = domainSolver[d1]->gf[gfIndex[d1]];
      GridFunction & gf2 = domainSolver[d2]->gf[gfIndex[d2]];

      CompositeGrid & cg1 = gf1.cg;
      assert( grid1>=0 && grid1<=cg1.numberOfComponentGrids());
      MappedGrid & mg1 = cg1[grid1];
      const IntegerArray & bc1 = mg1.boundaryCondition();
      const IntegerArray & share1 = mg1.sharedBoundaryFlag();

      CompositeGrid & cg2 = gf2.cg;
      assert( grid2>=0 && grid2<=cg2.numberOfComponentGrids());
      MappedGrid & mg2 = cg2[grid2];
      const IntegerArray & bc2 = mg2.boundaryCondition();
      const IntegerArray & share2 = mg2.sharedBoundaryFlag();

      IntegerArray gidLocal1(2,3), dimLocal1(2,3), bcLocal1(2,3);
      getLocalBoundsAndBoundaryConditions( gf1.u[grid1],gidLocal1,dimLocal1,bcLocal1 );

      IntegerArray gidLocal2(2,3), dimLocal2(2,3), bcLocal2(2,3);
      getLocalBoundsAndBoundaryConditions( gf2.u[grid2],gidLocal2,dimLocal2,bcLocal2 );

      const int numberOfDimensions = cg1.numberOfDimensions();

      const int orderOfAccuracyInSpace=2;  // **** do this for now ***
      const int tc1 = domainSolver[d1]->parameters.dbase.get<int >("tc");    // *** fix this *** assume T for now 
      const int tc2 = domainSolver[d2]->parameters.dbase.get<int >("tc");


      const int extra=0; // orderOfAccuracyInSpace/2;
      getBoundaryIndex(mg1.gridIndexRange(),side1,dir1,I1,I2,I3,extra);
      getBoundaryIndex(mg2.gridIndexRange(),side2,dir2,J1,J2,J3,extra);
  
      // check that the number of points in the tangential directions match -- eventually we will fix this
      for( int dir=1; dir<mg1.numberOfDimensions(); dir++ )
      {
	int dir1p = (dir1+dir) % mg1.numberOfDimensions();
	int dir2p = (dir2+dir) % mg2.numberOfDimensions();
	if( Iv[dir1p].getLength()!=Jv[dir2p].getLength() )
	{
	  printF("applyInterfaceBC:ERROR: The number of grid points on the two interfaces do not match\n"
	         " (d1,grid1,side1,dir1,bc1)=(%i,%i,%i,%i,%i) Iv=[%i,%i][%i,%i][%i,%i]\n"
		 " (d2,grid2,side2,dir2,bc2)=(%i,%i,%i,%i,%i) Jv=[%i,%i][%i,%i][%i,%i]\n",
		 d1,grid1,side1,dir1,mg1.boundaryCondition(side1,dir1),
                   I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound(),I3.getBase(),I3.getBound(),  
                 d2,grid2,side2,dir2,mg2.boundaryCondition(side2,dir2),
                   J1.getBase(),J1.getBound(),J2.getBase(),J2.getBound(),J3.getBase(),J3.getBound());
	  cout<<"grid names are "<<mg1.getName()<<" , "<<mg2.getName()<<endl;
	  Overture::abort("error");
	}

        // We need to interpolate the grid function if an interface has interpolation points on it:

        // **** NOTE: I think this is needed because we do not check the mask array when assigning
        //            interface points and thus over-write interpolation points **** FIX ME ***

        if( bc1(0,dir1p)==0 || bc1(1,dir1p)==0 )
	{
         interpolateThisDomain[d1]=true;
	}
        if( bc2(0,dir2p)==0 || bc2(1,dir2p)==0 )
	{
         interpolateThisDomain[d2]=true;
	}

      }

      realArray & u1 = gf1.u[grid1];
      realArray & u2 = gf2.u[grid2];

      #ifdef USE_PPP
        // First try for parallel -- assume aligned grids

        // copy values from u2 into an array u2b that is distributed in the same way as u1
        realArray u2b; u2b.partition(u1.getPartition());
        // the next line causes a bug in u2b.updateGhostBoundaries(); below -- doesn't like non-zero base
        // u2b.redim(u1.dimension(0),u1.dimension(1),u1.dimension(2),Range(tc2,tc2)); // note last arg
        u2b.redim(u1.dimension(0),u1.dimension(1),u1.dimension(2),u2.dimension(3)); // note last arg
        realSerialArray u2bLocal; getLocalArrayWithGhostBoundaries(u2b,u2bLocal);
	
        // **** we should extra values for extrpolation -- see mx/src/assignInterfaceBoundaryConditions.bC

        // stencil half width: (we copy points (-halfWidth,+halfWidth)
        const int halfWidth= orderOfAccuracyInSpace/2;
        Index Iv1[4], Iv2[4];
	getIndex(mg1.dimension(),Iv1[0],Iv1[1],Iv1[2]);
	Iv1[dir1]=Range(mg1.gridIndexRange(side1,dir1)-halfWidth,mg1.gridIndexRange(side1,dir1)+halfWidth);
	getIndex(mg2.dimension(),Iv2[0],Iv2[1],Iv2[2]);
	Iv2[dir2]=Range(mg2.gridIndexRange(side2,dir2)-halfWidth,mg2.gridIndexRange(side2,dir2)+halfWidth);

	Iv1[3]=tc2; Iv2[3]=tc2;

        const int nd=4;

        // NOTE: We also assume below that the grid metric match !  ** fix me ***

        // *wdh* 081122 -- I think we are assuming that the values come from the same side and dir !!
        //  -- otherwise we need to add a stride to the copy below I think --
	if( dir1!=dir2 )
	{
	  printF("Cgmx:assignInterfaceBC: Error: in parallel we assume that the interface satisfies\n"
		 " dir1==dir2 : you may have to remake the grid to satisfy this\n");
	  Overture::abort("error");
	}
        if( debug() & 8 )
	{
	  printF("interfaceBC: Iv1=[%i,%i][%i,%i][%i,%i][%i,%i]  Iv2=[%i,%i][%i,%i][%i,%i][%i,%i]\n",
		 Iv1[0].getBase(),Iv1[0].getBound(),Iv1[1].getBase(),Iv1[1].getBound(),
		 Iv1[2].getBase(),Iv1[2].getBound(),Iv1[3].getBase(),Iv1[3].getBound(),
		 Iv2[0].getBase(),Iv2[0].getBound(),Iv2[1].getBase(),Iv2[1].getBound(),
		 Iv2[2].getBase(),Iv2[2].getBound(),Iv2[3].getBase(),Iv2[3].getBound());
	}
	
        // *** DO NOT reverse the points ***
        // We just pretend that u2 pts are on the same side but shifted in index space *** 

        u2bLocal=0.; 
        ParallelUtility::copy(u2b,Iv1,u2,Iv2,nd);  // u2b(Iv1)=u2(Iv2)
        // u2b(Iv1[0],Iv1[1],Iv1[2],Iv1[3])=u2(Iv2[0],Iv2[1],Iv2[2],Iv2[3]);
	
        // copy values from u1 into an array u1b that is distributed in the same way as u2
        realArray u1b; u1b.partition(u2.getPartition());
        // u1b.redim(u2.dimension(0),u2.dimension(1),u2.dimension(2),Range(tc1,tc1));
        u1b.redim(u2.dimension(0),u2.dimension(1),u2.dimension(2),u1.dimension(3));
        realSerialArray u1bLocal; getLocalArrayWithGhostBoundaries(u1b,u1bLocal);
	
	Iv1[3]=tc1; Iv2[3]=tc1;
        u1bLocal=0.; 
        ParallelUtility::copy(u1b,Iv2,u1,Iv1,nd);  // u1b(Iv2)=u1(Iv1)
        // u1b(Iv2[0],Iv2[1],Iv2[2],Iv2[3])=u1(Iv1[0],Iv1[1],Iv1[2],Iv1[3]);

        u1b.updateGhostBoundaries(); // *********** these are currently needed ********************
        u2b.updateGhostBoundaries();

        realSerialArray u1Local; getLocalArrayWithGhostBoundaries(u1,u1Local);
        realSerialArray u2Local; getLocalArrayWithGhostBoundaries(u2,u2Local);
	
        int includeGhost=0;  // do NOT include parallel ghost since we can't apply the stencil there
        bool ok1 = ParallelUtility::getLocalArrayBounds(u1,u1Local,I1,I2,I3,includeGhost);
        bool ok2 = ParallelUtility::getLocalArrayBounds(u2,u2Local,J1,J2,J3,includeGhost);

	// ::display(u1Local,"interfaceBC: u1Local","%5.2f ");
	// ::display(u1bLocal,"interfaceBC:u1bLocal after copy","%5.2f ");

	// ::display(u2Local,"interfaceBC: u2Local","%5.2f ");
	// ::display(u2bLocal,"interfaceBC:u2bLocal after copy","%5.2f ");
	

      #else

        realSerialArray & u1Local = u1;
        realSerialArray & u2Local = u2;
	
      #endif

      // IN parallel :
//       #ifdef USE_PPP
//         const int includeGhost=1;
//         bool ok = ParallelUtility::getLocalArrayBounds(u1,u1Local,I1a,I2a,I3a,includeGhost);

//         realSerialArray u1Local; getLocalArrayWithGhostBoundaries(u1,u1Local);
//         realSerialArray u2Local; getLocalArrayWithGhostBoundaries(u2,u2Local);

//         J1b = I1a + J1.getBase() - I1.getBase();
	
//         realSerialArray u2b(J1b,J2b,J3b,R);  // copy of u2 overlapping u1Local
// 	// copy u2b from u2

//         realSerialArray u1b(I1b,I2b,I3b,R);  // copy of u1 overlapping u2Local
// 	// copy u1b from u1

//         // Now solve interface eqns for 
//         //     I  : (u1Local,u2b)   --> gives u1Local (throw away u2b)
//         //     II : (u1b,u2Local)   --> gives u2Local (throw away u1b)

//       #endif

      real t1=gf1.t;
      real t2=gf2.t;
      if( fabs(t1-t2) > REAL_EPSILON*100.*max(t1,t2) )
      {
	printf("applyInterfaceBC:WARNING: t1=%9.3e and t2=%9.3e are not the same, t1-t2=%8.2e\n",t1,t2,t1-t2);
      }
      
      real t= t1;

      // Apply a BC at the inteface


      bool useOpt=true;
      if( useOpt )
      {

	// optimised version 

	int n1a=I1.getBase(),n1b=I1.getBound(),
	  n2a=I2.getBase(),n2b=I2.getBound(),
	  n3a=I3.getBase(),n3b=I3.getBound();

	int m1a=J1.getBase(),m1b=J1.getBound(),
	  m2a=J2.getBase(),m2b=J2.getBound(),
	  m3a=J3.getBase(),m3b=J3.getBound();


	bool isRectangular1= mg1.isRectangular();
	real dx1[3]={0.,0.,0.}; //
	if( isRectangular1 )
	  mg1.getDeltaX(dx1);

	bool isRectangular2= mg2.isRectangular();
	real dx2[3]={0.,0.,0.}; //
	if( isRectangular2 )
	  mg2.getDeltaX(dx2);

	if( true )
	{ // for testing -- make rectangular grids look curvilinear ***************************************
	  isRectangular1=false;
	  isRectangular2=false;
	}
	if( !isRectangular1 )
	{
	  mg1.update(MappedGrid::THEinverseVertexDerivative);
	  mg2.update(MappedGrid::THEinverseVertexDerivative);
	}
		  
	assert(isRectangular1==isRectangular2);

	int useForcing = parameters.dbase.get<bool >("twilightZoneFlow");
        assert( parameters.dbase.get<bool >("twilightZoneFlow")==domainSolver[d1]->parameters.dbase.get<bool >("twilightZoneFlow"));
        assert( parameters.dbase.get<bool >("twilightZoneFlow")==domainSolver[d2]->parameters.dbase.get<bool >("twilightZoneFlow"));
	if( useForcing )
	{
	  mg1.update(MappedGrid::THEcenter);
	  mg2.update(MappedGrid::THEcenter);
	}
	

	int gridType = isRectangular1 ? 0 : 1;  // ******************************* fix this --> gridType[1,2]
	int orderOfExtrapolation=orderOfAccuracyInSpace+1;  // not used
	int useWhereMask=true;

        int materialInterfaceOption=0;  // unused 
        int interfaceInitialized=0;  // unused 
	int numberOfIterationsForInterfaceBC=0;
	real omegaForInterfaceIteration=.9;


        real ktc[2]  ={-1.,-1.};   // coefficient of thermal conductivity
	real kappa[2]={-1.,-1.};   //  coefficient of thermal diffusion 
	for( int dd=0; dd<=1; dd++ )
	{
	  const int d = dd==0 ? d1 : d2;
          ktc[dd] = domainSolver[d]->parameters.dbase.get<real>("thermalConductivity");

	  if( domainSolver[d]->parameters.dbase.has_key("kappa") )
	  {
	    kappa[dd]=domainSolver[d]->parameters.dbase.get<std::vector<real> >("kappa")[0];
	  } 
	  else if( domainSolver[d]->parameters.dbase.has_key("kThermal") )
	  {
	    kappa[dd]=domainSolver[d]->parameters.dbase.get<real>("kThermal");
	  }
	}
	if( kappa[0] <=0. || kappa[1]<=0. )
	{
	  printF("applyInterfaceBC:ERROR: a negative thermal diffusivity was found, kappa[0]=%e kappa[1]=%e\n",
                 kappa[0],kappa[1]);
	  Overture::abort("error");
	}
	if( ktc[0] <=0. || ktc[1]<=0. )
	{
	  printF("applyInterfaceBC:ERROR: a negative thermal conductivity was found, ktc[0]=%e ktc[1]=%e\n",
                 ktc[0],ktc[1]);
	  Overture::abort("error");
	}

        real ktc1=ktc[0], ktc2=ktc[1];

	if( t1<2.*dt || debug() & 2 )
	  printf("applyInterfaceBC: (d1,grid1,side1,dir1,kappa1,k1)=(%i,%i,%i,%i,%8.2e,%8.2e), "
		 "(d2,grid2,side2,dir2,kappa2,k2)=(%i,%i,%i,%i,%8.2e,%8.2e) \n",
		 d1,grid1,side1,dir1,kappa[0],ktc[0], d2,grid2,side2,dir2,kappa[1],ktc[1]);
	
        // We take the normal from one side the of the interface. normalSign1 and normalSign2
        // are used to flip the sign from one side to the other
        int normalSign1=1-2*side1;  
	int normalSign2=1-2*side2;

	int ierr=0;
	int ipar[]={ //
	  side1, dir1, grid1,
	  n1a,n1b,n2a,n2b,n3a,n3b,
	  side2, dir2, grid2,
	  m1a,m1b,m2a,m2b,m3a,m3b,
 	  gridType,            
 	  orderOfAccuracyInSpace,    
 	  orderOfExtrapolation,
 	  useForcing,          
 	  tc1,                  
 	  tc2,                  
 	  np,
 	  myid,
          normalSign1,
	  normalSign2,
 	  0,  // for use 
 	  0,  // for use 
 	  useWhereMask,       
 	  parameters.dbase.get<int >("debug"),
 	  numberOfIterationsForInterfaceBC,
 	  materialInterfaceOption,
	  interfaceInitialized
	};
		  
	real rpar[]={ //
	  dx1[0],
	  dx1[1],
	  dx1[2],
	  mg1.gridSpacing(0),
	  mg1.gridSpacing(1),
	  mg1.gridSpacing(2),
	  dx2[0],
	  dx2[1],
	  dx2[2],
	  mg2.gridSpacing(0),
	  mg2.gridSpacing(1),
	  mg2.gridSpacing(2),
	  t,    
	  (real &)domainSolver[d1]->parameters.dbase.get<OGFunction* >("exactSolution"),  // twilight zone pointer
	  (real &)domainSolver[d2]->parameters.dbase.get<OGFunction* >("exactSolution"),  // twilight zone pointer
	  dt,    
	  ktc1,
	  ktc2,
	  kappa[0],     // rpar[18] 
	  kappa[1],
	  0., // for later use 
	  0., // for later use 
	  omegaForInterfaceIteration
	};
		  
	real *u1p=u1Local.getDataPointer();
	real *prsxy1=isRectangular1 ? u1p : mg1.inverseVertexDerivative().getLocalArray().getDataPointer();
	real *pxy1= !useForcing ? u1p : mg1.center().getLocalArray().getDataPointer(); 
	int *mask1p=mg1.mask().getLocalArray().getDataPointer();

	real *u2p=u2Local.getDataPointer();
	real *prsxy2=isRectangular2 ? u2p : mg2.inverseVertexDerivative().getLocalArray().getDataPointer();
	real *pxy2= !useForcing ? u2p : mg2.center().getLocalArray().getDataPointer(); 
	int *mask2p=mg2.mask().getLocalArray().getDataPointer();


// 	real *rwk=interface.rwk;
// 	int *iwk=interface.iwk;
// 	assert( rwk!=NULL && iwk!=NULL );
        real rwk[1];
	int iwk[1];
      
//	const int ndf = max(interface.ndf1,interface.ndf2); 
	const int ndf = 0;

	// assign pointers into the work spaces
	int pa2=0,pa4=0,pa8=0, pipvt2=0,pipvt4=0,pipvt8=0;
	if( orderOfAccuracyInSpace==2 )
	{
	  pa2=0; 
	  pa4=pa2 + 2*2*2*ndf;
	  pa8=0;  // not used
	
	  pipvt2=0;
	  pipvt4=pipvt2 + 2*ndf; 
	  pipvt8=0;
	}
	else if( orderOfAccuracyInSpace==4 )
	{
	  pa2=0; // not used
	  pa4=0;
	  pa8=pa4+4*4*2*ndf;
	
	  pipvt2=0;
	  pipvt4=0;
	  pipvt8=pipvt4+4*ndf;
	
	}
       #ifndef USE_PPP
	interfaceCgCm( mg1.numberOfDimensions(), 
		       u1Local.getBase(0),u1Local.getBound(0),
		       u1Local.getBase(1),u1Local.getBound(1),
		       u1Local.getBase(2),u1Local.getBound(2),
		       mg1.gridIndexRange(0,0), *u1p, *mask1p,*prsxy1, *pxy1, bc1(0,0), 
		       u2Local.getBase(0),u2Local.getBound(0),
		       u2Local.getBase(1),u2Local.getBound(1),
		       u2Local.getBase(2),u2Local.getBound(2),
		       mg2.gridIndexRange(0,0), *u2p, *mask2p,*prsxy2, *pxy2, bc2(0,0), 
		       ipar[0], rpar[0], 
		       rwk[pa2],rwk[pa4],rwk[pa8], iwk[pipvt2],iwk[pipvt4],iwk[pipvt8],
		       ierr );
	if( false && debug() & 4 )
	{
	  ::display(u1Local(I1,I2,I3,tc1),"u1Local(I1,I2,I3,tc1) after interfaceCgCm","%5.2f ");
	  ::display(u2Local(J1,J2,J3,tc2),"u2Local(J1,J2,J3,tc2) after interfaceCgCm","%5.2f ");
	  getGhostIndex(mg1.gridIndexRange(),side1,dir1,I1,I2,I3,1);
	  getGhostIndex(mg2.gridIndexRange(),side2,dir2,J1,J2,J3,1);
	  ::display(u1Local(I1,I2,I3,tc1),"u1Local(Ig1,Ig2,Ig3,tc1) after interfaceCgCm","%5.2f ");
	  ::display(u2Local(J1,J2,J3,tc2),"u2Local(Jg1,Jg2,Jg3,tc2) after interfaceCgCm","%5.2f ");
	}
	
      #else
        // --- parallel version ---
        // --- For now we assume that the grids match exactly ---

        // In parallel we solve the interface equations on each processor -- thus we
        // solve the equations twice, one for each side of the interface

        assert( dir1==dir2 );

        // ******** get local gid bc ...
        if( ok1 )
	{
	  int ipar[]={ //
	    side1, dir1, grid1,
	    n1a,n1b,n2a,n2b,n3a,n3b,
	    side2, dir2, grid2,        // use grid1 info here
	    n1a,n1b,n2a,n2b,n3a,n3b,   // use grid1 info here
	    gridType,            
	    orderOfAccuracyInSpace,    
	    orderOfExtrapolation,
	    useForcing,          
	    tc1,                  
	    tc2,        
	    np,
	    myid,
	    normalSign1,
	    normalSign2,
	    0,  // for use 
	    0,  // for use 
	    useWhereMask,       
	    parameters.dbase.get<int >("debug"),
	    numberOfIterationsForInterfaceBC,
	    materialInterfaceOption,
	    interfaceInitialized
	  };
	  real rpar[]={ //
	    dx1[0],
	    dx1[1],
	    dx1[2],
	    mg1.gridSpacing(0),
	    mg1.gridSpacing(1),
	    mg1.gridSpacing(2),
	    dx2[0],
	    dx2[1],
	    dx2[2],
	    mg1.gridSpacing(0),
	    mg1.gridSpacing(1),
	    mg1.gridSpacing(2),
	    t,    
	    (real &)domainSolver[d1]->parameters.dbase.get<OGFunction* >("exactSolution"),  // twilight zone pointer
	    (real &)domainSolver[d2]->parameters.dbase.get<OGFunction* >("exactSolution"),  // twilight zone pointer
	    dt,    
	    ktc1,
	    ktc2,
  	    kappa[0],     // rpar[18] 
	    kappa[1],
	    0., // for later use 
	    0., // for later use 
	    omegaForInterfaceIteration
	  };
	  interfaceCgCm( mg1.numberOfDimensions(), 
			 u1Local.getBase(0),u1Local.getBound(0),
			 u1Local.getBase(1),u1Local.getBound(1),
			 u1Local.getBase(2),u1Local.getBound(2),
			 gidLocal1(0,0), *u1p, *mask1p,*prsxy1, *pxy1, bcLocal1(0,0), 
			 u2bLocal.getBase(0),u2bLocal.getBound(0),
			 u2bLocal.getBase(1),u2bLocal.getBound(1),
			 u2bLocal.getBase(2),u2bLocal.getBound(2),
			 // note: use grid1 mesh data here    -- ASSUMES GRIDS MATCH *FIX ME*
			 gidLocal1(0,0), *u2bLocal.getDataPointer(), *mask1p,*prsxy1, *pxy1, bcLocal1(0,0), 
			 ipar[0], rpar[0], 
			 rwk[pa2],rwk[pa4],rwk[pa8], iwk[pipvt2],iwk[pipvt4],iwk[pipvt8],
			 ierr );

    	  // ::display(u1Local,"interfaceBC: u1Local after interfaceCgCm","%5.2f ");
	  // ::display(u2bLocal,"interfaceBC:u2bLocal after interfaceCgCm","%5.2f ");
	}
	if( ok2 )
	{
	  int ipar[]={ //
	    side1, dir1, grid1,          // use grid2 info here
	    m1a,m1b,m2a,m2b,m3a,m3b,     // use grid2 info here
	    side2, dir2, grid2,
	    m1a,m1b,m2a,m2b,m3a,m3b,
	    gridType,            
	    orderOfAccuracyInSpace,    
	    orderOfExtrapolation,
	    useForcing,          
	    tc1,               
	    tc2,                  
            np,
	    myid,
	    normalSign1,
	    normalSign2,
	    0,  // for use 
	    0,  // for use 
	    useWhereMask,       
	    parameters.dbase.get<int >("debug"),
	    numberOfIterationsForInterfaceBC,
	    materialInterfaceOption,
	    interfaceInitialized
	  };
	  real rpar[]={ //
	    dx1[0],
	    dx1[1],
	    dx1[2],
	    mg2.gridSpacing(0),
	    mg2.gridSpacing(1),
	    mg2.gridSpacing(2),
	    dx2[0],
	    dx2[1],
	    dx2[2],
	    mg2.gridSpacing(0),
	    mg2.gridSpacing(1),
	    mg2.gridSpacing(2),
	    t,    
	    (real &)domainSolver[d1]->parameters.dbase.get<OGFunction* >("exactSolution"),  // twilight zone pointer
	    (real &)domainSolver[d2]->parameters.dbase.get<OGFunction* >("exactSolution"),  // twilight zone pointer
	    dt,    
	    ktc1,
	    ktc2,
  	    kappa[0],     // rpar[18] 
	    kappa[1],
	    0., // for later use 
	    0., // for later use 
	    omegaForInterfaceIteration
	  };
	  interfaceCgCm( mg1.numberOfDimensions(), 
			 u1bLocal.getBase(0),u1bLocal.getBound(0),
			 u1bLocal.getBase(1),u1bLocal.getBound(1),
			 u1bLocal.getBase(2),u1bLocal.getBound(2),
			 // note: use grid 2 mesh data here  -- ASSUMES GRIDS MATCH *FIX ME*
			 gidLocal2(0,0), *u1bLocal.getDataPointer(), *mask2p,*prsxy2, *pxy2, bcLocal2(0,0), 
			 u2Local.getBase(0),u2Local.getBound(0),
			 u2Local.getBase(1),u2Local.getBound(1),
			 u2Local.getBase(2),u2Local.getBound(2),
			 gidLocal2(0,0), *u2p, *mask2p,*prsxy2, *pxy2, bcLocal2(0,0), 
			 ipar[0], rpar[0], 
			 rwk[pa2],rwk[pa4],rwk[pa8], iwk[pipvt2],iwk[pipvt4],iwk[pipvt8],
			 ierr );

	  // ::display(u1bLocal,"interfaceBC:u1bLocal after interfaceCgCm","%5.2f ");
    	  // ::display(u2Local,"interfaceBC: u2Local after interfaceCgCm","%5.2f ");

	}
	
//         u1.updateGhostBoundaries(); // ********not needed ***********************
//         u2.updateGhostBoundaries();
	
      #endif

      }
      else
      {
	// Here is a simple interface condition: 
	//     
	//     u1=u2 on the interface 
	//     Extrapolate(u1) 

	const int tc1 = domainSolver[d1]->parameters.dbase.get<int >("tc");    // *** fix this *** assume T for now 
        Range C=Range(tc1,tc1);

	if( false )
	{
// 	  u1(I1,I2,I3,C)=u2(J1,J2,J3,C); 
// 	  u[grid1].applyBoundaryCondition(C,BCTypes::extrapolate,Parameters::interfaceBoundaryCondition,0.,t);
// 	  if( true )
// 	  {
// 	    u[grid1].periodicUpdate();  // *** fix this ***
// 	  }

	}
	else
	{ // 
	  // This is a fake BC: 

	  // do this for now:
	  if( J1.length()!=I1.length() || J2.length()!=I2.length() ) continue;
		    
	  // *** C=Range(parameters.dbase.get<int >("tc"),parameters.dbase.get<int >("tc"));

	  u2(J1,J2,J3,C)=u1(I1,I2,I3,C);   // boundary value of u2 = boundary value of u1

	  getGhostIndex(mg1.gridIndexRange(),side1,dir1,I1,I2,I3,0);
	  getGhostIndex(mg2.gridIndexRange(),side2,dir2,J1,J2,J3,-1);
	  u1(I1,I2,I3,C)=u2(J1,J2,J3,C);   // ghost value of u1 = first line in of u2

	  if( true )
	  {
	    gf1.u[grid1].periodicUpdate();  // *** fix this ***
	    gf2.u[grid2].periodicUpdate();  // *** fix this ***
	  }
                    
      
	}
      }
      

/* ----------------

                  // Now look for any refinement grids that also share this interface
                  if( cg.numberOfRefinementLevels()>1 )
		  {
                    for( int level=1; level<cg.numberOfRefinementLevels(); level++ )
		    {
		      GridCollection & rl = cg.refinementLevel[level];
		      for( int g1=0; g1<rl.numberOfGrids(); g1++ )
		      {
			const int grid1r=rl.gridNumber(g1); 
			const int baseGrid1=rl.baseGridNumber(g1); 
			if( baseGrid1==grid1 && 
                            cg[grid1r].boundaryCondition(side1,dir1)==Parameters::interfaceBoundaryCondition )
			{
			  // *** this refinement grid shares the interface *****


                          // I: interpolate from the coarser grid values of the same side of the interface
			  for( int l=0; l<level; l++ )
			  {
			  }
			  

                          // II: look for refinement grids at the same level on the opposite side of the interface
			  for( int g2=0; g2<rl.numberOfGrids(); g2++ )
			  {
			    const int grid2r=rl.gridNumber(g2); 
			    const int baseGrid2=rl.baseGridNumber(g2); 
			    if( baseGrid2==grid2 &&
				cg[grid2r].sharedBoundaryFlag(side2,dir2)==share1(side1,dir1) )
			    {
                              // grid2r is on the opposite side of the interface 

			    }
			  }
			}
		      }
		    }
		  }
		  
  ------------ */



// 		  if( tz!=NULL && pDebugFile!=NULL )
// 		  {
		    
// 		    OGFunction & e = *tz;

// 		    getGhostIndex(mg1.gridIndexRange(),side1,dir1,I1,I2,I3);

// 		    realArray err(I1,I2,I3);
		  
// 		    err=u1(I1,I2,I3,ex)-e(mg1,I1,I2,I3,ex,t+dt);
// 		    ::display(err,sPrintF("err in u1 (ex,ghost) after interface, t=%e",t+dt),pDebugFile,"%8.1e ");

// 		    getGhostIndex(mg1.gridIndexRange(),side1,dir1,I1,I2,I3,-1);
// 		    err=u1(I1,I2,I3,ex)-e(mg1,I1,I2,I3,ex,t+dt);
// 		    ::display(err,sPrintF("err in u1 (ex,line 1) after interface, t=%e",t+dt),pDebugFile,"%8.1e ");


// 		    getGhostIndex(mg2.gridIndexRange(),side2,dir2,I1,I2,I3);
                    
// 		    err=u2(I1,I2,I3,ex)-e(mg2,I1,I2,I3,ex,t+dt);
// 		    ::display(err,sPrintF("err in u2 (ex,ghost) after interface, t=%e",t+dt),pDebugFile,"%8.1e ");

// 		    getGhostIndex(mg2.gridIndexRange(),side2,dir2,I1,I2,I3,-1);
                    
// 		    err=u2(I1,I2,I3,ex)-e(mg2,I1,I2,I3,ex,t+dt);
// 		    ::display(err,sPrintF("err in u2 (ex,line 1) after interface, t=%e",t+dt),pDebugFile,"%8.1e ");

//		  }
      
    } // end for face
  } // end for inter
  

  // We need to interpolate the grid function if an interface has interpolation points on it:
  // **** NOTE: I think this is needed because we do not check the mask array when assigning
  //            interface points and thus over-write interpolation points **** FIX ME ***
  ForDomain(d)
  {
    if( interpolateThisDomain[d] )
    {
      if( debug() & 8 )
	printF("\n ++++++++++  Cgmp:assignInterfaceBoundaryConditions: Interpolate domain %i after assigning the "
	       "interface values +++++++++++++++\n",d);
      GridFunction & gf = domainSolver[d]->gf[gfIndex[d]];
      gf.u.interpolate();
    }
  }
  


  parameters.dbase.get<RealArray>("timing")(Parameters::timeForInterfaces)+=getCPU()-cpu0;

  return 0;
}


// //\begin{>>InterpolateRefinementsInclude.tex}{\subsection{interpolateRefinementBoundaries}} 
// int InterpolateRefinements::
// interpolateRefinementBoundaries( realGridCollectionFunction & u,
//                                  int levelToInterpolate /* = allLevels */,
//                                  const Range & C0 /* = nullRange */ )
// // ======================================================================================================
// // /Description:
// //    Interpolate the ghost values on refinement grids.
// //
// // /Note: This function assumes that grids are properly nested.
// // 
// // /levelToInterpolate: interpolate just this level (Note: nothing to do on level 0)
// // /C0 (input) : optionally specify which components to interpolate.
// //
// //\end{InterpolateRefinementsInclude.tex} 
// //=========================================================================================
// {
//   real time0=getCPU();
// 
//   if( levelToInterpolate==0 )
//     return 0;
//   
//   GridCollection & gc = *u.getGridCollection();
//   if( gc.numberOfRefinementLevels()<=1 )
//     return 0;
// 
//   if( false )
//   {
//     // use the PCS info
//     gc.updateParentChildSiblingInfo(); // this will only update if the info is out of date
// 
//     int returnValue= interpolateRefinementBoundaries(  *gc.getParentChildSiblingInfo(),u,levelToInterpolate,C0);
//     timeForRefinementBoundaries+=getCPU()-time0;
//     return returnValue;
//   }
//   
//   bool useMask = true;
// 
//   // debug = 3;
//   if( debug>0 )
//     openDebugFile();
// 
//   // const int numberOfDimensions=gc.numberOfDimensions();
//   assert( numberOfDimensions==gc.numberOfDimensions() );
//   
//   Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];
//   Index Jv[3], &J1=Jv[0], &J2=Jv[1], &J3=Jv[2];
//   
//   Range C = C0==nullRange ? Range(u.getComponentBase(0),u.getComponentBound(0)) : C0;
// 
//   int levelStart = levelToInterpolate==allLevels ? 1 : levelToInterpolate;
//   int levelEnd   = levelToInterpolate==allLevels ? gc.numberOfRefinementLevels()-1 : levelToInterpolate;
// 
//   // we assume the refinement ratio is the same for all grids
//   refinementRatio=gc.refinementLevel[1].refinementFactor(Range(0,2),0);
// 
//   // now try to interpolate all refinement grids
//   int level;
//   for( level=levelStart; level<=levelEnd; level++ )
//   {
//     GridCollection & rl = gc.refinementLevel[level];
//     for( int g=0; g<rl.numberOfComponentGrids(); g++ )
//     {
//       // interpolate this grid
//       const int grid=rl.gridNumber(g);
//       const int baseGrid=rl.baseGridNumber(g);
//       MappedGrid & mg=gc[grid];
//       const intArray & mask = mg.mask();
// 
//       realArray & u0 = u[grid];
// 	  
//       GridCollection & rlm1 = gc.refinementLevel[level-1];  // coarser level
// 
//       int c,g2;
//       // for now --first interpolate all ghost boundaries from the coarser level ** fix this **
//       int side,axis;
//       // if( true ) // ****************************************************
//       for( axis=0; axis<numberOfDimensions; axis++ )
//       {
// 	for( side=0; side<=1; side++ )
// 	{
//           if( mg.boundaryCondition(side,axis)==0 )
// 	  {
// 	    BOX box;
// 	    box.convert(IndexType(D_DECL(IndexType::NODE,IndexType::NODE,IndexType::NODE)));
//             int dir;
// 	    for( dir=0; dir<3; dir++ )
// 	    {
//               if( dir==axis )
// 	      {
//                 if( side==0 )
// 		{
// 		  Iv[dir]=Range(mg.indexRange(side,dir)-numberOfGhostLines,
// 				mg.indexRange(side,dir)-1);
// 		}
// 		else
// 		{
// 		  Iv[dir]=Range(mg.indexRange(side,dir)+1,
// 				mg.indexRange(side,dir)+numberOfGhostLines);
// 		}
// 		box.setSmall(dir,Iv[dir].getBase()-side);
// 		box.setBig  (dir,Iv[dir].getBound()+(1-side));
// 	      }
// 	      else if( dir<numberOfDimensions )
// 	      {
//                 // this will include the corner ghost point
//                 Iv[dir]=Range(mg.indexRange(Start,dir)-numberOfGhostLines,mg.indexRange(End,dir)+numberOfGhostLines);
// 		box.setSmall(dir,Iv[dir].getBase());
// 		box.setBig  (dir,Iv[dir].getBound());
// 	      }
//               else // unused dimensions 
// 	      {
//                 Iv[dir]=Range(mg.indexRange(Start,dir),mg.indexRange(End,dir));
// 		box.setSmall(dir,Iv[dir].getBase());
// 		box.setBig  (dir,Iv[dir].getBound());
// 	      }
// 	    }
// 	    
//             if( debug & 2 )
//             printf("interpolateRefinementBoundaries: g=%i side=%i axis=%i interp [%i,%i]x[%i,%i] "
// 		   "gID=[%i,%i]x[%i,%i] id=[%i,%i]x[%i,%i]\n",g,side,axis,I1.getBase(),I1.getBound(),
// 		   I2.getBase(),I2.getBound(),
//                    mg.gridIndexRange(0,0),mg.gridIndexRange(1,0),
// 		   mg.gridIndexRange(0,1),mg.gridIndexRange(1,1),
//                    mg.indexRange(0,0),mg.indexRange(1,0),
// 		   mg.indexRange(0,1),mg.indexRange(1,1));
// 	    
// 	    // If the box crosses a branch cut we try to interpolate it twice, the scond time we shift
// 	    // the box by period so we interpolate the part that was outside the fundamental period.
// 	    for( int split=0; split<=1; split++ )  // 
// 	    {
//               int periodicDirection=0, periodicShift=0;
// 	      if( split==1 )
// 	      {
// 		if( !boxWasAdjustedInPeriodicDirection( box, gc,baseGrid,level,periodicDirection,periodicShift ) )
// 		  break;
// 	      }
// 
// 	      for( g2=0; g2<rlm1.numberOfComponentGrids(); g2++ ) // check grids on coarser level
// 	      {
// 		if( rlm1.baseGridNumber(g2)==baseGrid )
// 		{
//                   // we build an extended box on the base level (includes ghost points)
// 		  BOX box2=level==1 ? buildBaseBox(rlm1[g2]) : rlm1[g2].box(); 
//                   // expand the coarse grid box if periodic
//                   for( dir=0; dir<numberOfDimensions; dir++ )
// 		  {
// 		    if( rlm1[g2].isPeriodic(dir) )
// 		    {
// 		      box2.setSmall(dir,box2.smallEnd(dir)-1);
// 		      box2.setBig(dir,box2.bigEnd(dir)+1);
// 		    }
// 		  }
// 		  
// 
// 		  box2.refine(refinementRatio(0));   // refine coarse grid box so we can intersect it with current box
// 		  if( box2.intersects(box) )
// 		  {
//                     // without Dan's "fix" to Box::intersect we also get intersections of width 1 wide -- when
//                     // we remove boundary points we may end up with no intersection. *wdh* 010922
// 
// 		    BOX intersection = intersects(box2,box);
// 		    bool ok=getIndex(intersection,side,axis,Iv)==0;
//                     if( !ok ) continue;  // no points defined to interpolate.
// 
// 		    int grid2=rlm1.gridNumber(g2);
// 		    // realArray u1(I1,I2,I3,C);
// 
// 		    // interp.interpolateCoarseToFine( u1,Iv,u[grid2],refinementRatio );
//                     if( split!=1 )
// 		    {
// 		      if( debug & 2 )
// 		      {
//                         fprintf(debugFile,
//                                "***interpRefBndry: Before interp from coarse, level=%i, grid=%i grid2=%i "
//                                "Iv=[%i,%i][%i,%i][%i,%i]\n",
//                                 level,grid,grid2,Iv[0].getBase(),Iv[0].getBound(),Iv[1].getBase(),Iv[1].getBound(),
//                                 Iv[2].getBase(),Iv[2].getBound());
// 			
// 			display(u[grid2],"u[grid2] (coarse grid) before interp from coarse",debugFile,"%4.1f");
// 		      }
// 
// 		      interp.interpolateFineFromCoarse( u0,mask,Iv,u[grid2],refinementRatio );
// 
// 		      if( debug & 2 )
// 		      {
// 			display(u0,"u0 (fine grid) after interp from coarse",debugFile,"%4.1f");
// 			display(u[grid2],"u[grid2] (coarse grid) after interp from coarse",debugFile,"%4.1f");
// 		      }
// 		      
// 		    }
// 		    else
// 		    {
// 		      realArray u1(I1,I2,I3,C);
// 		      interp.interpolateFineFromCoarse( u1,Iv,u[grid2],refinementRatio );
// 
// 		      J1=I1; J2=I2; J3=I3;
// 		      if( split==1 )
// 			Jv[periodicDirection]-=periodicShift;
// 
// 		      if( debug & 2 )
// 		      {
// 			printf("Interp bndry (side,axis)=(%i,%i) of grid %i from coarser grid %i I1=(%i,%i) I2=(%i,%i)\n",
// 			       side,axis,grid,grid2,I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound());
// 		     
// 			// displayMask(mask(J1,J2,J3),"mask(J1,J2,J3)");
// 		      }
// 		    
// 		      // display(u1,"u1");
// 
// 		      if( useMask )
// 		      {
// 			where( mask(J1,J2,J3)>0 )
// 			  for( c=C.getBase(); c<=C.getBound(); c++ )
// 			    u0(J1,J2,J3,c)=u1(I1,I2,I3,c);
// 		      }
// 		      else
// 		      {
// 			u0(J1,J2,J3,C)=u1(I1,I2,I3,C);
// 		      }
// 		    }
// 		    
// 		    if( intersection == box || level==1 ) // we are done in these cases (only 1 parent on level 0)
// 		      break;
// 		  }    
// 		}
// 	      }
// 	    }
// 	  }
// 	}  // for side
//       } // for axis
//       
//       
//       BOX initialBox = rl[g].box();
// 
//       // If the box crosses a branch cut we try to interpolate it twice, the scond time we shift
//       // the box by period so we interpolate the part that was outside the fundamental period.
//       for( int split=0; split<=1; split++ )  // 
//       {
//         int periodicDirection=0,periodicShift=0;
// 	if( split==1 )
// 	{
// 	  if( !boxWasAdjustedInPeriodicDirection( initialBox, gc,baseGrid,level,periodicDirection,periodicShift ) )
// 	    break;
// 	}
// 
// 	BOX box = initialBox;
// 	for( axis=0; axis<numberOfDimensions; axis++ )
// 	  box.grow(axis,numberOfGhostLines);  // grow this box to include all ghost lines to be interpolated
// 
//         // try to interpolate the ghost boundaries from the current level
// 	for( g2=0; g2<rl.numberOfComponentGrids(); g2++ )
// 	{
// 	  if( g2!=g && rl.baseGridNumber(g2)==baseGrid && box.intersects( rl[g2].box() ) )
// 	  {
// 	    BOX intersection = intersects(box,rl[g2].box());
// 	    getIndex(intersection,Iv);
// 	    int grid2=rl.gridNumber(g2);
//             // ** we actually copy the boundary points too here -- not necessary ***
// 	    if( debug & 2 )
// 	      printf("Interp bndry of grid %i from same level grid %i I1=(%i,%i) I2=(%i,%i)\n",
// 		     grid,grid2,I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound());
// 
//            
//             if( split==0 )
// 
// 	    {
// 	      if( useMask )
// 	      {
//                 if( true )
// 		{ // opt version (and P++ version)
//                   #ifdef USE_PPP
//                     // first copy (doing any necessary communication)
//                     realArray u2;  u2.partition(u0.getPartition()); 
//                     u2.redim(u0.dimension(0),u0.dimension(1),u0.dimension(2),C);
// 
//                     if( true )
// 		    {
// 		      // ** u2(I1,I2,I3,C)=1.; // we may need to do this to init the DARRAY ??
// 
//                       // Make the copy:
//                       // u2(I1,I2,I3,C)=u[grid2](I1,I2,I3,C);
// 
//                       const int numDims=4;
//                       Index Jv[4];
// 		      Jv[0]=I1; Jv[1]=I2; Jv[2]=I3; Jv[3]=C;
//                       
//                       ParallelUtility::copy( u2, Jv, u[grid2], Jv, numDims);
// 		    }
// 		    else
// 		    {
// 		      if( true || debug & 2 )
// 		      {
// 			printF("Interp bndry of grid %i from same level grid2 %i "
//                         "I1,I2,..C=[%i,%i][%i,%i][%i,%i][%i,%i]\n",
// 			       grid,grid2,I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound(),
//                                    I3.getBase(),I3.getBound(),C.getBase(),C.getBound());
//                         printF(" u[grid2]: [%i,%i][%i,%i][%i,%i][%i,%i]\n",u[grid2].getBase(0),u[grid2].getBound(0),
//                               u[grid2].getBase(1),u[grid2].getBound(1),u[grid2].getBase(2),u[grid2].getBound(2),
//                               u[grid2].getBase(3),u[grid2].getBound(3));
//                         printF(" u2      : [%i,%i][%i,%i][%i,%i][%i,%i]\n",u2.getBase(0),u2.getBound(0),
//                               u2.getBase(1),u2.getBound(1),u2.getBase(2),u2.getBound(2),
//                               u2.getBase(3),u2.getBound(3));
// 			
// 		      }
// 			
// 		      // u2(I1,I2,I3,C)=u[grid2](I1,I2,I3,C);
// 
//                       realArray ug2; ug2.partition(u[grid2].getPartition());
// 		      ug2.redim(u[grid2].dimension(0),u[grid2].dimension(1),
//                                 u[grid2].dimension(2),u[grid2].dimension(3));
// 		      
// 		      ug2=u[grid2];
// 		      ug2.updateGhostBoundaries();
// 
//                       u2.updateGhostBoundaries();
// 		      
// 
// 		      u2(I1,I2,I3,C)=ug2(I1,I2,I3,C);
// 
// 		    }
// 		    
//                     const realSerialArray & u0Local  =  u0.getLocalArrayWithGhostBoundaries();
//                     const realSerialArray & u2Local  =  u2.getLocalArrayWithGhostBoundaries();
//                     const intSerialArray & maskLocal = mask.getLocalArrayWithGhostBoundaries();
// 
//                     int i1a=max(Iv[0].getBase(),u0Local.getBase(0)), i1b=min(Iv[0].getBound(),u0Local.getBound(0));
//                     int i2a=max(Iv[1].getBase(),u0Local.getBase(1)), i2b=min(Iv[1].getBound(),u0Local.getBound(1));
//                     int i3a=max(Iv[2].getBase(),u0Local.getBase(2)), i3b=min(Iv[2].getBound(),u0Local.getBound(2));
// 
//                     if( i1a>i1b || i2a>i2b || i3a>i3b )
//                       continue;
// 		    
//                     Index J1=Range(i1a,i1b);
//                     Index J2=Range(i2a,i2b);
//                     Index J3=Range(i3a,i3b);
//                   #else
//                     const realSerialArray & u0Local  =  u0;
//                     const realSerialArray & u2Local  =  u[grid2];
//                     const intSerialArray & maskLocal = mask;
// 		  
//                     Index &J1 = Iv[0], &J2=Iv[1], &J3=Iv[2];
//                   #endif                   
//                    
// 		  real *u0p = u0Local.Array_Descriptor.Array_View_Pointer3;
// 		  const int u0Dim0=u0Local.getRawDataSize(0);
// 		  const int u0Dim1=u0Local.getRawDataSize(1);
// 		  const int u0Dim2=u0Local.getRawDataSize(2);
//                   #undef U0
//                   #define U0(i0,i1,i2,i3) u0p[i0+u0Dim0*(i1+u0Dim1*(i2+u0Dim2*(i3)))]
// 
// 		  real *u2p = u2Local.Array_Descriptor.Array_View_Pointer3;
// 		  const int u2Dim0=u2Local.getRawDataSize(0);
// 		  const int u2Dim1=u2Local.getRawDataSize(1);
// 		  const int u2Dim2=u2Local.getRawDataSize(2);
//                   #undef U2
//                   #define U2(i0,i1,i2,i3) u2p[i0+u2Dim0*(i1+u2Dim1*(i2+u2Dim2*(i3)))]
// 
//                   const int *maskp = maskLocal.Array_Descriptor.Array_View_Pointer2;
//                   const int maskDim0=maskLocal.getRawDataSize(0);
//                   const int maskDim1=maskLocal.getRawDataSize(1);
//                   const int md1=maskDim0, md2=md1*maskDim1; 
//                   #define MASK(i0,i1,i2) maskp[(i0)+(i1)*md1+(i2)*md2]
// 
//                   const int cBase=C.getBase(), cBound=C.getBound();
//                   int i1,i2,i3;
// 		  FOR_3D(i1,i2,i3,J1,J2,J3)
// 		  {
//                     if( MASK(i1,i2,i3)>0 )
// 		    {
// 		      for( int c=cBase; c<=cBound; c++ )
// 			U0(i1,i2,i3,c)=U2(i1,i2,i3,c);  // copy from same refinement level
// 		    }
// 		    
// 		  }
//                   #undef U0
//                   #undef U2
//                   #undef MASK
// 
// 		}
// 		else
// 		{
// 		  // old way
// 		  where( mask(I1,I2,I3)>0 )
// 		    for( c=C.getBase(); c<=C.getBound(); c++ )
// 		      u0(I1,I2,I3,c)=u[grid2](I1,I2,I3,c);  // copy from same refinement level
// 		}
// 		
// 	      }
// 	      else
// 	      {
// 	        u0(I1,I2,I3,C)=u[grid2](I1,I2,I3,C);  // copy from same refinement level
// 	      }
// 	      
// 	    }
// 	    else
// 	    {
// 	      J1=I1; J2=I2; J3=I3;
// 	      Jv[periodicDirection]-=periodicShift;
// 
// 	      // u0(J1,J2,J3,C)=u[grid2](I1,I2,I3,C);  // copy from same refinement level
// 	      if( useMask )
// 	      {
// 		where( mask(J1,J2,J3)>0 )
// 		  for( c=C.getBase(); c<=C.getBound(); c++ )
// 		    u0(J1,J2,J3,c)=u[grid2](I1,I2,I3,c); // copy from same refinement level
// 	      }		
// 	      else
// 	      {
// 		u0(J1,J2,J3,C)=u[grid2](I1,I2,I3,C);  // copy from same refinement level
// 	      }
// 	    }
// 	    
// 	    // could keep track of how many points are interpolated and break when done.
// 	  }
// 	}
// 
// /* ----
//         // **** BELOW is done elsewhere now ****
// 	// *** still have to interpolate remaining ghost points from the coarser grid.
//   
//         // now fill in coarse grid cells that lie under-neath this patch **** maybe this should be a separate step?
// 	if( false )
// 	{
// 	  
// 	  int ratio=refinementRatio(0); // *** fix this ***
// 
// 	  Box box0 = initialBox;
// 	  box0.coarsen(ratio);
// 
// 	  for( g2=0; g2<rlm1.numberOfComponentGrids(); g2++ )
// 	  {
// 	    if( rlm1.baseGridNumber(g2)==baseGrid && box0.intersects( rlm1[g2].box() ) )
// 	    {
// 	      BOX intersection = intersects(box0,rlm1[g2].box());
// 	      getIndex(intersection,Iv);
// 	      int grid2=rlm1.gridNumber(g2);
// 
// 	      J1=Range(I1.getBase()*ratio,I1.getBound()*ratio,ratio);
// 	      J2=Range(I2.getBase()*ratio,I2.getBound()*ratio,ratio);
// 	      J3=Range(I3.getBase()*ratio,I3.getBound()*ratio,ratio);
// 	  
// 	      if( split==1 )
// 		Jv[periodicDirection]-=periodicShift;
// 
// 	      if( debug & 2 )
// 		printf("Copy coarse grid %i from fine grid %i I1=(%i,%i) I2=(%i,%i)\n",
// 		       grid2,grid,I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound());
// 
// 	      if( useMask )
// 	      {
// 		where( gc[grid2].mask()(I1,I2,I3)>0 )
// 		  for( c=C.getBase(); c<=C.getBound(); c++ )
// 		    u[grid2](I1,I2,I3,c)=u0(J1,J2,J3,c);
// 	      }
// 	      else
// 	      {
// 		u[grid2](I1,I2,I3,C)=u0(J1,J2,J3,C);
// 	      }
// 
// 	      if( intersection == box || level==1 ) // we are done in these cases (only 1 parent on level 0)
// 		break;
// 	    }
// 	  }
// 	}
// ------------ */	
// 	
//       } // end for split
// 
//       #ifdef USE_PPP
//         u0.updateGhostBoundaries();
//       #endif
//       
//     } // end for g
//     
//   }
//   
//   timeForRefinementBoundaries+=getCPU()-time0;
//   
//   return 0;
// }
// 

// ===================================================================================================================
/// \brief Initialize the interface boundary conditions when they are solved by iteration.
/// \details When we iterate to solve the interface conditions we need to specify what sub-set
///   of the interface conditions we solve on each domain.
/// \param t (input) : current time
/// \param dt (input) : current time step
/// \param gfIndex (input) : Domain d should use the grid function : domainSolver[d]->gf[gfIndex[d]]
// ==================================================================================================================
int Cgmp::
initializeInterfaceBoundaryConditions( real t, real dt, std::vector<int> & gfIndex )
{
  real cpu0=getCPU();

  // ************* Much of this code is duplicated from assignInterfaceBoundaryConditions: FIX THIS *********

  if( !gridHasMaterialInterfaces ) return 0;

  InterfaceList & interfaceList = parameters.dbase.get<InterfaceList>("interfaceList");
  
  if( interfaceList.size()==0 )
  {
    // -- Initialize the list of interfaces --
    initializeInterfaces(gfIndex);

    if( interfaceList.size()>0 ) 
      gridHasMaterialInterfaces=true;
    else
    {
      gridHasMaterialInterfaces=false;
      return 0;
    }

    if( interfaceList.size()>0 )
    {
      printF("**** initializeInterfaceBoundaryConditions:initializeInterfaces: number of interfaces =%i\n",
	     interfaceList.size());
    }
  }
  const int np= max(1,Communication_Manager::numberOfProcessors());
  const int & myid = parameters.dbase.get<int>("myid");
  
  const int numberOfDomains=domainSolver.size();
  FILE *& interfaceFile =parameters.dbase.get<FILE* >("interfaceFile");

  const bool solveCoupledInterfaceEquations = parameters.dbase.get<bool>("solveCoupledInterfaceEquations");

  Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];
  Index Jv[3], &J1=Jv[0], &J2=Jv[1], &J3=Jv[2];
  

  // loop over interfaces
  for( int inter=0; inter < interfaceList.size(); inter++ )
  {
    InterfaceDescriptor & interfaceDescriptor = interfaceList[inter]; 

    // there may be multiple grid faces that lie on the interface:     
    for( int face=0; face<interfaceDescriptor.gridListSide1.size(); face++ )
    {
      GridFaceDescriptor & gridDescriptor1 = interfaceDescriptor.gridListSide1[face];
      GridFaceDescriptor & gridDescriptor2 = interfaceDescriptor.gridListSide2[face];
      
      const int d1=gridDescriptor1.domain, 
                grid1=gridDescriptor1.grid, side1=gridDescriptor1.side, dir1=gridDescriptor1.axis;
      const int d2=gridDescriptor2.domain, 
                grid2=gridDescriptor2.grid, side2=gridDescriptor2.side, dir2=gridDescriptor2.axis;

      assert( d1>=0 && d1<numberOfDomains && d2>=0 && d2<numberOfDomains );

      GridFunction & gf1 = domainSolver[d1]->gf[gfIndex[d1]];
      GridFunction & gf2 = domainSolver[d2]->gf[gfIndex[d2]];

      CompositeGrid & cg1 = gf1.cg;
      assert( grid1>=0 && grid1<=cg1.numberOfComponentGrids());
      MappedGrid & mg1 = cg1[grid1];
      const IntegerArray & bc1 = mg1.boundaryCondition();
      const IntegerArray & share1 = mg1.sharedBoundaryFlag();
      const IntegerArray & interfaceType1 = domainSolver[d1]->parameters.dbase.get<IntegerArray >("interfaceType");

      CompositeGrid & cg2 = gf2.cg;
      assert( grid2>=0 && grid2<=cg2.numberOfComponentGrids());
      MappedGrid & mg2 = cg2[grid2];
      const IntegerArray & bc2 = mg2.boundaryCondition();
      const IntegerArray & share2 = mg2.sharedBoundaryFlag();
      const IntegerArray & interfaceType2 = domainSolver[d2]->parameters.dbase.get<IntegerArray >("interfaceType");

      IntegerArray gidLocal1(2,3), dimLocal1(2,3), bcLocal1(2,3);
      getLocalBoundsAndBoundaryConditions( gf1.u[grid1],gidLocal1,dimLocal1,bcLocal1 );

      IntegerArray gidLocal2(2,3), dimLocal2(2,3), bcLocal2(2,3);
      getLocalBoundsAndBoundaryConditions( gf2.u[grid2],gidLocal2,dimLocal2,bcLocal2 );

      const int numberOfDimensions = cg1.numberOfDimensions();

      assert( interfaceType1(side1,dir1,grid1) == interfaceType2(side2,dir2,grid2) );

      if( interfaceType1(side1,dir1,grid1)==Parameters::heatFluxInterface )
      {
        // ********************************************
	// ********** Heat Flux Interface *************
        // ********************************************

	const int tc1 = domainSolver[d1]->parameters.dbase.get<int >("tc");    // *** fix this *** assume T for now 
	const int tc2 = domainSolver[d2]->parameters.dbase.get<int >("tc");

	real ktc[2]={1.,1.};

	ktc[0] = domainSolver[d1]->parameters.dbase.get<real>("thermalConductivity");
	ktc[1] = domainSolver[d2]->parameters.dbase.get<real>("thermalConductivity");
	if( ktc[0] <=0. || ktc[1]<=0. )
	{
	  printF("Cgmp::initializeInterfaceBoundaryConditions:ERROR: a negative thermal conductivity"
		 " was found, ktc[0]=%e ktc[1]=%e\n",ktc[0],ktc[1]);
	  Overture::abort("error");
	}

	real ktc1=ktc[0], ktc2=ktc[1];
      
	// ************ fix this ***********

	real ktd[2]={-1.,-1.}; //  coefficient of thermal diffusion 
	for( int dd=0; dd<=1; dd++ )
	{
	  const int d = dd==0 ? d1 : d2;
	  if( domainSolver[d]->parameters.dbase.has_key("kappa") )
	  {
	    ktd[dd]=domainSolver[d]->parameters.dbase.get<std::vector<real> >("kappa")[0];
	  } 
	  else if( domainSolver[d]->parameters.dbase.has_key("kThermal") )
	  {
	    ktd[dd]=domainSolver[d]->parameters.dbase.get<real>("kThermal");
	  }
	}

	fPrintF(interfaceFile,
		"Cgmp::initInterfaceBCs: interface %i : heat-flux interface (t=%e) (solveCoupled=%i):\n"
		"  (d1,grid1,side1,dir1,bc1)=(%i,%i,%i,%i,%i) kappa1=%9.3e (diffusion), k1=%9.3e (conduct.)\n"
		"  (d2,grid2,side2,dir2,bc2)=(%i,%i,%i,%i,%i) kappa2=%9.3e (diffusion), k2=%9.3e (conduct.)\n",
		inter,t,solveCoupledInterfaceEquations,
		d1,grid1,side1,dir1,mg1.boundaryCondition(side1,dir1),ktd[0],ktc1,
		d2,grid2,side2,dir2,mg2.boundaryCondition(side2,dir2),ktd[1],ktc2);

        const real kRatio = (ktc[0]/ktc[1])*sqrt(ktd[1]/ktd[0]);

	if( solveCoupledInterfaceEquations )
	{
	  // Apply Neumann BC's on both sides if we solve the coupled interface equations 
	  gridDescriptor1.interfaceBC=Parameters::neumannInterface;  
	  gridDescriptor1.a[0]=0.; gridDescriptor1.a[1]=ktc[0];
	  gridDescriptor2.interfaceBC=Parameters::neumannInterface;; 
	  gridDescriptor2.a[0]=0.; gridDescriptor2.a[1]=ktc[1];
	}
	// else if( ktc1 > ktc2 ) // ** this condition must match that in assignInterfaceRightHandSide **
	else if( parameters.dbase.get<bool>("useMixedInterfaceConditions") )
	{
	  // use mixed interface conditions
	  //   a11=1, a12 = k2*beta2 = k2*sqrt( 1/(theta2*kappa2*dt) + k2^2 )
	  //   a12=1, a22 = k1*beta1 = k1*sqrt( 1/(theta1*kappa1*dt) + k1^2 )

          const real theta1 = domainSolver[d1]->parameters.dbase.get<real>("implicitFactor");
          const real theta2 = domainSolver[d2]->parameters.dbase.get<real>("implicitFactor");
          const real kFact1=.1;
          const real kFact2=.1; 

          const real kbeta1 = ktc[0]*sqrt( 1./(ktd[0]*theta1*dt) + kFact1/(ktd[0]*dt) );
          const real kbeta2 = ktc[1]*sqrt( 1./(ktd[1]*theta2*dt) + kFact2/(ktd[1]*dt) );

	  real a11=1., a12=kbeta2;
	  real a21=1., a22=kbeta1;

	  if( true )
	  {
            // try this instead -- this should work better in the limiting cases when 
            //         kbeta1 << kbeta2   or   kbeta1 >> kbeta2
            // a12=kbeta2/kbeta1;
	    // a22=kbeta1/kbeta2;
            a12=kbeta2*kbeta2/kbeta1;
	    a22=kbeta1*kbeta1/kbeta2;

            if( true )
	    { // this next seems to work better -- make one side nearly a dirichlet condition: 
	      if( kbeta1>=kbeta2 )
	      {
		a22=kbeta1*kbeta1;
	      }
	      else
	      {
		a12=kbeta2*kbeta2;
	      }
	      
	    }
	    
	  }

	  gridDescriptor1.interfaceBC=Parameters::neumannInterface;   
	  gridDescriptor1.a[0]=a12; gridDescriptor1.a[1]=a11*ktc[0];

	  gridDescriptor2.interfaceBC=Parameters::neumannInterface;  
	  gridDescriptor2.a[0]=a22; gridDescriptor2.a[1]=a21*ktc[1];

	 fPrintF(interfaceFile,"Cgmp:initInterfaceBCs: use mixed-interface %i: a11=%8.2e, a12=%8.2e; a21=%8.2e, a22=%8.2e"
                 " (theta1=%8.2e, theta2=%8.2e, dt=%8.2e)\n",
		 inter,a11,a12,a21,a22, theta1,theta2,dt);
	}
	else if( kRatio > 1. ) // *wdh* 080720 
	{

	  // use Dirichlet-Neumann interface conditions
          if( t < dt )
  	    fPrintF(interfaceFile,"Cgmp:initInterfaceBCs: use Dirichlet(domain %i)-Neumann(domain %i) interface\n",d1,d2);

	  // Domain 1 : Neumann interface BC, Domain 2: Dirichlet interface BC
	  gridDescriptor1.interfaceBC=Parameters::neumannInterface;  
	  gridDescriptor1.a[0]=0.; gridDescriptor1.a[1]=ktc1;
	  gridDescriptor2.interfaceBC=Parameters::dirichletInterface; 
	  gridDescriptor2.a[0]=1.; gridDescriptor2.a[1]=0.;
	}
	else
	{
	  // Domain 2 : Neumann interface BC, Domain 1: Dirichlet interface BC
          if( t < dt )
   	    fPrintF(interfaceFile,"Cgmp:initInterfaceBCs: use Dirichlet(domain %i)-Neumann(domain %i) interface\n",d2,d1);

	  gridDescriptor1.interfaceBC=Parameters::dirichletInterface;  
	  gridDescriptor1.a[0]=1.; gridDescriptor1.a[1]=0.;
	  gridDescriptor2.interfaceBC=Parameters::neumannInterface;    
	  gridDescriptor2.a[0]=0.; gridDescriptor2.a[1]=ktc2;
	}
	domainSolver[d1]->setInterfaceBoundaryCondition( gridDescriptor1 );
	domainSolver[d2]->setInterfaceBoundaryCondition( gridDescriptor2 );
      }
      else if( interfaceType1(side1,dir1,grid1)==Parameters::tractionInterface )
      {
        // *******************************************
	// ********** Traction Interface *************
        // *******************************************


        // -- for now nothing needs to be done in this case.


      }
      else
      {
	printF("Cgmp::initializeInterfaceBoundaryConditions:ERROR:unexpected interfaceType=%i\n",
	       interfaceType1(side1,dir1,grid1));
	Overture::abort("error");
      }

    } // end for face
  } // end for inter

  parameters.dbase.get<RealArray>("timing")(Parameters::timeForInterfaces)+=getCPU()-cpu0;
  return 0;
}


// ===================================================================================================================
/// \brief Assign the RHS for the interface equations on domain d
/// \details Assign values in the BoundaryDataArray : domainSolver[d]->getBoundaryData(grid); on domain
///    d by evaluating the RHS for the interface equations. For example, the RHS may be the solution value "u"
///    or the normal component of the "stress", k u.n from the opposite side of the interface.
///   
/// \param d (input) : assign RHS for this domain.
/// \param t (input) : current time
/// \param dt (input) : current time step
/// \param correct (input) : correction step number.
/// \param gfIndex (input) : Domain d should use the grid function : domainSolver[d]->gf[gfIndex[d]]
// 
// ==================================================================================================================
int Cgmp::
assignInterfaceRightHandSide( int d, real t, real dt, int correct, std::vector<int> & gfIndex )
{
  real cpu0=getCPU();

  // ************* Much of this code is duplicated from assignInterfaceBoundaryConditions: FIX THIS *********

  if( !gridHasMaterialInterfaces ) return 0;
  
  InterfaceList & interfaceList = parameters.dbase.get<InterfaceList>("interfaceList");
  
  if( interfaceList.size()==0 )
  {
    // -- Initialize the list of interfaces --
    initializeInterfaces(gfIndex);

    if( interfaceList.size()>0 ) gridHasMaterialInterfaces=true;

    if( interfaceList.size()>0 )
    {
      printF("**** assignInterfaceBoundaryConditions:initializeInterfaces: number of interfaces =%i\n",
	     interfaceList.size());
    }
  }
  const int np= max(1,Communication_Manager::numberOfProcessors());
  const int & myid = parameters.dbase.get<int>("myid");
  
  const int numberOfDomains=domainSolver.size();
  int numberOfInterfaceHistoryValuesToSave = parameters.dbase.get<int>("numberOfInterfaceHistoryValuesToSave");
  int numberOfInterfaceIterateValuesToSave = parameters.dbase.get<int>("numberOfInterfaceIterateValuesToSave");
  FILE *& interfaceFile =parameters.dbase.get<FILE* >("interfaceFile");
  
  // real & omega = parameters.dbase.get<real>("interfaceOmega");

  Index Iv[4], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2], &I4=Iv[3];
  Index Jv[4], &J1=Jv[0], &J2=Jv[1], &J3=Jv[2], &J4=Jv[3];
  
  Index Iav[4], &Ia1=Iav[0], &Ia2=Iav[1], &Ia3=Iav[2], &Ia4=Iav[3];
  Index Jav[4], &Ja1=Jav[0], &Ja2=Jav[1], &Ja3=Jav[2], &Ja4=Jav[3];
  

  // ************************************************
  // ******* Apply the interface conditions. ********
  // ************************************************

  // loop over interfaces
  for( int inter=0; inter < interfaceList.size(); inter++ )
  {
    InterfaceDescriptor & interfaceDescriptor = interfaceList[inter]; 

    // There may be multiple grid faces that lie on the interface:     
    for( int face=0; face<interfaceDescriptor.gridListSide1.size(); face++ )
    {
      GridFaceDescriptor & gridDescriptor1 = interfaceDescriptor.gridListSide1[face];
      GridFaceDescriptor & gridDescriptor2 = interfaceDescriptor.gridListSide2[face];
      
      const int d1=gridDescriptor1.domain, 
                grid1=gridDescriptor1.grid, side1=gridDescriptor1.side, dir1=gridDescriptor1.axis;
      const int d2=gridDescriptor2.domain, 
                grid2=gridDescriptor2.grid, side2=gridDescriptor2.side, dir2=gridDescriptor2.axis;

      assert( d1>=0 && d1<numberOfDomains && d2>=0 && d2<numberOfDomains );

      // look for an interface on domain d : 
      if( d1!=d && d2!=d ) continue;
      

      GridFunction & gf1 = domainSolver[d1]->gf[gfIndex[d1]];
      GridFunction & gf2 = domainSolver[d2]->gf[gfIndex[d2]];

      CompositeGrid & cg1 = gf1.cg;
      assert( grid1>=0 && grid1<=cg1.numberOfComponentGrids());
      MappedGrid & mg1 = cg1[grid1];
      const IntegerArray & bc1 = mg1.boundaryCondition();
      const IntegerArray & share1 = mg1.sharedBoundaryFlag();
      const IntegerArray & interfaceType1 = domainSolver[d1]->parameters.dbase.get<IntegerArray >("interfaceType");
      const intArray & mask1 = mg1.mask();

      CompositeGrid & cg2 = gf2.cg;
      assert( grid2>=0 && grid2<=cg2.numberOfComponentGrids());
      MappedGrid & mg2 = cg2[grid2];
      const IntegerArray & bc2 = mg2.boundaryCondition();
      const IntegerArray & share2 = mg2.sharedBoundaryFlag();
      const IntegerArray & interfaceType2 = domainSolver[d2]->parameters.dbase.get<IntegerArray >("interfaceType");
      const intArray & mask2 = mg2.mask();

      IntegerArray gidLocal1(2,3), dimLocal1(2,3), bcLocal1(2,3);
      getLocalBoundsAndBoundaryConditions( gf1.u[grid1],gidLocal1,dimLocal1,bcLocal1 );

      IntegerArray gidLocal2(2,3), dimLocal2(2,3), bcLocal2(2,3);
      getLocalBoundsAndBoundaryConditions( gf2.u[grid2],gidLocal2,dimLocal2,bcLocal2 );

      const int numberOfDimensions = cg1.numberOfDimensions();

      const int orderOfAccuracyInSpace=2;  // **** do this for now ***

      #ifdef USE_PPP
        intSerialArray mask1Local; getLocalArrayWithGhostBoundaries(mask1,mask1Local);
        intSerialArray mask2Local; getLocalArrayWithGhostBoundaries(mask2,mask2Local);
      #else
        const intSerialArray & mask1Local = mask1;
        const intSerialArray & mask2Local = mask2;
      #endif

      if( ( debug() & 4 ) && correct==0 )
      {
	printF("Cgmp::assignInterfaceRightHandSide: interface found (t=%e):\n"
	       "  (d1,grid1,side1,dir1,bc1)=(%i,%i,%i,%i,%i)\n"
	       "  (d2,grid2,side2,dir2,bc2)=(%i,%i,%i,%i,%i)\n",t,
	       d1,grid1,side1,dir1,mg1.boundaryCondition(side1,dir1),
	       d2,grid2,side2,dir2,mg2.boundaryCondition(side2,dir2));
      }
      

      const int extra=0; // orderOfAccuracyInSpace/2;
      getBoundaryIndex(mg1.gridIndexRange(),side1,dir1,I1,I2,I3,extra);
      getBoundaryIndex(mg2.gridIndexRange(),side2,dir2,J1,J2,J3,extra);
      Ia1=I1, Ia2=I2, Ia3=I3; // save full range 
      Ja1=J1, Ja2=J2, Ja3=J3;
  
      // check that the number of points in the tangential directions match -- eventually we will fix this
      for( int dir=1; dir<mg1.numberOfDimensions(); dir++ )
      {
	int dir1p = (dir1+dir) % mg1.numberOfDimensions();
	int dir2p = (dir2+dir) % mg2.numberOfDimensions();
	if( Iv[dir1p].getLength()!=Jv[dir2p].getLength() )
	{
	  printF("assignInterfaceRightHandSide:ERROR: The number of grid points on the two interfaces do not match\n"
	         " (d1,grid1,side1,dir1,bc1)=(%i,%i,%i,%i,%i) Iv=[%i,%i][%i,%i][%i,%i]\n"
		 " (d2,grid2,side2,dir2,bc2)=(%i,%i,%i,%i,%i) Jv=[%i,%i][%i,%i][%i,%i]\n",
		 d1,grid1,side1,dir1,mg1.boundaryCondition(side1,dir1),
                   I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound(),I3.getBase(),I3.getBound(),  
                 d2,grid2,side2,dir2,mg2.boundaryCondition(side2,dir2),
                   J1.getBase(),J1.getBound(),J2.getBase(),J2.getBound(),J3.getBase(),J3.getBound());
	  Overture::abort("error");
	}
      }

      int includeGhost=0;  // is this right ? 
      bool ok1 = ParallelUtility::getLocalArrayBounds(mask1,mask1Local,I1,I2,I3,includeGhost);
      bool ok2 = ParallelUtility::getLocalArrayBounds(mask2,mask2Local,J1,J2,J3,includeGhost);


      GridFaceDescriptor info1(d1,grid1,side1,dir1);
      GridFaceDescriptor info2(d2,grid2,side2,dir2);

      // Domain da (=d1 or d2) is the domain "d" that we are assigning values to 
      // Domain db (=d1 or d2) is the domain "d" that we are getting values from
      const int da = d==d1 ? d1 : d2;
      const int db = d==d1 ? d2 : d1;
      const bool oka = d==d1 ? ok1 : ok2;
      const bool okb = d==d1 ? ok2 : ok1;
      
      GridFaceDescriptor & gridDescriptora = d==d1 ? gridDescriptor1 : gridDescriptor2;
      GridFaceDescriptor & gridDescriptorb = d==d1 ? gridDescriptor2 : gridDescriptor1;
      
      GridFaceDescriptor & infoa = d==d1 ? info1 : info2;
      GridFaceDescriptor & infob = d==d1 ? info2 : info1;

      RealArray u1,u2;
      RealArray & ua = d==d1 ? u1 : u2;
      RealArray & ub = d==d1 ? u2 : u1;
      info1.u=&u1;
      info2.u=&u2;

      real a0,a1;
      a0=gridDescriptora.a[0];
      a1=gridDescriptora.a[1];

      bool matchFlux= a1!=0.;

      Range Ia[4], Ib[4]; 

      assert( interfaceType1(side1,dir1,grid1) == interfaceType2(side2,dir2,grid2) );

      // find out from the domain we are assigning values to which interface data it wants: 
      int interfaceDataOptions=0;
      domainSolver[da]->getInterfaceDataOptions( infoa,interfaceDataOptions );

      if( interfaceType1(side1,dir1,grid1)==Parameters::heatFluxInterface )
      {
        // ********************************************
	// ********** Heat Flux Interface *************
        // ********************************************

	const int tc1 = domainSolver[d1]->parameters.dbase.get<int >("tc");  
	const int tc2 = domainSolver[d2]->parameters.dbase.get<int >("tc");

	real ktc[2]={1.,1.};

	ktc[0] = domainSolver[d1]->parameters.dbase.get<real>("thermalConductivity");
	ktc[1] = domainSolver[d2]->parameters.dbase.get<real>("thermalConductivity");
	if( ktc[0] <=0. || ktc[1]<=0. )
	{
	  printF("assignInterfaceRightHandSide:ERROR: a negative thermal conductivity was found, "
		 "ktc[0]=%e ktc[1]=%e\n",ktc[0],ktc[1]);
	  Overture::abort("error");
	}

	real ktd[2]={-1.,-1.}; //  coefficient of thermal diffusion 
	for( int dd=0; dd<=1; dd++ )
	{
	  const int d = dd==0 ? d1 : d2;
	  if( domainSolver[d]->parameters.dbase.has_key("kappa") )
	  {
	    ktd[dd]=domainSolver[d]->parameters.dbase.get<std::vector<real> >("kappa")[0];
	  } 
	  else if( domainSolver[d]->parameters.dbase.has_key("kThermal") )
	  {
	    ktd[dd]=domainSolver[d]->parameters.dbase.get<real>("kThermal");
	  }
	}

	real ktc1=ktc[0], ktc2=ktc[1];

	Range C1(tc1,tc1), C2(tc2,tc2);
        Iv[3]=C1; Jv[3]=C2;
	if( ok1 ) u1.redim(I1,I2,I3,C1);
	if( ok2 ) u2.redim(J1,J2,J3,C2);

	if( d==d1 )
	{
	  Ia[0]=I1; Ia[1]=I2; Ia[2]=I3; Ia[3]=Range(tc1,tc1);
	  Ib[0]=J1; Ib[1]=J2; Ib[2]=J3; Ib[3]=Range(tc2,tc2);
	}
	else
	{
	  Ia[0]=J1; Ia[1]=J2; Ia[2]=J3; Ia[3]=Range(tc2,tc2);
	  Ib[0]=I1; Ib[1]=I2; Ib[2]=I3; Ib[3]=Range(tc1,tc1);
	}
	
        // *wdh* 080717 -- just use:  (?)
        //       infob.a[0]=  gridDescriptora.a[0];
        //       infob.a[1]= -gridDescriptora.a[1];  // flip sign of normal to match 


	// -- fix this for a mixed BC ---
	if( parameters.dbase.get<bool>("useMixedInterfaceConditions") )
	{
          // *note* flip the sign of the normal to match normal from domain da: 
          // *note* flip k1 <--> k2 
          infob.a[0]=gridDescriptora.a[0]; infob.a[1]=-gridDescriptora.a[1]*ktc[db]/ktc[da];
	}
	else if( matchFlux )
	{
	  // *note* flip the sign of the normal to match normal from domain 1

          const real kRatio =  (ktc[0]/ktc[1])*sqrt(ktd[1]/ktd[0]);

	  // if( ktc1 > ktc2 ) // *wdh* 080522 
	  if( kRatio > 1. ) // *wdh* 080522 
	  {
  	    infob.a[0]=0.; infob.a[1]=-ktc2;
	  }
          else
	  {
  	    infob.a[0]=0.; infob.a[1]=-ktc1;
	  }


          // infob.a[1]=-ktc[db];  // **** use this ? **********************************************************************
	  
	}
	else
	{
	  infob.a[0]=1.; infob.a[1]=0.;
	}

	if( okb )
          domainSolver[db]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,infob,gridDescriptorb,gfIndex[db],t );

      }
      else if( interfaceType1(side1,dir1,grid1)==Parameters::tractionInterface )
      {
        // *******************************************
	// ********** Traction Interface *************
        // *******************************************

        // If domain "d" is an elastic-solid, we need to get the surface traction from the other domain
        // If domain "d" is a fluid, we need to get the interface position from the other domain

	// if( domainSolver[da]->getClassName()=="Cgsm" )

	const int uc1 = domainSolver[d1]->parameters.dbase.get<int >("uc");
	const int uc2 = domainSolver[d2]->parameters.dbase.get<int >("uc");

	// *wdh* 081201 Range C1(uc1,uc1+numberOfDimensions-1), C2(uc2,uc2+numberOfDimensions-1); // ************ fix me **************
        // for now add some extra space: 
	Range C1(uc1,uc1+2*numberOfDimensions-1), C2(uc2,uc2+2*numberOfDimensions-1); // ************ fix me **************
        Iv[3]=C1; Jv[3]=C2;
	if( ok1 ) u1.redim(I1,I2,I3,C1);
	if( ok2 ) u2.redim(J1,J2,J3,C2);

        if( ok1 ) u1=0.;
	if( ok2 ) u2=0.;

	if( d==d1 )
	{
	  Ia[0]=I1; Ia[1]=I2; Ia[2]=I3; Ia[3]=C1;
	  Ib[0]=J1; Ib[1]=J2; Ib[2]=J3; Ib[3]=C2;
	}
	else
	{
	  Ia[0]=J1; Ia[1]=J2; Ia[2]=J3; Ia[3]=C2;
	  Ib[0]=I1; Ib[1]=I2; Ib[2]=I3; Ib[3]=C1;
	}

        if( okb )
  	  domainSolver[db]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,infob,gridDescriptorb,gfIndex[db],t );


        // We need to adjust for the fact that the Domain solvers have different values for uc


      }
      else
      {
	printF("Cgmp::assignInterfaceRightHandSide:ERROR:unexpected interfaceType=%i\n",
	       interfaceType1(side1,dir1,grid1));
	Overture::abort("error");
      }


      //  --- Copy of u2 -> u1  ---
      // ** todo: In parallel ua may have a different distribution than ub ***
	// parallel :
        //     Allocate distributed arrays ubd(mg2),  ubd1(mg1)
        //        ubdLocal = ub; 
        //        copyArray ubd -> ubd1;
        //        uba <- ubd1Local 
        //        

      #ifdef USE_PPP 

	  // parallel :  we need to transfer the local ub arrays onto the parallel distribution of ua
    
          // **** --- check me ----

          Index D1,D2,D3;

          realArray u2d; 
	  u2d.partition(gf2.u[grid2].getPartition());
          getIndex(mg2.dimension(),D1,D2,D3);
	  u2d.redim(D1,D2,D3,Jv[3]);
          RealArray u2Local; getLocalArrayWithGhostBoundaries(u2d,u2Local);
	  
          realArray u1d;
	  u1d.partition(gf1.u[grid1].getPartition());
          getIndex(mg1.dimension(),D1,D2,D3);
	  u1d.redim(D1,D2,D3,Iv[3]);
          RealArray u1Local; getLocalArrayWithGhostBoundaries(u1d,u1Local);

          Iav[3]=Iv[3]; Jav[3]=Jv[3];

	  if( Ia1.getLength()==Ja1.getLength() && 
	      Ia2.getLength()==Ja2.getLength() &&
	      Ia3.getLength()==Ja3.getLength() )
	  {
	  }
	  else if( Ia1.getLength()==Ja1.getLength() && 
		   Ia2.getLength()==Ja2.getLength() &&
		   Ia3.getLength()==Ja3.getLength()  )
	  {
	    printF("Cgmp::assignInterfaceRightHandSide:ERROR: arrays on interface do not match conformally\n"
                   "  The first two Index's are switched -- finish me for parallel. \n");
	    Overture::abort("error");
	  }
	  else
	  {
	    printF("Cgmp::assignInterfaceRightHandSide:ERROR: arrays on interface do not match conformally\n");
            printF(" Ia1=[%i,%i] Ia2=[%i,%i] Ia3=[%i,%i]  Ja1=[%i,%i] Ja2=[%i,%i] Ja3=[%i,%i] \n",
		   Ia1.getBase(),Ia1.getBound(),Ia2.getBase(),Ia2.getBound(),Ia3.getBase(),Ia3.getBound(), 
		   Ja1.getBase(),Ja1.getBound(),Ja2.getBase(),Ja2.getBound(),Ja3.getBase(),Ja3.getBound());
	    Overture::abort("error");
	  }
	
	  if( d==d1 )
	  {
	    if( ok2) u2Local(Ib[0],Ib[1],Ib[2],Ib[3])=ub(Ib[0],Ib[1],Ib[2],Ib[3]);  
	    CopyArray::copyArray( u1d, Iav, u2d, Jav );  
  	    if( oka ) ua(Ia[0],Ia[1],Ia[2],Ia[3])=u1Local(Ia[0],Ia[1],Ia[2],Ia[3]);
	  }
	  else
	  {
	    if( ok1) u1Local(Ib[0],Ib[1],Ib[2],Ib[3])=ub(Ib[0],Ib[1],Ib[2],Ib[3]);  
	    CopyArray::copyArray( u2d, Jav, u1d, Iav );  
  	    if( oka ) ua(Ia[0],Ia[1],Ia[2],Ia[3])=u2Local(Ia[0],Ia[1],Ia[2],Ia[3]);
	  }
	  


      #else
        // serial 

        ua=0.;
	if( Ia[0].getLength()==Ib[0].getLength() && 
	    Ia[1].getLength()==Ib[1].getLength() &&
	    Ia[2].getLength()==Ib[2].getLength() )
	{
	  ua(Ia[0],Ia[1],Ia[2],Ia[3])=ub(Ib[0],Ib[1],Ib[2],Ib[3]);
	}
	else if( Ia[0].getLength()==Ib[1].getLength() && 
		 Ia[1].getLength()==Ib[0].getLength() &&
		 Ia[2].getLength()==Ib[2].getLength()  )
	{
	  int i1,i2,i3,i4,j1,j2,j3,j4;
	  // FOR_3D(i1,i2,i3,Ia[0],Ia[1],Ia[2]) 
	  // {
	  //   for( int i4=Ia[3].getBase(), j4=Ib[3].getBase(); i4<=Ia[3].getBound(); i4++, j4++ )
	  //     ua(i1,i2,i3,i4)=ub(i2,i1,i3,j4);  // note switch of i1,i2 in ub 
	  // }
	  FOR_4IJD(i1,i2,i3,i4,Ia[0],Ia[1],Ia[2],Ia[3],j1,j2,j3,j4,Ib[0],Ib[1],Ib[2],Ib[3])
	  {
            ua(i1,i2,i3,i4)=ub(j2,j1,j3,j4);  // note switch of j1,j2 in ub 
	  }
	  
	}
	else
	{
	  printF("Cgmp::assignInterfaceRightHandSide:ERROR: arrays on interface do not match conformally\n");
	  Overture::abort("error");
	}
      #endif
      
	if( debug() & 8 )
	{
	  ::display(ub,"RHS ub (from get)");
	  ::display(ua,"RHS ua (=ub) (for set)");
	}
	

      // *********************************************************
      // *********** Extrapolate the initial guess ***************
      // *********************************************************


      bool extrapolateFirstGuess=parameters.dbase.get<bool>("extrapolateInitialInterfaceValues");
      if( interfaceType1(side1,dir1,grid1)==Parameters::tractionInterface ||
          interfaceType2(side2,dir2,grid2)==Parameters::tractionInterface )
      {
        // we currently don't extrapolate for traction interfaces
	extrapolateFirstGuess=false;
      }
      

      // We should only need to extrapolate for one side the interface: 
      //   Extrapolate the side with the lower domain number -- this is assumed to appear first

      extrapolateFirstGuess = extrapolateFirstGuess && (d1<d2 ? d==d1 : d==d2);

      if( extrapolateFirstGuess && correct==0 )
      {
	// We have saved both u and k*u.n from previous times (for TZ we have saved u-ue, k*u.n - k*ue.n ).
        // We can extrapolate, in time, the RHS for the interface condition

        //  If we are solving the interface condition by iteration: (j= iteration number)
        //    a*k1*u.n[j+1] + b*u[j+1] = a*k2*v.n[j] + b*v[j]  

        // As a first guess we can use
        //    a*k1*u.n[0] + b*u[0] = 2.*f(t-dt) - f(t-2*dt) 
        // where
        //    f(t) = a*k2*v.n(t) + b*v(t)
        // OR
        //    f(t) = a*k1*u.n(t) + b*u(t)   <- this is better to use for parallel
       
	// a history of interface values at past times: 
        // **** For parallel: I think we can use gridDescriptora here and switch sign of a1 below ****
	// ** InterfaceDataHistory & idh = gridDescriptorb.interfaceDataHistory;   // values from the opposite side of the interface
 
        // consistency check: we currently don't extrapolate for traction interfaces
        assert( interfaceType1(side1,dir1,grid1)!=Parameters::tractionInterface );
        assert( interfaceType2(side2,dir2,grid2)!=Parameters::tractionInterface );
	

	InterfaceDataHistory & idh = gridDescriptora.interfaceDataHistory;   // values from the same side of the interface

	if( idh.interfaceDataList.size()>=2 )
	{
	  const int prev = ( idh.current -1 + numberOfInterfaceHistoryValuesToSave ) % 
	    numberOfInterfaceHistoryValuesToSave;
	    
	  real tc = idh.interfaceDataList[idh.current].t;
	  real tp = idh.interfaceDataList[prev       ].t;
	    
	    
	  // extrap in time: f(t) = (t-t1)/(t2-t1)*u2 + (t2-t)/(t2-t1)*u1

	  // 2nd order extrap in time:
	  real cex1=2., cex2=-1.;
	  real dtex = tc-tp;
	  if( dtex > dt*.1 )
	  { // adjust for variable time step (if dtex is not too small)
	    cex1=(t-tp)/dtex; cex2=(tc-t)/dtex;
	  }
	  if( debug() & 2 )
	    fPrintF(interfaceFile," +++interfaceRHS: interface %i: domain d=%i, t=%9.3e extrap RHS in time from "
		    " tc=%8.2e tp=%8.2e cex=(%8.2e,%8.2e)\n",
		    inter,d,t,tc,tp,cex1,cex2);
	    
	  if( oka )
	  {
	    // new way 
	    RealArray & uc = idh.interfaceDataList[idh.current].u;
	    RealArray & fc = idh.interfaceDataList[idh.current].f;
	    RealArray & up = idh.interfaceDataList[prev       ].u;
	    RealArray & fp = idh.interfaceDataList[prev       ].f;

            // const real a0 = gridDescriptora.a[0], a1=-gridDescriptora.a[1];  // note: flip sign of the normal
	    // ua(Ia[0],Ia[1],Ia[2],Ia[3]) = ( cex1*( a0*uc(Ib[0],Ib[1],Ib[2],Ib[3]) + a1*fc(Ib[0],Ib[1],Ib[2],Ib[3]) ) + 
	    // 				    cex2*( a0*up(Ib[0],Ib[1],Ib[2],Ib[3]) + a1*fp(Ib[0],Ib[1],Ib[2],Ib[3]) ) );

            const real a0 = gridDescriptora.a[0], a1=gridDescriptora.a[1]; 
	    ua(Ia[0],Ia[1],Ia[2],Ia[3]) = ( cex1*( a0*uc(Ia[0],Ia[1],Ia[2],Ia[3]) + a1*fc(Ia[0],Ia[1],Ia[2],Ia[3]) ) + 
					    cex2*( a0*up(Ia[0],Ia[1],Ia[2],Ia[3]) + a1*fp(Ia[0],Ia[1],Ia[2],Ia[3]) ) );

	  }

// 	  else
// 	  {
// 	    // old way
//             InterfaceDataHistory & idh = gridDescriptora.interfaceDataHistory;   
// 	    RealArray & uc = matchFlux ? idh.interfaceDataList[idh.current].f : idh.interfaceDataList[idh.current].u;
// 	    RealArray & up = matchFlux ? idh.interfaceDataList[prev       ].f : idh.interfaceDataList[prev       ].u;
// 	    ua(Ia[0],Ia[1],Ia[2],Ia[3]) = cex1*uc(Ia[0],Ia[1],Ia[2],Ia[3])+cex2*up(Ia[0],Ia[1],Ia[2],Ia[3]);
// 	  }
	  
	}
      }

      // *********************************************************
      // ************** Under-relaxed iteration ******************
      // *********************************************************

      // ***** NOTE: should we relax the first time through ??? --> I don't think so.

      bool underRelaxGuess=false;
      if( parameters.dbase.get<bool>("useMixedInterfaceConditions") )
      { // with mixed-conditions we under-relax the RHS for one side of the interface
	underRelaxGuess= d==d1;
      }
      else if( matchFlux )
      { // for D-N conditions we under-relax the Neumann condition
	underRelaxGuess=true;
      }
      if( interfaceType1(side1,dir1,grid1)==Parameters::tractionInterface ||
          interfaceType2(side2,dir2,grid2)==Parameters::tractionInterface )
      {
        // we currently don't under-relax for traction interfaces
	underRelaxGuess=false;
      }      
      if( underRelaxGuess  && correct>0 ) // *wdh* 080722 -- only relax for correct > 0 
      {
        // We have saved both u and k*u.n from previous times. These can be used to under-relax the 
        // iteration. If we are solving the interface condition by iteration: (j= iteration number)
        //    a*k1*u.n[j+1] + b*u[j+1] = a*k2*v.n[j] + b*v[j]  
        // The the relaxed iteration is 
        //    a*k1*u.n[j+1] + b*u[j+1] = omega*( a*k2*v.n[j] + b*v[j] ) + (1-omega)*( a*k1*u.n[j] + b*u[j] )
        // If omega=1 : no-relaxation. If omega=0, u[j+1] = u[j] 

        // consistency check: we currently don't under-relax for traction interfaces
        assert( interfaceType1(side1,dir1,grid1)!=Parameters::tractionInterface );
        assert( interfaceType2(side2,dir2,grid2)!=Parameters::tractionInterface );


	// interface values at past iterates for the current time:
	InterfaceDataHistory & idi = gridDescriptora.interfaceDataIterates;

	if( idi.interfaceDataList.size()>=1 )
	{
	  real tp = idi.interfaceDataList[idi.current].t;
          real & omega = interfaceDescriptor.interfaceOmega;
	  if( omega != 1. )
	  {
	    if( debug() & 4 )
	      fPrintF(interfaceFile,"+++interfaceRHS: interface %i: relax flux RHS, t=%9.3e, omega=%5.2f "
		      "(data iterates: current=%i,tp=%9.3e)\n",inter,t,omega,idi.current,tp);


	    if( interfaceType1(side1,dir1,grid1)==Parameters::heatFluxInterface )
	    {
	      // new way 
	      RealArray & up = idi.interfaceDataList[idi.current].u; 
	      RealArray & fp = idi.interfaceDataList[idi.current].f; 

	      real a0 = gridDescriptora.a[0], a1=gridDescriptora.a[1];  
              // fp already includes the factor of k so we need to divide a1 by this amount
              real ktca = domainSolver[da]->parameters.dbase.get<real>("thermalConductivity");
              a1 = a1/ktca;
	      if( oka )
	      {
		ua(Ia[0],Ia[1],Ia[2],Ia[3]) = ( omega*ua(Ia[0],Ia[1],Ia[2],Ia[3]) + 
			      (1.-omega)*( a0*up(Ia[0],Ia[1],Ia[2],Ia[3]) + a1*fp(Ia[0],Ia[1],Ia[2],Ia[3]) ) );
	      }
	    }
            else
	    {
	      printF("interfaceRHS:WARNING: relaxation not applied to a traction interface\n");
	    }
	    
// 	    else
// 	    {
// 	      // old way 
// 	      RealArray & up = matchFlux ? idi.interfaceDataList[idi.current].f : idi.interfaceDataList[idi.current].u; 
// 	      if( debug() & 8 )
// 	      {
// 		::display(ua(Ia[0],Ia[1],Ia[2],Ia[3]),"ua",interfaceFile,"%8.2e ");
// 		::display(up(Ia[0],Ia[1],Ia[2],Ia[3]),"up",interfaceFile,"%8.2e ");
// 	      }
	  
// 	      ua(Ia[0],Ia[1],Ia[2],Ia[3]) = omega*ua(Ia[0],Ia[1],Ia[2],Ia[3]) + (1.-omega)*up(Ia[0],Ia[1],Ia[2],Ia[3]);
//	    }
	  }
	}
      }

      infoa.a[0]=a0; infoa.a[1]= a1;  // These are used with "set" for TZ
      if( oka )
        domainSolver[da]->interfaceRightHandSide( setInterfaceRightHandSide,interfaceDataOptions,infoa,gridDescriptora,gfIndex[da],t );

      
    } // end for face
  } // end for inter


  parameters.dbase.get<RealArray>("timing")(Parameters::timeForInterfaces)+=getCPU()-cpu0;
  return 0;
  
}

// ===================================================================================================================
/// \brief Evaluate the residuals in the interface equation
/// \details 
///   
/// \param t (input) : current time
/// \param dt (input) : current time step
/// \param correct (input) : correction step number.
/// \param gfIndex (input) : Domain d should use the grid function : domainSolver[d]->gf[gfIndex[d]]
/// \param maxResidual (output) : Maximum residual in the interface equations for each interface.
/// \param saveInterfaceValues (input) : One of doNotSaveInterfaceValues, saveInterfaceTimeHistoryValues,
///    or saveInterfaceIterateValues to indicate whether interface values should be saved in a list for later use. 
// 
// ==================================================================================================================

int 
Cgmp::
getInterfaceResiduals( real t, real dt, std::vector<int> & gfIndex, std::vector<real> & maxResidual,
                       InterfaceValueEnum saveInterfaceValues /* =doNotSaveInterfaceValues */ )
{


  real cpu0=getCPU();

  // ************* Much of this code is duplicated from assignInterfaceBoundaryConditions: FIX THIS *********

  if( !gridHasMaterialInterfaces ) return 0;

  InterfaceList & interfaceList = parameters.dbase.get<InterfaceList>("interfaceList");
  assert( interfaceList.size()!=0 );

  const int np= max(1,Communication_Manager::numberOfProcessors());
  const int & myid = parameters.dbase.get<int>("myid");
  
  const int numberOfDomains=domainSolver.size();
  FILE *& interfaceFile =parameters.dbase.get<FILE* >("interfaceFile");

  int numberOfInterfaceHistoryValuesToSave = parameters.dbase.get<int>("numberOfInterfaceHistoryValuesToSave");
  int numberOfInterfaceIterateValuesToSave = parameters.dbase.get<int>("numberOfInterfaceIterateValuesToSave");

  Index Iv[4], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2], &I4=Iv[3];
  Index Jv[4], &J1=Jv[0], &J2=Jv[1], &J3=Jv[2], &J4=Jv[3];
  Index Kv[4], &K1=Kv[0], &K2=Kv[1], &K3=Kv[2], &K4=Kv[3];
  
  Index Iav[4], &Ia1=Iav[0], &Ia2=Iav[1], &Ia3=Iav[2], &Ia4=Iav[3];
  Index Jav[4], &Ja1=Jav[0], &Ja2=Jav[1], &Ja3=Jav[2], &Ja4=Jav[3];
  

  // ************************************************
  // ******* Check the Interfaces            ********
  // ************************************************

  if( maxResidual.size() < interfaceList.size() )
    maxResidual.resize(interfaceList.size(),0.);

  // loop over interfaces
  for( int inter=0; inter < interfaceList.size(); inter++ )
  {
    InterfaceDescriptor & interfaceDescriptor = interfaceList[inter]; 

    maxResidual[inter]=0.;
    
    // There may be multiple grid faces that lie on the interface:     
    for( int face=0; face<interfaceDescriptor.gridListSide1.size(); face++ )
    {
      GridFaceDescriptor & gridDescriptor1 = interfaceDescriptor.gridListSide1[face];
      GridFaceDescriptor & gridDescriptor2 = interfaceDescriptor.gridListSide2[face];
      
      const int d1=gridDescriptor1.domain, 
                grid1=gridDescriptor1.grid, side1=gridDescriptor1.side, dir1=gridDescriptor1.axis;
      const int d2=gridDescriptor2.domain, 
                grid2=gridDescriptor2.grid, side2=gridDescriptor2.side, dir2=gridDescriptor2.axis;

      assert( d1>=0 && d1<numberOfDomains && d2>=0 && d2<numberOfDomains );


      GridFunction & gf1 = domainSolver[d1]->gf[gfIndex[d1]];
      GridFunction & gf2 = domainSolver[d2]->gf[gfIndex[d2]];

      CompositeGrid & cg1 = gf1.cg;
      assert( grid1>=0 && grid1<=cg1.numberOfComponentGrids());
      MappedGrid & mg1 = cg1[grid1];
      const IntegerArray & bc1 = mg1.boundaryCondition();
      const IntegerArray & share1 = mg1.sharedBoundaryFlag();
      const IntegerArray & interfaceType1 = domainSolver[d1]->parameters.dbase.get<IntegerArray >("interfaceType");
      const intArray & mask1 = mg1.mask();
    
      CompositeGrid & cg2 = gf2.cg;
      assert( grid2>=0 && grid2<=cg2.numberOfComponentGrids());
      MappedGrid & mg2 = cg2[grid2];
      const IntegerArray & bc2 = mg2.boundaryCondition();
      const IntegerArray & share2 = mg2.sharedBoundaryFlag();
      const IntegerArray & interfaceType2 = domainSolver[d2]->parameters.dbase.get<IntegerArray >("interfaceType");
      const intArray & mask2 = mg2.mask();

      IntegerArray gidLocal1(2,3), dimLocal1(2,3), bcLocal1(2,3);
      getLocalBoundsAndBoundaryConditions( gf1.u[grid1],gidLocal1,dimLocal1,bcLocal1 );

      IntegerArray gidLocal2(2,3), dimLocal2(2,3), bcLocal2(2,3);
      getLocalBoundsAndBoundaryConditions( gf2.u[grid2],gidLocal2,dimLocal2,bcLocal2 );

      const int numberOfDimensions = cg1.numberOfDimensions();

      const int orderOfAccuracyInSpace=2;  // **** do this for now ***

      if( debug() & 8 )
      {
	printF("Cgmp::getInterfaceResiduals: interface found (t=%e):\n"
	       "  (d1,grid1,side1,dir1,bc1)=(%i,%i,%i,%i,%i)\n"
	       "  (d2,grid2,side2,dir2,bc2)=(%i,%i,%i,%i,%i)\n",t,
	       d1,grid1,side1,dir1,mg1.boundaryCondition(side1,dir1),
	       d2,grid2,side2,dir2,mg2.boundaryCondition(side2,dir2));
      }
      

      const int extra=0; // orderOfAccuracyInSpace/2;
      getBoundaryIndex(mg1.gridIndexRange(),side1,dir1,I1,I2,I3,extra);
      getBoundaryIndex(mg2.gridIndexRange(),side2,dir2,J1,J2,J3,extra);
      Ia1=I1, Ia2=I2, Ia3=I3; // save full range 
      Ja1=J1, Ja2=J2, Ja3=J3;
      
  
      #ifdef USE_PPP
        intSerialArray mask1Local; getLocalArrayWithGhostBoundaries(mask1,mask1Local);
        intSerialArray mask2Local; getLocalArrayWithGhostBoundaries(mask2,mask2Local);
      #else
        const intSerialArray & mask1Local = mask1;
        const intSerialArray & mask2Local = mask2;
      #endif
	

      // Trouble in parallel if mg1 has a different distribution from mg2
//       #ifdef USE_PPP   
//         Overture::abort("finish me for parallel Bill!");
//       #endif

      // check that the number of points in the tangential directions match -- eventually we will fix this
      for( int dir=1; dir<mg1.numberOfDimensions(); dir++ )
      {
	int dir1p = (dir1+dir) % mg1.numberOfDimensions();
	int dir2p = (dir2+dir) % mg2.numberOfDimensions();
	if( Iv[dir1p].getLength()!=Jv[dir2p].getLength() )
	{
	  printF("getInterfaceResiduals:ERROR: The number of grid points on the two interfaces do not match\n"
	         " (d1,grid1,side1,dir1,bc1)=(%i,%i,%i,%i,%i) Iv=[%i,%i][%i,%i][%i,%i]\n"
		 " (d2,grid2,side2,dir2,bc2)=(%i,%i,%i,%i,%i) Jv=[%i,%i][%i,%i][%i,%i]\n",
		 d1,grid1,side1,dir1,mg1.boundaryCondition(side1,dir1),
                   I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound(),I3.getBase(),I3.getBound(),  
                 d2,grid2,side2,dir2,mg2.boundaryCondition(side2,dir2),
                   J1.getBase(),J1.getBound(),J2.getBase(),J2.getBound(),J3.getBase(),J3.getBound());
	  Overture::abort("error");
	}
      }

      int includeGhost=0;  // is this right ? 
      bool ok1 = ParallelUtility::getLocalArrayBounds(mask1,mask1Local,I1,I2,I3,includeGhost);
      bool ok2 = ParallelUtility::getLocalArrayBounds(mask2,mask2Local,J1,J2,J3,includeGhost);

     
      assert( interfaceType1(side1,dir1,grid1) == interfaceType2(side2,dir2,grid2) );


      if( interfaceType1(side1,dir1,grid1)==Parameters::heatFluxInterface )
      {
        // ********************************************
	// ********** Heat Flux Interface *************
        // ********************************************


	const int tc1 = domainSolver[d1]->parameters.dbase.get<int >("tc");    
	const int tc2 = domainSolver[d2]->parameters.dbase.get<int >("tc");


	real ktc[2]={1.,1.};

	ktc[0] = domainSolver[d1]->parameters.dbase.get<real>("thermalConductivity");
	ktc[1] = domainSolver[d2]->parameters.dbase.get<real>("thermalConductivity");
	if( ktc[0] <=0. || ktc[1]<=0. )
	{
	  printF("getInterfaceResiduals:ERROR: a negative thermal conductivity was found, "
		 "ktc[0]=%e ktc[1]=%e\n",ktc[0],ktc[1]);
	  Overture::abort("error");
	}
	real ktc1=ktc[0], ktc2=ktc[1];

	// **** finish this ****
      
	GridFaceDescriptor info2(d2,grid2,side2,dir2);
	// info2.component=tc2;
	Range N2(tc2,tc2);
	RealArray u2;
        if( ok2 ) u2.redim(J1,J2,J3,N2);
	info2.u=&u2;

	GridFaceDescriptor info1(d1,grid1,side1,dir1);
	// info1.component=tc1;
	Range N1(tc1,tc1);
	RealArray u1;
        if( ok1 ) u1.redim(I1,I2,I3,N1);
	info1.u=&u1;

       
	// --- Evaluate [u] ---
        int interfaceDataOptions=Parameters::heatFluxInterfaceData;
	info1.a[0]=1.; info1.a[1]=0.;
        if( ok1 )
  	  domainSolver[d1]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,info1,gridDescriptor1,gfIndex[d1],t );

	info2.a[0]=1.; info2.a[1]=0.;
        if( ok2 )
  	  domainSolver[d2]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,info2,gridDescriptor2,gfIndex[d2],t );
	
      


	if( saveInterfaceValues==saveInterfaceTimeHistoryValues ||
	    saveInterfaceValues==saveInterfaceIterateValues )
	{ // *** save the interface solution values  ****
	  for( int iface=0; iface<=1; iface++ )
	  {
	    GridFaceDescriptor & gfd = iface==0 ? gridDescriptor1 : gridDescriptor2;
	    RealArray & ui = iface==0 ? u1 : u2;
	    InterfaceDataHistory & idh = 
	      (saveInterfaceValues==saveInterfaceTimeHistoryValues ? gfd.interfaceDataHistory
	       : gfd.interfaceDataIterates);
	
	    const int numToSave = 
	      (saveInterfaceValues==saveInterfaceTimeHistoryValues ? numberOfInterfaceHistoryValuesToSave :
	       numberOfInterfaceIterateValuesToSave );

	    if( idh.interfaceDataList.size()<numToSave )
	    {
	      idh.interfaceDataList.push_back(InterfaceData());
	      idh.current = idh.interfaceDataList.size()-1;
	    }
	    else
	    {
	      idh.current = (idh.current+1) % numToSave;
	    }
	    if( debug() & 4 )
	    {
	      if( saveInterfaceValues==saveInterfaceTimeHistoryValues )
	      {
		fPrintF(interfaceFile,"interfaceRes: interface %i (face=%i): save interface history data, "
                       "t=%9.2e, current=%i\n",inter,iface,t,idh.current);
	      }
	      else
	      {
		fPrintF(interfaceFile,"interfaceRes: interface %i (face=%i): save interface iterate data, "
                       "t=%9.2e, current=%i\n",inter,iface,t,idh.current);
	      }
	    }
	    
	    InterfaceData & id = idh.interfaceDataList[idh.current];
	    id.t=t;
	    id.u=ui;
	  }
	
	}
      


	if( false )
	{
	  ::display(u1(I1,I2,I3,N1) ,"interface-residual u1 for [u]");
	  ::display(u2(J1,J2,J3,N2) ,"interface-residual u2 for [u]");
	}
      
	// *fix this* for mask , parallel, etc.
	// real jumpInU = max(fabs(u1(I1,I2,I3,N1)-u2(J1,J2,J3,N2)));

        int i1,i2,i3,i4, j1,j2,j3,j4;
        real jumpInU=0.;

        #ifdef USE_PPP

	  // parallel :  we need to transfer the local u2 arrays onto the parallel distribution of u1
          // 
          //     Allocate distributed arrays u22d(mg2),  u21d(mg1)
          //        u22Local = u2; 
          //        copyArray u22d -> u21d;
          //        Use u21Local instead or u2 
          //        
    
          realArray u22d; 
	  u22d.partition(gf2.u[grid2].getPartition());
          Index D1,D2,D3;
          getIndex(mg2.dimension(),D1,D2,D3);
	  u22d.redim(D1,D2,D3,N2);
	  
          RealArray u22Local; getLocalArrayWithGhostBoundaries(u22d,u22Local);
          if( ok2 ) u22Local(J1,J2,J3,N2)=u2(J1,J2,J3,N2);  // fill the boundary of u22d 
	  
          realArray u21d;
	  u21d.partition(gf1.u[grid1].getPartition());
          getIndex(mg1.dimension(),D1,D2,D3);
	  u21d.redim(D1,D2,D3,N1);

          Iv[3]=N1; Jv[3]=N2;
          Iav[3]=N1; Jav[3]=N2;
          // copy the boundary of u22d to u21d :  u21d(I1,I2,I3,I4) = u22d(J1,J2,J3,J4) (requires communication)
	  CopyArray::copyArray( u21d, Iav, u22d, Jav );  
 
          RealArray u21Local; getLocalArrayWithGhostBoundaries(u21d,u21Local);
          
          int J1Base, J2Base, J3Base, J4Base; 
	  if( ok1 )
	  {
	    FOR_4D(i1,i2,i3,i4,I1,I2,I3,I4)
	    {
	      if( mask1Local(i1,i2,i3) > 0 )
	      {
		jumpInU=max(jumpInU,fabs(u1(i1,i2,i3,i4)-u21Local(i1,i2,i3,i4)));
	      }
	    }
	  }
	  jumpInU=ParallelUtility::getMaxValue(jumpInU);
	  
        #else
          // serial case:           

          FOR_4IJD(i1,i2,i3,i4,I1,I2,I3,N1,j1,j2,j3,j4,J1,J2,J3,N2)
	  {
            if( mask1Local(i1,i2,i3) > 0 )
	    {
	      jumpInU=max(jumpInU,fabs(u1(i1,i2,i3,i4)-u2(j1,j2,j3,j4)));
	    }
	  }
	#endif


	// --- Evaluate [ktc*u.n] ---
	info1.a[0]=0.; info1.a[1]= ktc1;
	if( ok1 )
  	  domainSolver[d1]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,info1,gridDescriptor1,gfIndex[d1],t );

	info2.a[0]=0.; info2.a[1]= ktc2;  // do not flip sign here (since we may save u2 below)
        if( ok2 )
  	  domainSolver[d2]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,info2,gridDescriptor2,gfIndex[d2],t );

	if( saveInterfaceValues==saveInterfaceTimeHistoryValues ||
	    saveInterfaceValues==saveInterfaceIterateValues )
	{ // *** save the interface RHS values  ****
	  for( int iface=0; iface<=1; iface++ )
	  {
	    GridFaceDescriptor & gfd = iface==0 ? gridDescriptor1 : gridDescriptor2;
	    RealArray & ui = iface==0 ? u1 : u2;
	    InterfaceDataHistory & idh = 
	      (saveInterfaceValues==saveInterfaceTimeHistoryValues ? gfd.interfaceDataHistory
	       : gfd.interfaceDataIterates);
	
	    InterfaceData & id = idh.interfaceDataList[idh.current];
	    id.t=t;
	    id.f=ui;
	  }
	
	}

	//  *note* flip the sign of the normal to match normal from domain 1
        if( ok2 ) 
          u2(J1,J2,J3,N2)=-u2(J1,J2,J3,N2);

        real jumpInUn=0.;

        #ifdef USE_PPP

	  // parallel :  we need to transfer the local u2 arrays onto the parallel distribution of u1
    
          if( ok2 ) 
            u22Local(J1,J2,J3,N2)=u2(J1,J2,J3,N2);  // fill the boundary of u22d 
	  
          // copy the boundary of u22d to u21d :  u21d(I1,I2,I3,I4) = u22d(J1,J2,J3,J4) (requires communication)
	  CopyArray::copyArray( u21d, Iav, u22d, Jav );  
 
          // *** NOTE: for now we do not check the consistency of the mask's ***** fix me !
	  if( ok1 )
	  {
	    FOR_4D(i1,i2,i3,i4,I1,I2,I3,I4)
	    {
	      if( mask1Local(i1,i2,i3) > 0 )
	      {
		jumpInUn=max(jumpInUn,fabs(u1(i1,i2,i3,i4)-u21Local(i1,i2,i3,i4)));
	      }
	    }
	  }
	  jumpInUn=ParallelUtility::getMaxValue(jumpInUn);

        #else
          // serial case:           

          // do not check the residual on end-pts with adjacent dirichlet BC's   *** fix me **
          // int dir1p1 = (dir1+1) % mg1.numberOfDimensions();
          // int dir2p1 = (dir2+1) % mg2.numberOfDimensions();
	  // Iv[dir1p1]=Range(Iv[dir1p1].getBase()+1,Iv[dir1p1].getBound()-1);
	  // Jv[dir2p1]=Range(Jv[dir2p1].getBase()+1,Jv[dir2p1].getBound()-1);
	  

          FOR_4IJ(i1,i2,i3,i4,I1,I2,I3,N1,j1,j2,j3,j4,J1,J2,J3,N2)
	  {
            // the mask's must both be positive at the same points : 
            if(  int(mask1Local(i1,i2,i3)>0) + int(mask2(j1,j2,j3)>0) == 1 )
	    {
	      printF("Cgmp::getInterfaceResiduals: mask's do not agree!\n"
		     " The interface equation solvers currently require the masks on opposite sides of\n"
		     " the interface to agree\n");
	      printF(" domain1=%i (%s) grid1=%i (%s) mask1(%i,%i,%i)=%i\n"
		     " domain2=%i (%s) grid2=%i (%s) mask2(%i,%i,%i)=%i\n",
		     d1,(const char*)domainSolver[d1]->getName(),grid1,(const char*)mg1.getName(),i1,i2,i3,mask1Local(i1,i2,i3), 
		     d2,(const char*)domainSolver[d2]->getName(),grid2,(const char*)mg2.getName(),j1,j2,j3,mask2(j1,j2,j3) );
	      Overture::abort("error");
	    }
	  
	  
	    if( mask1Local(i1,i2,i3) > 0 )
	    {
	      jumpInUn=max(jumpInUn,fabs(u1(i1,i2,i3,i4)-u2(j1,j2,j3,j4)));
	    }
	  }

        #endif

	if( false )
	{
	  ::display(u1(I1,I2,I3,N1) ,"interface-residual: nu1*u1.n");
	  ::display(u2(J1,J2,J3,N2) ,"interface-residual: nu2*u2.n");
	}

        real interfaceResidual=max(jumpInU,jumpInUn);
	if( debug() & 2 )
	{
	  fPrintF(interfaceFile,
		  "interface %i step=%i: residuals: [u]=%8.2e [k*u.n]=%8.2e (omega=%9.3e,tol=%8.2e)\n",
		  inter,parameters.dbase.get<int >("globalStepNumber"),jumpInU,jumpInUn,
                  parameters.dbase.get<real>("interfaceOmega"),interfaceDescriptor.interfaceTolerance);
	}

	maxResidual[inter]=max(maxResidual[inter],interfaceResidual);
      
      }
      else if( interfaceType1(side1,dir1,grid1)==Parameters::tractionInterface )
      {
        // *******************************************
	// ********** Traction Interface *************
        // *******************************************

        // For now we do not compute residuals, just save a time history 
	if( !(saveInterfaceValues==saveInterfaceTimeHistoryValues) )
	{
	  continue;
	}
 
	// --- Evaluate the interface data ---

	const int uc1 = domainSolver[d1]->parameters.dbase.get<int >("uc");
	const int uc2 = domainSolver[d2]->parameters.dbase.get<int >("uc");

	// *wdh* 081201 Range C1(uc1,uc1+numberOfDimensions-1), C2(uc2,uc2+numberOfDimensions-1); // ************ fix me **************
        // for now add some extra space: 
	Range C1(uc1,uc1+2*numberOfDimensions-1), C2(uc2,uc2+2*numberOfDimensions-1); // ************ fix me **************

	GridFaceDescriptor info2(d2,grid2,side2,dir2);
	RealArray u2;
        if( ok2 ) u2.redim(J1,J2,J3,C2);
	info2.u=&u2;

	GridFaceDescriptor info1(d1,grid1,side1,dir1);
	RealArray u1;
        if( ok1 ) u1.redim(I1,I2,I3,C1);
	info1.u=&u1;

	int interfaceDataOptions=0;

	if( ok1 )
	{
	  // find out from domain d2 what interface data it wants: 
	  domainSolver[d2]->getInterfaceDataOptions( info2,interfaceDataOptions );
          // evaluate the interface data:
	  domainSolver[d1]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,info1,gridDescriptor1,gfIndex[d1],t );
	}
      
        if( ok2 )
	{
	  // find out from domain d1 what interface data it wants:
	  domainSolver[d1]->getInterfaceDataOptions( info1,interfaceDataOptions );
          // evaluate the interface data:
  	  domainSolver[d2]->interfaceRightHandSide( getInterfaceRightHandSide,interfaceDataOptions,info2,gridDescriptor2,gfIndex[d2],t );
	}
	
	if( saveInterfaceValues==saveInterfaceTimeHistoryValues ||
	    saveInterfaceValues==saveInterfaceIterateValues )
	{ 
          // *** save the interface RHS values  ****

	  if( debug() & 4 )
	  {
	    if( saveInterfaceValues==saveInterfaceTimeHistoryValues )
	      printP("Cgmp::getInterfaceResiduals: saving time history of traction: interface=%i, t=%9.3e\n",inter,t);
            else
	      printP("Cgmp::getInterfaceResiduals: saving interface iterate of traction: interface=%i, t=%9.3e\n",inter,t);
	  }
	  
	  for( int iface=0; iface<=1; iface++ )
	  {
	    GridFaceDescriptor & gfd = iface==0 ? gridDescriptor1 : gridDescriptor2;
	    RealArray & ui = iface==0 ? u1 : u2;
	    InterfaceDataHistory & idh = 
	      (saveInterfaceValues==saveInterfaceTimeHistoryValues ? gfd.interfaceDataHistory : gfd.interfaceDataIterates);
	
	    const int numToSave = 
	      (saveInterfaceValues==saveInterfaceTimeHistoryValues ? numberOfInterfaceHistoryValuesToSave : numberOfInterfaceIterateValuesToSave );

	    if( idh.interfaceDataList.size()<numToSave )
	    { // add a new entry: 
	      idh.interfaceDataList.push_back(InterfaceData());
	      idh.current = idh.interfaceDataList.size()-1;
	    }
	    else
	    { // over-write oldest entry: 
	      idh.current = (idh.current+1) % numToSave;
	    }

	    InterfaceData & id = idh.interfaceDataList[idh.current];
	    id.t=t;
	    id.f=ui;
	  }
	
	} // end saveInterfaceValues

      }
      else
      {
	printF("Cgmp::getInterfaceResiduals:ERROR:unexpected interfaceType=%i\n",interfaceType1(side1,dir1,grid1));
	OV_ABORT("error");
      }


      
    } // end for face
  } // end for inter

  parameters.dbase.get<RealArray>("timing")(Parameters::timeForInterfaces)+=getCPU()-cpu0;

  return 0;
}
