#================================================================================================
#  cgmx example:  scattering from 3d curved interface between glass and a vacuum
#
# Usage:
#   
#  cgmx [-noplot] bump3d  -g=<name> -tf=<tFinal> -tp=<tPlot> -kx=<num> -ky=<num> -kz=<num> ...
#                          -diss=<> -debug=<num> -cons=[0/1] -varDiss=<0|1> -go=[run/halt/og]
# Arguments:
#  -kx= -ky= -kz= : integer wave number of the incident wave
#  -varDiss :  if 1, use variable dissipation (only add dissipation near interpolation points)
# 
# Notes: 
# refractive index from vacuum to glass is 1.5
#       $epsGlass = 1.5^2 = 2.25
# 
#   pts/wave-length = (1/ds)*(1/ky) = factor*100/ky 
# 
# NOTE: diss=.5 was too small for fine grid runs;  diss=5. works but a smaller value may also be ok. 
# 
# Examples:
#
#   cgmx bump3d -g=interfaceBump3de2.order2 -eps1=2.25 -eps2=1. -ky=5  -diss=2. -go=halt
#   mpirun -np 2 $cgmxp bump3d -g=interfaceBump3de2.order2 -eps1=2.25 -eps2=1. -ky=5  -diss=2. -go=halt
#   totalview srun -a -N1 -n2 -ppdebug $cgmxp noplot bump3d -g=interfaceBump3de2.order2 -eps1=2.25 -eps2=1. -ky=5  -diss=2. -go=halt
#================================================================================================
# 
$tFinal=1.4; $tPlot=.2; $diss=5.; $cfl=.9; $varDiss=0; $varDissSmooths=20; 
$kx=0; $ky=1; $kz=0; 
$eps1=1.; # inner
$eps2=1.; # outer 
$show=" "; $backGround="backGround";
$interfaceIterations=3;
$grid="interfaceBump3de1.order2";
$cons=1; $go="halt"; 
$xa=-100.; $xb=100.; $ya=-100.; $yb=-.1; $za=-100.; $zb=100.;  # initial condition bounding box
# ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"tf=f"=>\$tFinal,"diss=f"=>\$diss,"tp=f"=>\$tPlot,"show=s"=>\$show,"debug=i"=>\$debug, \
 "cfl=f"=>\$cfl, "bg=s"=>\$backGround,"bcn=s"=>\$bcn,"go=s"=>\$go,"noplot=s"=>\$noplot,\
  "dtMax=f"=>\$dtMax,"kx=i"=>\$kx,"ky=i"=>\$ky,"kz=i"=>\$kz,"eps1=f"=>\$eps1,"eps2=f"=>\$eps2, "cons=i"=>\$cons,\
   "ii=i"=>\$interfaceIterations,"varDiss=i"=>\$varDiss ,"varDissSmooths=i"=>\$varDissSmooths,"yb=f"=>\$yb);
# -------------------------------------------------------------------------------------------------
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
#
# 
$grid
#
NFDTD
planeWaveInitialCondition
$kya= abs($ky);
$yb = int( $yb*$kya -.9999 )/$kya;   # we need to clip the plane wave on a period
if( $ky < 0 ){ $ya=$yb; $yb=100.; }
initial condition bounding box $xa $xb $ya $yb $za $zb
# zeroInitialCondition
# ====
# planeWaveScatteredFieldInitialCondition
# ====
# twilightZone
#  degreeSpace, degreeTime  1 1
#
kx,ky,kz $kx $ky $kz
#
# bc: all=dirichlet
# bc: all=perfectElectricalConductor
# bc: all=perfectElectricalConductor
bc: all=abcEM2
if( $ky > 0 ){ $inbc = "lower(0,1)"; }else{ $inbc = "upper(1,1)"; }
# bc: lower(0,1)=planeWaveBoundaryCondition
bc: $inbc=planeWaveBoundaryCondition
# 
bc: upper(0,0)=symmetry
bc: upper(1,0)=symmetry
bc: upper(0,2)=symmetry
bc: upper(1,2)=symmetry
# 
bc: upperInterface(0,0)=symmetry
bc: upperInterface(1,0)=symmetry
bc: upperInterface(0,2)=symmetry
bc: upperInterface(1,2)=symmetry
# 
bc: lower(0,0)=symmetry
bc: lower(1,0)=symmetry
bc: lower(0,2)=symmetry
bc: lower(1,2)=symmetry
# 
bc: lowerInterface(0,0)=symmetry
bc: lowerInterface(1,0)=symmetry
bc: lowerInterface(0,2)=symmetry
bc: lowerInterface(1,2)=symmetry
# 
# -- we need to subtract out the incident field on the "inflow" boundary before
#    applying the radiation boundary condition: 
adjust boundaries for incident field 0 all
adjust boundaries for incident field 1 lower
# 
#      innerAnnulus
#      innerSquare
#      outerAnnulus
#      outerSquare
# NOTE: material interfaces have share>=100
#* coefficients $eps1 1. all (eps,mu,grid-name)
coefficients $eps1 1. lower          (eps,mu,grid-name)
coefficients $eps1 1. lowerInterface (eps,mu,grid-name)
coefficients $eps2 1. upper          (eps,mu,grid-name)
coefficients $eps2 1. upperInterface (eps,mu,grid-name)
#
interface BC iterations $interfaceIterations
#
# bc: Annulus=perfectElectricalConductor
tFinal $tFinal
tPlot  $tPlot
#
dissipation $diss
#*********************************
show file options...
  MXSF:compressed
  MXSF:open
    $show
 # MXSF:frequency to save 
  MXSF:frequency to flush 5 
exit
#**********************************
#
use variable dissipation $varDiss
number of variable dissipation smooths $varDissSmooths
use conservative difference $cons
debug 0
#
cfl $cfl 
plot errors 0
check errors 0
# plot energy density 1
# plot intensity 1
#
#specify probes
#  -1.544e-01 2.334e-02 0.000e+00
#  -1.926e-02 2.234e-02 0.000e+00
# done
#
continue
#
plot:Ex
contour
plot contour lines (toggle)
# vertical scale factor 0.
# min max -1.1 1.1
exit
$go
