*================================================================================================
*  cgmx example:  Scattering of a plane wave from a cylinder (compute the errors)
*
*   cgmx cic.planeWaveBC
* 
* Usage:
*   
*  cgmx [-noplot] cic.planeWaveBC -g=<name> -tf=<tFinal> -tp=<tPlot> -diss=<> -debug=<num> -cons=[0/1] -go=[run/halt/og]
*
* Examples:
*   cgmx cic.planeWaveBC -g=cice2.order4.hdf
*   cgmx cic.planeWaveBC -g=cice3.order4.hdf
*   cgmx cic.planeWaveBC -g=cice16.order4.hdf
* parallel: 
*   mpirun -np 2 $cgmxp cic.planeWaveBC -g=cice3.order4.hdf
*   srun -N1 -n1 -ppdebug $cgmxp noplot cic.planeWaveBC -g=cice2.order2.hdf -tf=.1
*   srun -N1 -n1 -ppdebug $cgmxp cic.planeWaveBC -g=cice2.order4.hdf
*
* NOTES:
*   (1) In this example we solve for the scattered field. 
* 
**** NOTE: when using the scattered field as an initial condition and BC use the options
*           (1) planeWaveBoundaryForcing : to assign the plane wave forcing on the boundary
*           (2) planeWaveScatteredFieldInitialCondition
*           (3) square=dirichlet 
*           (4) slow start interval -1.
*================================================================================================
* 
$tFinal=10.; $tPlot=.1; $diss=.1; $cfl=.9; 
$grid="sib1.order4.hdf";
$cons=0; $go="halt"; 
* ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"tf=f"=>\$tFinal,"diss=f"=>\$diss,"tp=f"=>\$tPlot,"show=s"=>\$show,"debug=i"=>\$debug, \
 "cfl=f"=>\$cfl, "bg=s"=>\$backGround,"bcn=s"=>\$bcn,"go=s"=>\$go,"noplot=s"=>\$noplot,\
  "dtMax=f"=>\$dtMax, "cons=i"=>\$cons );
* -------------------------------------------------------------------------------------------------
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
*
*
$grid
*
NFDTD
*
*  Initial condition:
* zeroInitialCondition
* planeWaveInitialCondition
planeWaveBoundaryForcing
* ====
planeWaveScatteredFieldInitialCondition
** planeWaveInitialCondition
scatteringFromADiskKnownSolution
* ====
*
kx,ky,kz 1 0 0
* Gaussian plane wave: 100. -1.5 0. 0.
** bc: square=perfectElectricalConductor
* bc: square=abcEM2
* bc: square=abcPML
* pml width,strength,power 25 30. 4
* pml width,strength,power 35 30. 4
* pml width,strength,power 45 30. 4
* pml width,strength,power 55 30. 4
* =====================================
* radius for checking errors 2.
* =====================================
bc: square=dirichlet
*******
*bc: square(0,0)=planeWaveBoundaryCondition
*bc: square(1,0)=planeWaveBoundaryCondition
bc: Annulus=perfectElectricalConductor
tFinal $tFinal
tPlot  $tPlot
* 
dissipation  $diss
************************
** slow start interval 1.
slow start interval -1.
************************
* divergence damping  0.0025   .02  1. .005 .001 .0025 .01
************************
cfl $cfl
*
use conservative divergence $cons 
* plot scattered field 1
plot errors 1
check errors 1
*
continue
$go 

