*================================================================================================
*  cgmx example:  scattering from a dielectric cylinder or sphere (and compare to the exact solution)
*
* Usage:
*   
*  cgmx [-noplot] dielectricCyl  -g=<name> -tf=<tFinal> -tp=<tPlot> -kx=<num> -ky=<num> -kz=<num> -show=<name> ...
*                                -eps1=<> -eps2=<> -interit=<> -diss=<> -debug=<num> -cons=[0/1] -go=[run/halt/og]
* Arguments:
*  -kx= -ky= -kz= : integer wave numbers of the incident wave
*  -interit : number of iterations to solve the interface equations 
*
* Examples: (see Makefile to build grids)
*  --NEW: build grids with domains
*   cgmx dielectricCyl -g=innerOutere2.order4.hdf -kx=2 -eps1=.25 -eps2=1. -go=halt
*   cgmx dielectricCyl -g=innerOutere4.order4.hdf -kx=2 -eps1=.25 -eps2=1. -go=halt
*   cgmx dielectricCyl -g=innerOutere8.order4.hdf -kx=2 -eps1=.25 -eps2=1. -go=halt
* 
*   srun -N1 -n1 -ppdebug $cgmxp noplot dielectricCyl -g=innerOutere2.order2.hdf -kx=2 -eps1=.25 -eps2=1. -tf=.1 -useNewInterface=1 -go=go
* 
* -- OLD: 
*   cgmx dielectricCyl -g=innerOuter4.order2.hdf -kx=2 -eps1=.25 -eps2=1. -go=halt
*   cgmx dielectricCyl -g=innerOuter4.order4.hdf -kx=2 -eps1=.25 -eps2=1. -go=halt
*   cgmx dielectricCyl -g=innerOuter8.order4.hdf -kx=2 -eps1=2.  -eps2=1. -go=halt
*   cgmx dielectricCyl -g=innerOuter16.order4.hdf -kx=2 -eps1=2. -eps2=1. -go=halt
*
*  -- dielectric sphere: eps1=inside eps2=outside (should be 1)
*   cgmx dielectricCyl -cyl=0 -g=solidSphereInABoxi1.order2 -kx=1 -eps1=.25 -eps2=1. -go=halt
*   cgmx dielectricCyl -cyl=0 -g=solidSphereInABoxi2.order2 -kx=1 -eps1=.25 -eps2=1. -go=halt
*   cgmx dielectricCyl -cyl=0 -g=solidSphereInABoxi1.order2 -kx=1 -eps1=1. -eps2=1. -go=halt -tp=.01
* 
*   cgmx dielectricCyl -cyl=0 -g=solidSphereInABoxe2.order2 -kx=1 -eps1=.25 -eps2=1. -go=halt 
*   cgmx dielectricCyl -cyl=0 -g=solidSphereInABoxe4.order2 -kx=1 -eps1=.25 -eps2=1. -go=halt
*
*   cgmx noplot dielectricCyl -cyl=0 -g=solidSphereInABoxFixedi1.order2 -kx=1 -eps1=.25 -eps2=1. -go=go -tf=1.
* 
* parallel: 
*  srun -N1 -n1 -ppdebug $cgmxp noplot dielectricCyl -cyl=0 -g=solidSphereInABoxe2.order2 -kx=1 -eps1=.25 -eps2=1. -tp=.1 -tf=.2 -go=go
*================================================================================================
* 
$tFinal=2.; $tPlot=.1;  $show=" ";
$cfl = .8; $diss=.5; $debug=0; 
$cyl=1;   # set to 0 for a sphere 
$kx=2; $ky=0; $kz=0;
$eps1=.25; # inner
$eps2=1.;  # outer 
$show=" "; $backGround="backGround"; $useNewInterface=0; 
$interfaceIterations=3;
$grid="innerOuter4.order4.hdf";
$cons=0; $go="halt"; 
* 
* ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"tf=f"=>\$tFinal,"diss=f"=>\$diss,"tp=f"=>\$tPlot,"show=s"=>\$show,"debug=i"=>\$debug, \
 "cfl=f"=>\$cfl, "bg=s"=>\$backGround,"bcn=s"=>\$bcn,"go=s"=>\$go,"noplot=s"=>\$noplot,\
 "interit=i"=>\$interfaceIterations,"cyl=i"=>\$cyl,"useNewInterface=i"=>\$useNewInterface,\
 "dtMax=f"=>\$dtMax,"kx=i"=>\$kx,"ky=i"=>\$ky,"kz=i"=>\$kz,"eps1=f"=>\$eps1,"eps2=f"=>\$eps2, "cons=i"=>\$cons );
* -------------------------------------------------------------------------------------------------
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
* 
*
$grid
*
NFDTD
** planeWaveInitialCondition
* ++ zeroInitialCondition
* ==== in 3d solve for the full field ===
$ic = $cyl ? "planeWaveScatteredFieldInitialCondition" : "planeWaveInitialCondition";
* $ic
planeWaveScatteredFieldInitialCondition
$known = $cyl ? "scatteringFromADielectricDiskKnownSolution" : "scatteringFromADielectricSphereKnownSolution";
$known
* ====
* twilightZone
*  degreeSpace, degreeTime  1 1
*
kx,ky,kz $kx $ky $kz
* 
use new interface routines $useNewInterface
*
bc: all=dirichlet
* --
*bc: all=abcEM2
*bc: outerBox(0,0)=planeWaveBoundaryCondition
* --
* ++ bc: all=perfectElectricalConductor
* ++ bc: outerSquare(0,0)=planeWaveBoundaryCondition
* 
*      innerAnnulus
*      innerSquare
*      outerAnnulus
*      outerSquare
* NOTE: material interfaces have share>=100
if( $cyl eq 1 ){ $cmd = \
  "coefficients $eps1 1. innerAnnulus (eps,mu,grid-name)\n" .\
  "coefficients $eps1 1. innerSquare (eps,mu,grid-name)\n" .\
  "coefficients $eps2 1. outerAnnulus (eps,mu,grid-name)\n" .\
  "coefficients $eps2 1. outerSquare (eps,mu,grid-name)\n"; }
if( $cyl eq 0 ){ $cmd = \
  "coefficients $eps1 1. innerNorthPole (eps,mu,grid-name)\n" .\
  "coefficients $eps1 1. innerSouthPole (eps,mu,grid-name)\n" .\
  "coefficients $eps1 1. innerBox       (eps,mu,grid-name)\n" .\
  "coefficients $eps2 1. outerNorthPole (eps,mu,grid-name)\n" .\
  "coefficients $eps2 1. outerSouthPole (eps,mu,grid-name)\n" .\
  "coefficients $eps2 1. outerBox       (eps,mu,grid-name)\n"; }
$cmd
*
interface BC iterations $interfaceIterations
*
* bc: Annulus=perfectElectricalConductor
tFinal $tFinal
tPlot  $tPlot
*
dissipation $diss
*
* use conservative difference $cons 
debug $debug
*
cfl $cfl 
plot errors 1
check errors 1
* 
**********************************
show file options...
  MXSF:compressed
  MXSF:open
    $show
  MXSF:frequency to flush 5
exit
***********************************
continue
*
plot:Ey
* 
$go
