#================================================================================================
#  cgmx example:  plane wave moving through a domain
#
# Usage:
#   
#  cgmx [-noplot] planeWave  -g=<name> -tf=<tFinal> -tp=<tPlot> -kx=<num> -ky=<num> -kz=<num> -show=<name> ...
#                                -eps1=<> -eps2=<> -interit=<> -diss=<> -debug=<num> -cons=[0/1] -go=[run/halt/og]
# Arguments:
#  -kx= -ky= -kz= : integer wave numbers of the incident wave
#  -interit : number of iterations to solve the interface equations 
#
# Examples: 
#   cgmx planeWave -g=square10 -kx=1 -go=halt
#   cgmx planeWave -g=box10 -kx=1 -go=halt
# 
#   cgmx planeWave -g=sise4.order4 -kx=1 -ky=1 -go=halt
#   cgmx planeWave -g=rsise4.order4 -kx=1 -ky=1 -divDamping=1. -go=halt
#
#  mpirun -np 2 $cgmxp noplot planeWave -g=square10 -kx=1 -show="planeWave.show" -tf=.1 -go=go
#================================================================================================
# 
$tFinal=2.; $tPlot=.1;  $show=" ";
$cfl = .8; $diss=.5; $debug=0; $divDamping=.0; 
$cyl=1;   # set to 0 for a sphere 
$kx=2; $ky=0; $kz=0;
$eps1=.25; # inner
$eps2=1.;  # outer 
$show=" "; $backGround="backGround"; $useNewInterface=0; 
$interfaceIterations=3;
$grid="innerOuter4.order4.hdf";
$cons=0; $go="halt"; 
# 
# ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"tf=f"=>\$tFinal,"diss=f"=>\$diss,"tp=f"=>\$tPlot,"show=s"=>\$show,"debug=i"=>\$debug, \
    "cfl=f"=>\$cfl, "bg=s"=>\$backGround,"bcn=s"=>\$bcn,"go=s"=>\$go,"noplot=s"=>\$noplot,\
    "divDamping=f"=>\$divDamping,\
    "interit=i"=>\$interfaceIterations,"cyl=i"=>\$cyl,"useNewInterface=i"=>\$useNewInterface,\
    "dtMax=f"=>\$dtMax,"kx=i"=>\$kx,"ky=i"=>\$ky,"kz=i"=>\$kz,"eps1=f"=>\$eps1,"eps2=f"=>\$eps2, "cons=i"=>\$cons );
# -------------------------------------------------------------------------------------------------
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
# 
#
$grid
#
NFDTD
planeWaveInitialCondition
planeWaveKnownSolution
# ++ zeroInitialCondition
# ====
# twilightZone
#  degreeSpace, degreeTime  1 1
#
kx,ky,kz $kx $ky $kz
# 
#
bc: all=dirichlet
# --
#bc: all=abcEM2
#bc: outerBox(0,0)=planeWaveBoundaryCondition
# --
# ++ bc: all=perfectElectricalConductor
# ++ bc: outerSquare(0,0)=planeWaveBoundaryCondition
# 
#
# bc: Annulus=perfectElectricalConductor
tFinal $tFinal
tPlot  $tPlot
#
dissipation $diss
divergence damping $divDamping
#
# use conservative difference $cons 
debug $debug
#
cfl $cfl 
plot errors 1
check errors 1
# 
#*********************************
show file options...
  MXSF:compressed
  MXSF:open
  $show
  MXSF:frequency to flush 5
exit
#**********************************
continue
#
plot:Ey
# 
$go
