*
* cgsm -- test the twilightzone
*
* Usage:
*   
*  cgsm [-noplot] tz -g=<name> -tz=<poly/trig> -degreex=<> -degreet=<> -tf=<tFinal> -tp=<tPlot> -ts=[me|fe|ie]...
*                    -bcn=[d|sf|slip|mixed] -diss=<> -order=<2/4> -debug=<num> -bg=<backGround> -cons=[0/1] ...
*                    -pv=[nc|c|g|h] -godunovOrder=[1|2] -mu=<> -lambda=<> -rho=<> -go=[run/halt/og]
* 
*  -diss : coeff of artificial diffusion 
*  -bcn : d=dirichlet, sf=stress-free, slip=slipWall
*  -go : run, halt, og=open graphics
*  -cons : 1= conservative difference 
*  -pv : "pde-variation" : nc=non-conservative, c=conservative, g=godunov, h=hemp
*  -ts : time-stepping method, me=modified-equation, fe=forward-Euler, ie=improved-Euler, ab=adams-bashforth
* 
* Examples:
* 
* -- dirichlet BC: 
*  cgsm tz -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=2. -go=halt              [exact]
*  cgsm tz -g=nonSquare10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=halt           [exact]
*  cgsm tz -g=rotatedSquare10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=halt       [exact]
*  cgsm noplot tz -g=nonSquare10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go      [exact]
*  cgsm noplot tz -g=sise2.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go     [exact]
*  cgsm noplot tz -g=rsis2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go            [exact]
*  cgsm tz -g=cice1.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=halt
*  cgsm tz noplot -g=cice1.order2.hdf -degreex=2 -degreet=2 -diss=0.1 -tf=.5 -tz=trig -go=go
*  cgsm tz noplot -g=cice2.order2.hdf -degreex=2 -degreet=2 -diss=0.1 -tf=.5 -tz=trig -go=go
*   -- 3d --
*  cgsm tz -g=box10.hdf -degreex=2 -degreet=2 -diss=0. -tf=1. -go=halt                 [exact]
*  cgsm tz -g=nonBox5.hdf -degreex=2 -degreet=2 -diss=0. -tf=1. -go=halt               [exact]
*  cgsm tz -g=rotatedBox10.hdf -degreex=2 -degreet=2 -diss=0. -tf=1. -go=halt          [exact]
*  cgsm tz -g=bib2e.hdf -degreex=2 -degreet=2 -diss=0. -tf=1. -go=halt                 [exact]
*  cgsm tz -g=rbibe2.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=1. -go=halt         [exact]
* 
* -- stress free BC:
*  cgsm tz -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -bcn=sf -go=halt
*  cgsm tz -g=cice1.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -bcn=sf -go=halt
* 
* -- slip wall BC:
*  cgsm tz -g=nonSquare5.hdf -degreex=2 -degreet=2 -diss=0. -tp=.01 -bcn=slip -debug=3 -go=halt  [exact]
*  cgsm tz -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=2. -bcn=slip -go=halt   [exact]
* 
* -- godunov: 
*  cgsm tz -g=square10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -go=halt      [exact]
*  cgsm tz -g=square10.hdf -pv=g -degreex=2 -degreet=0 -tf=.5 -tp=.01 -go=halt      [exact]
*  cgsm tz -g=nonSquare10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -go=halt    [exact]
*  cgsm tz -g=rotatedSquare10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -go=halt    [exact]
*  cgsm tz -g=sis.hdf -pv=g -degreex=1 -degreet=1 -tf=1. -tp=.01 -go=halt      [exact]
*  cgsm tz -g=rsis2.hdf -pv=g -degreex=1 -degreet=1 -tf=1. -tp=.05 -go=halt
*     --- slip-wall BC
*  cgsm tz -g=square10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -bcn=slip -go=halt 
*     -- traction-BC
*  cgsm tz -g=square10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -bcn=sf -go=halt      [exact]
*  cgsm tz -g=nonSquare10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -bcn=sf -go=halt      [exact]
*  cgsm tz -g=rotatedSquare10.hdf -pv=g -degreex=2 -degreet=0 -tf=.5 -tp=.01 -bcn=sf -go=halt    [exact]
*  cgsm tz -g=rotatedSquare10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -bcn=sf -go=halt      [exact]
*  cgsm tz -g=annulus1.order2 -pv=g -degreex=2 -degreet=2 -tf=.5 -tp=.05 -bcn=sf -go=halt 
*  cgsm tz -g=quarterAnnulus -pv=g -degreex=2 -degreet=2 -tf=.5 -tp=.05 -bcn=sf -go=halt   ** trouble **
*  cgsm tz -g=quarterAnnulus -pv=g -degreex=2 -degreet=2 -tf=.5 -tp=.05 -bcn=mixed -bg=quarter -go=halt  OK
*  cgsm tz -g=quarterAnnulus -pv=g -degreex=2 -degreet=2 -tf=.5 -tp=.05 -bcn=mixed2 -bg=quarter -go=halt  OK
*     -- mixed BC's : traction + displacement 
*  cgsm tz -g=square10.hdf -pv=g -degreex=1 -degreet=1 -tf=.5 -tp=.01 -bcn=mixed -go=halt   [exact]
*  cgsm tz -g=square10.hdf -pv=g -degreex=2 -degreet=0 -tf=.5 -tp=.01 -bcn=mixed -go=halt   [exact]
*  cgsm tz -g=nonSquare10.hdf -pv=g -degreex=2 -degreet=0 -tf=.5 -tp=.01 -bcn=mixed -go=halt   [exact]
*  cgsm tz -g=rotatedSquare10.hdf -pv=g -degreex=2 -degreet=0 -tf=.5 -tp=.01 -bcn=mixed -go=halt   [exact] *NO* 
*     -- 3D
*  cgsm tz -g=box10.hdf -degreex=1 -degreet=1 -diss=0. -tf=1. -go=halt -pv=g
* 
**** MOL time-stepping:
*  cgsm tz -g=square5.hdf -degreex=2 -degreet=2 -diss=0. -tf=2. -tp=.01 -ts=fe  -debug=1 -go=halt [exact]
*  cgsm tz -g=square5.hdf -degreex=2 -degreet=1 -diss=0. -tf=2. -tp=.01 -ts=ie  -debug=1 -go=halt -pv=g [exact]
* 
* parallel: 
*  mpirun -np 1 $cgsmp tz.cmd -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=halt 
*  mpirun -np 1 $cgsmp noplot tz.cmd -g=square5.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go -debug=15
*  mpirun-wdh -np 2 $cgsmp noplot tz.cmd -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go 
*  srun -N1 -n1 -ppdebug $cgsmp noplot tz.cmd -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=halt 
*  totalview srun -a -N1 -n4 -ppdebug $cgsmp noplot tz.cmd -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go
*  srun -ppdebug -N2 -n2 memcheck_all $cgsmp noplot tz.cmd -g=square10.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 
* 
*  mpirun-wdh -np 2 $cgsmp noplot tz -g=sise2.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.1 -go=go
*  mpirun-wdh -np 4 $cgsmp noplot tz -g=nonBox8.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go
*  mpirun-wdh -np 1 $cgsmp noplot tz -g=rbibe1.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.1 -go=go
*  totalview srun -a -N1 -n2 -ppdebug $cgsmp noplot tz -g=rbibe1.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.1 -go=go
*  mpirun-wdh -np 4 $cgsmp noplot tz -g=rbibe2.order2.hdf -degreex=2 -degreet=2 -diss=0. -tf=.5 -go=go
* 
* --- set default values for parameters ---
* 
$noplot=""; $backGround="square"; $grid="square10"; $mu=1.; $lambda=1.; $pv="nc";
$debug = 0;  $tPlot=.1; $diss=.1; $dissOrder=2; $bcn="d"; $cons=0; $godunovOrder=2;
$tz = "poly"; $degreex=2; $degreet=2; $fx=2.; $fy=$fx; $fz=$fx; $ft=$fx;  $ts="me"; 
$order = 2; $go="run"; 
$tFinal=1.; $cfl=.9; 
*
* ----------------------------- get command line arguments ---------------------------------------
GetOptions( "g=s"=>\$grid,"tf=f"=>\$tFinal,"degreex=i"=>\$degreex, "degreet=i"=>\$degreet,"diss=f"=>\$diss,\
 "tp=f"=>\$tPlot, "tz=s"=>\$tz, "show=s"=>\$show,"order=i"=>\$order,"debug=i"=>\$debug, \
 "cfl=f"=>\$cfl, "bg=s"=>\$backGround,"bcn=s"=>\$bcn,"go=s"=>\$go,"noplot=s"=>\$noplot,\
  "mu=f"=>\$mu,"lambda=f"=>\$lambda,"rho=f"=>\$rho,"dtMax=f"=>\$dtMax, "cons=i"=>\$cons,"pv=s"=>\$pv,\
  "godunovOrder=f"=>\$godunovOrder,"ts=s"=>\$ts );
* -------------------------------------------------------------------------------------------------
if( $solver eq "best" ){ $solver="choose best iterative solver"; }
if( $tz eq "poly" ){ $tz="polynomial"; }else{ $tz="trigonometric"; }
if( $order eq "2" ){ $order = "second order accurate"; }else{ $order = "fourth order accurate"; }
if( $model eq "ins" ){ $model = "incompressible Navier Stokes"; }else\
                     { $model = "incompressible Navier Stokes\n Boussinesq model"; }
* 
if( $pv eq "nc" ){ $pv = "non-conservative"; }
if( $pv eq "c" ){ $pv = "conservative"; $cons=1; }
if( $pv eq "g" ){ $pv = "godunov"; }
if( $pv eq "h" ){ $pv = "hemp"; }
*
if( $ts eq "me" ){ $ts = "modifiedEquationTimeStepping"; }
if( $ts eq "fe" ){ $ts = "forwardEuler"; }
if( $ts eq "ie" ){ $ts = "improvedEuler"; }
if( $ts eq "ab" ){ $ts = "adamsBashforth2"; }
* 
if( $bcn eq "d" ){ $bcn = "all=displacementBC"; }
if( $bcn eq "sf" ){ $bcn = "all=tractionBC\n"; }
if( $bcn eq "slip" ){ $bcn = "all=slipWall\n"; }
if( $bcn eq "mixed" ){ $bcn = "all=displacementBC\n $backGround(0,0)=tractionBC\n $backGround(1,0)=tractionBC"; }
if( $bcn eq "mixed2" ){ $bcn = "all=displacementBC\n $backGround(0,1)=tractionBC\n $backGround(1,1)=tractionBC"; }
if( $go eq "halt" ){ $go = "break"; }
if( $go eq "og" ){ $go = "open graphics"; }
if( $go eq "run" || $go eq "go" ){ $go = "movie mode\n finish"; }
*
$grid
* -new: set-up stage: 
linear elasticity
$pv
 continue
*  -- time stepping method : 
$ts
* 
SMPDE:lambda $lambda
SMPDE:mu $mu 
SMPDE:rho $rho
SMPDE:Godunov order of accuracy $godunovOrder
* 
OBTZ:$tz
OBTZ:twilight zone flow 1
OBTZ:degree in space $degreex
OBTZ:degree in time $degreet
OBTZ:frequencies (x,y,z,t) $fx $fy $fz $ft
* 
final time $tFinal
times to plot $tPlot
dissipation $diss
order of dissipation $dissOrder
cfl $cfl
use conservative difference $cons
*
* bc: $backGround=stressFree
*  -- for now we don't handle a corner where two stree free Bc's meet: 
*  bc: all=dirichlet
*  bc: $backGround(0,1)=stressFree
*  bc: $backGround(1,1)=stressFree
boundary conditions
  $bcn
done
*
debug $debug
check errors 1
plot errors 1
* 
continue
$go

