// ---------------------------------------------------------------------------
// Solve the equations of Solid Mechanics
//
//
// ---------------------------------------------------------------------------

#include "MappedGridOperators.h"
#include "PlotStuff.h"
// #include "SquareMapping.h"
// #include "AnnulusMapping.h"
// #include "MatrixTransform.h"
// #include "DataPointMapping.h"

// #include "OGTrigFunction.h"
// #include "OGPolyFunction.h"
#include "display.h"

#include "Cgsm.h"
#include "ParallelUtility.h"

#include "CgSolverUtil.h"

int 
getLineFromFile( FILE *file, char s[], int lim);

void display(realArray & u )
{
  printF("u.getlength(0)=%i\n",u.getLength(0));
  
  ::display(u,"u");
}



int 
main(int argc, char *argv[])
{
  Overture::start(argc,argv);
  // Use this to avoid un-necessary communication: 
  Optimization_Manager::setForceVSG_Update(Off);
  const int myid=Communication_Manager::My_Process_Number;

  int plotOption=true;
  bool smartRelease=false;
  bool reportMemory=false;
  bool loadBalance=false;
  
  aString commandFileName="";
  if( argc > 1 )
  { // look at arguments for "noplot" or some other name
    aString line;
    for( int i=1; i<argc; i++ )
    {
      line=argv[i];
      if( line=="-noplot" || line=="noplot" )
        plotOption=false;
      else if( line=="-nopause" || line=="-abortOnEnd" || line=="-nodirect" ||
               line=="-readCollective" || line=="-writeCollective" ||
               line=="nopause" || line=="abortOnEnd" || line=="nodirect" )
        continue; // these commands are processed by getGraphicsInterface below 
      else if( line=="memory" )
      {
	reportMemory=true;
        Diagnostic_Manager::setTrackArrayData(TRUE);
      }
      else if( line=="loadBalance" || line=="-loadBalance" )
      {
	loadBalance=true;
      }
      else if( line=="release" )
      {
        smartRelease=true;
        printF("*** turn on smart release of memory ***\n");
        Diagnostic_Manager::setSmartReleaseOfInternalMemory( ON );
      }
      else if( commandFileName=="" )
      {
        commandFileName=line;    
        printF("cgins: reading commands from file [%s]\n",(const char*)commandFileName);
      }
      
    }
  }
  else
    printF("Usage: `cgsm [options][file.cmd]' \n"
            "     options:                            \n" 
            "          noplot:   run without graphics \n" 
            "          nopause: do not pause \n" 
            "          abortOnEnd: abort if command file ends \n" 
            "          memory:   run with A++ memory tracking\n" 
            "          release:  run with A++ smart release of memeory\n"
            "     file.cmd: read this command file \n");

  GenericGraphicsInterface & ps = *Overture::getGraphicsInterface("cgsm",false,argc,argv);

  // By default start saving the command file called "cgsm.cmd"
  aString logFile="cgsm.cmd";
  ps.saveCommandFile(logFile);
  printF("User commands are being saved in the file `%s'\n",(const char *)logFile);

  ps.appendToTheDefaultPrompt("cgsm>");

  // read from a command file if given
  if( commandFileName!="" )
  {
    printF("read command file =%s\n",(const char*)commandFileName);
    ps.readCommandFile(commandFileName);
  }

  aString nameOfShowFile;

#if 0
  aString nameOfOGFile, nameOfShowFile;
  ps.inputString(nameOfOGFile,"Enter the name of the (old) overlapping grid file:");

  printF("Try to read the overlapping grid file : %s\n",(const char*)nameOfOGFile);
  
  // create and read in a CompositeGrid
  #ifdef USE_PPP
    // On Parallel machines always add at least this many ghost lines on local arrays
    const int numGhost=2;
    MappedGrid::setMinimumNumberOfDistributedGhostLines(numGhost);


  #endif
  // Reduce the width of extrapolation for interp neighbours
  // and increase parallel ghost boundary width to 2 -- this will mean that we should
  // get the same result independent of np
  GenericMappedGridOperators::setDefaultMaximumWidthForExtrapolateInterpolationNeighbours(3);

  CompositeGrid m;
  getFromADataBase(m,nameOfOGFile);
  CompositeGrid & cg = m.numberOfMultigridLevels()==1 ? m : m.multigridLevel[0];
  if( m.numberOfMultigridLevels() > 1 )
    m.update(CompositeGrid::THEmultigridLevel);
    
#else

  CompositeGrid cg;
  readOrBuildTheGrid(ps, cg, loadBalance);

#endif


  Interpolant interpolant(cg);
  interpolant.setImplicitInterpolationMethod(Interpolant::iterateToInterpolate);

  #ifdef USE_PPP
  if( !interpolant.interpolationIsExplicit() )
  {
    printf("cgsm:ERROR: The parallel composite grid interpolator needs explicit interpolation ****\n");
    Overture::abort();
  }
  #endif

  Ogshow *show=NULL;
  Cgsm & solver = *new Cgsm(cg,&ps,show,plotOption); 

  real time0 = getCPU();

  solver.setParametersInteractively();
  solver.timing(Cgsm::totalTime) += getCPU()-time0;

  solver.solve();

  delete &solver;  // do this so we close the show file and shutdown PETSc (if used).

  Overture::finish();          
  return 0;
}
