// This file automatically generated from divergence.bC with bpp.
#include "Cgsm.h"
#include "UnstructuredMapping.h"
#include "CompositeGridOperators.h"
#include "display.h"
#include "ParallelUtility.h"
#include "SmParameters.h"

#define FOR_3D(i1,i2,i3,I1,I2,I3) int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); for(i3=I3Base; i3<=I3Bound; i3++) for(i2=I2Base; i2<=I2Bound; i2++) for(i1=I1Base; i1<=I1Bound; i1++)

#define FOR_3(i1,i2,i3,I1,I2,I3) I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); for(i3=I3Base; i3<=I3Bound; i3++) for(i2=I2Base; i2<=I2Bound; i2++) for(i1=I1Base; i1<=I1Bound; i1++)

void Cgsm::
getMaxDivAndCurl( const int current, real t, 
                                    realCompositeGridFunction *pdiv /* =NULL */, int component /* =0 */,
                                    realCompositeGridFunction *pvor /* =NULL */, int vorComponent /* =0 */,
                                    realCompositeGridFunction *pDensity /* = NULL */, 
                                    int rhoComponent /* =0 */,
                                    bool computeMaxNorms /* = true */ )
// ======================================================================================
// Compute the maximum divergence, vorticity AND the max-norm of the solution
// 
//  /pdiv,component (input) : save the divergence in 'component' if this pointer is non-NULL.
//  /pvor,component (input) : save the vorticity starting at 'vorComponent' if this pointer is non-NULL.
//
// /pCharge, rhoComponent (input) : if there is a charge density then supply rho
//   in component "rhoComponent" of the grid function *pDensity
// ======================================================================================
{
    int & debug = parameters.dbase.get<int >("debug");
    FILE *& debugFile  =parameters.dbase.get<FILE* >("debugFile");
    FILE *& logFile    =parameters.dbase.get<FILE* >("logFile");
    FILE *& pDebugFile =parameters.dbase.get<FILE* >("pDebugFile");

    Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];

    Range all;

    int u1c,u2c,u3c;
    if( parameters.dbase.get<SmParameters::PDEVariation>("pdeVariation") == SmParameters::hemp )
    {
        u1c = parameters.dbase.get<int >("u1c");
        u2c = parameters.dbase.get<int >("u2c");
        u3c = parameters.dbase.get<int >("u3c");
    }
    else
    {
        u1c = parameters.dbase.get<int >("uc");
        u2c = parameters.dbase.get<int >("vc");
        u3c = parameters.dbase.get<int >("wc");
    }

    const int & numberOfComponents = parameters.dbase.get<int >("numberOfComponents");
    const int & uc =  u1c;
    const int & vc =  u2c;
    const int & wc =  u3c;
    const int & rc =  parameters.dbase.get<int >("rc");
    const int & tc =  parameters.dbase.get<int >("tc");
    const int & orderOfAccuracyInSpace = parameters.dbase.get<int>("orderOfAccuracy");
    const int & orderOfAccuracyInTime  = parameters.dbase.get<int>("orderOfTimeAccuracy");

//   if( useChargeDensity && pDensity==NULL )
//   {
//     // Compute the charge density if it is not supplied
//     if( pRho==NULL )
//       pRho=new realCompositeGridFunction(cg);

//     getChargeDensity(current,t,*pRho );
//     pDensity=pRho;
//     rhoComponent=0;
//   }


    Range C=numberOfComponents;
    if( computeMaxNorms )
    {
        solutionNorm.redim(C); 
        solutionNorm=0.;

        vorUMax=0.;
        divUMax=0.;
        gradUMax=0.;  // holds the max(|ux|,|vy|,|wz|) 
    }
    
    int i1Max=0, i2Max=0, i3Max=0, gridMax=0;  // keeps point where max occurs

    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
        realMappedGridFunction & u = gf[current].u[grid];


//      #ifdef USE_PPP
//        u.updateGhostBoundaries();
//      #endif

        MappedGrid & mg = cg[grid];
        const intArray & mask = mg.mask();
        #ifdef USE_PPP
            intSerialArray  maskLocal; getLocalArrayWithGhostBoundaries(mask,maskLocal);
            realSerialArray uLocal;    getLocalArrayWithGhostBoundaries(u,uLocal);
        #else
            const intSerialArray & maskLocal = mask;
            const realSerialArray & uLocal   = u;
        #endif

        getIndex(mg.gridIndexRange(),I1,I2,I3);
    // Here is the box where we apply the interior equations when there is a PML
        getBoundsForPML( mg,Iv ); 

        int includeGhost=0; // do not include ghost
        bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,I1,I2,I3,includeGhost);  // 
        if( !ok ) continue;  // no communication allowed after this point

        Range V = mg.numberOfDimensions()==2 ? 1 : 3;
        realSerialArray divLocal(uLocal.dimension(0),uLocal.dimension(1),uLocal.dimension(2)), 
                                        vorLocal(uLocal.dimension(0),uLocal.dimension(1),uLocal.dimension(2),V); 
        
        MappedGridOperators & mgop = (*cgop)[grid];

        int i1,i2,i3;


        const int ng=orderOfAccuracyInSpace/2;
        const int ng3 = mg.numberOfDimensions()==2 ? 0 : ng;
            
//     Index J1 = Range(max(I1.getBase(),divLocal.getBase(0)+ng ),min(I1.getBound(),divLocal.getBound(0)-ng ));
//     Index J2 = Range(max(I2.getBase(),divLocal.getBase(1)+ng ),min(I2.getBound(),divLocal.getBound(1)-ng ));
//     Index J3 = Range(max(I3.getBase(),divLocal.getBase(2)+ng3),min(I3.getBound(),divLocal.getBound(2)-ng3));

        const int *maskp = maskLocal.Array_Descriptor.Array_View_Pointer2;
        const int maskDim0=maskLocal.getRawDataSize(0);
        const int maskDim1=maskLocal.getRawDataSize(1);
        const int md1=maskDim0, md2=md1*maskDim1; 
#define MASK(i0,i1,i2) maskp[(i0)+(i1)*md1+(i2)*md2]

        real *divp = divLocal.Array_Descriptor.Array_View_Pointer2;
        const int divDim0=divLocal.getRawDataSize(0);
        const int divDim1=divLocal.getRawDataSize(1);
        const int d1=divDim0, d2=d1*divDim1; 
#define DIV(i0,i1,i2) divp[(i0)+(i1)*d1+(i2)*d2]

        real *vorp = vorLocal.Array_Descriptor.Array_View_Pointer3;
        const int vorDim0=vorLocal.getRawDataSize(0);
        const int vorDim1=vorLocal.getRawDataSize(1);
        const int vorDim2=vorLocal.getRawDataSize(2);
        const int vd1=vorDim0, vd2=vd1*vorDim1, vd3=vd2*vorDim2; 
#define VOR(i0,i1,i2,i3) vorp[(i0)+(i1)*vd1+(i2)*vd2+(i3)*vd3]

        real *up = uLocal.Array_Descriptor.Array_View_Pointer3;
        const int uDim0=uLocal.getRawDataSize(0);
        const int uDim1=uLocal.getRawDataSize(1);
        const int uDim2=uLocal.getRawDataSize(2);
#undef U
#define U(i0,i1,i2,i3) up[i0+uDim0*(i1+uDim1*(i2+uDim2*(i3)))]

        if( computeMaxNorms )
        {
            mgop.useConservativeApproximations(false);  // turn off since there are no conservative

            Range E(uc,uc+mg.numberOfDimensions()-1);
            realSerialArray udLocal(uLocal.dimension(0),uLocal.dimension(1),uLocal.dimension(2),E);

            real *udp = udLocal.Array_Descriptor.Array_View_Pointer3;
            const int udDim0=udLocal.getRawDataSize(0);
            const int udDim1=udLocal.getRawDataSize(1);
            const int udDim2=udLocal.getRawDataSize(2);
#undef UD
#define UD(i0,i1,i2,i3) udp[i0+udDim0*(i1+udDim1*(i2+udDim2*(i3)))]


            if( true )
            {

      	if( debug & 64 )
      	{
        	  display(uLocal,sPrintF("getMaxDivAndCurl: uLocal p=%i, t=%9.3e",myid,t),pDebugFile," %10.2e ");
      	}
            
      	divLocal=0.;
      	vorLocal=0.;
            
      	mgop.derivative(MappedGridOperators::xDerivative,uLocal,udLocal,I1,I2,I3,E);
	//display(div(I1,I2,I3)," compute div: ux","%6.2f ");
      	if( mg.numberOfDimensions()==2 )
      	{
        	  FOR_3D(i1,i2,i3,I1,I2,I3)
        	  {
          	    if( MASK(i1,i2,i3)>0 )
          	    { // find max of x-derivatives:
            	      gradUMax=max(gradUMax,max(fabs(UD(i1,i2,i3,uc)),fabs(UD(i1,i2,i3,vc))));
            	      DIV(i1,i2,i3)=UD(i1,i2,i3,uc);
                            VOR(i1,i2,i3,0)=UD(i1,i2,i3,vc); // v.x 
          	    }
        	  }
      	}
      	else
      	{
        	  FOR_3D(i1,i2,i3,I1,I2,I3)
        	  {
          	    if( MASK(i1,i2,i3)>0 )
          	    { // find max of x-derivatives:
            	      gradUMax=max(gradUMax,max(fabs(UD(i1,i2,i3,uc)),fabs(UD(i1,i2,i3,vc)),fabs(UD(i1,i2,i3,wc))));
            	      DIV(i1,i2,i3)=UD(i1,i2,i3,uc);
                            VOR(i1,i2,i3,1)=-UD(i1,i2,i3,wc); 
                            VOR(i1,i2,i3,2)= UD(i1,i2,i3,vc); 
          	    }
        	  }
      	}
            
      	mgop.derivative(MappedGridOperators::yDerivative,uLocal, udLocal,I1,I2,I3,E);
	//display(ud(I1,I2,I3)," compute div: uy","%6.2f ");

      	if( mg.numberOfDimensions()==2 )
      	{
        	  FOR_3D(i1,i2,i3,I1,I2,I3)
        	  {
          	    if( MASK(i1,i2,i3)>0 )
          	    {  // include max of y-derivatives:
            	      gradUMax=max(gradUMax,max(fabs(UD(i1,i2,i3,uc)),fabs(UD(i1,i2,i3,vc))));
            	      DIV(i1,i2,i3)+=UD(i1,i2,i3,vc);
                            VOR(i1,i2,i3,0)-=UD(i1,i2,i3,uc); // -u.y 
          	    }
        	  }
      	}
      	else
      	{
        	  FOR_3D(i1,i2,i3,I1,I2,I3)
        	  {
          	    if( MASK(i1,i2,i3)>0 )
          	    { // include max of y-derivatives:
            	      gradUMax=max(gradUMax,max(fabs(UD(i1,i2,i3,uc)),fabs(UD(i1,i2,i3,vc)),fabs(UD(i1,i2,i3,wc))));
            	      DIV(i1,i2,i3)+=UD(i1,i2,i3,vc);

                            VOR(i1,i2,i3,0) =UD(i1,i2,i3,wc);
                            VOR(i1,i2,i3,2)-=UD(i1,i2,i3,uc); 
          	    }
        	  }

        	  mgop.derivative(MappedGridOperators::zDerivative,uLocal,udLocal,I1,I2,I3,E);
        	  FOR_3(i1,i2,i3,I1,I2,I3)
        	  { 
          	    if( MASK(i1,i2,i3)>0 )
          	    { // include max of z-derivatives:
            	      gradUMax=max(gradUMax,fabs(UD(i1,i2,i3,uc)),fabs(UD(i1,i2,i3,vc)),fabs(UD(i1,i2,i3,wc)));
            	      DIV(i1,i2,i3)+=UD(i1,i2,i3,wc);

                            VOR(i1,i2,i3,0)-=UD(i1,i2,i3,vc);
                            VOR(i1,i2,i3,1)+=UD(i1,i2,i3,uc); 
          	    }
        	  }
      	}

      	if( debug & 64 )
      	{
        	  display(divLocal,sPrintF("getMaxDivAndCurl: divLocal p=%i, t=%9.3e",myid,t),pDebugFile," %10.2e ");
      	}
            }
            else
            {
      	divLocal=0.; 
      	mgop.derivative(MappedGridOperators::divergence,uLocal,divLocal,I1,I2,I3);
            }
            
        } // end if compute max norms
        
        bool useConservativeDivergence=false;
        if( !computeMaxNorms || useConservativeDivergence )
        {
      // just compute div(E) -- no max norms
            if( useConservativeDivergence )
      	mgop.useConservativeApproximations(useConservative);
            else
                mgop.useConservativeApproximations(false); 

            if( (debug & 4) && useConservativeDivergence )
                printF("getMaxDivergence: get conservative divergence t=%9.3e, useConservative=%i\n",t,useConservative);
            mgop.derivative(MappedGridOperators::divergence,uLocal,divLocal,I1,I2,I3);
            
        }
        mgop.useConservativeApproximations(useConservative);  // reset
        

    // display(div(I1,I2,I3),"div","%8.2e ");
        
        
    // display(div(I1,I2,I3)," div"," %10.2e ");

        if( computeMaxNorms )
        {
            real uMax[3]={0.,0.,0.}; //
            FOR_3D(i1,i2,i3,I1,I2,I3)
            {
      	if( MASK(i1,i2,i3)>0 )
      	{
        	  if( fabs(DIV(i1,i2,i3))>divUMax )
        	  {
          	    i1Max=i1; i2Max=i2; i3Max=i3; gridMax=grid;
          	    divUMax=fabs(DIV(i1,i2,i3));
        	  }
        	  if( fabs(VOR(i1,i2,i3,0))>vorUMax )
        	  {
          	    vorUMax=fabs(VOR(i1,i2,i3,0));
        	  }
        	  uMax[0]=max(fabs(U(i1,i2,i3,uc)),uMax[0]);
        	  uMax[1]=max(fabs(U(i1,i2,i3,vc)),uMax[1]);
      	}
      	else
      	{
        	  DIV(i1,i2,i3)=0.;
      	}
            
            }
            solutionNorm(uc)=max(solutionNorm(uc),uMax[0]);
            solutionNorm(vc)=max(solutionNorm(vc),uMax[1]);
        }
        
        if( pdiv!=NULL )
        {
      // save the divergence (for plotting probably)
            #ifdef USE_PPP
                realSerialArray div; getLocalArrayWithGhostBoundaries((*pdiv)[grid],div);
            #else
                realSerialArray & div =(*pdiv)[grid];
            #endif
            
            div(I1,I2,I3,component)=divLocal(I1,I2,I3);  // we could avoid storage for div in this case
        }
        if( pvor!=NULL )
        {
      // save the vorticity (for plotting probably)
            #ifdef USE_PPP
                realSerialArray vor; getLocalArrayWithGhostBoundaries((*pvor)[grid],vor);
            #else
                realSerialArray & vor =(*pvor)[grid];
            #endif
            vor(I1,I2,I3,V+vorComponent)=vorLocal(I1,I2,I3,V);  // we could avoid storage for vor in this case
        }
        

        if( computeMaxNorms && debug & 4 )
            printF("+++ processor=%i divUMax=%8.2e max at i=(%i,%i,%i) grid=%i \n",myid,
           	     divUMax,i1Max,i2Max,i3Max,gridMax);
            

    // Communication_Manager::Sync();
        
    } // end for grid
    if( computeMaxNorms )
    {
        divUMax=ParallelUtility::getMaxValue(divUMax);  // get max over all processors
        vorUMax=ParallelUtility::getMaxValue(vorUMax);  // get max over all processors
        gradUMax=ParallelUtility::getMaxValue(gradUMax);  // get max over all processors
    
        for( int c=C.getBase(); c<=C.getBound(); c++ )
            solutionNorm(c)=ParallelUtility::getMaxValue(solutionNorm(c)); // get max over all processors
    }
    

#undef DIV
#undef MASK
#undef U
#undef UD
}


void Cgsm::
getVelocityAndStress( const int current, real t, 
                  		      realCompositeGridFunction *pv /* =NULL */, int vComponent /* =0 */,
                  		      realCompositeGridFunction *ps /* =NULL */, int sComponent /* =0 */,
                  		      bool computeMaxNorms /* = true */ )
// ======================================================================================
/// \brief Compute the velocity and stress by differencing the displacements.
/// 
///  /pv,vComponent (input) : save the velocity in 'vComponent' if this pointer is non-NULL.
///  /ps,sComponent (input) : save the stress starting at 'sComponent' if this pointer is non-NULL.
///
// ======================================================================================
{
    int & debug = parameters.dbase.get<int >("debug");
    FILE *& debugFile  =parameters.dbase.get<FILE* >("debugFile");
    FILE *& logFile    =parameters.dbase.get<FILE* >("logFile");
    FILE *& pDebugFile =parameters.dbase.get<FILE* >("pDebugFile");

    Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];

    Range all;

    const int & numberOfComponents = parameters.dbase.get<int >("numberOfComponents");
    const int & uc =  parameters.dbase.get<int >("uc");
    const int & vc =  parameters.dbase.get<int >("vc");
    const int & wc =  parameters.dbase.get<int >("wc");
    const int & rc =  parameters.dbase.get<int >("rc");
    const int & tc =  parameters.dbase.get<int >("tc");
    const int & orderOfAccuracyInSpace = parameters.dbase.get<int>("orderOfAccuracy");
    const int & orderOfAccuracyInTime  = parameters.dbase.get<int>("orderOfTimeAccuracy");

    Range C=numberOfComponents;
    real velocityNorm=0., stressNorm=0.;
    
    const int prev = (current-1+numberOfTimeLevels) % numberOfTimeLevels;
    const real dti = deltaT>0. ? 1./deltaT : 1.;

    real & rho=parameters.dbase.get<real>("rho");
    RealArray & muGrid = parameters.dbase.get<RealArray>("muGrid");
    RealArray & lambdaGrid = parameters.dbase.get<RealArray>("lambdaGrid");

    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
        realMappedGridFunction & u  = gf[current].u[grid];
        realMappedGridFunction & up = gf[prev   ].u[grid];

        MappedGrid & mg = cg[grid];
        const intArray & mask = mg.mask();
        #ifdef USE_PPP
            intSerialArray  maskLocal; getLocalArrayWithGhostBoundaries(mask,maskLocal);
            realSerialArray uLocal;    getLocalArrayWithGhostBoundaries(u,uLocal);
            realSerialArray upLocal;   getLocalArrayWithGhostBoundaries(up,upLocal);
        #else
            const intSerialArray & maskLocal = mask;
            const realSerialArray & uLocal   = u;
            const realSerialArray & upLocal  = up;
        #endif

        getIndex(mg.gridIndexRange(),I1,I2,I3);
    // Here is the box where we apply the interior equations when there is a PML
        getBoundsForPML( mg,Iv ); 

        int includeGhost=0; // do not include ghost
        bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,I1,I2,I3,includeGhost);  // 
        if( !ok ) continue;  // no communication allowed after this point

        Range D1=uLocal.dimension(0), D2=uLocal.dimension(1), D3=uLocal.dimension(2);

        Range V = mg.numberOfDimensions();             // velocity components
        Range S = mg.numberOfDimensions()==2 ? 3 : 6;  // stress components (assumes symmetric tensor)
        Range E(uc,uc+mg.numberOfDimensions()-1);      // displacements

        realSerialArray vLocal(D1,D2,D3,V), 
                                        sLocal(D1,D2,D3,S); 
        
        MappedGridOperators & mgop = (*cgop)[grid];

        int i1,i2,i3;

        const real lambda = lambdaGrid(grid);
        const real mu = muGrid(grid);


        const int ng=orderOfAccuracyInSpace/2;
        const int ng3 = mg.numberOfDimensions()==2 ? 0 : ng;
            
        const int *maskp = maskLocal.Array_Descriptor.Array_View_Pointer2;
        const int maskDim0=maskLocal.getRawDataSize(0);
        const int maskDim1=maskLocal.getRawDataSize(1);
        const int md1=maskDim0, md2=md1*maskDim1; 
#define MASK(i0,i1,i2) maskp[(i0)+(i1)*md1+(i2)*md2]

        real *vp = vLocal.Array_Descriptor.Array_View_Pointer3;
        const int vDim0=vLocal.getRawDataSize(0);
        const int vDim1=vLocal.getRawDataSize(1);
        const int vDim2=vLocal.getRawDataSize(2);
#undef V
#define V(i0,i1,i2,i3) vp[i0+vDim0*(i1+vDim1*(i2+vDim2*(i3)))]

        real *sp = sLocal.Array_Descriptor.Array_View_Pointer3;
        const int sDim0=sLocal.getRawDataSize(0);
        const int sDim1=sLocal.getRawDataSize(1);
        const int sDim2=sLocal.getRawDataSize(2);
#undef S
#define S(i0,i1,i2,i3) sp[i0+sDim0*(i1+sDim1*(i2+sDim2*(i3)))]

        real *pu = uLocal.Array_Descriptor.Array_View_Pointer3;
        const int uDim0=uLocal.getRawDataSize(0);
        const int uDim1=uLocal.getRawDataSize(1);
        const int uDim2=uLocal.getRawDataSize(2);
#undef U
#define U(i0,i1,i2,i3) pu[i0+uDim0*(i1+uDim1*(i2+uDim2*(i3)))]

        real *upp = upLocal.Array_Descriptor.Array_View_Pointer3;
        const int upDim0=upLocal.getRawDataSize(0);
        const int upDim1=upLocal.getRawDataSize(1);
        const int upDim2=upLocal.getRawDataSize(2);
#undef UP
#define UP(i0,i1,i2,i3) upp[i0+upDim0*(i1+upDim1*(i2+upDim2*(i3)))]


        realSerialArray ux(D1,D2,D3,E),uy(D1,D2,D3,E), uz;


        vLocal=0.;
        sLocal=0.;
            
        mgop.derivative(MappedGridOperators::xDerivative,uLocal,ux,I1,I2,I3,E);
        mgop.derivative(MappedGridOperators::yDerivative,uLocal,uy,I1,I2,I3,E);

        if( mg.numberOfDimensions()==2 )
        {
            FOR_3D(i1,i2,i3,I1,I2,I3)
            {
      	if( MASK(i1,i2,i3)>0 )
      	{ 
	  // velocity: approximate by differences -- this is wrong at t=0 if there is no prev solution
        	  V(i1,i2,i3,0) = (U(i1,i2,i3,uc) - UP(i1,i2,i3,uc))*dti;
        	  V(i1,i2,i3,1) = (U(i1,i2,i3,vc) - UP(i1,i2,i3,vc))*dti;
        	  
	  // Stress:
        	  real div = ux(i1,i2,i3,uc) + uy(i1,i2,i3,vc);
        	  S(i1,i2,i3,0) = lambda*div + 2.*mu*(ux(i1,i2,i3,uc));  // s11
        	  S(i1,i2,i3,1) = mu*(uy(i1,i2,i3,uc)+ux(i1,i2,i3,vc));  // s12
        	  S(i1,i2,i3,2) = lambda*div + 2.*mu*(uy(i1,i2,i3,vc));  // s22

        	  velocityNorm=max(velocityNorm,fabs(V(i1,i2,i3,0)),fabs(V(i1,i2,i3,1)));
        	  stressNorm  =max(stressNorm  ,fabs(S(i1,i2,i3,0)),fabs(S(i1,i2,i3,1)),fabs(S(i1,i2,i3,2)));
      	}
            }
        }
        else
        {
            uz.redim(D1,D2,D3,E);
            mgop.derivative(MappedGridOperators::zDerivative,uLocal,uz,I1,I2,I3,E);

            FOR_3D(i1,i2,i3,I1,I2,I3)
            {
      	if( MASK(i1,i2,i3)>0 )
      	{ 
      	}
            }
        }
            
        if( pv!=NULL )
        {
      // save the velocity (for plotting probably)
            #ifdef USE_PPP
                realSerialArray v; getLocalArrayWithGhostBoundaries((*pv)[grid],v);
            #else
                realSerialArray & v =(*pv)[grid];
            #endif
            
            v(I1,I2,I3,V+vComponent)=vLocal(I1,I2,I3,V); // we could avoid storage for div in this case
        }
        if( ps!=NULL )
        {
      // save the stress (for plotting probably)
            #ifdef USE_PPP
                realSerialArray s; getLocalArrayWithGhostBoundaries((*ps)[grid],s);
            #else
                realSerialArray & s =(*ps)[grid];
            #endif
            s(I1,I2,I3,S+sComponent)=sLocal(I1,I2,I3,S);  // we could avoid storage 
        }
        

    // Communication_Manager::Sync();
        
    } // end for grid

    if( computeMaxNorms )
    {
        velocityNorm=ParallelUtility::getMaxValue(velocityNorm);  // get max over all processors
        stressNorm  =ParallelUtility::getMaxValue(stressNorm);  // get max over all processors
    }

    if( computeMaxNorms && (true || debug & 2) )
    {
        printF("getVelocityAndStress: |velocity|=%8.2e, |stress|=%8.2e at t=%8.2e \n",velocityNorm,stressNorm,t);
    }
    
    

#undef MASK
#undef U
#undef UP
#undef S
#undef V

}


