// ==================================================================================
// Macro to evaluate the traveling wave solution
// 
// evalSolution : true=eval solution, false=eval error
// ==================================================================================
#beginMacro getTravelingWaveSolution(evalSolution,U,ERR,X,t,I1,I2,I3)
//	printF("INFO: The traveling wave solutions are combinations of p and s solutions:\n"
//               "  [u1,u2] = ap* [ k1,k2] * G( k1*(x-xa)+k2*(y-ya) - cp*t )  (p-wave)\n"
//               "  [u1,u2] = as* [-k2,k1] * G( k1*(x-xa)+k2*(y-ya) - cs*t )  (s-wave\n",
//               "    where  G(xi)=0 for xi>0 and G(xi)=-1 for xi<0 \n");
{
  // hard code some numbers for now: 
  // real k1=1., k2=0., k3=0.;
  // real ap=1., xa=.5, ya=0.;
  real cp = sqrt( (lambda+2.*mu)/rho );
  real cs = sqrt( mu/rho );

  std::vector<real> & twd = parameters.dbase.get<std::vector<real> >("travelingWaveData");
  const int np = int(twd[0]);  // number of p wave solutions
  const int ns = int(twd[1]);  // number of s wave solutions

  // printF("**** travelingWave: cp=%8.2e, t=%8.2e v0=%8.2e *********\n",cp,t,v0);
  int i1,i2,i3;
  if( mg.numberOfDimensions()==2 )
  {
    real z0=0.;
    FOR_3D(i1,i2,i3,I1,I2,I3)
    {
      real x0 = X(i1,i2,i3,0);
      real y0 = X(i1,i2,i3,1);

      real u1=0., u2=0.;  // back-ground field

      int m=2;
      for (int n=0; n<np; n++ ) // p-wave solutions
      {
	real ap = twd[m++], k1=twd[m++], k2=twd[m++], k3=twd[m++], xa=twd[m++], ya=twd[m++], za=twd[m++];

	real xi = k1*(x0-xa) + k2*(y0-ya) - cp*t;

	if( xi<0 )
	{
	  u1 += -ap*k1*xi;
	  u2 += -ap*k2*xi;
	}
      }
      for (int n=0; n<ns; n++ ) // s-wave solutions
      {
	real as = twd[m++], k1=twd[m++], k2=twd[m++], k3=twd[m++], xa=twd[m++], ya=twd[m++], za=twd[m++];

	real xi = k1*(x0-xa) + k2*(y0-ya) - cs*t;

	if( xi<0 )
	{  // (-k2,k1)
	  u1 += +as*k2*xi;
	  u2 += -as*k1*xi;
	}
      }
      
      // printF(" (i1,i2)=(%i,%i) (x0,y0)=(%8.2e,%8.2e) xi=%8.2e (u1,u2)=(%8.2e,%8.2e)\n",i1,i2,x0,y0,xi,u1,u2);

      if( evalSolution )
      {
	U(i1,i2,i3,uc) =u1;
	U(i1,i2,i3,vc) =u2;
      }
      else
      {
	ERR(i1,i2,i3,uc) =U(i1,i2,i3,uc) - u1;
	ERR(i1,i2,i3,vc) =U(i1,i2,i3,vc) - u2;
      }
    }
  }
  else
  {
    OV_ABORT("Error: finish me");
  }
}

#endMacro








//   (Ex).t = (1/eps)*[  (Hz).y ]
//   (Ey).t = (1/eps)*[ -(Hz).x ]
//   (Hz).t = (1/mu) *[ (Ex).y - (Ey).x ]

#define exTrue(x,y,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*(-ky/(eps*cc))
#define eyTrue(x,y,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*( kx/(eps*cc))
#define hzTrue(x,y,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))

#define exLaplacianTrue(x,y,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*(+ky*(twoPi*twoPi*(kx*kx+ky*ky))/(eps*cc))
#define eyLaplacianTrue(x,y,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*(-kx*(twoPi*twoPi*(kx*kx+ky*ky))/(eps*cc))
#define hzLaplacianTrue(x,y,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*( -(twoPi*twoPi*(kx*kx+ky*ky) ) )

// define eyTrue(x,y,t) exp( -beta*SQR((x-x0)-c*(t)) )

// Here is a plane wave with the shape of a Gaussian
// xi = kx*(x)+ky*(y)-cc*(t)
// cc=  c*sqrt( kx*kx+ky*ky );
#define hzGaussianPulse(xi)  exp(-betaGaussianPlaneWave*((xi)*(xi)))
#define exGaussianPulse(xi)  hzGaussianPulse(xi)*(-ky/(eps*cc))
#define eyGaussianPulse(xi)  hzGaussianPulse(xi)*( kx/(eps*cc))

#define hzLaplacianGaussianPulse(xi)  ((4.*betaGaussianPlaneWave*betaGaussianPlaneWave*(kx*kx+ky*ky))*xi*xi-\
                                        (2.*betaGaussianPlaneWave*(kx*kx+ky*ky)))*exp(-betaGaussianPlaneWave*((xi)*(xi)))
#define exLaplacianGaussianPulse(xi)  hzLaplacianGaussianPulse(xi,t)*(-ky/(eps*cc))
#define eyLaplacianGaussianPulse(xi)  hzLaplacianGaussianPulse(xi,t)*( kx/(eps*cc))

// 3D
// E:
//   u.tt = (1/eps)*[ ((1/mu)*u.x).x + ((1/mu)*u.y).y + ((1/mu)*u.z).z ]
//   div(u)=0
// H
//   v.tt = (1/mu)*[ ((1/eps)*v.x).x + ((1/eps)*v.y).y + ((1/eps)*v.z).z ]
// Define macros for forcing functions


//
//   (Ex).t = (1/eps)*[ (Hz).y - (Hy).z ]
//   (Ey).t = (1/eps)*[ (Hx).z - (Hz).x ]
//   (Ez).t = (1/eps)*[ (Hy).x - (Hx).y ]
//   (Hx).t = (1/mu) *[ (Ey).z - (Ez).y ]
//   (Hy).t = (1/mu) *[ (Ez).x - (Ex).z ]
//   (Hz).t = (1/mu) *[ (Ex).y - (Ey).x ]

// ****************** finish this -> should `rotate' the 2d solution ****************

#define exTrue3d(x,y,z,t) sin(twoPi*(kx*(x)+ky*(y)+kz*(z)-cc*(t)))*(-ky/(eps*cc))
#define eyTrue3d(x,y,z,t) sin(twoPi*(kx*(x)+ky*(y)+kz*(z)-cc*(t)))*( kx/(eps*cc))
#define ezTrue3d(x,y,z,t) 0

#define hxTrue3d(x,y,z,t) 0
#define hyTrue3d(x,y,z,t) 0
#define hzTrue3d(x,y,z,t) sin(twoPi*(kx*(x)+ky*(y)+kz*(z)-cc*(t)))

#define exLaplacianTrue3d(x,y,z,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*(+ky*(twoPi*twoPi*(kx*kx+ky*ky))/(eps*cc))
#define eyLaplacianTrue3d(x,y,z,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*(-kx*(twoPi*twoPi*(kx*kx+ky*ky))/(eps*cc))
#define ezLaplacianTrue3d(x,y,z,t) 0

#define hxLaplacianTrue3d(x,y,z,t) 0
#define hyLaplacianTrue3d(x,y,z,t) 0
#define hzLaplacianTrue3d(x,y,z,t) sin(twoPi*(kx*(x)+ky*(y)-cc*(t)))*( -(twoPi*twoPi*(kx*kx+ky*ky) ) )


// ------------ macros for the plane material interface -------------------------
#defineMacro PMIex(x,y,t) (aa1*cos(twoPi*(kx*(x)+ky*(y)-w*(t))) + r*bb1*cos(twoPi*(-kx*(x)+ky*(y)-w*(t))))
#defineMacro PMIey(x,y,t) (aa2*cos(twoPi*(kx*(x)+ky*(y)-w*(t))) + r*bb2*cos(twoPi*(-kx*(x)+ky*(y)-w*(t))))
#defineMacro PMIhz(x,y,t) ((-twoPi/w)*( (a1*ky-a2*kx)*sin(twoPi*(kx*(x)+ky*(y)-w*(t))) \
                                    + (b1*ky+b2*kx)*r*sin(twoPi*(-kx*(x)+ky*(y)-w*(t)))) )

#defineMacro PMITex(x,y,t) (tau*dd1*cos(twoPi*(kappax*(x)+kappay*(y)-w*(t))))
#defineMacro PMITey(x,y,t) (tau*dd2*cos(twoPi*(kappax*(x)+kappay*(y)-w*(t))))
#defineMacro PMIThz(x,y,t) ((-twoPi*tau/w)*(d1*kappay-d2*kappax)*sin(twoPi*(kappax*(x)+kappay*(y)-w*(t))))

// OPTION: initialCondition, error, boundaryCondition
#beginMacro setPlaneMaterialInterfaceMacro(OPTION,J1,J2,J3)
if( numberOfDimensions==2 )
{
  int i1,i2,i3;

  // we assume there are just two grids with an interface in-between
  const real c1=cGrid(0);
  const real c2=cGrid(1);

  const real kNorm = sqrt(kx*kx+ky*ky);
  const real a1=-ky/kNorm, a2= kx/kNorm;
  const real b1=-ky/kNorm, b2=-kx/kNorm;
  const real w = c1*kNorm;

  real kappax,kappay,kappaNorm;
	
  const real cr=c2/c1;
  kappay=ky;
  // discrimant is assumed positive (otherwise internal reflection)
  kappax=sqrt( (kx*kx+ky*ky)/(cr*cr) - kappay*kappay );
	
  kappaNorm=sqrt(kappax*kappax+kappay*kappay);
  const real d1=-kappay/kappaNorm;
  const real d2= kappax/kappaNorm;
	
  const real cosTheta1=kx/kNorm;
  const real cosTheta2=kappax/kappaNorm;

   // reflection and transmission coefficients
  const real r = (c1*cosTheta1-c2*cosTheta2)/(c1*cosTheta1+c2*cosTheta2);
  const real tau = (2.*c2*cosTheta1)/(c1*cosTheta1+c2*cosTheta2);

  real tm=t-dt,x,y;
  const real x0=x0PlaneMaterialInterface[0], y0=x0PlaneMaterialInterface[1];
  // The solution is rotated according to the normal of the plane of the material interface
  real an0=normalPlaneMaterialInterface[0], an1=normalPlaneMaterialInterface[1];
  real aNorm = sqrt(an0*an0+an1*an1);
  an0/= aNorm; an1/=aNorm;
  real aa1=an0*a1-an1*a2, aa2= an1*a1+an0*a2;
  real bb1=an0*b1-an1*b2, bb2= an1*b1+an0*b2;
  real dd1=an0*d1-an1*d2, dd2= an1*d1+an0*d2;

  if( grid==0 )
  { // incident plus reflected wave.
    FOR_3D(i1,i2,i3,J1,J2,J3)
    {
      real xx1 = XEP(i1,i2,i3,0) -x0;
      real yy1 = XEP(i1,i2,i3,1) -y0;
      // rotate the coordinate system
      x = an0*xx1+an1*yy1;
      y =-an1*xx1+an0*yy1;

      real u1 = PMIex(x,y,t);
      real u2 = PMIey(x,y,t);
      real u3 = PMIhz(x,y,t);

      #If #OPTION eq "initialCondition"
       UEX(i1,i2,i3)= u1;
       UEY(i1,i2,i3)= u2;
       UHZ(i1,i2,i3)= u3;
 
       UMEX(i1,i2,i3)= PMIex(x,y,tm);
       UMEY(i1,i2,i3)= PMIey(x,y,tm);
       UMHZ(i1,i2,i3)= PMIhz(x,y,tm);
      #Elif #OPTION eq "error"
       ERREX(i1,i2,i3)=UEX(i1,i2,i3)-u1;
       ERREY(i1,i2,i3)=UEY(i1,i2,i3)-u2;
       ERRHZ(i1,i2,i3)=UHZ(i1,i2,i3)-u3;
      #Elif #OPTION eq "boundaryCondition"
       U(i1,i2,i3,ex)= u1;
       U(i1,i2,i3,ey)= u2;
       U(i1,i2,i3,hz)= u3;
      #End
    }
  }
  else
  {
    // transmitted wave
    FOR_3D(i1,i2,i3,J1,J2,J3)
    {
      real xx1 = XEP(i1,i2,i3,0) -x0;
      real yy1 = XEP(i1,i2,i3,1) -y0;
      // rotate the coordinate system
      x = an0*xx1+an1*yy1;
      y =-an1*xx1+an0*yy1;

      real u1 = PMITex(x,y,t);
      real u2 = PMITey(x,y,t);
      real u3 = PMIThz(x,y,t);

      #If #OPTION eq "initialCondition"
       UEX(i1,i2,i3)= u1;
       UEY(i1,i2,i3)= u2;
       UHZ(i1,i2,i3)= u3;
 
       UMEX(i1,i2,i3)= PMITex(x,y,tm);
       UMEY(i1,i2,i3)= PMITey(x,y,tm);
       UMHZ(i1,i2,i3)= PMIThz(x,y,tm);
      #Elif #OPTION eq "error"
       ERREX(i1,i2,i3)=UEX(i1,i2,i3)-u1;
       ERREY(i1,i2,i3)=UEY(i1,i2,i3)-u2;
       ERRHZ(i1,i2,i3)=UHZ(i1,i2,i3)-u3;
      #Elif #OPTION eq "boundaryCondition"
       U(i1,i2,i3,ex)= u1;
       U(i1,i2,i3,ey)= u2;
       U(i1,i2,i3,hz)= u3;
      #End
    }
  }
 }
#endMacro

//==================================================================================================
// Evaluate Tom Hagstom's exact solution defined as an integral of Guassian sources
// 
// OPTION: OPTION=solution or OPTION=error OPTION=bounary to compute the solution or the error or
//     the boundary condition
//
//==================================================================================================
#beginMacro getGaussianIntegralSolution(OPTION,VEX,VEY,VHZ,t)

if( initialConditionOption==gaussianIntegralInitialCondition )
{
  
  double wt,wx,wy;
  const int nsources=1;
  double xs[nsources], ys[nsources], tau[nsources], var[nsources], amp[nsources];
  xs[0]=0.;
  ys[0]=1.e-8*1./3.;  // should not be on a grid point
  tau[0]=-.95;
  var[0]=30.;
  amp[0]=1.;
   
  double period= 1.;  // period in y
  double time=t;
   
  int i1,i2,i3;

  FOR_3D(i1,i2,i3,J1,J2,J3)
  {
    double x=X(i1,i2,i3,0); 
    double y=X(i1,i2,i3,1);

    exmax(wt,wx,wy,nsources,xs[0],ys[0],tau[0],var[0],amp[0],period,x,y,time);

    #If #OPTION eq "solution"
      VEX(i1,i2,i3) = wy;
      VEY(i1,i2,i3) =-wx;
      VHZ(i1,i2,i3)= wt;
    #Elif #OPTION eq "error" 
      ERREX(i1,i2,i3) = VEX(i1,i2,i3) - wy;
      ERREY(i1,i2,i3) = VEY(i1,i2,i3) + wx;
      ERRHZ(i1,i2,i3) = VHZ(i1,i2,i3) - wt;

    #Else
      U(i1,i2,i3,ex) = wy;
      U(i1,i2,i3,ey) =-wx;
      U(i1,i2,i3,hz) = wt;
    #End
	
  }
}

#endMacro
