      subroutine smg3d( m,nd1a,nd1b,n1a,n1b,
     *                    nd2a,nd2b,n2a,n2b,
     *                    nd3a,nd3b,n3a,n3b,
     *                    ds1,ds2,ds3,dt,t,xy,rx,det,
     *                    u,up,mask,
     *                    nrparam,rparam,niparam,iparam,nrwrk,rwrk,
     *                    niwrk,iwrk,idebug,ier )
c
      implicit none
c.. ingoing declarations
      integer nd1a,nd1b,n1a,n1b,m
      integer nd2a,nd2b,n2a,n2b
      integer nd3a,nd3b,n3a,n3b
      integer mask(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b)
      integer nrparam,niparam,nrwrk,niwrk
      integer iparam(niparam)
      integer iwrk(niwrk)
      integer idebug,ier
      real ds1,ds2,ds3,dt,t
      real xy(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,3)
      real rx(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,3,3)
      real det(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b)
      real u(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,m)
      real up(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,m)
      real rparam(nrparam)
      real rwrk(nrwrk)
c
c.. locals
      integer i1,i2,i3,k
      integer md1a,md1b,md2a,md2b,md3a,md3b
      integer ngrid,lw,la1,laj,nreq
      real tol,almax(3)
      include 'smupcommons.h'
c
      write(6,*)'*** Entering smg3d ***'
c
c.. declarations for these are in smupcommons
      mu     = rparam(3)
      lam    = rparam(4)
      rho0   = rparam(5)
      eptz   = rparam(6)
      iorder = iparam(1)
      icart  = iparam(2)
      itz    = iparam(3)
c
c      iorder = 1
c
c.. set array dimensions for parallel
      md1a = max( nd1a,n1a-2 )
      md1b = min( nd1b,n1b+2 )
      md2a = max( nd2a,n2a-2 )
      md2b = min( nd2b,n2b+2 )
      md3a = max( nd3a,n3a-2 )
      md3b = min( nd3b,n3b+2 )
c
c.. sanity check
      if( m.ne.15 ) then
        write(6,*)'Error (smg3d) : m=15 is assumed'
        stop
      end if
c
c.. split up real workspace
      ngrid  = (md2b-md2a+1)*(md1b-md1a+1)
      lw   = 1
      la1  = lw+m*7*ngrid*2
      laj  = la1+9*2*ngrid*2
      nreq = laj+ngrid*2-1
      if( nreq.gt.nrwrk ) then
        ier = 4
        return
      end if
c
c.. filter out underflow
      tol = 1.0e-30
      do k = 1,m
        do i3 = md3a,md3b
        do i2 = md2a,md2b
        do i1 = md1a,md1b
          if( abs( u(i1,i2,i3,k) ).lt.tol ) u(i1,i2,i3,k) = 0.0
        end do
        end do
        end do
      end do
c
c.. make call to compute up
      call smup3d( m,nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *               n1a, n1b, n2a, n2b, n3a, n3b,
     *               md1a,md1b,md2a,md2b,md3a,md3b,
     *               ds1,ds2,ds3,dt,t,xy,rx,det,u,up,
     *               mask,almax,rwrk(lw),rwrk(la1),
     *               rwrk(laj),ier )
c
c.. add an artificial dissipation here if you want
cccccccccccccccccccccccccccccccccccccccccccccccccc
c
c.. compute real and imaginary parts of time stepping eigenvalues
      rparam(1) = 0.0
      rparam(2) = almax(1)/ds1+almax(2)/ds2+almax(3)/ds3
c
      write(6,*)'*** Leaving smg3d ***'
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smup3d( m,nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                     n1a, n1b, n2a, n2b, n3a, n3b,
     *                     md1a,md1b,md2a,md2b,md3a,md3b,
     *                     ds1,ds2,ds3,dt,t,xy,rx,det,u,up,
     *                     mask,almax,w,a1,aj,ier )
c
      implicit none
c.. ingoing declarations
      integer m,ier
      integer nd1a,nd1b,nd2a,nd2b,nd3a,nd3b
      integer n1a, n1b, n2a, n2b, n3a, n3b
      integer md1a,md1b,md2a,md2b,md3a,md3b
      integer mask(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b)
      real ds1,ds2,ds3,dt,t
      real xy(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,3)
      real rx(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,3,3)
      real det(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b)
      real u(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,m)
      real up(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,m)
      real almax(3)
      real w(m,md1a:md1b,md2a:md2b,7,2)
      real a1(3,3,md1a:md1b,md2a:md2b,2)
      real aj(md1a:md1b,md2a:md2b,2)
      include 'smupcommons.h'
c
c.. local declarations
      integer j,k,i1,i2,i3,i3m1
      integer i1p1,i2p1
      integer s1m,s1p,s2m,s2p,s3m,s3p,scent
      integer lay,laym1
      real met0(3,3),aj0
      real fx(12),htz(15),t1
      real ut(15),ux(12),uy(12),uz(12),u0(3),kap,x,y,z
      data s1m,s1p,s2m,s2p,s3m,s3p,scent / 1,2,3,4,5,6,7 /
c
c.. initialize max eigenvalues
      almax(1) = 0.0
      almax(2) = 0.0
      almax(3) = 0.0
c
c.. initialize up
      do k = 1,12
        do i3 = n3a,n3b
        do i2 = n2a,n2b
        do i1 = n1a,n1b
          up(i1,i2,i3,k) = 0.0
        end do
        end do
        end do
      end do
c
      i3 = n3a-1
      lay = 1 
c.. the "lay" variable indicates which layer we are on within w, a1, and aj
c     so that we don't have to do the costly copies to update the layer.
c
c.. set grid metrics and grid velocities
      call smupmetrics3d( nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                    md1a,md1b,md2a,md2b,md3a,md3b,
     *                    i3,rx,det,a1(1,1,md1a,md2a,lay),
     *                    aj(md1a,md2a,lay) )
      if( icart.eq.1 ) then
        ! if we are doing Cartesian then fill in all layers at once so we
        ! don't have to do it later
        call smupmetrics3d( nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                      md1a,md1b,md2a,md2b,md3a,md3b,
     *                      i3,rx,det,a1(1,1,md1a,md2a,2),
     *                      aj(md1a,md2a,2) )
      end if
c
c.. set first layer of "slope corrected" values
      call smupslope3d( m,nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                    md1a,md1b,md2a,md2b,md3a,md3b,
     *                    n1a, n1b, n2a, n2b, n3a, n3b,
     *                    i3,ds1,ds2,ds3,dt,t,xy,
     *                    a1(1,1,md1a,md2a,lay),
     *                    aj(nd1a,md2a,lay), 
     *                    mask(nd1a,nd2a,i3),u,
     *                    w(1,md1a,md2a,1,lay) )
c
c.. loop over planes i3=n3a,n3b+1
      do i3 = n3a,n3b+1
        i3m1 = i3-1
c
c.. update layer pointer
        laym1 = lay
        lay = lay+1
        if( lay.gt.2 ) lay = 1
c
c.. set metrics if needed
        if( icart.ne.1 ) then
          call smupmetrics3d( nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                        md1a,md1b,md2a,md2b,md3a,md3b,
     *                        i3,rx,det,a1(1,1,md1a,md2a,lay),
     *                        aj(md1a,md2a,lay) )
        end if
c
c.. slope correction for top layer of cells
        call smupslope3d( m,nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                      md1a,md1b,md2a,md2b,md3a,md3b,
     *                      n1a, n1b, n2a, n2b, n3a, n3b,
     *                      i3,ds1,ds2,ds3,dt,t,xy,
     *                      a1(1,1,md1a,md2a,lay),
     *                      aj(nd1a,md2a,lay), 
     *                      mask(nd1a,nd2a,i3),u,
     *                      w(1,md1a,md2a,1,lay) )

c
c..compute s3 flux and add it to up
        do i2 = n2a,n2b
        do i1 = n1a,n1b
          if( mask(i1,i2,i3).ne.0 .and. mask(i1,i2,i3m1).ne.0 ) then
            ! average metrics
            aj0 = 0.5*(aj(i1,i2,lay)+aj(i1,i2,laym1))
            met0(1,1) = 0.5*(a1(3,1,i1,i2,lay)+a1(3,1,i1,i2,laym1))
            met0(1,2) = 0.5*(a1(3,2,i1,i2,lay)+a1(3,2,i1,i2,laym1))
            met0(1,3) = 0.5*(a1(3,3,i1,i2,lay)+a1(3,3,i1,i2,laym1))
            met0(2,1) = 0.5*(a1(1,1,i1,i2,lay)+a1(1,1,i1,i2,laym1))
            met0(2,2) = 0.5*(a1(1,2,i1,i2,lay)+a1(1,2,i1,i2,laym1))
            met0(2,3) = 0.5*(a1(1,3,i1,i2,lay)+a1(1,3,i1,i2,laym1))
            met0(3,1) = 0.5*(a1(2,1,i1,i2,lay)+a1(2,1,i1,i2,laym1))
            met0(3,2) = 0.5*(a1(2,2,i1,i2,lay)+a1(2,2,i1,i2,laym1))
            met0(3,3) = 0.5*(a1(2,3,i1,i2,lay)+a1(2,3,i1,i2,laym1))

            ! solve Riemann problem
            call smupflux3d( m,aj0,met0,w(1,i1,i2,s3p,laym1),
     *                       w(1,i1,i2,s3m,lay),fx,almax(3) )
            do k = 1,12
              up(i1,i2,i3,k) = up(i1,i2,i3,k)+
     *           fx(k)/(ds3*aj(i1,i2,lay))
              up(i1,i2,i3m1,k) = up(i1,i2,i3m1,k)-
     *           fx(k)/(ds3*aj(i1,i2,laym1))
            end do
          end if
        end do
        end do
c
c.. if i3.le.n3b, then compute fluxes in the s1 and s2 directions
        if( i3.le.n3b ) then
c
c.. compute s1 fluxes
          do i2 = n2a,n2b
          do i1 = n1a-1,n1b
            i1p1 = i1+1
            if( mask(i1,i2,i3).ne.0.and.mask(i1p1,i2,i3).ne.0 ) then
              aj0 = 0.5*(aj(i1p1,i2,lay)+aj(i1,i2,lay))
              met0(1,1) = 0.5*(a1(1,1,i1p1,i2,lay)+a1(1,1,i1,i2,lay))
              met0(1,2) = 0.5*(a1(1,2,i1p1,i2,lay)+a1(1,2,i1,i2,lay))
              met0(1,3) = 0.5*(a1(1,3,i1p1,i2,lay)+a1(1,3,i1,i2,lay))
              met0(2,1) = 0.5*(a1(2,1,i1p1,i2,lay)+a1(2,1,i1,i2,lay))
              met0(2,2) = 0.5*(a1(2,2,i1p1,i2,lay)+a1(2,2,i1,i2,lay))
              met0(2,3) = 0.5*(a1(2,3,i1p1,i2,lay)+a1(2,3,i1,i2,lay))
              met0(3,1) = 0.5*(a1(3,1,i1p1,i2,lay)+a1(3,1,i1,i2,lay))
              met0(3,2) = 0.5*(a1(3,2,i1p1,i2,lay)+a1(3,2,i1,i2,lay))
              met0(3,3) = 0.5*(a1(3,3,i1p1,i2,lay)+a1(3,3,i1,i2,lay))

              ! solve Riemann problem
              call smupflux3d( m,aj0,met0,w(1,i1,i2,s1p,lay),
     *           w(1,i1p1,i2,s1m,lay),fx,almax(1) )
              do k = 1,12
                up(i1p1,i2,i3,k) = up(i1p1,i2,i3,k)+
     *             fx(k)/(ds1*aj(i1p1,i2,lay))
                up(i1,i2,i3,k) = up(i1,i2,i3,k)-
     *             fx(k)/(ds1*aj(i1,i2,lay))
              end do
            end if
          end do
          end do
c
c.. compute s2 fluxes
          do i2 = n2a-1,n2b
          do i1 = n1a,n1b
            i2p1 = i2+1
            if( mask(i1,i2,i3).ne.0.and.mask(i1,i2p1,i3).ne.0 ) then
              aj0 = 0.5*(aj(i1,i2p1,lay)+aj(i1,i2,lay))
              met0(1,1) = 0.5*(a1(2,1,i1,i2p1,lay)+a1(2,1,i1,i2,lay))
              met0(1,2) = 0.5*(a1(2,2,i1,i2p1,lay)+a1(2,2,i1,i2,lay))
              met0(1,3) = 0.5*(a1(2,3,i1,i2p1,lay)+a1(2,3,i1,i2,lay))
              met0(2,1) = 0.5*(a1(3,1,i1,i2p1,lay)+a1(3,1,i1,i2,lay))
              met0(2,2) = 0.5*(a1(3,2,i1,i2p1,lay)+a1(3,2,i1,i2,lay))
              met0(2,3) = 0.5*(a1(3,3,i1,i2p1,lay)+a1(3,3,i1,i2,lay))
              met0(3,1) = 0.5*(a1(1,1,i1,i2p1,lay)+a1(1,1,i1,i2,lay))
              met0(3,2) = 0.5*(a1(1,2,i1,i2p1,lay)+a1(1,2,i1,i2,lay))
              met0(3,3) = 0.5*(a1(1,3,i1,i2p1,lay)+a1(1,3,i1,i2,lay))

              ! solve Riemann problem
              call smupflux3d( m,aj0,met0,w(1,i1,i2,s2p,lay),
     *           w(1,i1,i2p1,s2m,lay),fx,almax(2) )
              do k = 1,12
                up(i1,i2p1,i3,k) = up(i1,i2p1,i3,k)+
     *             fx(k)/(ds2*aj(i1,i2p1,lay))
                up(i1,i2,i3,k) = up(i1,i2,i3,k)-
     *             fx(k)/(ds2*aj(i1,i2,lay))
              end do
            end if
          end do
          end do
c
c.. add free stream correction here ... FIX ME!! ...
c
c.. complete up
          do k = 13,15
            do i2 = n2a,n2b
            do i1 = n1a,n1b
              up(i1,i2,i3,k) = w(k-12,i1,i2,scent,lay)
            end do
            end do
          end do
        end if
c
c.. twilight zone
        if( itz.ne.0 ) then
          t1 = t
          if( iorder.eq.2 ) t1 = t1+0.5*dt
          do i2 = n2a,n2b
          do i1 = n1a,n1b
            call smuptz3d( m,xy(i1,i2,i3,1),xy(i1,i2,i3,2),
     *         xy(i1,i2,i3,3),t1,htz )
            do k = 1,15
              up(i1,i2,i3,k) = up(i1,i2,i3,k)+htz(k)
            end do
          end do
          end do
        end if
c
c.. end loop over layers
      end do
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smupflux3d( m,aj,met,wl,wr,fx,speed )
c
      implicit none
c.. ingoing declarations
      integer m
      real aj,met(3,3),wl(*),wr(*),fx(*),speed
      include 'smupcommons.h'
c
c.. local declarations
      integer j,k
      real norm(3),tan1(3),tan2(3)
      real al(12),el(12,12),er(12,12)
      real alpha(12),alp
c
c.. metrics
      norm(1) = met(1,1)
      norm(2) = met(1,2)
      norm(3) = met(1,3)
      tan1(1) = met(2,1)
      tan1(2) = met(2,2)
      tan1(3) = met(2,3)
      tan2(1) = met(3,1)
      tan2(2) = met(3,2)
      tan2(3) = met(3,3)

c
c.. get eigen structure. We could get away with part of this structure
c     as was done in smg2d, but in order to reduce coding errors
c     we will just go for the glory for now.      
      call smeig3d( norm,tan1,tan2,al,el,er )
c
c.. wave strengths (from the left state)
      do k = 1,12
        alpha(k) = 0.0
        do j = 1,12
          alpha(k) = alpha(k)+el(k,j)*(wr(j)-wl(j))
        end do
      end do
c
c.. flux (on the left)
c      do k = 1,12
c        wl(k) = 0.5*(wl(k)+wr(k))
c      end do
      call smflux3d( norm,aj,wl,fx )
c
c.. Riemann solution flux (computed from the left)
      do j = 1,12
        if( al(j).lt.0.0 ) then
          alp = aj*alpha(j)*al(j)
          do k = 1,12
            fx(k) = fx(k)+alp*er(k,j)
          end do
        end if
c
c.. fastest wave speed
        speed = max( abs(al(j)),speed )
      end do
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine smupslope3d( m,nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                          md1a,md1b,md2a,md2b,md3a,md3b,
     *                          n1a, n1b, n2a, n2b, n3a, n3b,
     *                          i3,ds1,ds2,ds3,dt,t,xy,
     *                          a1,aj,mask,u,w )
c
      implicit none
c.. ingoing declarations
      integer m,i3
      integer nd1a,nd1b,nd2a,nd2b,nd3a,nd3b
      integer md1a,md1b,md2a,md2b,md3a,md3b
      integer n1a, n1b, n2a, n2b, n3a, n3b
      integer mask(nd1a:nd1b,nd2a:nd2b)
      real ds1,ds2,ds3,dt,t
      real xy(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,3)
      real a1(3,3,md1a:md1b,md2a:md2b)
      real aj(md1a:md1b,md2a:md2b)
      real u(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,m)
      real w(m,md1a:md1b,md2a:md2b,7)
      include 'smupcommons.h'
c
c.. local declarations
      integer i1,i2,j,k,dir,ilimit
      real al(12),el(12,12),er(12,12)
      real norm(3),tan1(3),tan2(3)
      real htz(15)
      real alphal,alphar,alpha,tmp,tmp2,tmp3
      real s1,s2,s3
c
c      data ilimit / 1 / 
      data ilimit / 0 / 
c
c.. copy solution into w
      do dir = 1,7
        do k = 1,12
          do i2 = md2a,md2b
          do i1 = md1a,md1b
            w(k,i1,i2,dir) = u(i1,i2,i3,k)
          end do
          end do
        end do
      end do
c
      if( iorder.eq.1 ) return
c
      s1 = 0.5*dt/ds1
      s2 = 0.5*dt/ds2
      s3 = 0.5*dt/ds3
c
c.. loop over space
      do i2 = n2a-1,n2b+1
      do i1 = n1a-1,n1b+1
        if( mask(i1,i2).ne.0 ) then
c
c.. twilight zone stuff
          if( itz.ne.0 ) then
            call smuptz3d( m,xy(i1,i2,i3,1),xy(i1,i2,i3,2),
     *                     xy(i1,i2,i3,3),t,htz )
            do dir = 1,7
              do k = 1,12
                w(k,i1,i2,dir) = w(k,i1,i2,dir)+0.5*dt*htz(k)
              end do
            end do
          end if
c
c.. s1 direction
          norm(1) = a1(1,1,i1,i2)
          norm(2) = a1(1,2,i1,i2)
          norm(3) = a1(1,3,i1,i2)
          tan1(1) = a1(2,1,i1,i2)
          tan1(2) = a1(2,2,i1,i2)
          tan1(3) = a1(2,3,i1,i2)
          tan2(1) = a1(3,1,i1,i2)
          tan2(2) = a1(3,2,i1,i2)
          tan2(3) = a1(3,3,i1,i2)
          call smeig3d( norm,tan1,tan2,al,el,er )
c
          do j = 1,12
            alphal = 0.0
            alphar = 0.0
            do k = 1,12
              alphal = alphal+el(j,k)*(u(i1,i2,i3,k)-u(i1-1,i2,i3,k))
              alphar = alphar+el(j,k)*(u(i1+1,i2,i3,k)-u(i1,i2,i3,k))
            end do
c
            if( ilimit.eq.0 ) then
              alphal = 0.5*(alphal+alphar)
              alphar = alphal
            end if
c
            if( alphal*alphar.gt.0.0 ) then
              if( abs(alphal).lt.abs(alphar) ) then
                alpha = alphal
              else
                alpha = alphar
              end if
c
              tmp3 = s1*al(j)
              tmp = tmp3*alpha
              do k = 1,12
                tmp2 = tmp*er(k,j)
                w(k,i1,i2,1) = w(k,i1,i2,1)-(tmp3+0.5)*alpha*er(k,j)
                w(k,i1,i2,2) = w(k,i1,i2,2)-(tmp3-0.5)*alpha*er(k,j)
                w(k,i1,i2,3) = w(k,i1,i2,3)-tmp2
                w(k,i1,i2,4) = w(k,i1,i2,4)-tmp2
                w(k,i1,i2,5) = w(k,i1,i2,5)-tmp2
                w(k,i1,i2,6) = w(k,i1,i2,6)-tmp2
                w(k,i1,i2,7) = w(k,i1,i2,7)-tmp2
              end do
            end if
          end do
c
c.. s2 direction
          norm(1) = a1(2,1,i1,i2)
          norm(2) = a1(2,2,i1,i2)
          norm(3) = a1(2,3,i1,i2)
          tan1(1) = a1(3,1,i1,i2)
          tan1(2) = a1(3,2,i1,i2)
          tan1(3) = a1(3,3,i1,i2)
          tan2(1) = a1(1,1,i1,i2)
          tan2(2) = a1(1,2,i1,i2)
          tan2(3) = a1(1,3,i1,i2)
          call smeig3d( norm,tan1,tan2,al,el,er )
c
          do j = 1,12
            alphal = 0.0
            alphar = 0.0
            do k = 1,12
              alphal = alphal+el(j,k)*(u(i1,i2,i3,k)-u(i1,i2-1,i3,k))
              alphar = alphar+el(j,k)*(u(i1,i2+1,i3,k)-u(i1,i2,i3,k))
            end do
c
            if( ilimit.eq.0 ) then
              alphal = 0.5*(alphal+alphar)
              alphar = alphal
            end if
c
            if( alphal*alphar.gt.0.0 ) then
              if( abs(alphal).lt.abs(alphar) ) then
                alpha = alphal
              else
                alpha = alphar
              end if
c
              tmp3 = s2*al(j)
              tmp = tmp3*alpha
              do k = 1,12
                tmp2 = tmp*er(k,j)
                w(k,i1,i2,1) = w(k,i1,i2,1)-tmp2
                w(k,i1,i2,2) = w(k,i1,i2,2)-tmp2
                w(k,i1,i2,3) = w(k,i1,i2,3)-(tmp3+0.5)*alpha*er(k,j)
                w(k,i1,i2,4) = w(k,i1,i2,4)-(tmp3-0.5)*alpha*er(k,j)
                w(k,i1,i2,5) = w(k,i1,i2,5)-tmp2
                w(k,i1,i2,6) = w(k,i1,i2,6)-tmp2
                w(k,i1,i2,7) = w(k,i1,i2,7)-tmp2
              end do
            end if
          end do
c
c.. s3 direction
          norm(1) = a1(3,1,i1,i2)
          norm(2) = a1(3,2,i1,i2)
          norm(3) = a1(3,3,i1,i2)
          tan1(1) = a1(1,1,i1,i2)
          tan1(2) = a1(1,2,i1,i2)
          tan1(3) = a1(1,3,i1,i2)
          tan2(1) = a1(2,1,i1,i2)
          tan2(2) = a1(2,2,i1,i2)
          tan2(3) = a1(2,3,i1,i2)
          call smeig3d( norm,tan1,tan2,al,el,er )
c
          do j = 1,12
            alphal = 0.0
            alphar = 0.0
            do k = 1,12
              alphal = alphal+el(j,k)*(u(i1,i2,i3,k)-u(i1,i2,i3-1,k))
              alphar = alphar+el(j,k)*(u(i1,i2,i3+1,k)-u(i1,i2,i3,k))
            end do
c
            if( ilimit.eq.0 ) then
              alphal = 0.5*(alphal+alphar)
              alphar = alphal
            end if
c
            if( alphal*alphar.gt.0.0 ) then
              if( abs(alphal).lt.abs(alphar) ) then
                alpha = alphal
              else
                alpha = alphar
              end if
c
              tmp3 = s3*al(j)
              tmp = tmp3*alpha
              do k = 1,12
                tmp2 = tmp*er(k,j)
                w(k,i1,i2,1) = w(k,i1,i2,1)-tmp2
                w(k,i1,i2,2) = w(k,i1,i2,2)-tmp2
                w(k,i1,i2,3) = w(k,i1,i2,3)-tmp2
                w(k,i1,i2,4) = w(k,i1,i2,4)-tmp2
                w(k,i1,i2,5) = w(k,i1,i2,5)-(tmp3+0.5)*alpha*er(k,j)
                w(k,i1,i2,6) = w(k,i1,i2,6)-(tmp3-0.5)*alpha*er(k,j)
                w(k,i1,i2,7) = w(k,i1,i2,7)-tmp2
              end do
            end if
          end do
c
        end if
      end do
      end do

c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smeig3d( a,s,t,al,el,er )
c
      implicit none
c.. ingoing declarations
      real a(3),s(3),t(3),al(12),el(12,12),er(12,12)
      include 'smupcommons.h'
c
c.. locals
      integer i,j,k,row1,row2,col,ier
      real vm1(12),vm2(12),vm3(12)
      real vp1(12),vp2(12),vp3(12)
      real w1(12),w2(12),w3(12)
      real x1(12),x2(12),x3(12)
      real y1(12),y2(12),y3(12)
      real an(3),sn(3),tn(3)
      real c1,c2,kap,rad,alpha
      real sum,eps,jac
      real mult1,mult2,mult3,mult4,mult5
c
      eps = 1.0e-10
c
      kap = lam+2.0*mu
      c1 = sqrt( kap/rho0 )
      c2 = sqrt( mu/rho0 )
      mult1 = 1.0/(2.0*c1)
      mult2 = 1.0/(2.0*c2)
      mult3 = 1.0/(2.0*kap)
      mult4 = 1.0/(2.0*mu)
      mult5 = 1.0/kap
c
c.. do orthonormalization of a,s,t into an,sn,tn
c     while we are he we set eigenvalues too
      ! normalize an
      rad    = sqrt(a(1)**2+a(2)**2+a(3)**2)
      an(1)  = a(1)/rad
      an(2)  = a(2)/rad
      an(3)  = a(3)/rad
      al(1)  = -rad*c1
      al(2)  = -rad*c2
      al(3)  = -rad*c2
      al(4)  = 0.0
      al(5)  = 0.0
      al(6)  = 0.0
      al(7)  = 0.0
      al(8)  = 0.0
      al(9)  = 0.0
      al(10) = rad*c2
      al(11) = rad*c2
      al(12) = rad*c1
c
      ! set sn to be part of s which is orthoogonal to an
      alpha = an(1)*s(1)+an(2)*s(2)+an(3)*s(3)
      sn(1) = s(1)-alpha*an(1)
      sn(2) = s(2)-alpha*an(2)
      sn(3) = s(3)-alpha*an(3)
      ! normalize sn
      rad   = sqrt(sn(1)**2+sn(2)**2+sn(3)**2)
      sn(1) = sn(1)/rad
      sn(2) = sn(2)/rad
      sn(3) = sn(3)/rad
c
      ! set tn to be part of t which is orthoogonal to an and sn
      alpha = an(1)*t(1)+an(2)*t(2)+an(3)*t(3)
      tn(1) = t(1)-alpha*an(1)
      tn(2) = t(2)-alpha*an(2)
      tn(3) = t(3)-alpha*an(3)
      alpha = sn(1)*tn(1)+sn(2)*tn(2)+sn(3)*tn(3)
      tn(1) = tn(1)-alpha*sn(1)
      tn(2) = tn(2)-alpha*sn(2)
      tn(3) = tn(3)-alpha*sn(3)
      ! normalize tn
      rad   = sqrt(tn(1)**2+tn(2)**2+tn(3)**2)
      tn(1) = tn(1)/rad
      tn(2) = tn(2)/rad
      tn(3) = tn(3)/rad
c
c.. check to make sure these really are orthonormal AND right handed!
      jac = an(1)*(sn(2)*tn(3)-tn(2)*sn(3))+
     *      an(2)*(sn(3)*tn(1)-sn(1)*tn(3))+
     *      an(3)*(sn(1)*tn(2)-sn(2)*tn(1))
      if( jac.lt.0 ) then
        tn(1) = -tn(1)
        tn(2) = -tn(2)
        tn(3) = -tn(3)
      end if
      jac = an(1)*(sn(2)*tn(3)-tn(2)*sn(3))+
     *      an(2)*(sn(3)*tn(1)-sn(1)*tn(3))+
     *      an(3)*(sn(1)*tn(2)-sn(2)*tn(1))
      if( abs( 1.0-jac ).gt.eps ) then
        write(6,*)'error(smeig3d):orthonormal set not found!!',jac
        write(6,*)a(1),a(2),a(3)
        write(6,*)s(1),s(2),s(3)
        write(6,*)t(1),t(2),t(3)
        stop
      end if
c
c.. set up w,x,y vectors
      do i = 1,12
        w1(i) = 0.0
        w2(i) = 0.0
        w3(i) = 0.0
        x1(i) = 0.0
        x2(i) = 0.0
        x3(i) = 0.0
        y1(i) = 0.0
        y2(i) = 0.0
        y3(i) = 0.0        
      end do
      w1(7)  =  an(3)
      x1(8)  =  an(3)
      y1(9)  =  an(3)
c
      w1(10) =  -an(2)
      x1(11) =  -an(2)
      y1(12) =  -an(2)
cc
      w2(4)  = -an(3)
      x2(5)  = -an(3)
      y2(6)  = -an(3)
c
      w2(10) =  an(1)
      x2(11) =  an(1)
      y2(12) =  an(1)
cc
      w3(4)  =  an(2)
      x3(5)  =  an(2)
      y3(6)  =  an(2)
c
      w3(7)  = -an(1)
      x3(8)  = -an(1)
      y3(9)  = -an(1)
c
c.. set up v vectors
      vm1(1)  = 0.0
      vm1(2)  = an(3)*c2
      vm1(3)  = -an(2)*c2
      vm1(4)  = 0.0
      vm1(5)  = an(1)*an(3)*mu
      vm1(6)  = -an(1)*an(2)*mu
      vm1(7)  = an(1)*an(3)*mu
      vm1(8)  = 2.0*an(2)*an(3)*mu
      vm1(9)  = mu*(an(3)**2-an(2)**2)
      vm1(10) = -an(1)*an(2)*mu
      vm1(11) = mu*(an(3)**2-an(2)**2)
      vm1(12) = -2.0*an(2)*an(3)*mu
c      
      vm2(1)  = -an(3)*c2
      vm2(2)  = 0.0
      vm2(3)  = an(1)*c2
      vm2(4)  = -2.0*an(1)*an(3)*mu
      vm2(5)  = -an(2)*an(3)*mu
      vm2(6)  = (an(1)**2-an(3)**2)*mu
      vm2(7)  = -an(2)*an(3)*mu
      vm2(8)  = 0.0
      vm2(9)  = an(1)*an(2)*mu
      vm2(10) = (an(1)**2-an(3)**2)*mu
      vm2(11) = an(1)*an(2)*mu
      vm2(12) = 2.0*an(1)*an(3)*mu
c
      vm3(1)  = an(2)*c2
      vm3(2)  = -an(1)*c2
      vm3(3)  = 0.0
      vm3(4)  = 2.0*an(1)*an(2)*mu
      vm3(5)  = -(an(1)**2-an(2)**2)*mu
      vm3(6)  = an(2)*an(3)*mu
      vm3(7)  = -(an(1)**2-an(2)**2)*mu
      vm3(8)  = -2.0*an(1)*an(2)*mu
      vm3(9)  = -an(1)*an(3)*mu
      vm3(10) = an(2)*an(3)*mu
      vm3(11) = -an(1)*an(3)*mu
      vm3(12) = 0.0
c
      do i = 1,3
        vp1(i) = -vm1(i)
        vp2(i) = -vm2(i)
        vp3(i) = -vm3(i)
      end do
      do i = 4,12
        vp1(i) = vm1(i)
        vp2(i) = vm2(i)
        vp3(i) = vm3(i)
      end do
c
c.. left eigenvectors ...
      do i = 1,3
        el(1,i)  = mult1*an(i)
        el(2,i)  = mult2*tn(i)
        el(3,i)  = -mult2*sn(i)
        el(4,i)  = 0.0
        el(5,i)  = 0.0
        el(6,i)  = 0.0
        el(7,i)  = 0.0
        el(8,i)  = 0.0
        el(9,i)  = 0.0
        el(10,i) = -mult2*tn(i)
        el(11,i) = mult2*sn(i)
        el(12,i) = -mult1*an(i)
      end do
c
      do i = 1,3
        do j = 1,3
          col = 3*i+j
          el(1,col) = an(i)*an(j)*mult3
          el(2,col) = an(i)*tn(j)*mult4
          el(3,col) = -an(i)*sn(j)*mult4
          do k = 1,3
            el(3+k,col) = -mult5*(an(i)*an(j)*tn(k)*lam+
     *         an(i)*an(k)*tn(j)*kap)
            el(6+k,col) =  mult5*(an(i)*an(j)*sn(k)*lam+
     *         an(i)*an(k)*sn(j)*kap)
          end do
          el(10,col) = el(2,col)
          el(11,col) = el(3,col)
          el(12,col) = el(1,col)
        end do
      end do
c
      do i = 1,3
        row1 = i+3
        row2 = i+6
        do j = 1,3
          col = 3*j+i
          el(row1,col) = el(row1,3*j+i)+tn(j)
          el(row2,col) = el(row2,3*j+i)-sn(j)
        end do
      end do
c
c.. right eigenvectors ... some of this could go above, but this is clearer
      er(1,1)  = an(1)*c1
      er(2,1)  = an(2)*c1
      er(3,1)  = an(3)*c1
      er(4,1)  = 2.0*an(1)**2*mu+lam
      er(5,1)  = 2.0*an(1)*an(2)*mu
      er(6,1)  = 2.0*an(1)*an(3)*mu
      er(7,1)  = 2.0*an(1)*an(2)*mu
      er(8,1)  = 2.0*an(2)**2*mu+lam
      er(9,1)  = 2.0*an(2)*an(3)*mu
      er(10,1) = 2.0*an(1)*an(3)*mu
      er(11,1) = 2.0*an(2)*an(3)*mu
      er(12,1) = 2.0*an(3)**2*mu+lam
c
      er(1,12) = -er(1,1)
      er(2,12) = -er(2,1)
      er(3,12) = -er(3,1)
      do i = 4,12
        er(i,12) = er(i,1)
      end do
c
      do i = 1,12
        er(i,2)  = sn(1)*vm1(i)+sn(2)*vm2(i)+sn(3)*vm3(i)
        er(i,3)  = tn(1)*vm1(i)+tn(2)*vm2(i)+tn(3)*vm3(i)
        er(i,4)  = sn(1)*w1(i)+sn(2)*w2(i)+sn(3)*w3(i)
        er(i,5)  = sn(1)*x1(i)+sn(2)*x2(i)+sn(3)*x3(i)
        er(i,6)  = sn(1)*y1(i)+sn(2)*y2(i)+sn(3)*y3(i)
        er(i,7)  = tn(1)*w1(i)+tn(2)*w2(i)+tn(3)*w3(i)
        er(i,8)  = tn(1)*x1(i)+tn(2)*x2(i)+tn(3)*x3(i)
        er(i,9)  = tn(1)*y1(i)+tn(2)*y2(i)+tn(3)*y3(i)
        er(i,10) = sn(1)*vp1(i)+sn(2)*vp2(i)+sn(3)*vp3(i)
        er(i,11) = tn(1)*vp1(i)+tn(2)*vp2(i)+tn(3)*vp3(i)
      end do
c
c.. probably want to check that these really are inverses for debugging purposes
c
c      if( .true. ) then
      if( .false. ) then
      ier = 0
      do i = 1,12
        do j = 1,12
          sum = 0.0
          do k = 1,12
            sum = sum+el(i,k)*er(k,j)
          end do
          if( i.eq.j ) sum = sum-1.0
          if( abs( sum ).gt.eps ) then
            write(6,*)i,j,sum
            ier = 1
          end if
        end do
      end do
      if( ier.ne.0 ) then
        write(6,*)'error(smeig3d), metrics below!'
        write(6,*)an(1),an(2),an(3)
        write(6,*)sn(1),sn(2),sn(3)
        write(6,*)tn(1),tn(2),tn(3)
        pause
      end if
      end if
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smflux3d( an,aj,w,f )
c
      implicit none
c.. ingoing declarations
      real an(3),aj,w(12),f(12)
      include 'smupcommons.h'
c
c.. local declarations
      real a1,a2,a3,kap
c
      a1 = an(1)
      a2 = an(2)
      a3 = an(3)
c
      kap = lam+2.0*mu
c
      f(1)  = -aj*(a1*w(4)+a2*w(7)+a3*w(10))/rho0
      f(2)  = -aj*(a1*w(5)+a2*w(8)+a3*w(11))/rho0
      f(3)  = -aj*(a1*w(6)+a2*w(9)+a3*w(12))/rho0
      f(4)  = -aj*(a1*kap*w(1)+lam*(a2*w(2)+a3*w(3)))
      f(5)  = -aj*(mu*(a1*w(2)+a2*w(1)))
      f(6)  = -aj*(mu*(a1*w(3)+a3*w(1)))
      f(7)  = f(5)
      f(8)  = -aj*(lam*(a1*w(1)+a3*w(3))+kap*a2*w(2))
      f(9)  = -aj*(mu*(a2*w(3)+a3*w(2)))
      f(10) = f(6)
      f(11) = f(9)
      f(12) = -aj*(lam*(a1*w(1)+a2*w(2))+kap*a3*w(3))
c
      return
      end
c
c+++++++++++++++++++++++
c
      subroutine smupmetrics3d( nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,
     *                          md1a,md1b,md2a,md2b,md3a,md3b,
     *                          i3, rx,det,a1,aj )
c
      implicit none 
c.. ingoing declarations
      integer nd1a,nd1b,nd2a,nd2b,nd3a,nd3b
      integer md1a,md1b,md2a,md2b,md3a,md3b
      integer i3
      real rx(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,3,3)
      real det(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b)
      real a1(3,3,md1a:md1b,md2a:md2b)
      real aj(md1a:md1b,md2a:md2b)
      include 'smupcommons.h'
c
c.. local declarations
      integer i,j,i1,i2
c
      if( icart.eq.0 ) then
c
c.. non-Cartesian
        do i2 = md2a,md2b
        do i1 = md1a,md1b
          do i = 1,3
          do j = 1,3
            a1(i,j,i1,i2) = rx(i1,i2,i3,i,j)
          end do
          end do
          aj(i1,i2)=det(i1,i2,i3)
        end do
        end do
c
      else
c
c.. Cartesian
        do i2 = md2a,md2b
        do i1 = md1a,md1b
          do i = 1,3
          do j = 1,3
            if( i.eq.j ) then
              a1(i,j,i1,i2) = 1.0
            else
              a1(i,j,i1,i2) = 0.0
            end if
          end do
          end do
          aj(i1,i2) = 1.0
        end do
        end do
c
      end if
c
      return
      end
c
c++++++++++++++++++++
c
      subroutine smuptz3d( m,x,y,z,t,h )
c
      implicit none
c.. ingoing declarations
      integer m
      real x,y,z,t,h(15)
      include 'smupcommons.h'
c
c.. local declarations
      integer k
      real ut(15),ux(12),uy(12),uz(12),u0(3)
      real kap
c
      kap = lam+2.0*mu
c
      do k = 1,15
        call ogDeriv( eptz,1,0,0,0,x,y,z,t,k-1,ut(k) )
      end do
      do k = 1,12
        call ogDeriv( eptz,0,1,0,0,x,y,z,t,k-1,ux(k) )
        call ogDeriv( eptz,0,0,1,0,x,y,z,t,k-1,uy(k) )
        call ogDeriv( eptz,0,0,0,1,x,y,z,t,k-1,uz(k) )
      end do
      do k = 1,3
        call ogDeriv( eptz,0,0,0,0,x,y,z,t,k-1,u0(k) )
      end do
c
      h(1)  = ut(1) -(ux(4)+uy(7)+uz(10))/rho0
      h(2)  = ut(2) -(ux(5)+uy(8)+uz(11))/rho0
      h(3)  = ut(3) -(ux(6)+uy(9)+uz(12))/rho0
      h(4)  = ut(4) -(kap*ux(1)+lam*(uy(2)+uz(3)))
      h(5)  = ut(5) -mu*(ux(2)+uy(1))
      h(6)  = ut(6) -mu*(ux(3)+uz(1))
      h(7)  = ut(7) -mu*(ux(2)+uy(1))
      h(8)  = ut(8) -(lam*(ux(1)+uz(3))+kap*uy(2))
      h(9)  = ut(9) -mu*(uy(3)+uz(2))
      h(10) = ut(10)-mu*(ux(3)+uz(1))
      h(11) = ut(11)-mu*(uy(3)+uz(2))
      h(12) = ut(12)-(lam*(ux(1)+uy(2))+kap*uz(3))
      h(13) = ut(13)-u0(1)
      h(14) = ut(14)-u0(2)
      h(15) = ut(15)-u0(3)
c
      return
      end
c
c++++++++++++++++++++
c

