c *******************************************************************************
c   Solid Mechanics boundary conditions for the First-order system
c *******************************************************************************

c These next include files will define the macros that will define the difference approximations
c The actual macro is called below
c #Include "defineDiffNewerOrder2f.h"
c #Include "defineDiffNewerOrder4f.h"


#beginMacro beginLoops(n1a,n1b,n2a,n2b,n3a,n3b,na,nb)
do i3=n3a,n3b
do i2=n2a,n2b
do i1=n1a,n1b
do n=na,nb
  ! write(*,'(" periodic i1,i2,i3,n=",4i4)') i1,i2,i3,n
#endMacro

#beginMacro endLoops()
end do
end do
end do
end do
#endMacro


#beginMacro beginLoops2d()
 i3=n3a
 do i2=n2a,n2b
 do i1=n1a,n1b
#endMacro
#beginMacro endLoops2d()
 end do
 end do
#endMacro

! loop on the boundary but include ghost points in tangential directions
#beginMacro beginGhostLoops2d()
 i3=n3a
 do i2=nn2a,nn2b
 do i1=nn1a,nn1b
#endMacro

#beginMacro beginLoops3d()
 do i3=n3a,n3b
 do i2=n2a,n2b
 do i1=n1a,n1b
#endMacro
#beginMacro endLoops3d()

 end do
 end do
 end do
#endMacro

#beginMacro beginGhostLoops3d()
 do i3=nn3a,nn3b
 do i2=nn2a,nn2b
 do i1=nn1a,nn1b
#endMacro


#defineMacro extrap2(uu,k1,k2,k3,kc,ks1,ks2,ks3) \
            (2.*uu(k1,k2,k3,kc)-uu(k1+ks1,k2+ks2,k3+ks3,kc))

#defineMacro extrap3(uu,k1,k2,k3,kc,ks1,ks2,ks3) \
            (3.*uu(k1,k2,k3,kc)-3.*uu(k1+ks1,k2+ks2,k3+ks3,kc)\
            +   uu(k1+2*ks1,k2+2*ks2,k3+2*ks3,kc))

#defineMacro extrap4(uu,k1,k2,k3,kc,ks1,ks2,ks3) \
            (4.*uu(k1,k2,k3,kc)-6.*uu(k1+ks1,k2+ks2,k3+ks3,kc)\
            +4.*uu(k1+2*ks1,k2+2*ks2,k3+2*ks3,kc)-uu(k1+3*ks1,k2+3*ks2,k3+3*ks3,kc))

#defineMacro extrap5(uu,k1,k2,k3,kc,ks1,ks2,ks3) \
            (5.*uu(k1,k2,k3,kc)-10.*uu(k1+ks1,k2+ks2,k3+ks3,kc)\
            +10.*uu(k1+2*ks1,k2+2*ks2,k3+2*ks3,kc)-5.*uu(k1+3*ks1,k2+3*ks2,k3+3*ks3,kc)\
            +uu(k1+4*ks1,k2+4*ks2,k3+4*ks3,kc))

c ************************************************************************************************
c  This macro is used for looping over the faces of a grid to assign booundary conditions
c
c extra: extra points to assign
c          Case 1: extra=numberOfGhostPoints -- for assigning extended boundaries
c          Case 2: extra=-1 -- for assigning ghost points but not including extended boundaries
c numberOfGhostPoints : number of ghost points (1 for 2nd order, 2 for fourth-order ...)
c
c
c Output:
c  n1a,n1b,n2a,n2b,n3a,n3b : from gridIndexRange
c  nn1a,nn1b,nn2a,nn2b,nn3a,nn3b : includes "extra" points
c 
c ***********************************************************************************************
#beginMacro beginLoopOverSides(extra,numberOfGhostPoints)
 extra1a=extra
 extra1b=extra
 extra2a=extra
 extra2b=extra
 if( nd.eq.3 )then
   extra3a=extra
   extra3b=extra
 else
   extra3a=0
   extra3b=0
 end if
 if( boundaryCondition(0,0).lt.0 )then
   extra1a=max(0,extra1a) ! over-ride extra=-1 : assign ends in periodic directions (or internal parallel boundaries)
 else if( boundaryCondition(0,0).eq.0 )then
   extra1a=numberOfGhostPoints  ! include interpolation points since we assign ghost points outside these
 end if
 ! **NOTE** the bc on the right may be negative even it is not on the left (for parallel)
 if( boundaryCondition(1,0).lt.0 )then
   extra1b=max(0,extra1b) ! over-ride extra=-1 : assign ends in periodic directions
 else if( boundaryCondition(1,0).eq.0 )then
   extra1b=numberOfGhostPoints
 end if

 if( boundaryCondition(0,1).lt.0 )then
   extra2a=max(0,extra2a) ! over-ride extra=-1 : assign ends in periodic directions (or internal parallel boundaries)
 else if( boundaryCondition(0,1).eq.0 )then
   extra2a=numberOfGhostPoints  ! include interpolation points since we assign ghost points outside these
 end if
 ! **NOTE** the bc on the right may be negative even it is not on the left (for parallel)
 if( boundaryCondition(1,1).lt.0 )then
   extra2b=max(0,extra2b) ! over-ride extra=-1 : assign ends in periodic directions
 else if( boundaryCondition(1,1).eq.0 )then
   extra2b=numberOfGhostPoints
 end if

 if(  nd.eq.3 )then
  if( boundaryCondition(0,2).lt.0 )then
    extra3a=max(0,extra3a) ! over-ride extra=-1 : assign ends in periodic directions (or internal parallel boundaries)
  else if( boundaryCondition(0,2).eq.0 )then
    extra3a=numberOfGhostPoints  ! include interpolation points since we assign ghost points outside these
  end if
  ! **NOTE** the bc on the right may be negative even it is not on the left (for parallel)
  if( boundaryCondition(1,2).lt.0 )then
    extra3b=max(0,extra3b) ! over-ride extra=-1 : assign ends in periodic directions
  else if( boundaryCondition(1,2).eq.0 )then
    extra3b=numberOfGhostPoints
  end if
 end if

 do axis=0,nd-1
 do side=0,1

   if( boundaryCondition(side,axis).gt.0 )then

     ! write(*,'(" bcOpt: side,axis,bc=",3i2)') side,axis,boundaryCondition(side,axis)

     n1a=gridIndexRange(0,0)
     n1b=gridIndexRange(1,0)
     n2a=gridIndexRange(0,1)
     n2b=gridIndexRange(1,1)
     n3a=gridIndexRange(0,2)
     n3b=gridIndexRange(1,2)
     if( axis.eq.0 )then
       n1a=gridIndexRange(side,axis)
       n1b=gridIndexRange(side,axis)
     else if( axis.eq.1 )then
       n2a=gridIndexRange(side,axis)
       n2b=gridIndexRange(side,axis)
     else
       n3a=gridIndexRange(side,axis)
       n3b=gridIndexRange(side,axis)
     end if


     nn1a=gridIndexRange(0,0)-extra1a
     nn1b=gridIndexRange(1,0)+extra1b
     nn2a=gridIndexRange(0,1)-extra2a
     nn2b=gridIndexRange(1,1)+extra2b
     nn3a=gridIndexRange(0,2)-extra3a
     nn3b=gridIndexRange(1,2)+extra3b
     if( axis.eq.0 )then
       nn1a=gridIndexRange(side,axis)
       nn1b=gridIndexRange(side,axis)
     else if( axis.eq.1 )then
       nn2a=gridIndexRange(side,axis)
       nn2b=gridIndexRange(side,axis)
     else
       nn3a=gridIndexRange(side,axis)
       nn3b=gridIndexRange(side,axis)
     end if

     is=1-2*side

     is1=0
     is2=0
     is3=0
     if( axis.eq.0 )then
       is1=1-2*side
     else if( axis.eq.1 )then
       is2=1-2*side
     else if( axis.eq.2 )then
       is3=1-2*side
     else
       stop 5
     end if
     
     axisp1=mod(axis+1,nd)
     axisp2=mod(axis+2,nd)
     
     i3=n3a

!*      ! (js1,js2,js3) used to compute tangential derivatives
!*      js1=0
!*      js2=0
!*      js3=0
!*      if( axisp1.eq.0 )then
!*        js1=1-2*side
!*      else if( axisp1.eq.1 )then
!*        js2=1-2*side
!*      else if( axisp1.eq.2 )then
!*        js3=1-2*side
!*      else
!*        stop 5
!*      end if
!* 
!*      ! (ks1,ks2,ks3) used to compute second tangential derivative
!*      ks1=0
!*      ks2=0
!*      ks3=0
!*      if( axisp2.eq.0 )then
!*        ks1=1-2*side
!*      else if( axisp2.eq.1 )then
!*        ks2=1-2*side
!*      else if( axisp2.eq.2 )then
!*        ks3=1-2*side
!*      else
!*        stop 5
!*      end if

     if( debug.gt.7 )then
       write(*,'(" bcOpt: grid,side,axis=",3i3,", \
         loop bounds: n1a,n1b,n2a,n2b,n3a,n3b=",6i3)') grid,side,axis,\
         n1a,n1b,n2a,n2b,n3a,n3b

     end if
   end if ! if bc>0 
#endMacro

#beginMacro endLoopOverSides()
 end do ! end side
 end do ! end axis
#endMacro

! ====================================================================
! Look up an integer parameter from the data base
! ====================================================================
#beginMacro getIntParameter(name)
 ok=getInt(pdb,'name',name) 
 if( ok.eq.0 )then
   write(*,'("*** bcOptSmFOS:ERROR: unable to find name")') 
   stop 1122
 end if
#endMacro

! ====================================================================
! Look up a real parameter from the data base
! ====================================================================
#beginMacro getRealParameter(name)
 ok=getReal(pdb,'name',name) 
 if( ok.eq.0 )then
   write(*,'("*** bcOptSmFOS:ERROR: unable to find name")') 
   stop 1133
 end if
#endMacro

! ==================================================================================
!  Use this next macro to fill in values to the boundary forcing array 
!  (We cannot use the statement function bcf on the LHS of an expression)
! ==================================================================================
#defineMacro bcfa(side,axis,i1,i2,i3,m) bcf0(bcOffset(side,axis) + \
          (i1-dim(0,0,side,axis)+(dim(1,0,side,axis)-dim(0,0,side,axis)+1)* \
          (i2-dim(0,1,side,axis)+(dim(1,1,side,axis)-dim(0,1,side,axis)+1)* \
          (i3-dim(0,2,side,axis)+(dim(1,2,side,axis)-dim(0,2,side,axis)+1)*(m)))))

      subroutine bcOptSmFOS( nd, nd1a,nd1b,nd2a,nd2b,nd3a,nd3b,\
                          gridIndexRange, u, mask,rx, xy, det, boundaryCondition, \
                          addBoundaryForcing, interfaceType, dim, bcf00,bcf10,bcf01,bcf11,bcf02,bcf12,\
                          bcf0,bcOffset,ipar, rpar, pdb, ierr )
c ===================================================================================
c  Boundary conditions for solid mechanics : First Order System
c
c  gridType : 0=rectangular, 1=curvilinear
c
c  c2= mu/rho, c1=(mu+lambda)/rho;
c 
c The forcing for the boundary conditions can be accessed in two ways. One can either 
c use the arrays: 
c       bcf00(i1,i2,i3,m), bcf10(i1,i2,i3,m), bcf01(i1,i2,i3,m), bcf11(i1,i2,i3,m), 
c       bcf02(i1,i2,i3,m), bcf12(i1,i2,i3,m)
c which provide values for the 6 different faces in 6 different arrays. One can also
c access the same values using the single statement function
c         bcf(side,axis,i1,i2,i3,m)
c which is defined below. 
c ===================================================================================

      implicit none

      integer nd, \
              nd1a,nd1b,nd2a,nd2b,nd3a,nd3b, \
              ierr

      real u(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,0:*)
      integer mask(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b)
      real rx(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,0:nd-1,0:nd-1)
      real xy(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b,0:nd-1)
      real det(nd1a:nd1b,nd2a:nd2b,nd3a:nd3b)
      integer gridIndexRange(0:1,0:2),boundaryCondition(0:1,0:2)

      integer addBoundaryForcing(0:1,0:2)
      integer interfaceType(0:1,0:2,0:*)
      integer dim(0:1,0:2,0:1,0:2)

      real bcf00(dim(0,0,0,0):dim(1,0,0,0), dim(0,1,0,0):dim(1,1,0,0), dim(0,2,0,0):dim(1,2,0,0),0:*)
      real bcf10(dim(0,0,1,0):dim(1,0,1,0), dim(0,1,1,0):dim(1,1,1,0), dim(0,2,1,0):dim(1,2,1,0),0:*)
      real bcf01(dim(0,0,0,1):dim(1,0,0,1), dim(0,1,0,1):dim(1,1,0,1), dim(0,2,0,1):dim(1,2,0,1),0:*)
      real bcf11(dim(0,0,1,1):dim(1,0,1,1), dim(0,1,1,1):dim(1,1,1,1), dim(0,2,1,1):dim(1,2,1,1),0:*)
      real bcf02(dim(0,0,0,2):dim(1,0,0,2), dim(0,1,0,2):dim(1,1,0,2), dim(0,2,0,2):dim(1,2,0,2),0:*)
      real bcf12(dim(0,0,1,2):dim(1,0,1,2), dim(0,1,1,2):dim(1,1,1,2), dim(0,2,1,2):dim(1,2,1,2),0:*)

      real bcf0(0:*)
      integer*8 bcOffset(0:1,0:2)

      integer ipar(0:*)
      real rpar(0:*)

      double precision pdb  ! pointer to data base

c     --- local variables ----
      
      integer numberOfComponents

      real ue,ve,we,v1e,v2e,v3e,uet,vet,wet,v1et,v2et,v3et
      real uex,uey,uez, vex,vey,vez, wex,wey,wez
      real v1ex,v1ey,v1ez, v2ex,v2ey,v2ez, v3ex,v3ey,v3ez 
      real tau11,tau12,tau21,tau22
      real tau11x,tau11y,tau12x,tau12y,tau21x,tau21y,tau22x,tau22y
      real accel1,accel2,s11tilde,s12tilde,s21tilde,s22tilde

      integer side,axis,grid,gridType,orderOfAccuracy,orderOfExtrapolation,twilightZone,\
        uc,vc,wc,useWhereMask,debug,nn,n1,n2
      real dx(0:2),dr(0:2)
      real t,ep,dt,c1,c2,mu,lambda,alpha,beta,rho
      integer axisp1,axisp2,i1,i2,i3,is1,is2,is3,j1,j2,j3,js1,js2,js3,ks1,ks2,ks3,is,js,it,nit
      integer option,initialized

      integer numGhost,numberOfGhostPoints
      integer side1,side2
      integer n1a,n1b,n2a,n2b,n3a,n3b
      integer nn1a,nn1b,nn2a,nn2b,nn3a,nn3b
      integer extra1a,extra1b,extra2a,extra2b,extra3a,extra3b

      integer ok,getInt,getReal
      integer v1c,v2c,v3c,s11c,s12c,s13c,s21c,s22c,s23c,s31c,s32c,s33c

      real ux0,uy0,uz0,vx0,vy0,vz0,wx0,wy0,wz0

      real an1,an2,an3,aNormi,epsx,f1,f2,f3,b1,b2,b3
      real an11,an21,an12,an22,f11,f21,f12,f22,b11,b21,b12,b22,dot1,dot2
      real rad,fdot1,fdot2,fdot3,adet,a(0:1,0:1,0:1)

      real a11,a12,a21,a22,deti

      real err

      integer axis1,axis2
      real u1x,u2x,u1y,u2y,u1r,u2r,u1s,u2s,s11t,s12t,s21t,s22t
      real u1xe,u2xe,u1ye,u2ye,u1re,u2re,u1se,u2se,s11e,s12e,s21e,s22e
      real u1rr,u1rs,u1ss,u2rr,u2rs,u2ss,rxr,rxs,ryr,rys,sxr,sxs,syr,sys
      real u1xx,u1xy,u1yy,u2xx,u2xy,u2yy
      real u1xxe,u1xye,u1yye,u2xxe,u2xye,u2yye
      real s11xe,s12xe,s21xe,s22xe,s11ye,s12ye,s21ye,s22ye
      real u1xpe,u1ype,u2xpe,u2ype,u1xme,u2xme,u1yme,u2yme,s11pe,s12pe,s21pe,s22pe,s11me,s12me,s21me,s22me
      real u1xp,u2xp,u1yp,u2yp,u1xm,u2xm,u1ym,u2ym

      integer mc,icart
      real ds1,ds2

      ! boundary conditions parameters
      #Include "bcDefineFortranInclude.h"

      integer rectangular,curvilinear
      parameter(\
        rectangular=0,\
        curvilinear=1)


c     --- start statement function ----
      real bcf
      integer kd,m,n

      ! Here is the the generic boundary condition forcing array. It uses the bcOffset(side,axis) values as an
      ! an offset from the bcf0 array to access the bcf10, bcf01, bcf11, ... arrays
      bcf(side,axis,i1,i2,i3,m) = bcf0(bcOffset(side,axis) + \
          (i1-dim(0,0,side,axis)+(dim(1,0,side,axis)-dim(0,0,side,axis)+1)* \
          (i2-dim(0,1,side,axis)+(dim(1,1,side,axis)-dim(0,1,side,axis)+1)* \
          (i3-dim(0,2,side,axis)+(dim(1,2,side,axis)-dim(0,2,side,axis)+1)*(m)))))

c............... end statement functions

c       write(6,*)'bcs',(ipar(i1),i1=0,10),(rpar(i2),i2=0,12)
c       pause

c      mc=8
c      n1a=gridIndexRange(0,0)
c      n1b=gridIndexRange(1,0)
c      n2a=gridIndexRange(0,1)
c      n2b=gridIndexRange(1,1)
c      do i2=nd2a,nd2b
c      do i1=nd1a,nd1b
c        write(1,321)i1,i2,(u(i1,i2,nd3a,i3),i3=0,7)
c  321   format(2(1x,i3),8(1x,1pe10.3))
c      end do
c      end do
c      pause

      ierr=0

      nd                   =ipar(0)
      grid                 =ipar(1)
      uc                   =ipar(2)
      vc                   =ipar(3)
      wc                   =ipar(4)
      gridType             =ipar(5)
      orderOfAccuracy      =ipar(6)
      orderOfExtrapolation =ipar(7)
      twilightZone         =ipar(8)
      useWhereMask         =ipar(9)
      debug                =ipar(10)
     
      dx(0)                =rpar(0)
      dx(1)                =rpar(1)
      dx(2)                =rpar(2)
      dr(0)                =rpar(3)
      dr(1)                =rpar(4)
      dr(2)                =rpar(5)
      t                    =rpar(6)
      ep                   =rpar(7) ! pointer for exact solution
      dt                   =rpar(8)
      mu                   =rpar(9)
      lambda               =rpar(10)
      c1                   =rpar(11)
      c2                   =rpar(12)
     
      epsx=1.e-20 ! for aNormi -- fix me ---

c       write(6,*)'bcs'
c       pause
      ! debug = 15 ! *** turn on temporarily ***


      ! look up the component numbers for the velocity and stress from the C++ data base
      getIntParameter(v1c)
      getIntParameter(v2c)
      getIntParameter(v3c)

      getIntParameter(s11c)
      getIntParameter(s12c)
      getIntParameter(s13c)
      getIntParameter(s21c)
      getIntParameter(s22c)
      getIntParameter(s23c)
      getIntParameter(s31c)
      getIntParameter(s32c)
      getIntParameter(s33c)

      getRealParameter(rho)

      if( debug.gt.3 )then
        write(*,'(" bcOptSmFOS: mu,lambda,rho,c1,c2=",5f10.5," gridType=",i2)') mu,lambda,rho,c1,c2,gridType
           ! '
      end if

      if( debug.gt.7 )then
        write(*,'(" bcOptSmFOS: **START** grid=",i4," uc,vc,wc=",3i2)') grid,uc,vc,wc
           ! '
      end if
      if( debug.gt.7 )then
       n1a=gridIndexRange(0,0)
       n1b=gridIndexRange(1,0)
       n2a=gridIndexRange(0,1)
       n2b=gridIndexRange(1,1)
       n3a=gridIndexRange(0,2)
       n3b=gridIndexRange(1,2)
       write(*,'(" bcOptSmFOS: grid=",i3,",n1a,n1b,n2a,n2b,n3a,n3b=",6i3)') grid,n1a,n1b,n2a,n2b,n3a,n3b
        ! ' 
       ! write(*,*) 'bcOptSmFOS: u=',((((u(i1,i2,i3,m),m=0,nd-1),i1=n1a,n1b),i2=n2a,n2b),i3=n3a,n3b)
      end if
     

      if( .false. )then

        ! *** Don: call your BC routine here ****

        mc=8
        n1a=gridIndexRange(0,0)
        n1b=gridIndexRange(1,0)
        n2a=gridIndexRange(0,1)
        n2b=gridIndexRange(1,1)
        ds1=dr(0)
        ds2=dr(1)
        call smgbcs (mc,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,ds1,ds2,t,xy,u,boundaryCondition)

        return

      end if


      numGhost=orderOfAccuracy/2


c variables: u(j1,j2,1:2)=velocity
c            u(j1,j2,3:6)=stress (S11,S12,S21,S22)
c            u(j1,j2,7:8)=displacement


c      write(*,'(" -------- bcOptSmFOS: use new BCs orderOfExtrapolation=",i3)') orderOfExtrapolation
c      ! ' 

      numberOfComponents= nd + nd + nd*nd ! displacement, velocity and stress components 

      if( nd.eq.2 )then

        ! *********************************** 
        ! **************** 2D ***************
        ! *********************************** 


c*******
c******* Fill in forcing arrays if they are not provided ***********
c*******

        beginLoopOverSides(numGhost,numGhost)

         if( boundaryCondition(side,axis).eq.displacementBC )then

          if( addBoundaryForcing(side,axis).eq.0 )then
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then
             ! given displacements:
             bcfa(side,axis,i1,i2,i3,uc)  =0.   
             bcfa(side,axis,i1,i2,i3,vc)  =0.
             ! given velocities:
             bcfa(side,axis,i1,i2,i3,v1c) =0.   
             bcfa(side,axis,i1,i2,i3,v2c) =0.
             ! given acceleration: 
             bcfa(side,axis,i1,i2,i3,s11c)=0.   
             bcfa(side,axis,i1,i2,i3,s12c)=0.
            end if
           endLoops2d()
          else if( twilightZone.ne.0 )then
    !       write(6,*)'start..'
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,uc,ue)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,vc,ve)

             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v1c,v1e)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v2c,v2e)

             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s11c,tau11x)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s21c,tau21y)
             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s12c,tau12x)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s22c,tau22y)

             bcfa(side,axis,i1,i2,i3,uc)=ue
             bcfa(side,axis,i1,i2,i3,vc)=ve
             bcfa(side,axis,i1,i2,i3,v1c)=v1e
             bcfa(side,axis,i1,i2,i3,v2c)=v2e

             bcfa(side,axis,i1,i2,i3,s11c) = tau11x+tau21y
             bcfa(side,axis,i1,i2,i3,s12c) = tau12x+tau22y

      !       write(6,'(2(1x,i2),6(1x,1pe12.5))')i1,i2,ue,ve,v1e,v2e,tau11x+tau21y,tau12x+tau22y

      ! write(*,'(" i1,i2=",2i3," set ue,bcf(uc)=",2e10.2)') i1,i2,ue,bcf(side,axis,i1,i2,i3,uc)     
            end if
           endLoops2d()
      !     write(6,*)'end'
      !     pause
          end if

         else if( boundaryCondition(side,axis).eq.tractionBC )then

          if( addBoundaryForcing(side,axis).eq.0 )then
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then

             ! given traction (for the traction BC)
             bcfa(side,axis,i1,i2,i3,s11c)=0.   
             bcfa(side,axis,i1,i2,i3,s12c)=0.

             ! given traction (for determining displacements). Normally this is equal to the above
             ! traction values except when using twilight-zone
             bcfa(side,axis,i1,i2,i3,uc)=0.   
             bcfa(side,axis,i1,i2,i3,vc)=0.

             ! given rate of change of traction (for determining the velocity)
             bcfa(side,axis,i1,i2,i3,v1c)=0.   
             bcfa(side,axis,i1,i2,i3,v2c)=0.
            end if
           endLoops2d()
          else if( twilightZone.ne.0 )then
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then

             ! (an1,an2) = outward normal
             if( gridType.eq.rectangular )then
               if( axis.eq.0 )then
                 an1=-is
                 an2=0.
               else 
                 an1=0.
                 an2=-is
               end if
             else
               aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
               an1=-is*rx(i1,i2,i3,axis,0)*aNormi
               an2=-is*rx(i1,i2,i3,axis,1)*aNormi
             end if
             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,uc,uex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,uc,uey)
             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,vc,vex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,vc,vey)

             bcfa(side,axis,i1,i2,i3,uc ) = an1*( (lambda+2.*mu)*uex+lambda*vey ) + an2*( mu*(uey+vex) )
             bcfa(side,axis,i1,i2,i3,vc ) = an1*( mu*(uey+vex) ) + an2*( lambda*uex + (lambda+2.*mu)*vey )

             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v1c,v1ex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v1c,v1ey)
             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v2c,v2ex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v2c,v2ey)

             bcfa(side,axis,i1,i2,i3,v1c) = an1*( (lambda+2.*mu)*v1ex+lambda*v2ey ) + an2*( mu*(v1ey+v2ex) )
             bcfa(side,axis,i1,i2,i3,v2c) = an1*( mu*(v1ey+v2ex) ) + an2*( lambda*v1ex + (lambda+2.*mu)*v2ey )

             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s11c,tau11)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s21c,tau21)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s12c,tau12)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s22c,tau22)

             bcfa(side,axis,i1,i2,i3,s11c) = an1*tau11+an2*tau21   ! note : n_j sigma_ji  : sum over first index 
             bcfa(side,axis,i1,i2,i3,s12c) = an1*tau12+an2*tau22

            end if
           endLoops2d()

          else
           ! fill in the traction BC into the stress components  *wdh* 081109
           ! (this is needed since for TZ flow these values are different)
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then
             bcfa(side,axis,i1,i2,i3,s11c)=bcfa(side,axis,i1,i2,i3,uc)
             bcfa(side,axis,i1,i2,i3,s12c)=bcfa(side,axis,i1,i2,i3,vc)
            end if
           endLoops2d()
          end if

         else if( boundaryCondition(side,axis).eq.slipWall )then

c          write(6,*)side,axis,addBoundaryForcing(side,axis)
c          pause

          if( addBoundaryForcing(side,axis).eq.0 )then
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then

             ! given tangential stress (often zero)
             bcfa(side,axis,i1,i2,i3,s11c)=0.

             ! time rate of change of tangential stress
             bcfa(side,axis,i1,i2,i3,s12c)=0.

             ! given normal displacement
             bcfa(side,axis,i1,i2,i3,uc)=0.   

             ! time rate of change of normal displacement
             bcfa(side,axis,i1,i2,i3,v1c)=0.
            end if
           endLoops2d()
          else if( twilightZone.ne.0 )then
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then

             ! (an1,an2) = outward normal,  (-an2,an1) = tangent
             if( gridType.eq.rectangular )then
               if( axis.eq.0 )then
                 an1=-is
                 an2=0.
               else 
                 an1=0.
                 an2=-is
               end if
             else
               aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
               an1=-is*rx(i1,i2,i3,axis,0)*aNormi
               an2=-is*rx(i1,i2,i3,axis,1)*aNormi
             end if

             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,uc,ue)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,vc,ve)

             bcfa(side,axis,i1,i2,i3,uc) = an1*ue+an2*ve

             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v1c,ue)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v2c,ve)

             bcfa(side,axis,i1,i2,i3,v1c) = an1*ue+an2*ve

             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s11c,tau11)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s21c,tau21)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s12c,tau12)
             call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s22c,tau22)

             bcfa(side,axis,i1,i2,i3,s11c) = an1*(-an2*tau11+an1*tau12)+an2*(-an2*tau21+an1*tau22)

             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,uc,uex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,uc,uey)
             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,vc,vex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,vc,vey)

             tau11=(lambda+2.*mu)*uex+lambda*vey
             tau12=mu*(uey+vex)
             tau21=tau12
             tau22=lambda*uex+(lambda+2.*mu)*vey
             bcfa(side,axis,i1,i2,i3,s21c) = an1*(-an2*tau11+an1*tau12)+an2*(-an2*tau21+an1*tau22)

             call ogDeriv(ep,1,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s11c,tau11)
             call ogDeriv(ep,1,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s21c,tau21)
             call ogDeriv(ep,1,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s12c,tau12)
             call ogDeriv(ep,1,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,s22c,tau22)

             bcfa(side,axis,i1,i2,i3,s12c) = an1*(-an2*tau11+an1*tau12)+an2*(-an2*tau21+an1*tau22)

             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v1c,v1ex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v1c,v1ey)
             call ogDeriv(ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v2c,v2ex)
             call ogDeriv(ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,v2c,v2ey)

             tau11=(lambda+2.*mu)*v1ex+lambda*v2ey
             tau12=mu*(v1ey+v2ex)
             tau21=tau12
             tau22=lambda*v1ex+(lambda+2.*mu)*v2ey
             bcfa(side,axis,i1,i2,i3,s22c) = an1*(-an2*tau11+an1*tau12)+an2*(-an2*tau21+an1*tau22)

            end if
           endLoops2d()

          else
           ! fill in the traction BC into the stress components  *wdh* 081109
           ! (this is needed since for TZ flow these values are different)
           beginLoops2d()
            if (mask(i1,i2,i3).ne.0) then
             bcfa(side,axis,i1,i2,i3,s11c)=bcfa(side,axis,i1,i2,i3,uc)
             bcfa(side,axis,i1,i2,i3,s12c)=bcfa(side,axis,i1,i2,i3,vc)
            end if
           endLoops2d()
          end if

         end if ! bc 
        endLoopOverSides()


c*******
c******* Primary Dirichlet boundary conditions ***********
c*******

        beginLoopOverSides(numGhost,numGhost)
         if( boundaryCondition(side,axis).eq.displacementBC )then

          ! *************** Displacement BC *****************
          ! ..step 0: Dirichlet bcs for displacement and velocity

          beginLoops2d()
           if (mask(i1,i2,i3).ne.0) then
            u(i1,i2,i3,uc) =bcf(side,axis,i1,i2,i3,uc)    ! given displacements
            u(i1,i2,i3,vc) =bcf(side,axis,i1,i2,i3,vc)
            u(i1,i2,i3,v1c)=bcf(side,axis,i1,i2,i3,v1c)   ! given velocities
            u(i1,i2,i3,v2c)=bcf(side,axis,i1,i2,i3,v2c)
            
            !call ogDeriv(ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.,t,uc,ue)
            !write(*,'(" i1,i2=",2i3," u,ue=",2e10.2)') i1,i2,u(i1,i2,i3,uc),ue      

           end if
          endLoops2d()

         else if( boundaryCondition(side,axis).eq.tractionBC )then
           ! ********* Traction BC ********
           ! put "dirichlet parts of the traction BC here 
          if( gridType.eq.rectangular )then

           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              ! set normal components of the stress, n=(-is,0)
              u(i1,i2,i3,s11c) =-is*bcf(side,axis,i1,i2,i3,s11c) 
              u(i1,i2,i3,s12c) =-is*bcf(side,axis,i1,i2,i3,s12c)
             end if
            endLoops2d()
           else
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              ! set normal components of the stress, n=(0,-is)
              u(i1,i2,i3,s21c) =-is*bcf(side,axis,i1,i2,i3,s11c) 
              u(i1,i2,i3,s22c) =-is*bcf(side,axis,i1,i2,i3,s12c)
             end if
            endLoops2d()
           end if

          else  ! curvilinear 

            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              f1=bcf(side,axis,i1,i2,i3,s11c)              ! given traction forces
              f2=bcf(side,axis,i1,i2,i3,s12c)
              ! (an1,an2) = outward normal 
              aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
              an1=-is*rx(i1,i2,i3,axis,0)*aNormi
              an2=-is*rx(i1,i2,i3,axis,1)*aNormi
              b1=f1-(an1*u(i1,i2,i3,s11c)+an2*u(i1,i2,i3,s21c))
              b2=f2-(an1*u(i1,i2,i3,s12c)+an2*u(i1,i2,i3,s22c))

              u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+an1*b1
              u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+an1*b2

              u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+an2*b1
              u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+an2*b2
             end if
            endLoops2d()

          end if  ! end gridType

         else if( boundaryCondition(side,axis).eq.slipWall )then
           ! ********* SlipWall BC ********
           ! put "dirichlet parts of the slipwall BC here 
          if( gridType.eq.rectangular )then

           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              ! set n.tau.t and the normal component of displacement, n=(-is,0), t=(0,-is)
              u(i1,i2,i3,s12c) = bcf(side,axis,i1,i2,i3,s11c)
              u(i1,i2,i3,uc) = -is*bcf(side,axis,i1,i2,i3,uc)
              u(i1,i2,i3,v1c) = -is*bcf(side,axis,i1,i2,i3,v1c)
             end if
            endLoops2d()
           else
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              ! set n.tau.t and the normal component of displacement, n=(0,-is), t=(+is,0)
              u(i1,i2,i3,s21c) = -bcf(side,axis,i1,i2,i3,s11c)
              u(i1,i2,i3,vc) = -is*bcf(side,axis,i1,i2,i3,uc)
              u(i1,i2,i3,v2c) = -is*bcf(side,axis,i1,i2,i3,v1c)
             end if
            endLoops2d()
           end if

          else  ! curvilinear 

            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              f1=bcf(side,axis,i1,i2,i3,s11c)              ! given tangential traction force
              f2=bcf(side,axis,i1,i2,i3,uc)                ! given normal displacement
              f3=bcf(side,axis,i1,i2,i3,v1c)               ! given normal velocity
              ! (an1,an2) = outward normal and (-an2,an1) = unit tangent
              aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
              an1=-is*rx(i1,i2,i3,axis,0)*aNormi
              an2=-is*rx(i1,i2,i3,axis,1)*aNormi
              b1=f1-an1*(-u(i1,i2,i3,s11c)*an2+u(i1,i2,i3,s12c)*an1)-an2*(-u(i1,i2,i3,s21c)*an2+u(i1,i2,i3,s22c)*an1)
              b2=f2-an1*u(i1,i2,i3,uc)-an2*u(i1,i2,i3,vc)
              b3=f3-an1*u(i1,i2,i3,v1c)-an2*u(i1,i2,i3,v2c)

              u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)-an1*b1*an2
              u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+an1*b1*an1
              u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)-an2*b1*an2
              u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+an2*b1*an1

              u(i1,i2,i3,uc)=u(i1,i2,i3,uc)+an1*b2
              u(i1,i2,i3,vc)=u(i1,i2,i3,vc)+an2*b2

              u(i1,i2,i3,v1c)=u(i1,i2,i3,v1c)+an1*b3
              u(i1,i2,i3,v2c)=u(i1,i2,i3,v2c)+an2*b3
             end if
            endLoops2d()

          end if  ! end gridType

         end if ! bc 
        endLoopOverSides()

c*******
c******* Extrapolate to the first ghost cells (only for physical sides) ********
c*******

        beginLoopOverSides(numGhost,numGhost)
         if( boundaryCondition(side,axis).gt.0 )then
          beginLoops2d()
           if (mask(i1,i2,i3).ne.0) then
            do n=0,numberOfComponents-1
              u(i1-is1,i2-is2,i3,n)=extrap3(u,i1,i2,i3,n,is1,is2,is3)
            end do
           end if
          endLoops2d()
         end if ! bc 
        endLoopOverSides()


c*******
c******* Fix up components of stress in the corners (such as n1a,n2a) ********
c*******

c Note: it does not appear to be possible to set the components of stress in the corner
c       if one of the sides is a slipwall bc

        axis1=0
        axis2=1
        i3=gridIndexRange(0,2)
        if (gridType.eq.rectangular) then
          do side1=0,1
            i1=gridIndexRange(side1,axis1)
            do side2=0,1
              i2=gridIndexRange(side2,axis2)
              if (mask(i1,i2,i3).ne.0) then
                if (boundaryCondition(side1,axis1).eq.displacementBC.and.boundaryCondition(side2,axis2).eq.displacementBC) then

c Cartesian grid, pure displacement/velocity bcs
                  u1x=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dx(0))
                  u2x=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dx(0))
                  u1y=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dx(1))
                  u2y=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dx(1))
                  u(i1,i2,i3,s11c)=lambda*(u1x+u2y)+2.0*mu*u1x
                  u(i1,i2,i3,s12c)=mu*(u1y+u2x)
                  u(i1,i2,i3,s21c)=mu*(u1y+u2x)
                  u(i1,i2,i3,s22c)=lambda*(u1x+u2y)+2.0*mu*u2y

c                elseif (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

c Cartesian grid, pure traction bcs
c                 No fix is needed since the normals for (side1,axis1) and (side2,axis2) are orthogonal

                elseif (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.displacementBC) then

c Cartesian grid, mix bcs, case 1
                  u1x=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dx(0))
                  u2x=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dx(0))
                  u1y=(u(i1,i2,i3,s12c)-mu*u2x)/mu
                  u2y=(u(i1,i2,i3,s11c)-(lambda+2.0*mu)*u1x)/lambda
                  u(i1,i2,i3,s21c)=mu*(u1y+u2x)
                  u(i1,i2,i3,s22c)=lambda*(u1x+u2y)+2.0*mu*u2y
c                  write(6,*)'here (1), side1,side2=',side1,side2
c                  write(6,*)boundaryCondition(0,0),boundaryCondition(1,0),boundaryCondition(0,1),boundaryCondition(1,1)
c                  pause

                elseif (boundaryCondition(side1,axis1).eq.displacementBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

c Cartesian grid, mix bcs, case 2
                  u1y=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dx(1))
                  u2y=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dx(1))
                  u1x=(u(i1,i2,i3,s22c)-(lambda+2.0*mu)*u2y)/lambda
                  u2x=(u(i1,i2,i3,s21c)-mu*u1y)/mu
                  u(i1,i2,i3,s11c)=lambda*(u1x+u2y)+2.0*mu*u1x
                  u(i1,i2,i3,s12c)=mu*(u1y+u2x)
c                  write(6,*)'here (2), side1,side2=',side1,side2
c                  pause

                end if
              end if
            end do
          end do
        else
          do side1=0,1
            i1=gridIndexRange(side1,axis1)
            do side2=0,1
              i2=gridIndexRange(side2,axis2)
              if (mask(i1,i2,i3).ne.0) then
                if (boundaryCondition(side1,axis1).eq.displacementBC.and.boundaryCondition(side2,axis2).eq.displacementBC) then

c non-Cartesian grid, pure displacement/velocity bcs
                  u1r=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dr(0))
                  u2r=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dr(0))
                  u1s=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dr(1))
                  u2s=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dr(1))
                  u1x=u1r*rx(i1,i2,i3,0,0)+u1s*rx(i1,i2,i3,1,0)
                  u2x=u2r*rx(i1,i2,i3,0,0)+u2s*rx(i1,i2,i3,1,0)
                  u1y=u1r*rx(i1,i2,i3,0,1)+u1s*rx(i1,i2,i3,1,1)
                  u2y=u2r*rx(i1,i2,i3,0,1)+u2s*rx(i1,i2,i3,1,1)
                  u(i1,i2,i3,s11c)=lambda*(u1x+u2y)+2.0*mu*u1x
                  u(i1,i2,i3,s12c)=mu*(u1y+u2x)
                  u(i1,i2,i3,s21c)=mu*(u1y+u2x)
                  u(i1,i2,i3,s22c)=lambda*(u1x+u2y)+2.0*mu*u2y

                elseif (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

c non-Cartesian grid, pure traction bcs (needed since grid lines may not be orthogonal)
                  is=1-2*side1
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis1,0)**2+rx(i1,i2,i3,axis1,1)**2))
                  an11=-is*rx(i1,i2,i3,axis1,0)*aNormi            ! normals for axis1,side1
                  an21=-is*rx(i1,i2,i3,axis1,1)*aNormi
                  f11=bcf(side1,axis1,i1,i2,i3,s11c)              ! given traction forces for axis1,side1
                  f21=bcf(side1,axis1,i1,i2,i3,s12c)

                  is=1-2*side2
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis2,0)**2+rx(i1,i2,i3,axis2,1)**2))
                  an12=-is*rx(i1,i2,i3,axis2,0)*aNormi            ! normals for axis2,side2
                  an22=-is*rx(i1,i2,i3,axis2,1)*aNormi
                  f12=bcf(side2,axis2,i1,i2,i3,s11c)              ! given traction forces for axis2,side2
                  f22=bcf(side2,axis2,i1,i2,i3,s12c)

                  b11=f11-(an11*u(i1,i2,i3,s11c)+an21*u(i1,i2,i3,s21c))
                  b21=f21-(an11*u(i1,i2,i3,s12c)+an21*u(i1,i2,i3,s22c))

                  dot1=an11*an12+an21*an22                        ! cosine of the angle between the normals
                  dot2=an21*an12-an11*an22                        ! cosine of the angle between tangent(1) and normal(2)
                  b12=(f12-(an12*u(i1,i2,i3,s11c)+an22*u(i1,i2,i3,s21c))-dot1*b11)/dot2
                  b22=(f22-(an12*u(i1,i2,i3,s12c)+an22*u(i1,i2,i3,s22c))-dot1*b21)/dot2

                  u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+an11*b11+an21*b12
                  u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+an11*b21+an21*b22

                  u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+an21*b11-an11*b12
                  u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+an21*b21-an11*b22

                elseif (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.displacementBC) then

c non-Cartesian grid, mix bcs, case 1
                  is=1-2*side1
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis1,0)**2+rx(i1,i2,i3,axis1,1)**2))
                  an1=-is*rx(i1,i2,i3,axis1,0)*aNormi
                  an2=-is*rx(i1,i2,i3,axis1,1)*aNormi

                  u1r=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dr(0))
                  u2r=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dr(0))
                  a11=an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,0)+an2*mu*rx(i1,i2,i3,axis2,1)
                  a12=an1*lambda*rx(i1,i2,i3,axis2,1)+an2*mu*rx(i1,i2,i3,axis2,0)
                  b1=an1*u(i1,i2,i3,s11c)+an2*u(i1,i2,i3,s21c)
     *                 -(an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,0)+an2*mu*rx(i1,i2,i3,axis1,1))*u1r
     *                 -(an1*lambda*rx(i1,i2,i3,axis1,1)+an2*mu*rx(i1,i2,i3,axis1,0))*u2r
                  a21=an1*mu*rx(i1,i2,i3,axis2,1)+an2*lambda*rx(i1,i2,i3,axis2,0)
                  a22=an1*mu*rx(i1,i2,i3,axis2,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,1)
                  b2=an1*u(i1,i2,i3,s12c)+an2*u(i1,i2,i3,s22c)
     *                 -(an1*mu*rx(i1,i2,i3,axis1,1)+an2*lambda*rx(i1,i2,i3,axis1,0))*u1r
     *                 -(an1*mu*rx(i1,i2,i3,axis1,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,1))*u2r
                  deti=1.0/(a11*a22-a21*a12)
                  u1s=( b1*a22-b2*a12)*deti
                  u2s=(-b1*a21+b2*a11)*deti

                  u1x=rx(i1,i2,i3,0,0)*u1r+rx(i1,i2,i3,1,0)*u1s
                  u1y=rx(i1,i2,i3,0,1)*u1r+rx(i1,i2,i3,1,1)*u1s
                  u2x=rx(i1,i2,i3,0,0)*u2r+rx(i1,i2,i3,1,0)*u2s
                  u2y=rx(i1,i2,i3,0,1)*u2r+rx(i1,i2,i3,1,1)*u2s

                  u(i1,i2,i3,s11c)=(lambda+2.0*mu)*u1x+lambda*u2y
                  u(i1,i2,i3,s21c)=mu*(u1y+u2x)
                  u(i1,i2,i3,s12c)=u(i1,i2,i3,s21c)
                  u(i1,i2,i3,s22c)=(lambda+2.0*mu)*u2y+lambda*u1x

                elseif (boundaryCondition(side1,axis1).eq.displacementBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

c non-Cartesian grid, mix bcs, case 1
                  is=1-2*side2
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis2,0)**2+rx(i1,i2,i3,axis2,1)**2))
                  an1=-is*rx(i1,i2,i3,axis2,0)*aNormi
                  an2=-is*rx(i1,i2,i3,axis2,1)*aNormi

                  u1s=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dr(1))
                  u2s=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dr(1))
                  a11=an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,0)+an2*mu*rx(i1,i2,i3,axis1,1)
                  a12=an1*lambda*rx(i1,i2,i3,axis1,1)+an2*mu*rx(i1,i2,i3,axis1,0)
                  b1=an1*u(i1,i2,i3,s11c)+an2*u(i1,i2,i3,s21c)
     *                 -(an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,0)+an2*mu*rx(i1,i2,i3,axis2,1))*u1s
     *                 -(an1*lambda*rx(i1,i2,i3,axis2,1)+an2*mu*rx(i1,i2,i3,axis2,0))*u2s
                  a21=an1*mu*rx(i1,i2,i3,axis1,1)+an2*lambda*rx(i1,i2,i3,axis1,0)
                  a22=an1*mu*rx(i1,i2,i3,axis1,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,1)
                  b2=an1*u(i1,i2,i3,s12c)+an2*u(i1,i2,i3,s22c)
     *                 -(an1*mu*rx(i1,i2,i3,axis2,1)+an2*lambda*rx(i1,i2,i3,axis2,0))*u1s
     *                 -(an1*mu*rx(i1,i2,i3,axis2,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,1))*u2s
                  deti=1.0/(a11*a22-a21*a12)
                  u1r=( b1*a22-b2*a12)*deti
                  u2r=(-b1*a21+b2*a11)*deti

                  u1x=rx(i1,i2,i3,0,0)*u1r+rx(i1,i2,i3,1,0)*u1s
                  u1y=rx(i1,i2,i3,0,1)*u1r+rx(i1,i2,i3,1,1)*u1s
                  u2x=rx(i1,i2,i3,0,0)*u2r+rx(i1,i2,i3,1,0)*u2s
                  u2y=rx(i1,i2,i3,0,1)*u2r+rx(i1,i2,i3,1,1)*u2s

                  u(i1,i2,i3,s11c)=(lambda+2.0*mu)*u1x+lambda*u2y
                  u(i1,i2,i3,s21c)=mu*(u1y+u2x)
                  u(i1,i2,i3,s12c)=u(i1,i2,i3,s21c)
                  u(i1,i2,i3,s22c)=(lambda+2.0*mu)*u2y+lambda*u1x

                end if
              end if
            end do
          end do
        end if

c..add on TZ flow contribution (if necessary)
        if (twilightZone.ne.0) then
          do side1=0,1
            i1=gridIndexRange(side1,axis1)
            do side2=0,1
              i2=gridIndexRange(side2,axis2)
              if (mask(i1,i2,i3).ne.0) then
                if (boundaryCondition(side1,axis1).eq.displacementBC.and.boundaryCondition(side2,axis2).eq.displacementBC) then

c pure displacement/velocity bcs
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)
                  call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                  call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xe)
                  call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1ye)
                  call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                  u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+s11e-(lambda*(u1xe+u2ye)+2.0*mu*u1xe)
                  u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+s12e-mu*(u1ye+u2xe)
                  u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+s21e-mu*(u1ye+u2xe)
                  u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+s22e-(lambda*(u1xe+u2ye)+2.0*mu*u2ye)

                elseif (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

c pure traction bcs
c                  No TZ forcing needed here.  For the Cartesian case, no fix was done, and for the
c                  non-Cartesian case, the forcing was already included in the bcf array.

                elseif (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.displacementBC) then

c mix bcs, case 1
                  call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                  call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xe)
                  call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1ye)
                  call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)

                  if (gridType.eq.rectangular) then

c   Cartesian case
                    u1ye=(s12e-mu*u2xe)/mu
                    u2ye=(s11e-(lambda+2.0*mu)*u1xe)/lambda
                    u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+s21e-mu*(u1ye+u2xe)
                    u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+s22e-lambda*(u1xe+u2ye)-2.0*mu*u2ye

                  else

c   non-Cartesian case
                    is=1-2*side1
                    aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis1,0)**2+rx(i1,i2,i3,axis1,1)**2))
                    an1=-is*rx(i1,i2,i3,axis1,0)*aNormi
                    an2=-is*rx(i1,i2,i3,axis1,1)*aNormi

                    deti=1.0/(rx(i1,i2,i3,axis2,1)*rx(i1,i2,i3,axis1,0)-rx(i1,i2,i3,axis2,0)*rx(i1,i2,i3,axis1,1))
                    u1re=(rx(i1,i2,i3,axis2,1)*u1xe-rx(i1,i2,i3,axis2,0)*u1ye)*deti
                    u2re=(rx(i1,i2,i3,axis2,1)*u2xe-rx(i1,i2,i3,axis2,0)*u2ye)*deti
                    a11=an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,0)+an2*mu*rx(i1,i2,i3,axis2,1)
                    a12=an1*lambda*rx(i1,i2,i3,axis2,1)+an2*mu*rx(i1,i2,i3,axis2,0)
                    b1=an1*s11e+an2*s21e
     *                   -(an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,0)+an2*mu*rx(i1,i2,i3,axis1,1))*u1re
     *                   -(an1*lambda*rx(i1,i2,i3,axis1,1)+an2*mu*rx(i1,i2,i3,axis1,0))*u2re
                    a21=an1*mu*rx(i1,i2,i3,axis2,1)+an2*lambda*rx(i1,i2,i3,axis2,0)
                    a22=an1*mu*rx(i1,i2,i3,axis2,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,1)
                    b2=an1*s12e+an2*s22e
     *                   -(an1*mu*rx(i1,i2,i3,axis1,1)+an2*lambda*rx(i1,i2,i3,axis1,0))*u1re
     *                   -(an1*mu*rx(i1,i2,i3,axis1,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,1))*u2re
                    deti=1.0/(a11*a22-a21*a12)
                    u1se=( b1*a22-b2*a12)*deti
                    u2se=(-b1*a21+b2*a11)*deti

                    u1xe=rx(i1,i2,i3,0,0)*u1re+rx(i1,i2,i3,1,0)*u1se
                    u1ye=rx(i1,i2,i3,0,1)*u1re+rx(i1,i2,i3,1,1)*u1se
                    u2xe=rx(i1,i2,i3,0,0)*u2re+rx(i1,i2,i3,1,0)*u2se
                    u2ye=rx(i1,i2,i3,0,1)*u2re+rx(i1,i2,i3,1,1)*u2se

                    u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+s11e-(lambda+2.0*mu)*u1xe-lambda*u2ye
                    u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+s21e-mu*(u1ye+u2xe)
                    u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+s12e-mu*(u1ye+u2xe)
                    u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+s22e-(lambda+2.0*mu)*u2ye-lambda*u1xe

                  end if

                elseif (boundaryCondition(side1,axis1).eq.displacementBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

c mix bcs, case 2
                  call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                  call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xe)
                  call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1ye)
                  call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)

                  if (gridType.eq.rectangular) then

c   Cartesian case
                    u1xe=(s22e-(lambda+2.0*mu)*u2ye)/lambda
                    u2xe=(s21e-mu*u1ye)/mu
                    u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+s11e-lambda*(u1xe+u2ye)-2.0*mu*u1xe
                    u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+s12e-mu*(u1ye+u2xe)

                  else

c   non-Cartesian case
                    is=1-2*side2
                    aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis2,0)**2+rx(i1,i2,i3,axis2,1)**2))
                    an1=-is*rx(i1,i2,i3,axis2,0)*aNormi
                    an2=-is*rx(i1,i2,i3,axis2,1)*aNormi

                    deti=1.0/(rx(i1,i2,i3,axis1,0)*rx(i1,i2,i3,axis2,1)-rx(i1,i2,i3,axis1,1)*rx(i1,i2,i3,axis2,0))
                    u1se=(rx(i1,i2,i3,axis1,0)*u1ye-rx(i1,i2,i3,axis1,1)*u1xe)*deti
                    u2se=(rx(i1,i2,i3,axis1,0)*u2ye-rx(i1,i2,i3,axis1,1)*u2xe)*deti
                    a11=an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,0)+an2*mu*rx(i1,i2,i3,axis1,1)
                    a12=an1*lambda*rx(i1,i2,i3,axis1,1)+an2*mu*rx(i1,i2,i3,axis1,0)
                    b1=an1*s11e+an2*s21e
     *                   -(an1*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,0)+an2*mu*rx(i1,i2,i3,axis2,1))*u1se
     *                   -(an1*lambda*rx(i1,i2,i3,axis2,1)+an2*mu*rx(i1,i2,i3,axis2,0))*u2se
                    a21=an1*mu*rx(i1,i2,i3,axis1,1)+an2*lambda*rx(i1,i2,i3,axis1,0)
                    a22=an1*mu*rx(i1,i2,i3,axis1,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis1,1)
                    b2=an1*s12e+an2*s22e
     *                   -(an1*mu*rx(i1,i2,i3,axis2,1)+an2*lambda*rx(i1,i2,i3,axis2,0))*u1se
     *                   -(an1*mu*rx(i1,i2,i3,axis2,0)+an2*(lambda+2.0*mu)*rx(i1,i2,i3,axis2,1))*u2se
                    deti=1.0/(a11*a22-a21*a12)
                    u1re=( b1*a22-b2*a12)*deti
                    u2re=(-b1*a21+b2*a11)*deti

                    u1xe=rx(i1,i2,i3,0,0)*u1re+rx(i1,i2,i3,1,0)*u1se
                    u1ye=rx(i1,i2,i3,0,1)*u1re+rx(i1,i2,i3,1,1)*u1se
                    u2xe=rx(i1,i2,i3,0,0)*u2re+rx(i1,i2,i3,1,0)*u2se
                    u2ye=rx(i1,i2,i3,0,1)*u2re+rx(i1,i2,i3,1,1)*u2se

                    u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+s11e-(lambda+2.0*mu)*u1xe-lambda*u2ye
                    u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+s21e-mu*(u1ye+u2xe)
                    u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+s12e-mu*(u1ye+u2xe)
                    u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+s22e-(lambda+2.0*mu)*u2ye-lambda*u1xe

                  end if

                end if
              end if
            end do
          end do
        end if



          if( .false. )then

            ! fix up corner stresses

            icart=0
            if (gridType.eq.rectangular) icart=1

            mc=8
            n1a=gridIndexRange(0,0)
            n1b=gridIndexRange(1,0)
            n2a=gridIndexRange(0,1)
            n2b=gridIndexRange(1,1)
            ds1=dr(0)
            ds2=dr(1)
            call cstress (mc,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,ds1,ds2,t,xy,mask,rx,u,icart)

          end if


c*******
c******* Secondary Neumann boundary conditions (compatibility conditions) ********
c*******

        beginLoopOverSides(numGhost,numGhost)

         if( boundaryCondition(side,axis).eq.displacementBC )then

          if( gridType.eq.rectangular )then

           ! ********* DISPLACEMENT : Cartesian Grid **********

           !      u(j1,j2,3:6)=stress (S11,S12,S21,S22)
           ! Use:
           !   s11_x + s21_y = rho * u_tt  (from momentum eqn)
           !   s12_x + s22_y = rho * v_tt  (from momentum eqn)

           ! *** What about s12 = s21 ***
           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              accel1=rho*bcf(side,axis,i1,i2,i3,s11c)   ! rho times given acceleration
              accel2=rho*bcf(side,axis,i1,i2,i3,s12c)
              u(i1-is1,i2,i3,s11c)=u(i1+is1,i2,i3,s11c)-2.*is*dx(0)*(accel1-(u(i1,i2+1,i3,s21c)-u(i1,i2-1,i3,s21c))/(2.*dx(1)))
              u(i1-is1,i2,i3,s12c)=u(i1+is1,i2,i3,s12c)-2.*is*dx(0)*(accel2-(u(i1,i2+1,i3,s22c)-u(i1,i2-1,i3,s22c))/(2.*dx(1)))
             end if
            endLoops2d()
           else
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              accel1=rho*bcf(side,axis,i1,i2,i3,s11c)  ! rho times given acceleration
              accel2=rho*bcf(side,axis,i1,i2,i3,s12c)
              u(i1,i2-is2,i3,s21c)=u(i1,i2+is2,i3,s21c)-2.*is*dx(1)*(accel1-(u(i1+1,i2,i3,s11c)-u(i1-1,i2,i3,s11c))/(2.*dx(0)))
              u(i1,i2-is2,i3,s22c)=u(i1,i2+is2,i3,s22c)-2.*is*dx(1)*(accel2-(u(i1+1,i2,i3,s12c)-u(i1-1,i2,i3,s12c))/(2.*dx(0)))
             end if
            endLoops2d()
           end if

          else

           ! *********** DISPLACEMENT : Curvilinear Grid ****************

           ! To compute (s11,s21) use: 
           !    (1)   D_r[ J*(rx,ry).(s11,s21)] + D_s[J*(sx,sy).(s11,s21)] = J * rho * u_tt  (normal component from momentum eqn)
           !    (2)   Use extrapolated values to get  J*(sx,sy).(s11,s21)(-1)   ("tangential component" from extrapolation)
           ! To give 2 equations for (s11,s21) on the ghost point:
           !   (J rx) s11(-1) + (J ry) s21(-1) = f1 = s11tilde  (from momentum eqn)
           !   (J sx) s11(-1) + (J sy) s21(-1) = f2 = s21tilde  (from extrapolated values)
           ! Solve:
           !     s11(-1) = sy*f1 - ry*f2 
           !     s21(-1) =-sx*f1 + rx*f2
           !
           ! A similar expression holds for (s12,s22) 
           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              accel1=rho*bcf(side,axis,i1,i2,i3,s11c)   ! rho times given acceleration
              accel2=rho*bcf(side,axis,i1,i2,i3,s12c)

              s11tilde=det(i1+is1,i2,i3)*(rx(i1+is1,i2,i3,0,0)*u(i1+is1,i2,i3,s11c)+rx(i1+is1,i2,i3,0,1)*u(i1+is1,i2,i3,s21c))
     *              -2.*dr(0)*is*(det(i1,i2,i3)*accel1
     *                         -(det(i1,i2+1,i3)*(rx(i1,i2+1,i3,1,0)*u(i1,i2+1,i3,s11c)+rx(i1,i2+1,i3,1,1)*u(i1,i2+1,i3,s21c))
     *                          -det(i1,i2-1,i3)*(rx(i1,i2-1,i3,1,0)*u(i1,i2-1,i3,s11c)+rx(i1,i2-1,i3,1,1)*u(i1,i2-1,i3,s21c)))/(2.*dr(1)))
              s21tilde=det(i1-is1,i2,i3)*(rx(i1-is1,i2,i3,1,0)*u(i1-is1,i2,i3,s11c)+rx(i1-is1,i2,i3,1,1)*u(i1-is1,i2,i3,s21c))
              u(i1-is1,i2,i3,s11c)= rx(i1-is1,i2,i3,1,1)*s11tilde-rx(i1-is1,i2,i3,0,1)*s21tilde
              u(i1-is1,i2,i3,s21c)=-rx(i1-is1,i2,i3,1,0)*s11tilde+rx(i1-is1,i2,i3,0,0)*s21tilde

              s12tilde=det(i1+is1,i2,i3)*(rx(i1+is1,i2,i3,0,0)*u(i1+is1,i2,i3,s12c)+rx(i1+is1,i2,i3,0,1)*u(i1+is1,i2,i3,s22c))
     *              -2.*dr(0)*is*(det(i1,i2,i3)*accel2
     *                         -(det(i1,i2+1,i3)*(rx(i1,i2+1,i3,1,0)*u(i1,i2+1,i3,s12c)+rx(i1,i2+1,i3,1,1)*u(i1,i2+1,i3,s22c))
     *                          -det(i1,i2-1,i3)*(rx(i1,i2-1,i3,1,0)*u(i1,i2-1,i3,s12c)+rx(i1,i2-1,i3,1,1)*u(i1,i2-1,i3,s22c)))/(2.*dr(1)))
              s22tilde=det(i1-is1,i2,i3)*(rx(i1-is1,i2,i3,1,0)*u(i1-is1,i2,i3,s12c)+rx(i1-is1,i2,i3,1,1)*u(i1-is1,i2,i3,s22c))
              u(i1-is1,i2,i3,s12c)= rx(i1-is1,i2,i3,1,1)*s12tilde-rx(i1-is1,i2,i3,0,1)*s22tilde
              u(i1-is1,i2,i3,s22c)=-rx(i1-is1,i2,i3,1,0)*s12tilde+rx(i1-is1,i2,i3,0,0)*s22tilde
             end if
            endLoops2d()
           else ! axis .eq. 1
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then

              accel1=rho*bcf(side,axis,i1,i2,i3,s11c)   ! rho times given acceleration
              accel2=rho*bcf(side,axis,i1,i2,i3,s12c)

              s11tilde=det(i1,i2-is2,i3)*(rx(i1,i2-is2,i3,0,0)*u(i1,i2-is2,i3,s11c)+rx(i1,i2-is2,i3,0,1)*u(i1,i2-is2,i3,s21c))
c*dws              s11tilde=det(i1,i2+is2,i3)*(rx(i1,i2+is2,i3,0,0)*u(i1,i2+is2,i3,s11c)+rx(i1,i2+is2,i3,0,1)*u(i1,i2+is2,i3,s21c))
              s21tilde=det(i1,i2+is2,i3)*(rx(i1,i2+is2,i3,1,0)*u(i1,i2+is2,i3,s11c)+rx(i1,i2+is2,i3,1,1)*u(i1,i2+is2,i3,s21c))
     *              -2.*dr(1)*is*(det(i1,i2,i3)*accel1
     *                         -(det(i1+1,i2,i3)*(rx(i1+1,i2,i3,0,0)*u(i1+1,i2,i3,s11c)+rx(i1+1,i2,i3,0,1)*u(i1+1,i2,i3,s21c))
     *                          -det(i1-1,i2,i3)*(rx(i1-1,i2,i3,0,0)*u(i1-1,i2,i3,s11c)+rx(i1-1,i2,i3,0,1)*u(i1-1,i2,i3,s21c)))/(2.*dr(0)))
              u(i1,i2-is2,i3,s11c)= rx(i1,i2-is2,i3,1,1)*s11tilde-rx(i1,i2-is2,i3,0,1)*s21tilde
              u(i1,i2-is2,i3,s21c)=-rx(i1,i2-is2,i3,1,0)*s11tilde+rx(i1,i2-is2,i3,0,0)*s21tilde
!      write(*,'(" i1,i2,i3,is1,is2,is,s11,s21(0,1)=",3i4,3i3,2e10.2)') i1,i2,i3,is1,is2,is,u(i1,i2+is2,i3,s11c),u(i1,i2+is2,i3,s21c)
!      write(*,'(" det(0,1),s11t,s12t,s11,s21=",5e10.2)') det(i1,i2+is2,i3),s11tilde, s21tilde,u(i1,i2-is2,i3,s11c),u(i1,i2-is2,i3,s21c)
              s12tilde=det(i1,i2-is2,i3)*(rx(i1,i2-is2,i3,0,0)*u(i1,i2-is2,i3,s12c)+rx(i1,i2-is2,i3,0,1)*u(i1,i2-is2,i3,s22c))
c*dws              s12tilde=det(i1,i2+is2,i3)*(rx(i1,i2+is2,i3,0,0)*u(i1,i2+is2,i3,s12c)+rx(i1,i2+is2,i3,0,1)*u(i1,i2+is2,i3,s22c))
              s22tilde=det(i1,i2+is2,i3)*(rx(i1,i2+is2,i3,1,0)*u(i1,i2+is2,i3,s12c)+rx(i1,i2+is2,i3,1,1)*u(i1,i2+is2,i3,s22c))
     *              -2.*dr(1)*is*(det(i1,i2,i3)*accel2
     *                         -(det(i1+1,i2,i3)*(rx(i1+1,i2,i3,0,0)*u(i1+1,i2,i3,s12c)+rx(i1+1,i2,i3,0,1)*u(i1+1,i2,i3,s22c))
     *                          -det(i1-1,i2,i3)*(rx(i1-1,i2,i3,0,0)*u(i1-1,i2,i3,s12c)+rx(i1-1,i2,i3,0,1)*u(i1-1,i2,i3,s22c)))/(2.*dr(0)))
              u(i1,i2-is2,i3,s12c)= rx(i1,i2-is2,i3,1,1)*s12tilde-rx(i1,i2-is2,i3,0,1)*s22tilde
              u(i1,i2-is2,i3,s22c)=-rx(i1,i2-is2,i3,1,0)*s12tilde+rx(i1,i2-is2,i3,0,0)*s22tilde
             end if
            endLoops2d()

           end if ! axis 

          end if ! end gridType


         else if( boundaryCondition(side,axis).eq.tractionBC )then

           ! **************** TRACTION : Neumann type conditions ******************

          if( gridType.eq.rectangular )then

           ! ********* TRACTION : Cartesian Grid **********

           ! Assign displacements on the ghost points from given tractions on the boundary
           !   s11 = lambda ( u.x + v.y ) + 2 mu u.x 
           !   s12 = s21 = mu ( u.y + v.x )
           !   s22 = lambda ( u.x + v.y ) + 2 mu v.y 
           ! an1*s11 + an2*s12 = f1 
           ! an1*s21 + an2*s22 = f2 
           
           ! Assign velocities on the ghost points from given time derivatives of the tractions on the boundary

           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              f1=bcf(side,axis,i1,i2,i3,uc)               ! given traction forces
              f2=bcf(side,axis,i1,i2,i3,vc)
              fdot1=bcf(side,axis,i1,i2,i3,v1c)           ! rate of change of traction forces
              fdot2=bcf(side,axis,i1,i2,i3,v2c)
              u(i1-is1,i2,i3,uc)=u(i1+is1,i2,i3,uc)+2.*dx(0)*(f1+is*lambda*(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.*dx(1)))/(lambda+2.*mu)
              u(i1-is1,i2,i3,vc)=u(i1+is1,i2,i3,vc)+2.*dx(0)*(f2 +is*mu*(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.*dx(1)))/mu
              u(i1-is1,i2,i3,v1c)=u(i1+is1,i2,i3,v1c)+2.*dx(0)*(fdot1+is*lambda*(u(i1,i2+1,i3,v2c)-u(i1,i2-1,i3,v2c))/(2.*dx(1)))/(lambda+2.*mu)
              u(i1-is1,i2,i3,v2c)=u(i1+is1,i2,i3,v2c)+2.*dx(0)*(fdot2 +is*mu*(u(i1,i2+1,i3,v1c)-u(i1,i2-1,i3,v1c))/(2.*dx(1)))/mu
             end if
            endLoops2d()
           else
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              f1=bcf(side,axis,i1,i2,i3,uc)              ! given traction forces
              f2=bcf(side,axis,i1,i2,i3,vc)
              fdot1=bcf(side,axis,i1,i2,i3,v1c)           ! rate of change of traction forces
              fdot2=bcf(side,axis,i1,i2,i3,v2c)
              u(i1,i2-is2,i3,uc)=u(i1,i2+is2,i3,uc)+2.*dx(1)*(f1 +is*mu*(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.*dx(0)))/mu
              u(i1,i2-is2,i3,vc)=u(i1,i2+is2,i3,vc)+2.*dx(1)*(f2+is*lambda*(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.*dx(0)))/(lambda+2.*mu)
              u(i1,i2-is2,i3,v1c)=u(i1,i2+is2,i3,v1c)+2.*dx(1)*(fdot1 +is*mu*(u(i1+1,i2,i3,v2c)-u(i1-1,i2,i3,v2c))/(2.*dx(0)))/mu
              u(i1,i2-is2,i3,v2c)=u(i1,i2+is2,i3,v2c)+2.*dx(1)*(fdot2+is*lambda*(u(i1+1,i2,i3,v1c)-u(i1-1,i2,i3,v1c))/(2.*dx(0)))/(lambda+2.*mu)
             end if
            endLoops2d()
           end if

          else

           ! *********** TRACTION : Curvilinear Grid ****************

           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              f1=-is*bcf(side,axis,i1,i2,i3,uc)              ! given traction forces (adjust here for sign of normal)
              f2=-is*bcf(side,axis,i1,i2,i3,vc)
              fdot1=-is*bcf(side,axis,i1,i2,i3,v1c)          ! rate of change of traction forces
              fdot2=-is*bcf(side,axis,i1,i2,i3,v2c)
              rad=sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2)
              a(0,0,0)=(lambda+2.*mu)*rx(i1,i2,i3,0,0)**2+mu*rx(i1,i2,i3,0,1)**2
              a(0,1,0)=(lambda+mu)*rx(i1,i2,i3,0,0)*rx(i1,i2,i3,0,1)
              a(1,0,0)=a(0,1,0)
              a(1,1,0)=mu*rx(i1,i2,i3,0,0)**2+(lambda+2.*mu)*rx(i1,i2,i3,0,1)**2
              adet=a(0,0,0)*a(1,1,0)-a(0,1,0)*a(1,0,0)
              a(0,0,1)=(lambda+2.*mu)*rx(i1,i2,i3,0,0)*rx(i1,i2,i3,1,0)+mu*rx(i1,i2,i3,0,1)*rx(i1,i2,i3,1,1)
              a(0,1,1)=lambda*rx(i1,i2,i3,0,0)*rx(i1,i2,i3,1,1)+mu*rx(i1,i2,i3,0,1)*rx(i1,i2,i3,1,0)
              a(1,0,1)=lambda*rx(i1,i2,i3,0,1)*rx(i1,i2,i3,1,0)+mu*rx(i1,i2,i3,0,0)*rx(i1,i2,i3,1,1)
              a(1,1,1)=mu*rx(i1,i2,i3,0,0)*rx(i1,i2,i3,1,0)+(lambda+2.*mu)*rx(i1,i2,i3,0,1)*rx(i1,i2,i3,1,1)
              u(i1-is1,i2,i3,uc)=u(i1+is1,i2,i3,uc)-2.*is*dr(0)*(rad*( a(1,1,0)*f1-a(0,1,0)*f2)
     *                       -( a(1,1,0)*a(0,0,1)-a(0,1,0)*a(1,0,1))*(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.*dr(1))
     *                       -( a(1,1,0)*a(0,1,1)-a(0,1,0)*a(1,1,1))*(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.*dr(1)))/adet
              u(i1-is1,i2,i3,vc)=u(i1+is1,i2,i3,vc)-2.*is*dr(0)*(rad*(-a(1,0,0)*f1+a(0,0,0)*f2)
     *                       -(-a(1,0,0)*a(0,0,1)+a(0,0,0)*a(1,0,1))*(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.*dr(1))
     *                       -(-a(1,0,0)*a(0,1,1)+a(0,0,0)*a(1,1,1))*(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.*dr(1)))/adet
              u(i1-is1,i2,i3,v1c)=u(i1+is1,i2,i3,v1c)-2.*is*dr(0)*(rad*( a(1,1,0)*fdot1-a(0,1,0)*fdot2)
     *                       -( a(1,1,0)*a(0,0,1)-a(0,1,0)*a(1,0,1))*(u(i1,i2+1,i3,v1c)-u(i1,i2-1,i3,v1c))/(2.*dr(1))
     *                       -( a(1,1,0)*a(0,1,1)-a(0,1,0)*a(1,1,1))*(u(i1,i2+1,i3,v2c)-u(i1,i2-1,i3,v2c))/(2.*dr(1)))/adet
              u(i1-is1,i2,i3,v2c)=u(i1+is1,i2,i3,v2c)-2.*is*dr(0)*(rad*(-a(1,0,0)*fdot1+a(0,0,0)*fdot2)
     *                       -(-a(1,0,0)*a(0,0,1)+a(0,0,0)*a(1,0,1))*(u(i1,i2+1,i3,v1c)-u(i1,i2-1,i3,v1c))/(2.*dr(1))
     *                       -(-a(1,0,0)*a(0,1,1)+a(0,0,0)*a(1,1,1))*(u(i1,i2+1,i3,v2c)-u(i1,i2-1,i3,v2c))/(2.*dr(1)))/adet
             end if
            endLoops2d()
           else ! axis .eq. 1
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
              f1=-is*bcf(side,axis,i1,i2,i3,uc)              ! given traction forces
              f2=-is*bcf(side,axis,i1,i2,i3,vc)
              fdot1=-is*bcf(side,axis,i1,i2,i3,v1c)           ! rate of change of traction forces
              fdot2=-is*bcf(side,axis,i1,i2,i3,v2c)
              rad=sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2)
              a(0,0,0)=(lambda+2.*mu)*rx(i1,i2,i3,1,0)**2+mu*rx(i1,i2,i3,1,1)**2
              a(0,1,0)=(lambda+mu)*rx(i1,i2,i3,1,0)*rx(i1,i2,i3,1,1)
              a(1,0,0)=a(0,1,0)
              a(1,1,0)=mu*rx(i1,i2,i3,1,0)**2+(lambda+2.*mu)*rx(i1,i2,i3,1,1)**2
              adet=a(0,0,0)*a(1,1,0)-a(0,1,0)*a(1,0,0)
              a(0,0,1)=(lambda+2.*mu)*rx(i1,i2,i3,1,0)*rx(i1,i2,i3,0,0)+mu*rx(i1,i2,i3,1,1)*rx(i1,i2,i3,0,1)
              a(0,1,1)=lambda*rx(i1,i2,i3,1,0)*rx(i1,i2,i3,0,1)+mu*rx(i1,i2,i3,1,1)*rx(i1,i2,i3,0,0)
              a(1,0,1)=lambda*rx(i1,i2,i3,1,1)*rx(i1,i2,i3,0,0)+mu*rx(i1,i2,i3,1,0)*rx(i1,i2,i3,0,1)
              a(1,1,1)=mu*rx(i1,i2,i3,1,0)*rx(i1,i2,i3,0,0)+(lambda+2.*mu)*rx(i1,i2,i3,1,1)*rx(i1,i2,i3,0,1)
              u(i1,i2-is2,i3,uc)=u(i1,i2+is2,i3,uc)-2.*is*dr(1)*(rad*( a(1,1,0)*f1-a(0,1,0)*f2)
     *                       -( a(1,1,0)*a(0,0,1)-a(0,1,0)*a(1,0,1))*(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.*dr(0))
     *                       -( a(1,1,0)*a(0,1,1)-a(0,1,0)*a(1,1,1))*(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.*dr(0)))/adet
              u(i1,i2-is2,i3,vc)=u(i1,i2+is2,i3,vc)-2.*is*dr(1)*(rad*(-a(1,0,0)*f1+a(0,0,0)*f2)
     *                       -(-a(1,0,0)*a(0,0,1)+a(0,0,0)*a(1,0,1))*(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.*dr(0))
     *                       -(-a(1,0,0)*a(0,1,1)+a(0,0,0)*a(1,1,1))*(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.*dr(0)))/adet
              u(i1,i2-is2,i3,v1c)=u(i1,i2+is2,i3,v1c)-2.*is*dr(1)*(rad*( a(1,1,0)*fdot1-a(0,1,0)*fdot2)
     *                       -( a(1,1,0)*a(0,0,1)-a(0,1,0)*a(1,0,1))*(u(i1+1,i2,i3,v1c)-u(i1-1,i2,i3,v1c))/(2.*dr(0))
     *                       -( a(1,1,0)*a(0,1,1)-a(0,1,0)*a(1,1,1))*(u(i1+1,i2,i3,v2c)-u(i1-1,i2,i3,v2c))/(2.*dr(0)))/adet
              u(i1,i2-is2,i3,v2c)=u(i1,i2+is2,i3,v2c)-2.*is*dr(1)*(rad*(-a(1,0,0)*fdot1+a(0,0,0)*fdot2)
     *                       -(-a(1,0,0)*a(0,0,1)+a(0,0,0)*a(1,0,1))*(u(i1+1,i2,i3,v1c)-u(i1-1,i2,i3,v1c))/(2.*dr(0))
     *                       -(-a(1,0,0)*a(0,1,1)+a(0,0,0)*a(1,1,1))*(u(i1+1,i2,i3,v2c)-u(i1-1,i2,i3,v2c))/(2.*dr(0)))/adet
             end if
            endLoops2d()

           end if ! axis 

          end if ! end gridType

         else if( boundaryCondition(side,axis).eq.slipWall )then

           ! **************** SLIPWALL : Neumann type conditions ******************

          if( gridType.eq.rectangular )then

           ! ********* SLIPWALL : Cartesian Grid **********

           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
               u(i1-is1,i2,i3,vc )=u(i1+is1,i2,i3,vc )-2.*is*dx(0)*( bcf(side,axis,i1,i2,i3,s21c)/mu-(u(i1,i2+1,i3,uc )-u(i1,i2-1,i3,uc ))/(2.*dx(1)))
               u(i1-is1,i2,i3,v2c)=u(i1+is1,i2,i3,v2c)-2.*is*dx(0)*( bcf(side,axis,i1,i2,i3,s22c)/mu-(u(i1,i2+1,i3,v1c)-u(i1,i2-1,i3,v1c))/(2.*dx(1)))
             end if
            endLoops2d()
           else
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
               u(i1,i2-is2,i3,uc )=u(i1,i2+is2,i3,uc )-2.*is*dx(1)*(-bcf(side,axis,i1,i2,i3,s21c)/mu-(u(i1+1,i2,i3,vc )-u(i1-1,i2,i3,vc ))/(2.*dx(0)))
               u(i1,i2-is2,i3,v1c)=u(i1,i2+is2,i3,v1c)-2.*is*dx(1)*(-bcf(side,axis,i1,i2,i3,s22c)/mu-(u(i1+1,i2,i3,v2c)-u(i1-1,i2,i3,v2c))/(2.*dx(0)))
             end if
            endLoops2d()
           end if

          else

           ! *********** SLIPWALL : Curvilinear Grid ****************

c              ! (an1,an2) = outward normal and (-an2,an1) = unit tangent
c              aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
c              an1=-is*rx(i1,i2,i3,axis,0)*aNormi
c              an2=-is*rx(i1,i2,i3,axis,1)*aNormi


           if( axis.eq.0 )then
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
               b1=2.*rx(i1,i2,i3,axis,0)*rx(i1,i2,i3,axis,1)
               b2=rx(i1,i2,i3,axis,0)**2-rx(i1,i2,i3,axis,1)**2
               b3=1./(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2)

               f1=bcf(side,axis,i1,i2,i3,s21c)/mu-b3*((-b1*rx(i1,i2,i3,1,0)+b2*rx(i1,i2,i3,1,1))*(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))
     *                                               +( b1*rx(i1,i2,i3,1,1)+b2*rx(i1,i2,i3,1,0))*(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc)))/(2.*dr(1))
               f1=-rx(i1,i2,i3,axis,1)*u(i1+is1,i2,i3,uc)+rx(i1,i2,i3,axis,0)*u(i1+is1,i2,i3,vc)-2.*is*dr(0)*f1
               f2= rx(i1,i2,i3,axis,0)*u(i1-is1,i2,i3,uc)+rx(i1,i2,i3,axis,1)*u(i1-is1,i2,i3,vc)
               u(i1-is1,i2,i3,uc)=b3*(-rx(i1,i2,i3,axis,1)*f1+rx(i1,i2,i3,axis,0)*f2)
               u(i1-is1,i2,i3,vc)=b3*( rx(i1,i2,i3,axis,0)*f1+rx(i1,i2,i3,axis,1)*f2)

               f1=bcf(side,axis,i1,i2,i3,s22c)/mu-b3*((-b1*rx(i1,i2,i3,1,0)+b2*rx(i1,i2,i3,1,1))*(u(i1,i2+1,i3,v1c)-u(i1,i2-1,i3,v1c))
     *                                               +( b1*rx(i1,i2,i3,1,1)+b2*rx(i1,i2,i3,1,0))*(u(i1,i2+1,i3,v2c)-u(i1,i2-1,i3,v2c)))/(2.*dr(1))
               f1=-rx(i1,i2,i3,axis,1)*u(i1+is1,i2,i3,v1c)+rx(i1,i2,i3,axis,0)*u(i1+is1,i2,i3,v2c)-2.*is*dr(0)*f1
               f2= rx(i1,i2,i3,axis,0)*u(i1-is1,i2,i3,v1c)+rx(i1,i2,i3,axis,1)*u(i1-is1,i2,i3,v2c)
               u(i1-is1,i2,i3,v1c)=b3*(-rx(i1,i2,i3,axis,1)*f1+rx(i1,i2,i3,axis,0)*f2)
               u(i1-is1,i2,i3,v2c)=b3*( rx(i1,i2,i3,axis,0)*f1+rx(i1,i2,i3,axis,1)*f2)

             end if
            endLoops2d()
           else ! axis .eq. 1
            beginLoops2d()
             if (mask(i1,i2,i3).ne.0) then
               b1=2.*rx(i1,i2,i3,axis,0)*rx(i1,i2,i3,axis,1)
               b2=rx(i1,i2,i3,axis,0)**2-rx(i1,i2,i3,axis,1)**2
               b3=1./(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2)

               f1=bcf(side,axis,i1,i2,i3,s21c)/mu-b3*((-b1*rx(i1,i2,i3,0,0)+b2*rx(i1,i2,i3,0,1))*(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))
     *                                               +( b1*rx(i1,i2,i3,0,1)+b2*rx(i1,i2,i3,0,0))*(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc)))/(2.*dr(0))
               f1=-rx(i1,i2,i3,axis,1)*u(i1,i2+is2,i3,uc)+rx(i1,i2,i3,axis,0)*u(i1,i2+is2,i3,vc)-2.*is*dr(1)*f1
               f2= rx(i1,i2,i3,axis,0)*u(i1,i2-is2,i3,uc)+rx(i1,i2,i3,axis,1)*u(i1,i2-is2,i3,vc)
               u(i1,i2-is2,i3,uc)=b3*(-rx(i1,i2,i3,axis,1)*f1+rx(i1,i2,i3,axis,0)*f2)
               u(i1,i2-is2,i3,vc)=b3*( rx(i1,i2,i3,axis,0)*f1+rx(i1,i2,i3,axis,1)*f2)

               f1=bcf(side,axis,i1,i2,i3,s22c)/mu-b3*((-b1*rx(i1,i2,i3,0,0)+b2*rx(i1,i2,i3,0,1))*(u(i1+1,i2,i3,v1c)-u(i1-1,i2,i3,v1c))
     *                                               +( b1*rx(i1,i2,i3,0,1)+b2*rx(i1,i2,i3,0,0))*(u(i1+1,i2,i3,v2c)-u(i1-1,i2,i3,v2c)))/(2.*dr(0))
               f1=-rx(i1,i2,i3,axis,1)*u(i1,i2+is2,i3,v1c)+rx(i1,i2,i3,axis,0)*u(i1,i2+is2,i3,v2c)-2.*is*dr(1)*f1
               f2= rx(i1,i2,i3,axis,0)*u(i1,i2-is2,i3,v1c)+rx(i1,i2,i3,axis,1)*u(i1,i2-is2,i3,v2c)
               u(i1,i2-is2,i3,v1c)=b3*(-rx(i1,i2,i3,axis,1)*f1+rx(i1,i2,i3,axis,0)*f2)
               u(i1,i2-is2,i3,v2c)=b3*( rx(i1,i2,i3,axis,0)*f1+rx(i1,i2,i3,axis,1)*f2)

             end if
            endLoops2d()

           end if ! axis 

          end if ! end gridType

         end if ! bc 
        endLoopOverSides()

c              n1a=gridIndexRange(0,0)
c              n1b=gridIndexRange(1,0)
c              n2a=gridIndexRange(0,1)
c              n2b=gridIndexRange(1,1)
c              i3=gridIndexRange(0,2)
c              do i1=n1a-1,n1b+1
c              do i2=n2a-1,n2b+1
c                write(33,"(2(1x,i3),6(1x,1pe9.2))")i1,i2,u(i1,i2,i3,uc),u(i1,i2,i3,vc),u(i1,i2,i3,s11c),u(i1,i2,i3,s12c),u(i1,i2,i3,s21c),u(i1,i2,i3,s22c)
c              end do
c              end do
c              pause


c*******
c******* Secondary Dirichlet conditions for the tangential components of stress (tractionBC only) ********
c*******

        beginLoopOverSides(numGhost,numGhost)
          if( boundaryCondition(side,axis).eq.tractionBC )then
            if( gridType.eq.rectangular )then

              if( axis.eq.0 )then
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    u1x=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dx(0))
                    u1y=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dx(1))
                    u2x=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dx(0))
                    u2y=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dx(1))
                    u(i1,i2,i3,s21c)=mu*(u1y+u2x)
                    u(i1,i2,i3,s22c)=lambda*(u1x+u2y)+2.0*mu*u2y
                  end if
                endLoops2d()
              else
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    u1x=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dx(0))
                    u1y=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dx(1))
                    u2x=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dx(0))
                    u2y=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dx(1))
                    u(i1,i2,i3,s11c)=lambda*(u1x+u2y)+2.0*mu*u1x
                    u(i1,i2,i3,s12c)=mu*(u1y+u2x)
                  end if
                endLoops2d()
              end if

            else  ! curvilinear 

              beginLoops2d()
                if (mask(i1,i2,i3).ne.0) then
                  u1r=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dr(0))
                  u1s=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dr(1))
                  u2r=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dr(0))
                  u2s=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dr(1))
                  u1x=u1r*rx(i1,i2,i3,0,0)+u1s*rx(i1,i2,i3,1,0)
                  u1y=u1r*rx(i1,i2,i3,0,1)+u1s*rx(i1,i2,i3,1,1)
                  u2x=u2r*rx(i1,i2,i3,0,0)+u2s*rx(i1,i2,i3,1,0)
                  u2y=u2r*rx(i1,i2,i3,0,1)+u2s*rx(i1,i2,i3,1,1)

                  s11t=lambda*(u1x+u2y)+2.0*mu*u1x
                  s12t=mu*(u1y+u2x)
                  s21t=s12t
                  s22t=lambda*(u1x+u2y)+2.0*mu*u2y
                  ! (an1,an2) = outward normal => (an2,-an1) = tangent vector
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
                  an1=-is*rx(i1,i2,i3,axis,0)*aNormi
                  an2=-is*rx(i1,i2,i3,axis,1)*aNormi
                  b1=an2*(s11t-u(i1,i2,i3,s11c))-an1*(s21t-u(i1,i2,i3,s21c))
                  b2=an2*(s12t-u(i1,i2,i3,s12c))-an1*(s22t-u(i1,i2,i3,s22c))

                  u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+an2*b1
                  u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+an2*b2

                  u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)-an1*b1
                  u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)-an1*b2

                end if
              endLoops2d()

            end if  ! end gridType

          end if ! bc 
        endLoopOverSides()

c set tangential components of stress on the boundary  (TZ forcing, if necessary)
        if (twilightZone.ne.0) then
          beginLoopOverSides(numGhost,numGhost)
            if( boundaryCondition(side,axis).eq.tractionBC )then
              if( gridType.eq.rectangular )then

              if( axis.eq.0 )then
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xe)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1ye)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)
                    u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+s21e-mu*(u1ye+u2xe)
                    u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+s22e-lambda*(u1xe+u2ye)-2.0*mu*u2ye
                  end if
                endLoops2d()
              else
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xe)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1ye)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                    u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+s11e-lambda*(u1xe+u2ye)-2.0*mu*u1xe
                    u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+s12e-mu*(u1ye+u2xe)
                  end if
                endLoops2d()
              end if

              else  ! curvilinear 

                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xe)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1ye)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)

                    ! (an1,an2) = outward normal => (an2,-an1) = tangent vector
                    aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
                    an1=-is*rx(i1,i2,i3,axis,0)*aNormi
                    an2=-is*rx(i1,i2,i3,axis,1)*aNormi
                    b1=an2*(s11e-lambda*(u1xe+u2ye)-2.0*mu*u1xe)-an1*(s21e-mu*(u1ye+u2xe))
                    b2=an2*(s12e-mu*(u1ye+u2xe))-an1*(s22e-lambda*(u1xe+u2ye)-2.0*mu*u2ye)

                    u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+an2*b1
                    u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+an2*b2

                    u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)-an1*b1
                    u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)-an1*b2

                  end if
                endLoops2d()

              end if  ! end gridType

            end if ! bc 
          endLoopOverSides()
        end if


c*******
c******* Re-compute all components of stress at tractionBC-tractionBC corners ********
c*******

        axis1=0
        axis2=1
        i3=gridIndexRange(0,2)
        if (gridType.eq.rectangular) then

c Cartesian grid (not sure if this case is necessary to do)
          do side1=0,1
            i1=gridIndexRange(side1,axis1)
            do side2=0,1
              i2=gridIndexRange(side2,axis2)
              if (mask(i1,i2,i3).ne.0) then
                if (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

                  is=1-2*side1
                  u(i1,i2,i3,s11c) =-is*bcf(side1,axis1,i1,i2,i3,s11c) 
                  u(i1,i2,i3,s12c) =-is*bcf(side1,axis1,i1,i2,i3,s12c)

                  is=1-2*side2
                  u(i1,i2,i3,s21c) =-is*bcf(side2,axis2,i1,i2,i3,s11c) 
                  u(i1,i2,i3,s22c) =-is*bcf(side2,axis2,i1,i2,i3,s12c)

                end if
              end if
            end do
          end do

        else

c non-Cartesian grid
          do side1=0,1
            i1=gridIndexRange(side1,axis1)
            do side2=0,1
              i2=gridIndexRange(side2,axis2)
              if (mask(i1,i2,i3).ne.0) then
                if (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then

                  is=1-2*side1
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis1,0)**2+rx(i1,i2,i3,axis1,1)**2))
                  an11=-is*rx(i1,i2,i3,axis1,0)*aNormi            ! normals for axis1,side1
                  an21=-is*rx(i1,i2,i3,axis1,1)*aNormi
                  f11=bcf(side1,axis1,i1,i2,i3,s11c)              ! given traction forces for axis1,side1
                  f21=bcf(side1,axis1,i1,i2,i3,s12c)

                  is=1-2*side2
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis2,0)**2+rx(i1,i2,i3,axis2,1)**2))
                  an12=-is*rx(i1,i2,i3,axis2,0)*aNormi            ! normals for axis2,side2
                  an22=-is*rx(i1,i2,i3,axis2,1)*aNormi
                  f12=bcf(side2,axis2,i1,i2,i3,s11c)              ! given traction forces for axis2,side2
                  f22=bcf(side2,axis2,i1,i2,i3,s12c)

                  b11=f11-(an11*u(i1,i2,i3,s11c)+an21*u(i1,i2,i3,s21c))
                  b21=f21-(an11*u(i1,i2,i3,s12c)+an21*u(i1,i2,i3,s22c))

                  dot1=an11*an12+an21*an22                        ! cosine of the angle between the normals
                  dot2=an21*an12-an11*an22                        ! cosine of the angle between tangent(1) and normal(2)
                  b12=(f12-(an12*u(i1,i2,i3,s11c)+an22*u(i1,i2,i3,s21c))-dot1*b11)/dot2
                  b22=(f22-(an12*u(i1,i2,i3,s12c)+an22*u(i1,i2,i3,s22c))-dot1*b21)/dot2

                  u(i1,i2,i3,s11c)=u(i1,i2,i3,s11c)+an11*b11+an21*b12
                  u(i1,i2,i3,s12c)=u(i1,i2,i3,s12c)+an11*b21+an21*b22

                  u(i1,i2,i3,s21c)=u(i1,i2,i3,s21c)+an21*b11-an11*b12
                  u(i1,i2,i3,s22c)=u(i1,i2,i3,s22c)+an21*b21-an11*b22

c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)
c                  s11e=u(i1,i2,i3,s11c)-s11e
c                  s12e=u(i1,i2,i3,s12c)-s12e
c                  s21e=u(i1,i2,i3,s21c)-s21e
c                  s22e=u(i1,i2,i3,s22c)-s22e

c           write(6,"(2(1x,i2),4(1x,1pe10.3))")side1,side2,s11e,s12e,s21e,s22e

                end if
              end if
            end do
          end do
        end if


c*******
c******* Secondary Dirichlet conditions for stress (slipWall only) ********
c*******

c TZ forcing (if necessary)
        if (twilightZone.ne.0) then
          beginLoopOverSides(numGhost,numGhost)
            if( boundaryCondition(side,axis).eq.slipWall )then
              if( gridType.eq.rectangular )then

                if( axis.eq.0 )then
                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                      call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)
                      bcfa(side,axis,i1,i2,i3,s11c)=s11e-lambda*(u1xe+u2ye)-2.0*mu*u1xe
                      bcfa(side,axis,i1,i2,i3,s21c)=s21e-s12e
                      bcfa(side,axis,i1,i2,i3,s22c)=s22e-lambda*(u1xe+u2ye)-2.0*mu*u2ye
                    end if
                  endLoops2d()
                else
                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                      call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                      call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)
                      bcfa(side,axis,i1,i2,i3,s11c)=s11e-lambda*(u1xe+u2ye)-2.0*mu*u1xe
                      bcfa(side,axis,i1,i2,i3,s12c)=s12e-s21e
                      bcfa(side,axis,i1,i2,i3,s22c)=s22e-lambda*(u1xe+u2ye)-2.0*mu*u2ye
                    end if
                  endLoops2d()
                end if

              else  ! curvilinear 

                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xe)
                    call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xe)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1ye)
                    call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2ye)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21e)
                    call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22e)

                    s11t=lambda*(u1xe+u2ye)+2.0*mu*u1xe
                    s12t=mu*(u1ye+u2xe)
                    s21t=s12t
                    s22t=lambda*(u1xe+u2ye)+2.0*mu*u2ye
                    ! (an1,an2) = outward normal and (-an2,an1) = unit tangent
                    aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
                    an1=-is*rx(i1,i2,i3,axis,0)*aNormi
                    an2=-is*rx(i1,i2,i3,axis,1)*aNormi
                    b1=an1*(-s11t*an2+s12t*an1)+an2*(-s21t*an2+s22t*an1)

                    bcfa(side,axis,i1,i2,i3,vc)=b1

                    bcfa(side,axis,i1,i2,i3,s11c)=s11e-s11t
                    bcfa(side,axis,i1,i2,i3,s12c)=s12e-s12t
                    bcfa(side,axis,i1,i2,i3,s21c)=s21e-s21t
                    bcfa(side,axis,i1,i2,i3,s22c)=s22e-s22t
                  end if
                endLoops2d()

              end if  ! end gridType

            end if ! bc 
          endLoopOverSides()
        else
          beginLoopOverSides(numGhost,numGhost)
            if( boundaryCondition(side,axis).eq.slipWall )then
              beginLoops2d()
                if (mask(i1,i2,i3).ne.0) then
                  ! the true bc for stress is moved for convenience applying TZ flow
                  bcfa(side,axis,i1,i2,i3,vc)=bcfa(side,axis,i1,i2,i3,s11c)  
                  bcfa(side,axis,i1,i2,i3,s11c)=0.
                  bcfa(side,axis,i1,i2,i3,s12c)=0.
                  bcfa(side,axis,i1,i2,i3,s21c)=0.
                  bcfa(side,axis,i1,i2,i3,s22c)=0.
                end if
              endLoops2d()
            end if ! bc 
          endLoopOverSides()
        end if

        beginLoopOverSides(numGhost,numGhost)
          if( boundaryCondition(side,axis).eq.slipWall )then
            if( gridType.eq.rectangular )then

              if( axis.eq.0 )then
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    u1x=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dx(0))
                    u2y=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dx(1))
                    u(i1,i2,i3,s11c)=bcfa(side,axis,i1,i2,i3,s11c)+lambda*(u1x+u2y)+2.0*mu*u1x
                    u(i1,i2,i3,s21c)=bcfa(side,axis,i1,i2,i3,s21c)+u(i1,i2,i3,s12c)
                    u(i1,i2,i3,s22c)=bcfa(side,axis,i1,i2,i3,s22c)+lambda*(u1x+u2y)+2.0*mu*u2y
                  end if
                endLoops2d()
              else
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    u1x=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dx(0))
                    u2y=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dx(1))
                    u(i1,i2,i3,s11c)=bcfa(side,axis,i1,i2,i3,s11c)+lambda*(u1x+u2y)+2.0*mu*u1x
                    u(i1,i2,i3,s12c)=bcfa(side,axis,i1,i2,i3,s12c)+u(i1,i2,i3,s21c)
                    u(i1,i2,i3,s22c)=bcfa(side,axis,i1,i2,i3,s22c)+lambda*(u1x+u2y)+2.0*mu*u2y
                  end if
                endLoops2d()
              end if

            else  ! curvilinear 

              beginLoops2d()
                if (mask(i1,i2,i3).ne.0) then
                  u1r=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dr(0))
                  u1s=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dr(1))
                  u2r=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dr(0))
                  u2s=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dr(1))
                  u1x=u1r*rx(i1,i2,i3,0,0)+u1s*rx(i1,i2,i3,1,0)
                  u1y=u1r*rx(i1,i2,i3,0,1)+u1s*rx(i1,i2,i3,1,1)
                  u2x=u2r*rx(i1,i2,i3,0,0)+u2s*rx(i1,i2,i3,1,0)
                  u2y=u2r*rx(i1,i2,i3,0,1)+u2s*rx(i1,i2,i3,1,1)

                  s11t=lambda*(u1x+u2y)+2.0*mu*u1x
                  s12t=mu*(u1y+u2x)
                  s21t=s12t
                  s22t=lambda*(u1x+u2y)+2.0*mu*u2y
                  ! (an1,an2) = outward normal and (-an2,an1) = unit tangent
                  aNormi=1./max(epsx,sqrt(rx(i1,i2,i3,axis,0)**2+rx(i1,i2,i3,axis,1)**2))
                  an1=-is*rx(i1,i2,i3,axis,0)*aNormi
                  an2=-is*rx(i1,i2,i3,axis,1)*aNormi
                  b1=bcf(side,axis,i1,i2,i3,vc)-an1*(-s11t*an2+s12t*an1)-an2*(-s21t*an2+s22t*an1)

                  u(i1,i2,i3,s11c)=bcfa(side,axis,i1,i2,i3,s11c)+s11t-an1*b1*an2
                  u(i1,i2,i3,s12c)=bcfa(side,axis,i1,i2,i3,s12c)+s12t+an1*b1*an1
                  u(i1,i2,i3,s21c)=bcfa(side,axis,i1,i2,i3,s21c)+s21t-an2*b1*an2
                  u(i1,i2,i3,s22c)=bcfa(side,axis,i1,i2,i3,s22c)+s22t+an2*b1*an1
                end if
              endLoops2d()

            end if  ! end gridType

          end if ! bc 
        endLoopOverSides()


c*******
c******* re-extrapolation components of stress to first ghost line ********
c*******

        beginLoopOverSides(numGhost,numGhost)
          if( boundaryCondition(side,axis).eq.tractionBC.or.boundaryCondition(side,axis).eq.slipWall )then
            beginLoops2d()
              if (mask(i1,i2,i3).ne.0) then
                u(i1-is1,i2-is2,i3,s11c)=extrap3(u,i1,i2,i3,s11c,is1,is2,is3)
                u(i1-is1,i2-is2,i3,s12c)=extrap3(u,i1,i2,i3,s12c,is1,is2,is3)
                u(i1-is1,i2-is2,i3,s21c)=extrap3(u,i1,i2,i3,s21c,is1,is2,is3)
                u(i1-is1,i2-is2,i3,s22c)=extrap3(u,i1,i2,i3,s22c,is1,is2,is3)
              end if
            endLoops2d()

          end if ! bc 
        endLoopOverSides()


c*******
c******* re-extrapolation solution to first ghost line near corners ********
c*******

        i3=gridIndexRange(0,2)
        do side1=0,1
          i1=gridIndexRange(side1,axis1)
          is1=1-2*side1
          do side2=0,1
            i2=gridIndexRange(side2,axis2)
            is2=1-2*side2

c extrapolate in the i1 direction
c*wdh       if (boundaryCondition(side1,axis1).eq.displacementBC) then
            if (boundaryCondition(side1,axis1).eq.displacementBC.and.boundaryCondition(side2,axis2).gt.0) then
              if (mask(i1,i2,i3).ne.0) then
                do n=0,numberOfComponents-1
                  u(i1-is1,i2,i3,n)=extrap3(u,i1,i2,i3,n,is1,0,0)
                end do
              end if
            end if

c extrapolate in the i2 direction
c*wdh       if (boundaryCondition(side2,axis2).eq.displacementBC) then
            if (boundaryCondition(side2,axis2).eq.displacementBC.and.boundaryCondition(side1,axis1).gt.0) then
              if (mask(i1,i2,i3).ne.0) then
                do n=0,numberOfComponents-1
                  u(i1,i2-is2,i3,n)=extrap3(u,i1,i2,i3,n,0,is2,0)
                end do
              end if
            end if

c extrapolate in the diagonal direction
            if (boundaryCondition(side1,axis1).gt.0.and.boundaryCondition(side2,axis2).gt.0) then
              if (mask(i1,i2,i3).ne.0) then
                do n=0,numberOfComponents-1
                  u(i1-is1,i2-is2,i3,n)=extrap3(u,i1,i2,i3,n,is1,is2,0)
                end do
              end if
            end if
          end do
        end do


c*******
c******* Extrapolation of stress to the first ghost line (for the tractionBC case only) ********
c*******

c******* currently this extrapolation does not work for the case when a tractionBC side ********
c        includes tractionBC-tractionBC corner and the grid is not Cartesian

       if( .true. ) then ! -----------
c         write(*,'(" --bcFOS: set stress on the ghost pt --")') 

c fix up periodic boundaries (if necessary)
        if (boundaryCondition(0,0).eq.tractionBC.or.boundaryCondition(1,0).eq.tractionBC) then
          if (boundaryCondition(0,1).lt.0) then
            if (boundaryCondition(1,1).lt.0) then
              n1a=gridIndexRange(0,0)
              n1b=gridIndexRange(1,0)
              n2a=gridIndexRange(0,1)
              n2b=gridIndexRange(1,1)
              do i1=n1a-1,n1b+1
                u(i1,n2a-1,0,uc)=u(i1,n2b-1,0,uc)
                u(i1,n2a-1,0,vc)=u(i1,n2b-1,0,vc)
                u(i1,n2b+1,0,uc)=u(i1,n2a+1,0,uc)
                u(i1,n2b+1,0,vc)=u(i1,n2a+1,0,vc)
              end do
            else
              write(*,'("periodic bcs must be on opposite sides???")')
              stop
            end if
          end if
        elseif (boundaryCondition(0,1).eq.tractionBC.or.boundaryCondition(1,1).eq.tractionBC) then
          if (boundaryCondition(0,0).lt.0) then
            if (boundaryCondition(1,0).lt.0) then
              n1a=gridIndexRange(0,0)
              n1b=gridIndexRange(1,0)
              n2a=gridIndexRange(0,1)
              n2b=gridIndexRange(1,1)
              do i2=n2a-1,n2b+1
                u(n1a-1,i2,0,uc)=u(n1b-1,i2,0,uc)
                u(n1a-1,i2,0,vc)=u(n1b-1,i2,0,vc)
                u(n1b+1,i2,0,uc)=u(n1a+1,i2,0,uc)
                u(n1b+1,i2,0,vc)=u(n1a+1,i2,0,vc)
              end do
            else
              write(*,'("periodic bcs must be on opposite sides???")')
              stop
            end if
          end if
        end if

c        if( boundaryCondition(0,0).lt.0 )then
c         ! perioidic fix -- finish me ---
c         n1a=gridIndexRange(0,0)
c         n1b=gridIndexRange(1,0)
c         n2a=gridIndexRange(0,1)
c         n2b=gridIndexRange(1,1)

c         do i2=n2a-1,n2b+1
c          u(n1a-1,i2,0,uc)=u(n1b-1,i2,0,uc)
c          u(n1a-1,i2,0,vc)=u(n1b-1,i2,0,vc)

c          u(n1b+1,i2,0,uc)=u(n1a+1,i2,0,uc)
c          u(n1b+1,i2,0,vc)=u(n1a+1,i2,0,vc)
c         end do
c        end if


        ! *** determine the stress on the ghost points from the "normal-derivative" of the stress: :
        !               s_ij = lambda( u.x + v.y ) + 2*mu u.x
        ! 
        beginLoopOverSides(numGhost,numGhost)
          if( boundaryCondition(side,axis).eq.tractionBC )then
            if( gridType.eq.rectangular )then

              if( axis.eq.0 )then
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    u1xx=(u(i1+1,i2,i3,uc)-2.*u(i1,i2,i3,uc)+u(i1-1,i2,i3,uc))/dx(0)**2
                    u1xy=(u(i1+1,i2+1,i3,uc)-u(i1-1,i2+1,i3,uc)-u(i1+1,i2-1,i3,uc)+u(i1-1,i2-1,i3,uc))/(4.*dx(0)*dx(1))
                    u2xx=(u(i1+1,i2,i3,vc)-2.*u(i1,i2,i3,vc)+u(i1-1,i2,i3,vc))/dx(0)**2
                    u2xy=(u(i1+1,i2+1,i3,vc)-u(i1-1,i2+1,i3,vc)-u(i1+1,i2-1,i3,vc)+u(i1-1,i2-1,i3,vc))/(4.*dx(0)*dx(1))
                    u(i1-is1,i2,i3,s11c)=u(i1+is1,i2,i3,s11c)-2.*is*dx(0)*((lambda+2.*mu)*u1xx+lambda*u2xy)
                    u(i1-is1,i2,i3,s12c)=u(i1+is1,i2,i3,s12c)-2.*is*dx(0)*mu*(u1xy+u2xx)
c                    u(i1-is1,i2,i3,s21c)=u(i1-is1,i2,i3,s12c)  
                    u(i1-is1,i2,i3,s21c)=u(i1+is1,i2,i3,s21c)-2.*is*dx(0)*mu*(u1xy+u2xx)
                    u(i1-is1,i2,i3,s22c)=u(i1+is1,i2,i3,s22c)-2.*is*dx(0)*(lambda*u1xx+(lambda+2.*mu)*u2xy)
                  end if
                endLoops2d()
              else ! axis .eq. 1
                beginLoops2d()
                  if (mask(i1,i2,i3).ne.0) then
                    u1xy=(u(i1+1,i2+1,i3,uc)-u(i1-1,i2+1,i3,uc)-u(i1+1,i2-1,i3,uc)+u(i1-1,i2-1,i3,uc))/(4.*dx(0)*dx(1))
                    u1yy=(u(i1,i2+1,i3,uc)-2.*u(i1,i2,i3,uc)+u(i1,i2-1,i3,uc))/dx(1)**2
                    u2xy=(u(i1+1,i2+1,i3,vc)-u(i1-1,i2+1,i3,vc)-u(i1+1,i2-1,i3,vc)+u(i1-1,i2-1,i3,vc))/(4.*dx(0)*dx(1))
                    u2yy=(u(i1,i2+1,i3,vc)-2.*u(i1,i2,i3,vc)+u(i1,i2-1,i3,vc))/dx(1)**2
                    u(i1,i2-is2,i3,s11c)=u(i1,i2+is2,i3,s11c)-2.*is*dx(1)*((lambda+2.*mu)*u1xy+lambda*u2yy)
                    u(i1,i2-is2,i3,s12c)=u(i1,i2+is2,i3,s12c)-2.*is*dx(1)*mu*(u1yy+u2xy)
c                    u(i1,i2-is2,i3,s21c)=u(i1,i2-is2,i3,s12c)
                    u(i1,i2-is2,i3,s21c)=u(i1,i2+is2,i3,s21c)-2.*is*dx(1)*mu*(u1yy+u2xy)
                    u(i1,i2-is2,i3,s22c)=u(i1,i2+is2,i3,s22c)-2.*is*dx(1)*(lambda*u1xy+(lambda+2.*mu)*u2yy)
                  end if
                endLoops2d()
              end if ! axis 

            else  ! curvilinear

         if (.false.) then    ! temporary : controls which discretization is used (option 1)

              if( axis.eq.0 )then
                if( boundaryCondition(0,1).ne.tractionBC.and.boundaryCondition(1,1).ne.tractionBC) then
                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      u1xp= (rx(i1+1,i2,i3,0,0)+rx(i1,i2,i3,0,0))*(u(i1+1,i2,i3,uc)-u(i1,i2,i3,uc))/(2.*dr(0))
     *                     +(rx(i1+1,i2,i3,1,0)+rx(i1,i2,i3,1,0))*(u(i1+1,i2+1,i3,uc)+u(i1,i2+1,i3,uc)-u(i1+1,i2-1,i3,uc)-u(i1,i2-1,i3,uc))/(8.*dr(1))
                      u2xp= (rx(i1+1,i2,i3,0,0)+rx(i1,i2,i3,0,0))*(u(i1+1,i2,i3,vc)-u(i1,i2,i3,vc))/(2.*dr(0))
     *                     +(rx(i1+1,i2,i3,1,0)+rx(i1,i2,i3,1,0))*(u(i1+1,i2+1,i3,vc)+u(i1,i2+1,i3,vc)-u(i1+1,i2-1,i3,vc)-u(i1,i2-1,i3,vc))/(8.*dr(1))
                      u1yp= (rx(i1+1,i2,i3,0,1)+rx(i1,i2,i3,0,1))*(u(i1+1,i2,i3,uc)-u(i1,i2,i3,uc))/(2.*dr(0))
     *                     +(rx(i1+1,i2,i3,1,1)+rx(i1,i2,i3,1,1))*(u(i1+1,i2+1,i3,uc)+u(i1,i2+1,i3,uc)-u(i1+1,i2-1,i3,uc)-u(i1,i2-1,i3,uc))/(8.*dr(1))
                      u2yp= (rx(i1+1,i2,i3,0,1)+rx(i1,i2,i3,0,1))*(u(i1+1,i2,i3,vc)-u(i1,i2,i3,vc))/(2.*dr(0))
     *                     +(rx(i1+1,i2,i3,1,1)+rx(i1,i2,i3,1,1))*(u(i1+1,i2+1,i3,vc)+u(i1,i2+1,i3,vc)-u(i1+1,i2-1,i3,vc)-u(i1,i2-1,i3,vc))/(8.*dr(1))
                      u1xm= (rx(i1,i2,i3,0,0)+rx(i1-1,i2,i3,0,0))*(u(i1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.*dr(0))
     *                     +(rx(i1,i2,i3,1,0)+rx(i1-1,i2,i3,1,0))*(u(i1,i2+1,i3,uc)+u(i1-1,i2+1,i3,uc)-u(i1,i2-1,i3,uc)-u(i1-1,i2-1,i3,uc))/(8.*dr(1))
                      u2xm= (rx(i1,i2,i3,0,0)+rx(i1-1,i2,i3,0,0))*(u(i1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.*dr(0))
     *                     +(rx(i1,i2,i3,1,0)+rx(i1-1,i2,i3,1,0))*(u(i1,i2+1,i3,vc)+u(i1-1,i2+1,i3,vc)-u(i1,i2-1,i3,vc)-u(i1-1,i2-1,i3,vc))/(8.*dr(1))
                      u1ym= (rx(i1,i2,i3,0,1)+rx(i1-1,i2,i3,0,1))*(u(i1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.*dr(0))
     *                     +(rx(i1,i2,i3,1,1)+rx(i1-1,i2,i3,1,1))*(u(i1,i2+1,i3,uc)+u(i1-1,i2+1,i3,uc)-u(i1,i2-1,i3,uc)-u(i1-1,i2-1,i3,uc))/(8.*dr(1))
                      u2ym= (rx(i1,i2,i3,0,1)+rx(i1-1,i2,i3,0,1))*(u(i1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.*dr(0))
     *                     +(rx(i1,i2,i3,1,1)+rx(i1-1,i2,i3,1,1))*(u(i1,i2+1,i3,vc)+u(i1-1,i2+1,i3,vc)-u(i1,i2-1,i3,vc)-u(i1-1,i2-1,i3,vc))/(8.*dr(1))
                      u(i1-is1,i2,i3,s11c)=u(i1+is1,i2,i3,s11c)-2.*is*( (lambda+2.*mu)*(u1xp-u1xm) + lambda*(u2yp-u2ym) )
                      u(i1-is1,i2,i3,s12c)=u(i1+is1,i2,i3,s12c)-2.*is*( mu*(u1yp-u1ym+u2xp-u2xm) )
                      u(i1-is1,i2,i3,s21c)=u(i1+is1,i2,i3,s21c)-2.*is*( mu*(u1yp-u1ym+u2xp-u2xm) )
                      u(i1-is1,i2,i3,s22c)=u(i1+is1,i2,i3,s22c)-2.*is*( lambda*(u1xp-u1xm) + (lambda+2.*mu)*(u2yp-u2ym) )
                    end if
                  endLoops2d()
                end if
              else ! axis .eq. 1
                if( boundaryCondition(0,0).ne.tractionBC.and.boundaryCondition(1,0).ne.tractionBC) then
                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      u1xp= (rx(i1,i2+1,i3,0,0)+rx(i1,i2,i3,0,0))*(u(i1+1,i2+1,i3,uc)+u(i1+1,i2,i3,uc)-u(i1-1,i2+1,i3,uc)-u(i1-1,i2,i3,uc))/(8.*dr(0))
     *                     +(rx(i1,i2+1,i3,1,0)+rx(i1,i2,i3,1,0))*(u(i1,i2+1,i3,uc)-u(i1,i2,i3,uc))/(2.*dr(1))
                      u2xp= (rx(i1,i2+1,i3,0,0)+rx(i1,i2,i3,0,0))*(u(i1+1,i2+1,i3,vc)+u(i1+1,i2,i3,vc)-u(i1-1,i2+1,i3,vc)-u(i1-1,i2,i3,vc))/(8.*dr(0))
     *                     +(rx(i1,i2+1,i3,1,0)+rx(i1,i2,i3,1,0))*(u(i1,i2+1,i3,vc)-u(i1,i2,i3,vc))/(2.*dr(1))
                      u1yp= (rx(i1,i2+1,i3,0,1)+rx(i1,i2,i3,0,1))*(u(i1+1,i2+1,i3,uc)+u(i1+1,i2,i3,uc)-u(i1-1,i2+1,i3,uc)-u(i1-1,i2,i3,uc))/(8.*dr(0))
     *                     +(rx(i1,i2+1,i3,1,1)+rx(i1,i2,i3,1,1))*(u(i1,i2+1,i3,uc)-u(i1,i2,i3,uc))/(2.*dr(1))
                      u2yp= (rx(i1,i2+1,i3,0,1)+rx(i1,i2,i3,0,1))*(u(i1+1,i2+1,i3,vc)+u(i1+1,i2,i3,vc)-u(i1-1,i2+1,i3,vc)-u(i1-1,i2,i3,vc))/(8.*dr(0))
     *                     +(rx(i1,i2+1,i3,1,1)+rx(i1,i2,i3,1,1))*(u(i1,i2+1,i3,vc)-u(i1,i2,i3,vc))/(2.*dr(1))
                      u1xm= (rx(i1,i2,i3,0,0)+rx(i1,i2-1,i3,0,0))*(u(i1+1,i2,i3,uc)+u(i1+1,i2-1,i3,uc)-u(i1-1,i2,i3,uc)-u(i1-1,i2-1,i3,uc))/(8.*dr(0))
     *                     +(rx(i1,i2,i3,1,0)+rx(i1,i2-1,i3,1,0))*(u(i1,i2,i3,uc)-u(i1,i2-1,i3,uc))/(2.*dr(1))
                      u2xm= (rx(i1,i2,i3,0,0)+rx(i1,i2-1,i3,0,0))*(u(i1+1,i2,i3,vc)+u(i1+1,i2-1,i3,vc)-u(i1-1,i2,i3,vc)-u(i1-1,i2-1,i3,vc))/(8.*dr(0))
     *                     +(rx(i1,i2,i3,1,0)+rx(i1,i2-1,i3,1,0))*(u(i1,i2,i3,vc)-u(i1,i2-1,i3,vc))/(2.*dr(1))
                      u1ym= (rx(i1,i2,i3,0,1)+rx(i1,i2-1,i3,0,1))*(u(i1+1,i2,i3,uc)+u(i1+1,i2-1,i3,uc)-u(i1-1,i2,i3,uc)-u(i1-1,i2-1,i3,uc))/(8.*dr(0))
     *                     +(rx(i1,i2,i3,1,1)+rx(i1,i2-1,i3,1,1))*(u(i1,i2,i3,uc)-u(i1,i2-1,i3,uc))/(2.*dr(1))
                      u2ym= (rx(i1,i2,i3,0,1)+rx(i1,i2-1,i3,0,1))*(u(i1+1,i2,i3,vc)+u(i1+1,i2-1,i3,vc)-u(i1-1,i2,i3,vc)-u(i1-1,i2-1,i3,vc))/(8.*dr(0))
     *                     +(rx(i1,i2,i3,1,1)+rx(i1,i2-1,i3,1,1))*(u(i1,i2,i3,vc)-u(i1,i2-1,i3,vc))/(2.*dr(1))
                      u(i1,i2-is2,i3,s11c)=u(i1,i2+is2,i3,s11c)-2.*is*( (lambda+2.*mu)*(u1xp-u1xm) + lambda*(u2yp-u2ym) )
                      u(i1,i2-is2,i3,s12c)=u(i1,i2+is2,i3,s12c)-2.*is*( mu*(u1yp-u1ym+u2xp-u2xm) )
                      u(i1,i2-is2,i3,s21c)=u(i1,i2+is2,i3,s21c)-2.*is*( mu*(u1yp-u1ym+u2xp-u2xm) )
                      u(i1,i2-is2,i3,s22c)=u(i1,i2+is2,i3,s22c)-2.*is*( lambda*(u1xp-u1xm) + (lambda+2.*mu)*(u2yp-u2ym) )
                    end if
                  endLoops2d()
                end if
              end if ! axis

           else        ! option 2

              if( axis.eq.0 )then
                if( boundaryCondition(0,1).ne.tractionBC.and.boundaryCondition(1,1).ne.tractionBC) then
                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      u1r=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dr(0))
                      u1s=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dr(1))
                      u2r=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dr(0))
                      u2s=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dr(1))
                      u1rr=(u(i1+1,i2,i3,uc)-2.*u(i1,i2,i3,uc)+u(i1-1,i2,i3,uc))/dr(0)**2
                      u1rs=(u(i1+1,i2+1,i3,uc)-u(i1-1,i2+1,i3,uc)-u(i1+1,i2-1,i3,uc)+u(i1-1,i2-1,i3,uc))/(4.*dr(0)*dr(1))
                      u2rr=(u(i1+1,i2,i3,vc)-2.*u(i1,i2,i3,vc)+u(i1-1,i2,i3,vc))/dr(0)**2
                      u2rs=(u(i1+1,i2+1,i3,vc)-u(i1-1,i2+1,i3,vc)-u(i1+1,i2-1,i3,vc)+u(i1-1,i2-1,i3,vc))/(4.*dr(0)*dr(1))
                      rxr=(rx(i1+1,i2,i3,0,0)-rx(i1-1,i2,i3,0,0))/(2.*dr(0))
                      ryr=(rx(i1+1,i2,i3,0,1)-rx(i1-1,i2,i3,0,1))/(2.*dr(0))
                      sxr=(rx(i1+1,i2,i3,1,0)-rx(i1-1,i2,i3,1,0))/(2.*dr(0))
                      syr=(rx(i1+1,i2,i3,1,1)-rx(i1-1,i2,i3,1,1))/(2.*dr(0))
                      u(i1-is1,i2,i3,s11c)=u(i1+is1,i2,i3,s11c)-2.*is*dr(0)*(
     *                                     +(lambda+2.*mu)*(u1rr*rx(i1,i2,i3,0,0)+u1rs*rx(i1,i2,i3,1,0)+u1r*rxr+u1s*sxr)
     *                                             +lambda*(u2rr*rx(i1,i2,i3,0,1)+u2rs*rx(i1,i2,i3,1,1)+u2r*ryr+u2s*syr))
                      u(i1-is1,i2,i3,s12c)=u(i1+is1,i2,i3,s12c)-2.*is*dr(0)*(
     *                                                +mu*((u1rr*rx(i1,i2,i3,0,1)+u1rs*rx(i1,i2,i3,1,1)+u1r*ryr+u1s*syr)
     *                                                    +(u2rr*rx(i1,i2,i3,0,0)+u2rs*rx(i1,i2,i3,1,0)+u2r*rxr+u2s*sxr)))
c                      u(i1-is1,i2,i3,s21c)=u(i1-is1,i2,i3,s12c)  
                      u(i1-is1,i2,i3,s21c)=u(i1+is1,i2,i3,s21c)-2.*is*dr(0)*(
     *                                                +mu*((u1rr*rx(i1,i2,i3,0,1)+u1rs*rx(i1,i2,i3,1,1)+u1r*ryr+u1s*syr)
     *                                                    +(u2rr*rx(i1,i2,i3,0,0)+u2rs*rx(i1,i2,i3,1,0)+u2r*rxr+u2s*sxr)))
                      u(i1-is1,i2,i3,s22c)=u(i1+is1,i2,i3,s22c)-2.*is*dr(0)*(
     *                                             +lambda*(u1rr*rx(i1,i2,i3,0,0)+u1rs*rx(i1,i2,i3,1,0)+u1r*rxr+u1s*sxr)
     *                                     +(lambda+2.*mu)*(u2rr*rx(i1,i2,i3,0,1)+u2rs*rx(i1,i2,i3,1,1)+u2r*ryr+u2s*syr))
                    end if
                  endLoops2d()
                end if
              else ! axis .eq. 1
                if( boundaryCondition(0,0).ne.tractionBC.and.boundaryCondition(1,0).ne.tractionBC) then
                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      u1r=(u(i1+1,i2,i3,uc)-u(i1-1,i2,i3,uc))/(2.0*dr(0))
                      u1s=(u(i1,i2+1,i3,uc)-u(i1,i2-1,i3,uc))/(2.0*dr(1))
                      u2r=(u(i1+1,i2,i3,vc)-u(i1-1,i2,i3,vc))/(2.0*dr(0))
                      u2s=(u(i1,i2+1,i3,vc)-u(i1,i2-1,i3,vc))/(2.0*dr(1))
                      u1rs=(u(i1+1,i2+1,i3,uc)-u(i1-1,i2+1,i3,uc)-u(i1+1,i2-1,i3,uc)+u(i1-1,i2-1,i3,uc))/(4.*dr(0)*dr(1))
                      u1ss=(u(i1,i2+1,i3,uc)-2.*u(i1,i2,i3,uc)+u(i1,i2-1,i3,uc))/dr(1)**2
                      u2rs=(u(i1+1,i2+1,i3,vc)-u(i1-1,i2+1,i3,vc)-u(i1+1,i2-1,i3,vc)+u(i1-1,i2-1,i3,vc))/(4.*dr(0)*dr(1))
                      u2ss=(u(i1,i2+1,i3,vc)-2.*u(i1,i2,i3,vc)+u(i1,i2-1,i3,vc))/dr(1)**2
                      rxs=(rx(i1,i2+1,i3,0,0)-rx(i1,i2-1,i3,0,0))/(2.*dr(1))
                      rys=(rx(i1,i2+1,i3,0,1)-rx(i1,i2-1,i3,0,1))/(2.*dr(1))
                      sxs=(rx(i1,i2+1,i3,1,0)-rx(i1,i2-1,i3,1,0))/(2.*dr(1))
                      sys=(rx(i1,i2+1,i3,1,1)-rx(i1,i2-1,i3,1,1))/(2.*dr(1))
                      u(i1,i2-is2,i3,s11c)=u(i1,i2+is2,i3,s11c)-2.*is*dr(1)*(
     *                                     +(lambda+2.*mu)*(u1rs*rx(i1,i2,i3,0,0)+u1ss*rx(i1,i2,i3,1,0)+u1r*rxs+u1s*sxs)
     *                                             +lambda*(u2rs*rx(i1,i2,i3,0,1)+u2ss*rx(i1,i2,i3,1,1)+u2r*rys+u2s*sys))
                      u(i1,i2-is2,i3,s12c)=u(i1,i2+is2,i3,s12c)-2.*is*dr(1)*(
     *                                                +mu*((u1rs*rx(i1,i2,i3,0,1)+u1ss*rx(i1,i2,i3,1,1)+u1r*rys+u1s*sys)
     *                                                    +(u2rs*rx(i1,i2,i3,0,0)+u2ss*rx(i1,i2,i3,1,0)+u2r*rxs+u2s*sxs)))
c                      u(i1,i2-is2,i3,s21c)=u(i1,i2-is2,i3,s12c)
                      u(i1,i2-is2,i3,s21c)=u(i1,i2+is2,i3,s21c)-2.*is*dr(1)*(
     *                                                +mu*((u1rs*rx(i1,i2,i3,0,1)+u1ss*rx(i1,i2,i3,1,1)+u1r*rys+u1s*sys)
     *                                                    +(u2rs*rx(i1,i2,i3,0,0)+u2ss*rx(i1,i2,i3,1,0)+u2r*rxs+u2s*sxs)))
                      u(i1,i2-is2,i3,s22c)=u(i1,i2+is2,i3,s22c)-2.*is*dr(1)*(
     *                                             +lambda*(u1rs*rx(i1,i2,i3,0,0)+u1ss*rx(i1,i2,i3,1,0)+u1r*rxs+u1s*sxs)
     *                                     +(lambda+2.*mu)*(u2rs*rx(i1,i2,i3,0,1)+u2ss*rx(i1,i2,i3,1,1)+u2r*rys+u2s*sys))
                    end if
                  endLoops2d()
                end if
              end if ! axis

           end if    ! end of options (temporary)

            end if  ! end gridType
          end if ! bc 
        endLoopOverSides()

c add on TZ forcing (if necessary)
        if (twilightZone.ne.0) then
          beginLoopOverSides(numGhost,numGhost)
            if( boundaryCondition(side,axis).eq.tractionBC )then
              if( gridType.eq.rectangular )then

                if( axis.eq.0 )then

                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      call ogDeriv (ep,0,2,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xxe)
                      call ogDeriv (ep,0,2,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xxe)
                      call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xye)
                      call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xye)
                      call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11xe)
                      call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12xe)
                      call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21xe)
                      call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22xe)
                      u(i1-is1,i2,i3,s11c)=u(i1-is1,i2,i3,s11c)-2.*is*dx(0)*(s11xe-(lambda+2.*mu)*u1xxe-lambda*u2xye)
                      u(i1-is1,i2,i3,s12c)=u(i1-is1,i2,i3,s12c)-2.*is*dx(0)*(s12xe-mu*(u1xye+u2xxe))
                      u(i1-is1,i2,i3,s21c)=u(i1-is1,i2,i3,s21c)-2.*is*dx(0)*(s21xe-mu*(u1xye+u2xxe))
                      u(i1-is1,i2,i3,s22c)=u(i1-is1,i2,i3,s22c)-2.*is*dx(0)*(s22xe-lambda*u1xxe-(lambda+2.*mu)*u2xye)
                    end if
                  endLoops2d()
                else ! axis .eq. 1
                  beginLoops2d()
                    if (mask(i1,i2,i3).ne.0) then
                      call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xye)
                      call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xye)
                      call ogDeriv (ep,0,0,2,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1yye)
                      call ogDeriv (ep,0,0,2,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2yye)
                      call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11ye)
                      call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12ye)
                      call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21ye)
                      call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22ye)
                      u(i1,i2-is2,i3,s11c)=u(i1,i2-is2,i3,s11c)-2.*is*dx(1)*(s11ye-(lambda+2.*mu)*u1xye-lambda*u2yye)
                      u(i1,i2-is2,i3,s12c)=u(i1,i2-is2,i3,s12c)-2.*is*dx(1)*(s12ye-mu*(u1yye+u2xye))
                      u(i1,i2-is2,i3,s21c)=u(i1,i2-is2,i3,s21c)-2.*is*dx(1)*(s21ye-mu*(u1yye+u2xye))
                      u(i1,i2-is2,i3,s22c)=u(i1,i2-is2,i3,s22c)-2.*is*dx(1)*(s22ye-lambda*u1xye-(lambda+2.*mu)*u2yye)
                    end if
                  endLoops2d()
                end if ! axis 

              else  ! curvilinear 

          if (.false.) then     ! temporary : difference ways to apply the TZ forcing

                if( axis.eq.0 )then
                  if( boundaryCondition(0,1).ne.tractionBC.and.boundaryCondition(1,1).ne.tractionBC) then
                    beginLoops2d()
                      if (mask(i1,i2,i3).ne.0) then
                        call ogDeriv (ep,0,1,0,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,uc,u1xpe)
                        call ogDeriv (ep,0,1,0,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,vc,u2xpe)
                        call ogDeriv (ep,0,0,1,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,uc,u1ype)
                        call ogDeriv (ep,0,0,1,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,vc,u2ype)
                        call ogDeriv (ep,0,1,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,uc,u1xme)
                        call ogDeriv (ep,0,1,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,vc,u2xme)
                        call ogDeriv (ep,0,0,1,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,uc,u1yme)
                        call ogDeriv (ep,0,0,1,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,vc,u2yme)
                        call ogDeriv (ep,0,0,0,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,s11c,s11pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,s12c,s12pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,s21c,s21pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1+is1,i2,i3,0),xy(i1+is1,i2,i3,1),0.0,t,s22c,s22pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s11c,s11me)
                        call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s12c,s12me)
                        call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s21c,s21me)
                        call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s22c,s22me)
                        u(i1-is1,i2,i3,s11c)=u(i1-is1,i2,i3,s11c)+s11me-s11pe+(lambda+2.*mu)*(u1xpe-u1xme)+lambda*(u2ype-u2yme)
                        u(i1-is1,i2,i3,s12c)=u(i1-is1,i2,i3,s12c)+s12me-s12pe+mu*(u1ype-u1yme+u2xpe-u2xme)
                        u(i1-is1,i2,i3,s21c)=u(i1-is1,i2,i3,s21c)+s21me-s21pe+mu*(u1ype-u1yme+u2xpe-u2xme)
                        u(i1-is1,i2,i3,s22c)=u(i1-is1,i2,i3,s22c)+s22me-s22pe+lambda*(u1xpe-u1xme)+(lambda+2.*mu)*(u2ype-u2yme)
c                  call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s11c,s11e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s12c,s12e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s21c,s21e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1-is1,i2,i3,0),xy(i1-is1,i2,i3,1),0.0,t,s22c,s22e)
c                  s11e=u(i1-is1,i2,i3,s11c)-s11e
c                  s12e=u(i1-is1,i2,i3,s12c)-s12e
c                  s21e=u(i1-is1,i2,i3,s21c)-s21e
c                  s22e=u(i1-is1,i2,i3,s22c)-s22e

c           write(6,"(2(1x,i2),4(1x,1pe10.3))")i1-is1,i2,s11e,s12e,s21e,s22e
                      end if
                    endLoops2d()
                  end if
                else ! axis .eq. 1
                  if( boundaryCondition(0,0).ne.tractionBC.and.boundaryCondition(1,0).ne.tractionBC) then
                    beginLoops2d()
                      if (mask(i1,i2,i3).ne.0) then
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,uc,u1xpe)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,vc,u2xpe)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,uc,u1ype)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,vc,u2ype)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,uc,u1xme)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,vc,u2xme)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,uc,u1yme)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,vc,u2yme)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,s11c,s11pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,s12c,s12pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,s21c,s21pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2+is2,i3,0),xy(i1,i2+is2,i3,1),0.0,t,s22c,s22pe)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s11c,s11me)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s12c,s12me)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s21c,s21me)
                        call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s22c,s22me)
                        u(i1,i2-is2,i3,s11c)=u(i1,i2-is2,i3,s11c)+s11me-s11pe+(lambda+2.*mu)*(u1xpe-u1xme)+lambda*(u2ype-u2yme)
                        u(i1,i2-is2,i3,s12c)=u(i1,i2-is2,i3,s12c)+s12me-s12pe+mu*(u1ype-u1yme+u2xpe-u2xme)
                        u(i1,i2-is2,i3,s21c)=u(i1,i2-is2,i3,s21c)+s21me-s21pe+mu*(u1ype-u1yme+u2xpe-u2xme)
                        u(i1,i2-is2,i3,s22c)=u(i1,i2-is2,i3,s22c)+s22me-s22pe+lambda*(u1xpe-u1xme)+(lambda+2.*mu)*(u2ype-u2yme)
c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s11c,s11e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s12c,s12e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s21c,s21e)
c                  call ogDeriv (ep,0,0,0,0,xy(i1,i2-is2,i3,0),xy(i1,i2-is2,i3,1),0.0,t,s22c,s22e)
c                  s11e=u(i1,i2-is2,i3,s11c)-s11e
c                  s12e=u(i1,i2-is2,i3,s12c)-s12e
c                  s21e=u(i1,i2-is2,i3,s21c)-s21e
c                  s22e=u(i1,i2-is2,i3,s22c)-s22e

c           write(6,"(2(1x,i2),4(1x,1pe10.3))")i1,i2-is2,s11e,s12e,s21e,s22e
                      end if
                    endLoops2d()
                  end if
                end if ! axis 

             else     ! options

                if( axis.eq.0 )then
                  if( boundaryCondition(0,1).ne.tractionBC.and.boundaryCondition(1,1).ne.tractionBC) then
                    beginLoops2d()
                      if (mask(i1,i2,i3).ne.0) then
                        call ogDeriv (ep,0,2,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xxe)
                        call ogDeriv (ep,0,2,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xxe)
                        call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xye)
                        call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xye)
                        call ogDeriv (ep,0,0,2,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1yye)
                        call ogDeriv (ep,0,0,2,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2yye)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11xe)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12xe)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21xe)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22xe)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11ye)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12ye)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21ye)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22ye)
                        u(i1-is1,i2,i3,s11c)=u(i1-is1,i2,i3,s11c)-2.*is*dr(0)*det(i1,i2,i3)*(
     *                                        +rx(i1,i2,i3,1,1)*(s11xe-(lambda+2.*mu)*u1xxe-lambda*u2xye)
     *                                        -rx(i1,i2,i3,1,0)*(s11ye-(lambda+2.*mu)*u1xye-lambda*u2yye))
                        u(i1-is1,i2,i3,s12c)=u(i1-is1,i2,i3,s12c)-2.*is*dr(0)*det(i1,i2,i3)*(
     *                                        +rx(i1,i2,i3,1,1)*(s12xe-mu*(u1xye+u2xxe))
     *                                        -rx(i1,i2,i3,1,0)*(s12ye-mu*(u1yye+u2xye)))
                        u(i1-is1,i2,i3,s21c)=u(i1-is1,i2,i3,s21c)-2.*is*dr(0)*det(i1,i2,i3)*(
     *                                        +rx(i1,i2,i3,1,1)*(s21xe-mu*(u1xye+u2xxe))
     *                                        -rx(i1,i2,i3,1,0)*(s21ye-mu*(u1yye+u2xye)))
                        u(i1-is1,i2,i3,s22c)=u(i1-is1,i2,i3,s22c)-2.*is*dr(0)*det(i1,i2,i3)*(
     *                                        +rx(i1,i2,i3,1,1)*(s22xe-lambda*u1xxe-(lambda+2.*mu)*u2xye)
     *                                        -rx(i1,i2,i3,1,0)*(s22ye-lambda*u1xye-(lambda+2.*mu)*u2yye))
                      end if
                    endLoops2d()
                  end if
                else ! axis .eq. 1
                  if( boundaryCondition(0,0).ne.tractionBC.and.boundaryCondition(1,0).ne.tractionBC) then
                    beginLoops2d()
                      if (mask(i1,i2,i3).ne.0) then
                        call ogDeriv (ep,0,2,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xxe)
                        call ogDeriv (ep,0,2,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xxe)
                        call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1xye)
                        call ogDeriv (ep,0,1,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2xye)
                        call ogDeriv (ep,0,0,2,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u1yye)
                        call ogDeriv (ep,0,0,2,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u2yye)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11xe)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12xe)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21xe)
                        call ogDeriv (ep,0,1,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22xe)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s11c,s11ye)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s12c,s12ye)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s21c,s21ye)
                        call ogDeriv (ep,0,0,1,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,s22c,s22ye)
                        u(i1,i2-is2,i3,s11c)=u(i1,i2-is2,i3,s11c)-2.*is*dr(1)*det(i1,i2,i3)*(
     *                                        -rx(i1,i2,i3,0,1)*(s11xe-(lambda+2.*mu)*u1xxe-lambda*u2xye)
     *                                        +rx(i1,i2,i3,0,0)*(s11ye-(lambda+2.*mu)*u1xye-lambda*u2yye))
                        u(i1,i2-is2,i3,s12c)=u(i1,i2-is2,i3,s12c)-2.*is*dr(1)*det(i1,i2,i3)*(
     *                                        -rx(i1,i2,i3,0,1)*(s12xe-mu*(u1xye+u2xxe))
     *                                        +rx(i1,i2,i3,0,0)*(s12ye-mu*(u1yye+u2xye)))
                        u(i1,i2-is2,i3,s21c)=u(i1,i2-is2,i3,s21c)-2.*is*dr(1)*det(i1,i2,i3)*(
     *                                        -rx(i1,i2,i3,0,1)*(s21xe-mu*(u1xye+u2xxe))
     *                                        +rx(i1,i2,i3,0,0)*(s21ye-mu*(u1yye+u2xye)))
                        u(i1,i2-is2,i3,s22c)=u(i1,i2-is2,i3,s22c)-2.*is*dr(1)*det(i1,i2,i3)*(
     *                                        -rx(i1,i2,i3,0,1)*(s22xe-lambda*u1xxe-(lambda+2.*mu)*u2xye)
     *                                        +rx(i1,i2,i3,0,0)*(s22ye-lambda*u1xye-(lambda+2.*mu)*u2yye))
                      end if
                    endLoops2d()
                  end if
                end if ! axis

            end if   ! end options

              end if  ! end gridType
            end if ! bc 
          endLoopOverSides()
        end if


c*******
c******* An atempt to fix up the tractionBC-tractionBC corner problem ********
c******* (helps but is not satisfactory)

c..fix up values for stress in the first ghost line near corners
c  and extrapolate all components in the corner ghost points.
        i3=gridIndexRange(0,2)
        do side1=0,1
          i1=gridIndexRange(side1,axis1)
          is1=1-2*side1
          do side2=0,1
            i2=gridIndexRange(side2,axis2)
            is2=1-2*side2

c extrapolate in the i1 direction
            if (boundaryCondition(side1,axis1).eq.tractionBC.and.boundaryCondition(side2,axis2).eq.tractionBC) then
              if (mask(i1,i2,i3).ne.0) then
                u(i1-is1,i2,i3,s11c)=extrap3(u,i1,i2,i3,s11c,is1,0,0)
                u(i1-is1,i2,i3,s12c)=extrap3(u,i1,i2,i3,s12c,is1,0,0)
                u(i1-is1,i2,i3,s21c)=extrap3(u,i1,i2,i3,s21c,is1,0,0)
                u(i1-is1,i2,i3,s22c)=extrap3(u,i1,i2,i3,s22c,is1,0,0)
              end if
            end if

c extrapolate in the i2 direction
            if (boundaryCondition(side2,axis2).eq.tractionBC.and.boundaryCondition(side1,axis1).eq.tractionBC) then
              if (mask(i1,i2,i3).ne.0) then
                u(i1,i2-is2,i3,s11c)=extrap3(u,i1,i2,i3,s11c,0,is2,0)
                u(i1,i2-is2,i3,s12c)=extrap3(u,i1,i2,i3,s12c,0,is2,0)
                u(i1,i2-is2,i3,s21c)=extrap3(u,i1,i2,i3,s21c,0,is2,0)
                u(i1,i2-is2,i3,s22c)=extrap3(u,i1,i2,i3,s22c,0,is2,0)
              end if
            end if

c extrapolate in the diagonal direction
            if (boundaryCondition(side1,axis1).gt.0.and.boundaryCondition(side2,axis2).gt.0) then
              if (mask(i1,i2,i3).ne.0) then
                do n=0,numberOfComponents-1
                  u(i1-is1,i2-is2,i3,n)=extrap3(u,i1,i2,i3,n,is1,is2,0)
                end do
              end if
            end if
          end do
        end do

       end if ! ---

        if ( .false. ) then
          n1a=gridIndexRange(0,0)
          n1b=gridIndexRange(1,0)
          n2a=gridIndexRange(0,1)
          n2b=gridIndexRange(1,1)

          i1=n1a-1
          i2=n2a
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
  
          i1=n1a-1
          i2=n2b
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
 
          i1=n1b+1
          i2=n2a
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
  
          i1=n1b+1
          i2=n2b
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
 
          i1=n1a
          i2=n2a-1
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
      
          i1=n1b
          i2=n2a-1
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
 
          i1=n1a
          i2=n2b+1
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
      
          i1=n1b
          i2=n2b+1
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,uc,u(i1,i2,i3,uc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,vc,u(i1,i2,i3,vc))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v1c,u(i1,i2,i3,v1c))
          call ogDeriv (ep,0,0,0,0,xy(i1,i2,i3,0),xy(i1,i2,i3,1),0.0,t,v2c,u(i1,i2,i3,v2c))
        end if


c*******
c******* Extrapolation to the second ghost line ********
c*******

        beginLoopOverSides(numGhost,numGhost)
         if( boundaryCondition(side,axis).gt.0 )then
          beginLoops2d()
           if (mask(i1,i2,i3).ne.0) then
            do n=0,numberOfComponents-1
              u(i1-2*is1,i2-2*is2,i3,n)=extrap3(u,i1-is1,i2-is2,i3,n,is1,is2,is3)
            end do
           end if
          endLoops2d()
         end if ! bc 
        endLoopOverSides()

c..extrapolate the 2nd ghost line near the corners.
        i3=gridIndexRange(0,2)
        do side1=0,1
          i1=gridIndexRange(side1,axis1)
          is1=1-2*side1
          do side2=0,1
            i2=gridIndexRange(side2,axis2)
            is2=1-2*side2

c extrapolate in the i1 direction
            if (boundaryCondition(side1,axis1).gt.0) then
              if (mask(i1,i2,i3).ne.0) then
                do n=0,numberOfComponents-1
                  u(i1-2*is1,i2-is2,i3,n)=extrap3(u,i1-is1,i2-is2,i3,n,is1,0,0)
                end do
              end if
            end if

c extrapolate in the i2 direction
            if (boundaryCondition(side2,axis2).gt.0) then
              if (mask(i1,i2,i3).ne.0) then
                do n=0,numberOfComponents-1
                  u(i1-is1,i2-2*is2,i3,n)=extrap3(u,i1-is1,i2-is2,i3,n,0,is2,0)
                end do
              end if
            end if

c extrapolate in the diagonal direction
            if (boundaryCondition(side1,axis1).gt.0.and.boundaryCondition(side2,axis2).gt.0) then
              if (mask(i1,i2,i3).ne.0) then
                do n=0,numberOfComponents-1
                  u(i1-2*is1,i2-2*is2,i3,n)=extrap3(u,i1-is1,i2-is2,i3,n,is1,is2,0)
                end do
              end if
            end if
          end do
        end do

      if( .false. )then

        ! *** Don: call your BC routine here ****

        icart=0
        if (gridType.eq.rectangular) icart=1

        mc=8
        n1a=gridIndexRange(0,0)
        n1b=gridIndexRange(1,0)
        n2a=gridIndexRange(0,1)
        n2b=gridIndexRange(1,1)
        ds1=dr(0)
        ds2=dr(1)
        call corner (mc,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,ds1,ds2,t,xy,mask,rx,u,icart)

        if (abs(t-.5).lt.1.e-8) then
          call smgerr (mc,nd1a,nd1b,n1a,n1b,nd2a,nd2b,n2a,n2b,ds1,ds2,t,xy,u)
        end if

        return

      end if

      else if( nd.eq.3 )then

       !    *************************
       !    ********** 3D ***********
       !    *************************

        stop 3366


      else
        ! unknown nd 
        stop 8826 
      end if 

      return
      end
